Package: scRNAseqApp
Title: A single-cell RNAseq Shiny app-package
Version: 1.12.0
Authors@R: 
    c(person(given = "Jianhong",
             family = "Ou",
             role = c("aut", "cre"),
             email = "jou@morgridge.org",
             comment=c(ORCID="0000-0002-8652-2488")))
Description: The scRNAseqApp is a Shiny app package designed for
        interactive visualization of single-cell data. It is an
        enhanced version derived from the ShinyCell, repackaged to
        accommodate multiple datasets. The app enables users to
        visualize data containing various types of information
        simultaneously, facilitating comprehensive analysis.
        Additionally, it includes a user management system to regulate
        database accessibility for different users.
License: GPL-3
Encoding: UTF-8
Roxygen: list(markdown = TRUE)
RoxygenNote: 7.3.3
VignetteBuilder: knitr
biocViews: Visualization, SingleCell, RNASeq
Depends: R (>= 4.3.0)
Imports: bibtex, bslib, circlize, ComplexHeatmap, colourpicker,
        data.table, desc, DBI, DT, fs, GenomicRanges, GenomeInfoDb,
        ggdendro, ggforce, ggnewscale, ggplot2, ggrepel, ggridges,
        grDevices, grid, gridExtra, htmltools, IRanges, jsonlite,
        Matrix, magrittr, methods, patchwork, plotly, RColorBrewer,
        RefManageR, reshape2, rhdf5, Rsamtools, RSQLite, rtracklayer,
        S4Vectors, scales, scrypt, Seurat, SeuratObject, shiny,
        shinyhelper, shinymanager, slingshot, SingleCellExperiment,
        sortable, stats, tools, xfun, xml2, utils
Suggests: rmarkdown, knitr, testthat, BiocStyle, shinytest2
Enhances: celldex, future, SingleR, SummarizedExperiment, tricycle,
        terra
URL: https://github.com/jianhong/scRNAseqApp
BugReports: https://github.com/jianhong/scRNAseqApp/issues
Config/pak/sysreqs: cmake libfontconfig1-dev libfreetype6-dev
        libglpk-dev make libbz2-dev libicu-dev liblzma-dev libpng-dev
        libuv1-dev libxml2-dev libssl-dev perl python3 xz-utils
        zlib1g-dev
Repository: https://bioc-release.r-universe.dev
Date/Publication: 2026-04-28 12:59:37 UTC
RemoteUrl: https://github.com/bioc/scRNAseqApp
RemoteRef: RELEASE_3_23
RemoteSha: 96c8478c988d445bc6e16f5fb13f936abcc81fe0
NeedsCompilation: no
Packaged: 2026-04-30 06:28:39 UTC; root
Author: Jianhong Ou [aut, cre] (ORCID: <https://orcid.org/0000-0002-8652-2488>)
Maintainer: Jianhong Ou <jou@morgridge.org>
Built: R 4.6.0; ; 2026-04-30 07:16:47 UTC; unix

# Generated by roxygen2: do not edit by hand

export(APPconf)
export(availableThemes)
export(createAppConfig)
export(createDataSet)
export(extractFragmentNameMapList)
export(scInit)
export(scRNAseqApp)
exportClasses(APPconf)
exportMethods("$")
exportMethods("$<-")
exportMethods("[")
exportMethods("[[")
exportMethods("[[<-")
exportMethods(as.character)
exportMethods(as.list)
exportMethods(lapply)
exportMethods(markers)
exportMethods(show)
exportMethods(unlist)
import(shiny)
importFrom(ComplexHeatmap,Heatmap)
importFrom(ComplexHeatmap,HeatmapAnnotation)
importFrom(ComplexHeatmap,Legend)
importFrom(ComplexHeatmap,draw)
importFrom(ComplexHeatmap,merge_dendrogram)
importFrom(ComplexHeatmap,pindex)
importFrom(ComplexHeatmap,rowAnnotation)
importFrom(DBI,dbAppendTable)
importFrom(DBI,dbClearResult)
importFrom(DBI,dbConnect)
importFrom(DBI,dbDisconnect)
importFrom(DBI,dbGetQuery)
importFrom(DBI,dbListFields)
importFrom(DBI,dbListTables)
importFrom(DBI,dbSendQuery)
importFrom(DBI,dbWriteTable)
importFrom(DBI,sqlInterpolate)
importFrom(DT,DTOutput)
importFrom(DT,JS)
importFrom(DT,datatable)
importFrom(DT,formatRound)
importFrom(DT,renderDT)
importFrom(GenomeInfoDb,`seqlevelsStyle<-`)
importFrom(GenomeInfoDb,keepSeqlevels)
importFrom(GenomeInfoDb,seqinfo)
importFrom(GenomeInfoDb,seqlevelsStyle)
importFrom(GenomeInfoDb,seqnames)
importFrom(GenomicRanges,GRanges)
importFrom(GenomicRanges,GRangesList)
importFrom(GenomicRanges,`end<-`)
importFrom(GenomicRanges,`start<-`)
importFrom(GenomicRanges,`strand<-`)
importFrom(GenomicRanges,coverage)
importFrom(GenomicRanges,end)
importFrom(GenomicRanges,gaps)
importFrom(GenomicRanges,makeGRangesFromDataFrame)
importFrom(GenomicRanges,reduce)
importFrom(GenomicRanges,seqnames)
importFrom(GenomicRanges,start)
importFrom(GenomicRanges,strand)
importFrom(GenomicRanges,width)
importFrom(IRanges,IRanges)
importFrom(IRanges,Views)
importFrom(IRanges,nearest)
importFrom(IRanges,ranges)
importFrom(IRanges,shift)
importFrom(IRanges,subsetByOverlaps)
importFrom(IRanges,tile)
importFrom(IRanges,viewMeans)
importFrom(Matrix,sparseMatrix)
importFrom(RColorBrewer,brewer.pal)
importFrom(RColorBrewer,brewer.pal.info)
importFrom(RSQLite,SQLite)
importFrom(RefManageR,GetBibEntryWithDOI)
importFrom(RefManageR,GetPubMedByID)
importFrom(RefManageR,PrintBibliography)
importFrom(Rsamtools,TabixFile)
importFrom(Rsamtools,scanTabix)
importFrom(Rsamtools,seqnamesTabix)
importFrom(S4Vectors,DataFrame)
importFrom(S4Vectors,SimpleList)
importFrom(S4Vectors,mcols)
importFrom(S4Vectors,split)
importFrom(Seurat,CellCycleScoring)
importFrom(Seurat,CreateSeuratObject)
importFrom(Seurat,FindAllMarkers)
importFrom(Seurat,FindVariableFeatures)
importFrom(Seurat,Read10X)
importFrom(Seurat,ScaleData)
importFrom(Seurat,as.SingleCellExperiment)
importFrom(SeuratObject,Assays)
importFrom(SeuratObject,CreateDimReducObject)
importFrom(SeuratObject,DefaultAssay)
importFrom(SeuratObject,DefaultBoundary)
importFrom(SeuratObject,DefaultFOV)
importFrom(SeuratObject,Embeddings)
importFrom(SeuratObject,GetAssayData)
importFrom(SeuratObject,GetTissueCoordinates)
importFrom(SeuratObject,Idents)
importFrom(SeuratObject,Images)
importFrom(SeuratObject,Misc)
importFrom(SeuratObject,Molecules)
importFrom(SeuratObject,Reductions)
importFrom(SeuratObject,VariableFeatures)
importFrom(SeuratObject,`DefaultAssay<-`)
importFrom(SeuratObject,`Idents<-`)
importFrom(SeuratObject,`Misc<-`)
importFrom(SingleCellExperiment,reducedDim)
importFrom(bibtex,read.bib)
importFrom(bslib,bs_current_theme)
importFrom(bslib,bs_get_variables)
importFrom(bslib,bs_theme)
importFrom(bslib,bs_themer)
importFrom(circlize,colorRamp2)
importFrom(colourpicker,colourInput)
importFrom(data.table,":=")
importFrom(data.table,.I)
importFrom(data.table,.N)
importFrom(data.table,.SD)
importFrom(data.table,as.data.table)
importFrom(data.table,data.table)
importFrom(data.table,dcast.data.table)
importFrom(data.table,fread)
importFrom(data.table,rbindlist)
importFrom(data.table,uniqueN)
importFrom(desc,desc)
importFrom(desc,desc_get_maintainer)
importFrom(fs,path_sanitize)
importFrom(ggdendro,dendro_data)
importFrom(ggforce,geom_arc_bar)
importFrom(ggforce,geom_bezier)
importFrom(ggforce,geom_circle)
importFrom(ggforce,geom_mark_hull)
importFrom(ggnewscale,new_scale_fill)
importFrom(ggplot2,.data)
importFrom(ggplot2,GeomPolygon)
importFrom(ggplot2,GeomViolin)
importFrom(ggplot2,aes)
importFrom(ggplot2,after_stat)
importFrom(ggplot2,coord_cartesian)
importFrom(ggplot2,coord_equal)
importFrom(ggplot2,coord_fixed)
importFrom(ggplot2,coord_flip)
importFrom(ggplot2,element_blank)
importFrom(ggplot2,element_line)
importFrom(ggplot2,element_rect)
importFrom(ggplot2,element_text)
importFrom(ggplot2,facet_grid)
importFrom(ggplot2,geom_bar)
importFrom(ggplot2,geom_boxplot)
importFrom(ggplot2,geom_col)
importFrom(ggplot2,geom_density_2d)
importFrom(ggplot2,geom_density_2d_filled)
importFrom(ggplot2,geom_hline)
importFrom(ggplot2,geom_jitter)
importFrom(ggplot2,geom_line)
importFrom(ggplot2,geom_point)
importFrom(ggplot2,geom_polygon)
importFrom(ggplot2,geom_raster)
importFrom(ggplot2,geom_rect)
importFrom(ggplot2,geom_segment)
importFrom(ggplot2,geom_text)
importFrom(ggplot2,geom_tile)
importFrom(ggplot2,geom_violin)
importFrom(ggplot2,ggplot)
importFrom(ggplot2,ggplot_build)
importFrom(ggplot2,ggplot_gtable)
importFrom(ggplot2,ggproto)
importFrom(ggplot2,ggsave)
importFrom(ggplot2,guide_colorbar)
importFrom(ggplot2,guide_legend)
importFrom(ggplot2,guides)
importFrom(ggplot2,labs)
importFrom(ggplot2,layer)
importFrom(ggplot2,layer_scales)
importFrom(ggplot2,position_dodge)
importFrom(ggplot2,scale_color_gradient2)
importFrom(ggplot2,scale_color_gradientn)
importFrom(ggplot2,scale_color_manual)
importFrom(ggplot2,scale_colour_manual)
importFrom(ggplot2,scale_fill_gradientn)
importFrom(ggplot2,scale_fill_manual)
importFrom(ggplot2,scale_x_continuous)
importFrom(ggplot2,scale_y_continuous)
importFrom(ggplot2,scale_y_discrete)
importFrom(ggplot2,theme)
importFrom(ggplot2,theme_classic)
importFrom(ggplot2,theme_minimal)
importFrom(ggplot2,unit)
importFrom(ggplot2,vars)
importFrom(ggplot2,xlab)
importFrom(ggplot2,xlim)
importFrom(ggplot2,ylab)
importFrom(ggplot2,ylim)
importFrom(ggrepel,geom_text_repel)
importFrom(ggridges,geom_density_ridges)
importFrom(ggridges,theme_ridges)
importFrom(grDevices,col2rgb)
importFrom(grDevices,colorRampPalette)
importFrom(grDevices,dev.off)
importFrom(grDevices,hcl.colors)
importFrom(grDevices,pdf)
importFrom(grDevices,pdfFonts)
importFrom(grDevices,rgb)
importFrom(grid,arrow)
importFrom(grid,convertWidth)
importFrom(grid,current.viewport)
importFrom(grid,gpar)
importFrom(grid,grid.circle)
importFrom(grid,grid.lines)
importFrom(grid,grid.xaxis)
importFrom(grid,grid.yaxis)
importFrom(grid,grobName)
importFrom(grid,grobTree)
importFrom(grid,seekViewport)
importFrom(grid,unit)
importFrom(grid,unit.c)
importFrom(gridExtra,arrangeGrob)
importFrom(gridExtra,grid.arrange)
importFrom(htmltools,htmlDependency)
importFrom(jsonlite,fromJSON)
importFrom(jsonlite,parse_json)
importFrom(magrittr,"%>%")
importFrom(methods,`slot<-`)
importFrom(methods,as)
importFrom(methods,getPackageName)
importFrom(methods,is)
importFrom(methods,new)
importFrom(methods,prototype)
importFrom(methods,representation)
importFrom(methods,setAs)
importFrom(methods,setClass)
importFrom(methods,setClassUnion)
importFrom(methods,setMethod)
importFrom(methods,show)
importFrom(methods,slot)
importFrom(methods,slotNames)
importFrom(patchwork,wrap_plots)
importFrom(plotly,event_data)
importFrom(plotly,event_register)
importFrom(plotly,ggplotly)
importFrom(plotly,layout)
importFrom(plotly,plot_ly)
importFrom(plotly,plotlyOutput)
importFrom(plotly,renderPlotly)
importFrom(plotly,toWebGL)
importFrom(reshape2,melt)
importFrom(rhdf5,H5Fclose)
importFrom(rhdf5,H5Fopen)
importFrom(rhdf5,H5Lexists)
importFrom(rhdf5,h5createDataset)
importFrom(rhdf5,h5createFile)
importFrom(rhdf5,h5createGroup)
importFrom(rhdf5,h5ls)
importFrom(rhdf5,h5read)
importFrom(rhdf5,h5write)
importFrom(rtracklayer,export)
importFrom(rtracklayer,import)
importFrom(scales,rescale)
importFrom(scales,zero_range)
importFrom(scrypt,hashPassword)
importFrom(shiny,Progress)
importFrom(shiny,downloadHandler)
importFrom(shiny,getDefaultReactiveDomain)
importFrom(shiny,markdown)
importFrom(shinyhelper,helper)
importFrom(shinyhelper,observe_helpers)
importFrom(shinymanager,auth_ui)
importFrom(shinymanager,check_credentials)
importFrom(shinymanager,create_db)
importFrom(shinymanager,fab_button)
importFrom(shinymanager,pwd_ui)
importFrom(shinymanager,read_db_decrypt)
importFrom(shinymanager,secure_app)
importFrom(shinymanager,secure_server)
importFrom(shinymanager,use_language)
importFrom(slingshot,SlingshotDataSet)
importFrom(slingshot,getLineages)
importFrom(slingshot,slingClusterLabels)
importFrom(slingshot,slingLineages)
importFrom(slingshot,slingMST)
importFrom(slingshot,slingParams)
importFrom(sortable,rank_list)
importFrom(stats,as.dendrogram)
importFrom(stats,as.formula)
importFrom(stats,chisq.test)
importFrom(stats,dist)
importFrom(stats,hclust)
importFrom(stats,prcomp)
importFrom(stats,quantile)
importFrom(stats,rnorm)
importFrom(stats,weighted.mean)
importFrom(stats,wilcox.test)
importFrom(tools,file_ext)
importFrom(utils,adist)
importFrom(utils,bibentry)
importFrom(utils,combn)
importFrom(utils,download.file)
importFrom(utils,getFromNamespace)
importFrom(utils,globalVariables)
importFrom(utils,head)
importFrom(utils,packageDescription)
importFrom(utils,packageVersion)
importFrom(utils,read.csv)
importFrom(utils,read.delim)
importFrom(utils,read.table)
importFrom(utils,write.csv)
importFrom(utils,write.table)
importFrom(xfun,base64_uri)
importFrom(xfun,is_abs_path)
importFrom(xml2,as_list)
importFrom(xml2,read_xml)
importMethodsFrom(Matrix,summary)

# scRNAseqApp

**scRNAseqApp** is a Bioconductor Shiny application for interactive exploration and visualization of single‑cell RNA‑seq (scRNA‑seq), scATAC-seq, multiomics or spatial data. It enables researchers to publish, explore, and share single‑cell datasets through a web‑based interface with rich plotting and filtering functionality.

---

## Key Features

* Interactive visualization of single‑cell datasets
* Support for multiple datasets within a single app
* Common single‑cell plots:

  * UMAP / t‑SNE
  * Violin plots
  * Dot plots
  * Heatmaps
* Metadata‑based filtering and grouping
* Downloadable figures and tables
* User authentication and management (optional)
* Suitable for local use or deployment on Shiny Server

---

## Installation

Install from Bioconductor:

```r
if (!require("BiocManager", quietly = TRUE))
    install.packages("BiocManager")

BiocManager::install("scRNAseqApp")
```

---

## Quick Start

### 1. Initialize an App Directory

Create and initialize an application folder:

```r
publish_folder <- tempdir()
scInit(app_path = publish_folder)
```

This sets up the required directory structure, database, and configuration files.
A typical app directory looks like:

```
MyApp/
├── data/                  # Stored datasets
├── www/
│   ├── database.sqlite    # User/app database
│   ├── counter.tsv        # Usage tracking
│   └── ...                # Static app assets
└── app.R
```

Note: When deploying, ensure the `www/` directory and database files are writable by the Shiny server user.

---

### 2. Launch the App

```r
scRNAseqApp(app_path = publish_folder)
```

The app will open in your default web browser.

---

## Adding Data

scRNAseqApp works primarily with **Seurat objects**.

---

### Option A: Add Data Programmatically (Recommended for Deployment)

```r
appconf <- createAppConfig(
    title = "My scRNA‑seq App",
    destinationFolder = "MyApp",
    species = "Homo sapiens",
    doi = "10.XXXX/your_doi",
    datatype = "scRNAseq",
    abstract = "Description of the dataset"
)

createDataSet(
    appconf,
    seurat_object,
    datafolder = file.path(publish_folder, "data")
)
```

This method is reproducible and recommended for shared or public apps.

---

### Option B: Upload Through the App (Admin Mode)

1. Open the app
2. Switch user to **administrator**
3. Navigate to the **Upload Data** tab
4. Upload a Seurat object

This method will use local computational resources, not suggested for large dataset.

---

## Deployment on Shiny Server

1. Install **scRNAseqApp** on the server.
2. Initialize the app directory using `scInit()`.
3. Set the correct permissions for:

   * `www/database.sqlite`
   * `www/counter.tsv`
   * Other files under `www/` if user management is enabled.
4. Add data by copying the `destinationFolder` (e.g., `MyApp` in the example above) into the `data/` folder on the server.  
   If the newly copied data are not available to public users, first try reloading the app,  
   then check the accessibility and permissions of the folder.
5. Removing the destination folder will result in deletion of the dataset.

The app can then be served like any standard Shiny application

---

## Notes

* Designed for **exploration and sharing**, not for heavy upstream computation
* Preprocessing (normalization, clustering, dimensionality reduction) should be done **before** loading data into the app
* Best used with well‑annotated metadata

---

For full details, see the official Bioconductor vignette:

```
vignette("scRNAseqApp")
```

---

### Video Tutorials
Comprehensive video tutorials demonstrating the usage of **scRNAseqApp** are available on the following YouTube playlist:  
https://youtube.com/playlist?list=PL0uThf-sipbnepUDzVmAHpNWUy65F4HOm


## Contributing

Contributions, suggestions, and bug reports are welcome!
Please open an issue or submit a pull request on [GitHub](https://github.com/jianhong/scRNAseqApp/issues).

# scripts to generate the data in extdata folder

## gene to symbol list: gn2sym.rds

The `gn2sym.rds` is a list with 2 elements: unique and multiple.
The unique element is a character vector of gene symbols with gene name as its
name.
The multiple is a list of character vectors with gene name as the name of the
list and gene symbols as elements for each gene name.
All the gene name and gene symbol were converted to lower case.
The script `gn2sym.R` is used to generate the `gn2sym` list.
The list were generated from packages:
"org.Ce.eg.db", "org.Cf.eg.db", "org.Dm.eg.db", "org.Dr.eg.db",
"org.Gg.eg.db", "org.Hs.eg.db", "org.Mm.eg.db", "org.Mmu.eg.db",
"org.Pt.eg.db", "org.Rn.eg.db", "org.Ss.eg.db", and "org.Xl.eg.db".
The data were created by `gn2sym.R` at the data Dec 23, 2022.

## cell cycle gene list: cc.genes.list.rds

The `cc.genes.list.rds` is a list of genes involved in cell cycles.
The first level of the list is the organisms which include 
Bos taurus, Caenorhabditis elegans, Drosophila melanogaster, Danio rerio,
Gallus gallus, Mus musculus, Macaca mulatta, Pan troglodytes,
Rattus norvegicus, Saccharomyces cerevisiae, Sus scrofa, and Homo sapiens.
The second level of the list is a list with names 'ensembl' and 'symbol'.
The elements in 'ensembl' is the ensembl ids and that in 'symbol' is the
gene symbols.
The third level of the list is list with names 's.genes' and 'g2m.genes'
with the gene ids related to 'S' and 'g2m' phase.
All the ids were converted from cell cycle genes (2019 update) of
Seurat package by biomart server of ensembl.org.
The data were created by `cell_cycle_gene_list.R` at the date Dec 23, 2022.

## sample data: pbmc_small

The `pbmc_samll` is the toy data to demo the APP. It was created by the script
`pbmc_small.R`.

## sample data: xenium_small

The `xenium_small` is the toy data to demo the APP for cell borders. See
`xenium_small.R`.

library(scRNAseqApp)
library(Seurat)
appconf <- createAppConfig(
  title="pbmc_small",
  destinationFolder = "pbmc_small",
  species = "Homo sapiens",
  doi="10.1038/nbt.3192",
  datatype = "scRNAseq",
  keywords = c("pbmc", "testdata"))
target='inst/extdata/data'
dir.create(target)
createDataSet(appconf, pbmc_small, datafolder=target)

pkgs <- available.packages(repos = repositories())
orgPkgs <- rownames(pkgs)[grepl("org.", rownames(pkgs), fixed = TRUE)]
BiocManager::install(orgPkgs)
orgPkgs <- c("org.Ce.eg.db", "org.Cf.eg.db", "org.Dm.eg.db", "org.Dr.eg.db",
             "org.Gg.eg.db", "org.Hs.eg.db", "org.Mm.eg.db", "org.Mmu.eg.db",
             "org.Pt.eg.db", "org.Rn.eg.db", "org.Ss.eg.db", "org.Xl.eg.db")
gns <- lapply(orgPkgs, function(org){
  library(org, character.only = TRUE)
  symbol <- as.list(get(sub(".db", "SYMBOL", org)))
  symbol <- vapply(symbol, `[`, FUN.VALUE = character(1L), i=1)
  genename <- as.list(get(sub(".db", "GENENAME", org)))
  names(genename) <- symbol[names(genename)]
  gn <- rep(names(genename), lengths(genename))
  names(gn) <- unlist(genename)
  gn
})

gnn <- unlist(lapply(gns, names))
sym <- unlist(gns)
gnn <- tolower(gnn)
sym <- tolower(sym)
gn <- split(sym, gnn)
gn <- lapply(gn, unique)
gn <- lapply(gn, function(.ele) .ele[!grepl("^(loc|gm)\\d+", .ele)])
table(lengths(gn))
gn1 <- unlist(gn[lengths(gn)==1])
gn2 <- gn[lengths(gn)>1]
res <- list('unique'=gn1, 'multiple'=gn2)
saveRDS(res, 'inst/extdata/gn2sym.rds')

library(biomaRt)
library(Seurat)
orgs <- c("Bos taurus",
          "Caenorhabditis elegans",
          "Drosophila melanogaster",
          "Danio rerio",
          "Gallus gallus",
          "Homo sapiens",
          "Mus musculus",
          "Macaca mulatta",
          "Pan troglodytes",
          "Rattus norvegicus",
          "Saccharomyces cerevisiae",
          "Sus scrofa")
orgs1 <- do.call(rbind, strsplit(orgs, " "))
orgs1 <- orgs1[, c(1, 2)]
homologN <- paste0(tolower(substr(orgs1[, 1], start = 1, stop = 1)), orgs1[, 2])
attrs <- lapply(homologN, function(.ele) paste0(.ele, c("_homolog_ensembl_gene","_homolog_associated_gene_name")))
names(attrs) <- orgs
s.genes <- cc.genes.updated.2019$s.genes
g2m.genes <- cc.genes.updated.2019$g2m.genes

mart <- useMart("ensembl", "hsapiens_gene_ensembl")
genes <- c(s.genes, g2m.genes)
genes <- getBM(attributes = c("ensembl_gene_id", "hgnc_symbol"),
               filters = "hgnc_symbol",
               values = genes, mart = mart)
genemap <- list()
for(i in seq_along(attrs)){
  tryCatch(
    {
      .genemap <- getBM(attributes = c("ensembl_gene_id",
                                       attrs[[i]]),
                        filters = "ensembl_gene_id",
                        values = genes$ensembl_gene_id, mart = mart)
      genemap[[names(attrs)[i]]] <- merge(genes, .genemap)
    },
    error = function(.e) message(.e)
  )
}
genemap[["Homo sapiens"]] <- cbind(genes, genes)
colnames(genemap[["Homo sapiens"]])[3:4] <- attrs[["Homo sapiens"]]
cc.genes.list.ensembl <- mapply(genemap, attrs[names(genemap)], FUN=function(.ele, .hn){
  s_genes <- .ele[.ele$hgnc_symbol %in% s.genes, .hn[1]]
  s_genes <- s_genes[s_genes!=""]
  g2m_genes <- .ele[.ele$hgnc_symbol %in% g2m.genes, .hn[1]]
  g2m_genes <- g2m_genes[g2m_genes!=""]
  list(s.genes=s_genes, g2m.genes=g2m_genes)
}, SIMPLIFY = FALSE)
cc.genes.list.symbol <- mapply(genemap, attrs[names(genemap)], FUN=function(.ele, .hn){
  s_genes <- .ele[.ele$hgnc_symbol %in% s.genes, .hn[2]]
  s_genes <- s_genes[s_genes!=""]
  g2m_genes <- .ele[.ele$hgnc_symbol %in% g2m.genes, .hn[2]]
  g2m_genes <- g2m_genes[g2m_genes!=""]
  list(s.genes=s_genes, g2m.genes=g2m_genes)
}, SIMPLIFY = FALSE)

cc.genes.list <- list("ensembl"=cc.genes.list.ensembl,
                      "symbol"=cc.genes.list.symbol)
cc.genes.list <- ChIPpeakAnno:::swapList(cc.genes.list)

saveRDS(cc.genes.list, "../inst/extdata/cc.genes.list.rds")

library(Seurat)
library(future)
plan("multisession", workers = 4)
library(ggplot2)
library(scRNAseqApp)
library(spacexr)
library(SPOTlight)
library(scran)

options(future.globals.maxSize = 10 * 1024^3)
# curl -O https://cf.10xgenomics.com/samples/xenium/1.0.2/Xenium_V1_FF_Mouse_Brain_Coronal_Subset_CTX_HP/Xenium_V1_FF_Mouse_Brain_Coronal_Subset_CTX_HP_outs.zip
# unzip Xenium_V1_FF_Mouse_Brain_Coronal_Subset_CTX_HP_outs.zip
path <- '~/Downloads/Xenium_V1'
xenium.obj <- LoadXenium(path, fov = "fov", segmentations = "cell")
xenium.obj <- subset(xenium.obj, subset = nCount_Xenium > 0)
xenium.obj <- SCTransform(xenium.obj, assay = "Xenium")
xenium.obj <- RunPCA(xenium.obj, npcs = 30, features = rownames(xenium.obj))
xenium.obj <- RunUMAP(xenium.obj, dims = 1:30)
xenium.obj <- FindNeighbors(xenium.obj, reduction = "pca", dims = 1:30)
xenium.obj <- FindClusters(xenium.obj, resolution = 0.3)
# curl -O https://www.dropbox.com/s/cuowvm4vrf65pvq/allen_cortex.rds?dl=1
allen.cortex.ref <- readRDS("~/Downloads/Xenium_V1/allen_cortex.rds")
allen.cortex.ref <- UpdateSeuratObject(allen.cortex.ref)

query.counts <- GetAssayData(xenium.obj, assay = "Xenium", layer = "counts")
coords <- GetTissueCoordinates(xenium.obj$fov, which = "centroids")
rownames(coords) <- coords$cell
coords$cell <- NULL
query <- SpatialRNA(coords, query.counts, colSums(query.counts))
Idents(allen.cortex.ref) <- "subclass"
# remove CR cells because there aren't enough of them for annotation
allen.cortex.ref <- subset(allen.cortex.ref, subset = subclass != "CR")
counts <- GetAssayData(allen.cortex.ref, assay = "RNA", slot = "counts")
cluster <- as.factor(allen.cortex.ref$subclass)
names(cluster) <- colnames(allen.cortex.ref)
nUMI <- allen.cortex.ref$nCount_RNA
names(nUMI) <- colnames(allen.cortex.ref)
nUMI <- colSums(counts)
levels(cluster) <- gsub("/", "-", levels(cluster))
reference <- Reference(counts, cluster, nUMI)
# run RCTD with many cores
RCTD <- create.RCTD(query, reference, max_cores = 8)
RCTD <- run.RCTD(RCTD, doublet_mode = "doublet")
annotations.df <- RCTD@results$results_df
annotations <- annotations.df$first_type
names(annotations) <- rownames(annotations.df)
xenium.obj$predicted.celltype <- annotations

# SPOTLight
Idents(allen.cortex.ref) <- 'subclass'
allen.cortex.ref <- NormalizeData(allen.cortex.ref, normalization.method = "LogNormalize", scale.factor = 10000)
allen.cortex.ref <- FindVariableFeatures(allen.cortex.ref, selection.method = "vst", nfeatures = 2000)
genes <- rownames(allen.cortex.ref)[!grepl(pattern = "^Rp[l|s]|Mt", x = rownames(allen.cortex.ref))]
allen.cortex.ref <- ScaleData(allen.cortex.ref, features = genes)
markers <- FindAllMarkers(allen.cortex.ref, logfc.threshold = 1, only.pos = TRUE, min.pct=0.3, return.thresh=0.05)
markers <- markers[markers$p_val_adj<0.05&abs(markers$avg_log2FC)>1, ]
dim(markers)
table(markers$cluster)
hvg <- VariableFeatures(allen.cortex.ref)

DefaultAssay(xenium.obj) <- 'Xenium'
xenium.obj <- NormalizeData(xenium.obj, normalization.method = "LogNormalize", scale.factor = 10000)

res <- SPOTlight(
    x = GetAssayData(allen.cortex.ref, assay = 'RNA', layer = 'data'),
    y = GetAssayData(xenium.obj, assay = 'Xenium', layer = 'data'),
    groups = as.character(allen.cortex.ref$subclass),
    mgs = markers,
    hvg = hvg,
    weight_id = "avg_log2FC",
    group_id = "cluster",
    gene_id = "gene")
head(res)
mat <- res$mat
colnames(mat) <- paste0('SPOTLight_', colnames(mat))
identical(rownames(res$mat), colnames(xenium.obj))
xenium.obj <- AddMetaData(xenium.obj, mat)

head(xenium.obj)

saveRDS(xenium.obj, '~/Downloads/Xenium_V1/tmp.rds')
write.csv(markers, '~/Downloads/Xenium_V1/markers.csv')
markers <- markers[markers$p_val_adj<0.005&abs(markers$avg_log2FC)>2, ]
markers <- split(markers$gene, markers$cluster)
markers <- lapply(markers, head, n=100)

qc <- function(mat) {
    # Indices of mitochondrial genes
    mito_idx <- grep("^MT-", rownames(mat))
    
    # Calculate QC metrics
    res <- mat |>
        scuttle::perCellQCMetrics(subsets = list("Mito" = mito_idx)) |>
        scuttle::perCellQCFilters(sub.fields = "subsets_Mito_percent")
    
    # Remove low quality cells
    mat[, !res$discard]
}
imputed_gex <- GetAssayData(xenium.obj, assay = "Xenium", layer = "counts") |>
    qc() |>
    scuttle::normalizeCounts() |>
    Seqtometry::impute()
future::plan("sequential")
# Allot 1 GiB of memory for global variables
options(future.globals.maxSize = 1024 ^ 3)
scores <- markers |>
    Seqtometry::score(imputed_gex, signatures = _)
colnames(scores) <- paste0('SeqtometryScore_', colnames(scores))
colnames(scores) <- make.names(colnames(scores), allow_ = TRUE)
xenium.obj <- AddMetaData(xenium.obj, metadata = scores[, -1])
markers <- lapply(markers, head, n=3)
markers <- unlist(markers)
markers <- unname(markers)
markers <- unique(markers)
DefaultAssay(xenium.obj) <- 'SCT'
appconf <- createAppConfig(
            title="xenium_small",
            destinationFolder = "xenium_small",
            species = "Homo sapiens",
            doi="10.1038/nbt.3192",
            datatype = "spatial",
            markers =markers[markers %in% rownames(xenium.obj)])
unlink('~/Downloads/Xenium_V1/xenium_small', recursive = TRUE)
createDataSet(appconf, seu = xenium.obj,
              datafolder = path, boundaries = 'segmentations')

#' @importFrom DT DTOutput
editUI <- function (id) {
    ns <- NS(id)
    tagList(column(
        width = 10,
        offset = 1,
        fluidRow(
            column(
                width = 6,
                selectInput(ns("dir"),
                            label = "Available data",
                            choices = c()),
                actionButton(
                    ns("refresh"),
                    label = "Refresh available data",
                    icon = icon("refresh"),
                    inline = TRUE
                ),
                textInput(ns('search'),
                             label="search",
                             value='',
                             placeholder='keywords to search')
            ),
            column(
                width = 6,
                actionButton(
                    ns('edit'),
                    label = "Apply change",
                    icon = icon("edit")),
                actionButton(
                    ns('reset'),
                    label = "Rollback",
                    icon = icon("refresh")
                ),
                actionButton(
                    ns('delete'),
                    label = "Delete",
                    icon = icon("remove")
                ),
                textOutput(ns('message'))
            )
        ),
        fluidRow(
            column(
                width = 6,
                textAreaInput(
                    ns("title"),
                    label = "Title for the data"),
                checkboxInput(
                    ns("token"),
                    label = "Use token to access the data",
                    value = FALSE
                ),
                checkboxInput(
                    ns("locker"),
                    label = "Require privilege for the data",
                    value = TRUE),
                selectInput(
                    ns("datatype"),
                    label = "Data type",
                    choices = c("scRNAseq", "scATACseq", "scMultiome"),
                    selected = "scRNAseq"
                ),
                selectInput(
                    ns("meta1"),
                    label = "Primary default meta to show",
                    choices = c()),
                selectInput(
                    ns("meta2"),
                    label = "Secondary default meta to show",
                    choices = c()),
                textInput(
                    ns("gene1"),
                    label = "Primary default gene to show"),
                textInput(
                    ns("gene2"),
                    label = "Secondary default gene to show"),
                selectInput(
                    ns("dimred1"),
                    label = "Primary default reduction dim to show",
                    choices = c()),
                selectInput(
                    ns("dimred2"),
                    label = "Secondary default reduction dim to show",
                    choices = c()),
                selectInput(
                    ns("grp1"),
                    label = "Primary default info to show",
                    choices = c()),
                selectInput(
                    ns("grp2"),
                    label = "Secondary default info to show",
                    choices = c())
            ),
            column(
                width = 6,
                selectInput(
                    ns("species"),
                    label = "Species",
                    choices = c(.globals$supported_organisms, 'others'),
                    selected = NULL
                ),
                textInput(
                    ns("species2"),
                    label = "Other species"),
                checkboxGroupInput(
                    ns("meta_to_include"),
                    label = "Columns to include from the metadata"),
                textAreaInput(
                    ns("multigene"),
                    label = "Default genes to show in bubbleplot/heatmap"),
                textAreaInput(
                    ns("keywords"),
                    label = "The key words for this work"),
                textAreaInput(
                    ns("abstract"),
                    label = "The abstract for this work"),
                textAreaInput(
                    ns("reference"),
                    label = "BibTeX string"),
                textInput(
                    ns("doi"),
                    label = "DOI"),
                textInput(
                    ns("pmid"),
                    label = "PubMed ID")
            )
        ),
        hr(),
        fluidRow(
            column(
                width = 3,
                h4("Assign cell cycle"),
                actionButton(
                    ns("tricycle"),
                    label = "Tricycle")
            ),
            column(
                width = 3,
                h4("Assign cell type"),
                selectInput(
                    ns("celldex"),
                    label = NULL,
                    choices = c(
                        "HumanPrimaryCellAtlasData",
                        "BlueprintEncodeData",
                        "MouseRNAseqData",
                        "ImmGenData",
                        "DatabaseImmuneCellExpressionData",
                        "NovershternHematopoieticData",
                        "MonacoImmuneData"
                    ),
                    selected = "HumanPrimaryCellAtlasData"
                ),
                actionButton(
                    ns("singler"),
                    label = "singleR")
            ),
            column(
                width = 3,
                h4("Trajectories"),
                actionButton(
                    ns("slingshot"),
                    label = "slingshot")
            ),
            column(
                width = 3,
                h4("Cell communications"),
                p('N/A')
            )
        ),
        hr(),
        fluidRow(
            h4("Danger Zone"),
            DTOutput(ns('conf')))
    ))
}
#' @importFrom DT renderDT JS
#' @importFrom S4Vectors SimpleList DataFrame
editServer <- function(id) {
    moduleServer(id, function(input, output, session) {
        global <- reactiveValues(
            sc1conf_data = NULL,
            sc1conf_orig = NULL,
            sc1meta = NULL,
            sc1def = NULL,
            appconf = NULL,
            markers = NULL,
            metaAdditional = NULL,
            ref = NULL,
            locker = TRUE,
            token = NULL,
            tokenList = NULL
        )
        updateSelectInput(
            session,
            "dir",
            choices = getDataSets(privilege = "all"),
            selected = getDataSets(privilege = "all")[1])
        observeEvent(input$refresh, {
            ds <- getDataSets(privilege = "all")
            updateSelectInput(
                session,
                "dir",
                choices = ds,
                selected = ds[1])
        })
        observeEvent(input$search, {
            if(nchar(input$search)>3){
                ds <- getDataSets(privilege = "all")
                key <- strsplit(input$search, '[^0-9a-zA-Z]')[[1]]
                ad <- vapply(ds, FUN=function(.ele){
                    .ele <- strsplit(.ele, '[^0-9a-zA-Z]')[[1]]
                    sum(apply(adist(key, .ele), 2, min))
                }, FUN.VALUE = numeric(1L))
                updateSelectInput(
                    session,
                    "dir",
                    choices = ds,
                    selected = ds[which.min(ad)[1]]
                )
            }
        })
        observeEvent(input$multigene, {
            if (!is.list(global$markers)) {
                global$markers <- strsplit(input$multigene, "\\s+|,|;")[[1]]
            }
        })
        observeEvent(input$dir, {
            if (!is.null(input$dir)) {
                if (input$dir %in% getDataSets(privilege = "all")) {
                    adminProcess({
                        global$sc1meta <- readData("sc1meta", input$dir)
                        global$sc1conf_orig <-
                            readData("sc1conf", input$dir)
                        global$sc1conf_orig$fID <-
                            formatfID_CL(global$sc1conf_orig$fID)
                        global$sc1conf_orig$fCL <-
                            formatfID_CL(global$sc1conf_orig$fCL)
                        global$sc1conf_data <- global$sc1conf_orig
                        cellInfo <- global$sc1conf_data$ID
                        dimred <-
                            gsub("_", "", cellInfo[global$sc1conf_data$dimred])
                        grp <- cellInfo[global$sc1conf_data$grp]
                        output$conf <- renderDT({
                            datatable(
                                global$sc1conf_data,
                                editable = TRUE,
                                options = list(columnDefs = list(
                                    list(
                                        targets = c(6, 7),
                                        render = JS(
                                            "function(data, type, row, meta){",
                                            " if(type === 'display'){",
                                            paste(
                                                "  return data",
                                                "? '<input type=\"checkbox\"",
                                                "checked/>'",
                                                ": '<input",
                                                "type=\"checkbox\"/>';"
                                            ),
                                            " }",
                                            " return data;",
                                            "}"
                                        )
                                    )
                                ))
                            )
                        })
                        updateCheckboxGroupInput(
                            session,
                            "meta_to_include",
                            label = "Columns to include from the metadata",
                            choices = cellInfo,
                            selected = cellInfo
                        )
                        global$sc1def <- readData("sc1def", input$dir)
                        updateTextInput(
                            session,
                            "gene1",
                            value = global$sc1def$gene1)
                        updateTextInput(
                            session,
                            "gene2",
                            value = global$sc1def$gene2)
                        updateTextAreaInput(
                            session,
                            "multigene",
                            value = paste(
                                global$sc1def$genes,
                                collapse = ", ")
                        )
                        updateSelectInput(
                            session,
                            "meta1",
                            choices = cellInfo,
                            selected = global$sc1def$meta1
                        )
                        updateSelectInput(
                            session,
                            "meta2",
                            choices = cellInfo,
                            selected = global$sc1def$meta2
                        )
                        updateSelectInput(
                            session,
                            "dimred1",
                            choices = dimred,
                            selected = global$sc1def$dimred[1]
                        )
                        updateSelectInput(
                            session,
                            "dimred2",
                            choices = dimred,
                            selected = global$sc1def$dimred[2]
                        )
                        updateSelectInput(
                            session,
                            "grp1",
                            choices = grp,
                            selected = global$sc1def$grp1
                        )
                        updateSelectInput(
                            session,
                            "grp2",
                            choices = grp,
                            selected = global$sc1def$grp2
                        )
                        global$appconf <- readData("appconf", input$dir)
                        global$ref <- global$appconf$ref$entry
                        
                        if (length(global$appconf$markers)) {
                            global$markers <- global$appconf$markers
                        }
                        updateTextAreaInput(
                            session,
                            "keywords",
                            value = global$appconf$keywords)
                        updateTextAreaInput(
                            session,
                            "title",
                            value = global$appconf$title)
                        updateTextInput(
                            session, "dir",
                            value = global$appconf$id)
                        updateSelectInput(
                            session,
                            "datatype",
                            selected = global$appconf$type)
                        updateSelectInput(
                            session,
                            "species",
                            selected = global$appconf$species)
                        if (!global$appconf$species %in%
                            .globals$supported_organisms) {
                            updateTextInput(
                                session,
                                "species2",
                                value = global$appconf$species)
                        } else {
                            updateTextInput(
                                session,
                                "species2",
                                value = ""
                            )
                        }
                        updateTextInput(
                            session,
                            "reference",
                            value = global$appconf$ref$bib)
                        updateTextInput(
                            session,
                            "abstract",
                            value = ifelse(
                                is.null(global$appconf$ref$entry$abstract),
                                '',global$appconf$ref$entry$abstract)
                        )
                        updateTextInput(
                            session,
                            "doi",
                            value = global$appconf$ref$doi)
                        updateTextInput(
                            session,
                            "pmid",
                            value = global$appconf$ref$pmid)
                        global$locker <- checkLocker(input$dir)
                        updateCheckboxInput(
                            session, "locker",
                            value = global$locker)
                        global$tokenList <- getTokenList()
                        if(any(global$tokenList==input$dir)){
                            global$token <- names(global$tokenList)[
                                which(global$tokenList==input$dir)]
                        }
                        if(length(global$token)){
                            updateCheckboxInput(
                                session, "token",
                                value = TRUE)
                        }
                    }, "Loading data", "Done loading!", duration = 1)
                }
            }
        })
        observeEvent(input$delete, {
            showModal(modalDialog(
                tagList(p(
                    "Are you sure you want to delete the data: ",
                    input$dir)
                ),
                title = paste("Delete data", input$dir),
                footer = tagList(
                    actionButton(
                        NS(id, "confirmDelete"),
                        "Delete"),
                    modalButton("Cancel")
                )
            ))
        })
        observeEvent(input$confirmDelete, {
            req(input$dir)
            unlink(
                file.path(.globals$datafolder, input$dir),
                recursive = TRUE)
            updateSelectInput(
                session,
                "dir",
                choices = getDataSets(privilege = "all"),
                selected = getDataSets(privilege = "all")[1])
            removeModal()
        })
        observeEvent(input$doi, {
            updateRefById(
                "reference",
                "doi",
                GetBibEntryWithDOI,
                input,
                output,
                session)
        })
        observeEvent(input$pmid, {
            global$ref <- updateRefById(
                "reference",
                "pmid",
                GetPubMedByID,
                input,
                output,
                session)
        })
        
        observeEvent(input$conf_cell_edit, {
            row  <- input$conf_cell_edit$row
            clmn <- input$conf_cell_edit$col
            global$sc1conf_data[row, clmn] <- input$conf_cell_edit$value
            saveData(global$sc1conf_data, input$dir, "sc1conf")
        })
        
        observeEvent(input$reset, {
            adminProcess({
                global$sc1conf_data <- global$sc1conf_orig
                conf_orig <- global$sc1conf_orig
                conf_orig$fID <- formatfID_CL(conf_orig$fID, rev = TRUE)
                conf_orig$fCL <- formatfID_CL(conf_orig$fCL, rev = TRUE)
                saveData(global$sc1meta, input$dir, "sc1meta")
                saveData(conf_orig, input$dir, "sc1conf")
                saveData(global$appconf, input$dir, "appconf")
                saveData(global$sc1def, input$dir, "sc1def")
                if (global$locker) {
                    setLocker(input$dir)
                } else{
                    removeLocker(input$dir)
                }
                if (length(global$token)) {
                    if(nchar(global$token)>=.globals$tokenMinLen){
                        setToken(input$dir, global$token)
                    }else{
                        removeToken(input$dir)
                        global$token <- NULL
                        global$tokenList <- getTokenList()
                    }
                } else {
                    removeToken(input$dir)
                    global$token <- NULL
                    global$tokenList <- getTokenList()
                }
            }, "Starting rollback", "Rollback done!")
        })
        
        getUpdatedMetaAndConf <- function() {
            meta <- global$sc1meta
            conf_new <- global$sc1conf_data
            conf_orig <- global$sc1conf_orig
            if (!identical(conf_new$ID, conf_orig$ID)) {
                colnames(meta)[match(conf_orig$ID, colnames(meta))] <-
                    conf_new$ID
            }
            conf_orig$fID <- formatfID_CL(conf_orig$fID, rev = TRUE)
            conf_orig$fCL <- formatfID_CL(conf_orig$fCL, rev = TRUE)
            conf_new$fID <- formatfID_CL(conf_new$fID, rev = TRUE)
            conf_new$fCL <- formatfID_CL(conf_new$fCL, rev = TRUE)
            if (!identical(conf_new$fID, conf_orig$fID)) {
                o_fID <- strsplit(conf_orig$fID, "\\|")
                n_fID <- strsplit(conf_new$fID, "\\|")
                if (!identical(lengths(o_fID), lengths(n_fID))) {
                    adminMsg("fID lengths are not identical!", "error")
                    return(NULL)
                }
                mapply(
                    o_fID,
                    n_fID,
                    conf_new$ID,
                    FUN = function(.old, .new, .id) {
                        if (!all(.old == .new, na.rm = TRUE)) {
                            for (i in which(.old != .new)) {
                                meta[meta[, .id] == .old[i], .id] <-
                                    .new[i]## need to check
                            }
                        }
                    }
                )
            }
            if (!identical(conf_new$fCL, conf_orig$fCL)) {
                o_fCL <- strsplit(conf_orig$fCL, "\\|")
                n_fCL <- strsplit(conf_new$fCL, "\\|")
                n_fID <- strsplit(conf_new$fID, "\\|")
                if (!identical(lengths(o_fCL), lengths(n_fCL))) {
                    adminMsg("fCL lengths are not identical!", "error")
                    return(NULL)
                }
                if (!identical(lengths(n_fID), lengths(n_fCL))) {
                    adminMsg(
                        "fCL lengths are not identical with fID!",
                        "error")
                    return(NULL)
                }
            }
            if (!is.null(global$metaAdditional)) {
                meta <- cbind(meta, do.call(cbind, global$metaAdditional))
            }
            return(list(meta = meta, config = conf_new))
        }
        getExprMetaConf <- function() {
            new_meta_conf <- getUpdatedMetaAndConf()
            meta <- new_meta_conf$meta
            rownames(meta) <- meta$sampleID
            meta$sampleID <- NULL
            gene <- readData("sc1gene", input$dir)
            expr <- readDataMatrix(input$dir, names(gene), rownames(meta))
            return(list(
                expr = expr,
                meta = meta,
                config = new_meta_conf$config
            ))
        }
        toSingleCellExperiment <- function(expr_meta_conf) {
            askNamespace("SingleCellExperiment", "S4Vectors")
            SingleCellExperiment::SingleCellExperiment(
                assays = S4Vectors::SimpleList(
                    counts = expr_meta_conf$expr,
                    logcounts = log2(expr_meta_conf$expr + 1)
                ),
                colData = S4Vectors::DataFrame(expr_meta_conf$meta),
                rowData = S4Vectors::DataFrame(
                    gene = rownames(expr_meta_conf$expr),
                    row.names = rownames(expr_meta_conf$expr)
                )
            )
        }
        findReductionMethod <- function() {
            global$sc1conf_data$ID[
                global$sc1conf_data$UI %in% global$sc1def$dimred]
        }
        getReductionData <- function() {
            data <- global$sc1meta[, mget(findReductionMethod())]
            rownames(data) <- global$sc1meta$sampleID
            data
        }
        findGrpIDs <- function() {
            x <- isolate(global$sc1conf_data$ID[global$sc1conf_data$grp])
            names(x) <- x
            x
        }
        addAditionalMeta <- function(additionalMeta) {
            global$metaAdditional <- c(
                global$metaAdditional,
                additionalMeta)
            cellInfo <- c(
                global$sc1conf_data$ID,
                names(global$metaAdditional))
            updateCheckboxGroupInput(
                session,
                "meta_to_include",
                choices = cellInfo,
                selected = cellInfo)
        }
        observeEvent(input$edit, {
            adminProcess({
                if (!identical(global$sc1conf_data, global$sc1conf_orig)) {
                    ## update meta if conf changed
                    new_meta_conf <- getUpdatedMetaAndConf()
                    saveData(new_meta_conf$meta, input$dir, "sc1meta")
                    saveData(
                        new_meta_conf$config,
                        input$dir,
                        "sc1conf")
                }
                sc1conf <- readData("sc1conf", input$dir)
                sc1conf <- sc1conf[sc1conf$ID %in% input$meta_to_include,]
                saveData(sc1conf, input$dir, "sc1conf")
                if(nchar(input$abstract)>1){
                    global$ref$abstract <- input$abstract
                }else{
                    global$ref$abstract <- NULL
                }
                updateAppConf(input, reactive({
                    global
                }))
                updateDef(input)
                if (input$locker) {
                    setLocker(input$dir)
                } else{
                    removeLocker(input$dir)
                }
                if (input$token) {
                    setToken(input$dir, global$token)
                } else {
                    removeToken(input$dir)
                    global$token <- NULL
                    global$tokenList <- getTokenList()
                }
            }, "Starting Update.", "Update done!")
        })
        observeEvent(input$tricycle, {
            if (!input$species %in% c("Homo sapiens", "Mus musculus")) {
                adminMsg("Only support human and mouse data.", type = "error")
                return()
            }
            askNamespace("tricycle", "SingleCellExperiment", "S4Vectors")
            adminProcess({
                expr_meta_conf <- getExprMetaConf()
                gname.type <-
                    ifelse(
                        grepl("ENS", rownames(expr_meta_conf$expr)[1]),
                        "ENSEMBL",
                        "SYMBOL")
                species <- ifelse(
                    input$species == "Mus musculus",
                    "mouse",
                    "human")
                exp <- toSingleCellExperiment(expr_meta_conf)
                tricyle <- addTricycle(
                    exp = exp,
                    gname.type = gname.type,
                    species = species,
                    meta = expr_meta_conf$meta
                )
                addAditionalMeta(tricyle)
            }, "Starting Tricycle!", "Tricycle done!")
        })
        observeEvent(input$singler, {
            askNamespace(
                "SingleR",
                "celldex",
                "SingleCellExperiment",
                "S4Vectors")
            if (!input$species %in% c(
                "Homo sapiens",
                "Mus musculus",
                "Danio rerio")) {
                adminMsg(
                    "Only support human, mouse and fish data.",
                    type = "error")
                return()
            }
            adminProcess({
                try(attachNamespace("celldex"))
                ref <-
                    get(input$celldex,
                        envir = as.environment("package:celldex"))()
                expr_meta_conf <- getExprMetaConf()
                sce <- toSingleCellExperiment(expr_meta_conf)
                pred <- SingleR::SingleR(
                    test = sce,
                    ref = ref,
                    labels = ref$label.fine
                )
                ## assign singler to meta data
                addAditionalMeta(list(SingleR_labels = pred$labels))
            }, "Starting SingleR!", "SingleR done!")
        })
        observeEvent(input$slingshot, {
            adminProcess({
                askNamespace("slingshot")
                misc_slingshot <- NULL
                dimred <- getReductionData()
                # by group points
                grp_ids <- lapply(findGrpIDs(), function(id) {
                    data <- global$sc1meta[, get(id)]
                    names(data) <- global$sc1meta$sampleID
                    data
                })
                lineages <- addSlingshot(dimred, grp_ids)
                message("Assign slingshot to miscellaneous data")
                ## assign slingshot to meta data
                writeMisc(lineages, input$dir, "slingshot")
            }, "Starting SlingShot", "SlingShot done!")
        })
    })
}

#' Create a dataset
#' Create a dataset from a Seurat object. The function will try to find
#' the markers in the Misc data named as 'markers'.
#' The misc data should be output of function `FindAllMarkers`.
#' @param datafolder app data folder
#' @param appconf a list object represent the information about the dataset
#' @param seu a Seurat object
#' @param config config file for makeShinyFiles
#' @param contrast The contrast group
#' @param LOCKER Set locker if the file is required login
#' @param assayName assay in single-cell data object to use for plotting
#'   gene expression, which must match one of the following:
#'   \itemize{
#'     \item{Seurat objects}: "RNA" or "integrated" assay,
#'       default is "RNA"
#'   }
#' @param gexSlot layer in single-cell assay to plot.
#' Default is to use the "data" layer
#' @param atacAssayName assay in single-cell data object to use for plotting
#' open chromatin.
#' @param atacSlot layer in single-cell atac assay to plot.
#' Default is to use the "data" layer
#' @param default.symbol character(1L) specifying the default rownames to be used. If use default, the gene symbols will be the row names of the assay. If one column name of the meta.feature of the assay is supplied, the function will try to extract the symbols from the meta.feature slot of the assay. 
#' @param theme color theme. default is "Paired" from ColorBrewer palettes.
#' @param binSize number of bps for each bin for ATAC fragment coverage. Used
#' to reduce the file size of bigwig.
#' @param fragmentNameMapList list of named character vector. 
#' The name map list must be the same order as the fragment list in the object.
#' For each element of the list,
#' the names of the vector are the name of the fragment and
#' the vector contains the cell names (column names of the assay). 
#' You can try \link{extractFragmentNameMapList}.
#' @importFrom SeuratObject Reductions Idents Assays DefaultAssay GetAssayData
#'  `DefaultAssay<-` VariableFeatures Misc `Misc<-` Embeddings `Idents<-`
#' @importFrom Seurat FindAllMarkers FindVariableFeatures ScaleData
#' @return The updated Seurat object.
#' @export
#' @examples
#' library(Seurat)
#' if(interactive()){
#'     appconf <- createAppConfig(
#'         title="pbmc_small",
#'         destinationFolder = "pbmc_small",
#'         species = "Homo sapiens",
#'         doi="10.1038/nbt.3192",
#'         datatype = "scRNAseq")
#'     createDataSet(appconf, pbmc_small, datafolder=tempdir())
#' }
createDataSet <- function(
        appconf,
        seu,
        config,
        contrast,
        assayName,
        gexSlot = c("data", "scale.data", "counts"),
        atacAssayName,
        atacSlot = c("data", "scale.data", "counts"),
        LOCKER = FALSE,
        datafolder = "data",
        default.symbol = 'rownames',
        theme = "Paired",
        binSize = 1,
        fragmentNameMapList) {
    stopifnot(file.exists(datafolder))
    stopifnot(is(seu, "Seurat"))
    stopifnot(is(appconf, "APPconf"))
    gexSlot <- match.arg(gexSlot)
    if (missing(config)) {
        config <- createConfig(seu, theme=theme)
    }
    pf <- file.path(datafolder, appconf$id)
    if(!file.exists(pf)){
        dir.create(pf, recursive = TRUE)
    }else{
        stop(pf, ' already exists')
    }
    ## markers
    markers <- appconf$markers
    if(length(markers)==0){
        if (!is.null(Misc(seu, "markers"))) {
            ## the markers is available at Misc(seu, "markers") slot
            markers <- Misc(seu, "markers")
        }
    }
    if(length(markers)==0){
        cellInfo <- colnames(seu[[]])
        assays <- Assays(seu)
        stopifnot(
            "Please input a seurat object with 'SCT' or 'RNA' assay" =
                any(c("SCT", "RNA") %in% assays))
        if (!DefaultAssay(seu) %in% c("SCT", "RNA")) {
            DefaultAssay(seu) <- match.arg(
                assays,
                choices = c("SCT", "RNA"),
                several.ok = TRUE)[1]
        }
        if (missing(assayName)) {
            assayName <- DefaultAssay(seu)
        }
        assayName <- assayName[1]
        stopifnot(
            "The assayName is not in input object" = assayName %in% assays)
        if (length(extAssayData(
            seu,
            assay = assayName,
            slot = "scale.data")) == 0) {
            seu <- FindVariableFeatures(
                seu,
                selection.method = "vst",
                nfeatures = 1000)
            seu <- ScaleData(seu)
        }
        top10 <- head(VariableFeatures(seu), 10)
        if (!missing(contrast)) {
            if (contrast[1] %in% cellInfo) {
                appconf$groupCol <- contrast[1]
            } else{
                stop("The input contrast is not in seu object")
            }
        } else{
            if (!is.null(appconf$groupCol)) {
                if (appconf$groupCol[1] %in% cellInfo) {
                    contrast <- appconf$groupCol[1]
                } else{
                    appconf$groupCol <- appconf$groupCol[-1]
                }
            } else{
                contrast <- NULL
            }
        }
        if (length(markers) == 0) {
            if (!is.factor(Idents(seu))) {
                if (!is.null(contrast)) {
                    Idents(seu) <- contrast
                } else{
                    grp <- cellInfo[grepl(
                        'cluster|cell(.*)type',
                        cellInfo,
                        ignore.case = TRUE)]
                    grp_d <- adist('celltype', grp)
                    Idents(seu) <- grp[which.min(grp_d)][1]
                }
            }
            markers <- FindAllMarkers(
                seu,
                only.pos = TRUE,
                min.pct = .25,
                logfc.threshold = .25
            )
            if (length(markers)) {
                Misc(seu, "markers") <- markers
            }
            if(length(markers$cluster)>0){
                appconf$markers <- split(markers, markers$cluster)
            }
        }
        if (length(markers) == 0) {
            markers <- top10
        } else{
            if (is.list(markers) && !is.data.frame(markers)) {
                markers <- as.data.frame(markers[[1]])
            }
            markers <- split(markers, markers$cluster)
            markers <-
                lapply(markers, head, n = min(5, ceiling(50 / length(markers))))
        }
    }
    if(!is.character(markers)){
        if(is.list(markers)){
            if(is.data.frame(markers)){
                markers <- list(markers)
            }
            ## list of data.frame
            markers <- lapply(markers, function(.ele){
                if(is.data.frame(.ele)){
                    cn <- grepl(
                        "^(gene|symbol)",
                        colnames(.ele),
                        ignore.case = TRUE)
                    if(any(cn)){
                        cn <- which(cn)[1]
                        return(as.character(.ele[, cn, drop=TRUE]))
                    }else{
                        return(rownames(.ele))
                    }
                }else{
                    if(is.character(.ele)){
                        return(.ele)
                    }else{
                        return(NULL)
                    }
                }
            })
            markers <- unique(unlist(markers, use.names = FALSE))
        }
    }
    if(length(markers)==0 || !is.character(markers)){
        stop("Can not locate the markers for the inputs.")
    }
    
    ## make shiny app
    makeShinyFiles(
        seu,
        scConf = config,
        assayName = assayName,
        gexSlot = gexSlot,
        atacAssayName = atacAssayName,
        atacSlot = atacSlot,
        appDir = pf,
        defaultGene1 = markers[1],
        defaultGene2 = markers[2],
        default.multigene = markers,
        default.symbol = default.symbol,
        binSize = binSize,
        fragmentNameMapList = fragmentNameMapList
    )
    
    .globals$datafolder <- datafolder
    saveAppConf(appconf)
    ## save misc data
    for (slot in names(Misc(seu))) {
        writeMisc(Misc(seu, slot), appconf$id, slot)
    }
    ## "Locker"
    if (LOCKER) {
        writeLines("", file.path(pf, "LOCKER"))
    }
    return(seu)
}


#' Create a metadata to describe the dataset
#' @description The function will return a APPconf object which contain
#' the reference, keywords for the dataset.
#' @param title The title of the dataset
#' @param destinationFolder The destination folder name of the dataset without
#'  the root folder of the datasets. The data will be saved as
#'  `appdataFolder/destinationFolder`
#' @param species The species of the dataset
#' @param doi,pmid The DOI or PMID of the reference
#' @param bibentry An object of bibentry
#' @param datatype character(1). Type of the data, scRNAseq, scATACseq or
#'  scMultiome.
#' @param markers A list of data.frame with gene symbols as rownames or
#'  a character vector.
#' @param keywords The keywords for the dataset.
#' For example the condition, cell type, tissue information
#' The keywords will be used for whole database search
#' @param abstract The abstract of the reference.
#' @return An object of \link{APPconf} object
#' @importFrom RefManageR GetBibEntryWithDOI GetPubMedByID
#' @export
#' @examples
#' if(interactive()){
#'     config <- createAppConfig(
#'         title="pbmc_small",
#'         destinationFolder = "pbmc_small",
#'         species = "Homo sapiens",
#'         doi="10.1038/nbt.3192",
#'         datatype = "scRNAseq")
#' }

createAppConfig <-
    function(
        title,
        destinationFolder,
        species,
        doi,
        pmid,
        bibentry,
        datatype = c("scRNAseq", "scATACseq", "scMultiome"),
        markers,
        keywords,
        abstract) {
        ## markers is a list of dataframe, rownames is the gene symbols
        if (!missing(markers)) {
            if (is.character(markers)) {
                markers <- markers[!is.na(markers)]
                markers <- markers[markers != ""]
                markers <- unique(markers)
                markers <- t(t(markers))
                rownames(markers) <- markers
                markers <- list(markers = as.data.frame(markers))
            }
        } else{
            markers <- list()
        }
        stopifnot(is.list(markers))
        lapply(markers, function(.ele) {
            stopifnot(
                "markers must be a list of data.frame" =
                    is.data.frame(.ele))
            stopifnot(
                "markers must be a list of data.frame
                with gene symbols as rownames" =
                    length(rownames(.ele)) == nrow(.ele)
            )
        })
        datatype <- match.arg(datatype)
        stopifnot(is.character(title))
        stopifnot(is.character(destinationFolder))
        stopifnot(is.character(species))
        if (!missing(keywords)) {
            stopifnot(is.character(keywords))
        } else
            keywords <- character(0L)
        pmidNotFromDOI <- TRUE
        if (!missing(doi)) {
            stopifnot(is.character(doi))
            if(missing(bibentry) || !is(bibentry, "bibentry")){
                bibentry <- GetBibEntryWithDOI(doi)
            }
            if (missing(pmid)){
                pmid <- idConverter(doi, type = "pmid")
                pmidNotFromDOI <- FALSE
            }
        }
        if (pmidNotFromDOI & !missing(pmid)) {
            stopifnot(is.character(pmid))
            if(missing(bibentry) || !is(bibentry, "bibentry")){
                bibentry <- GetPubMedByID(pmid)
            }
            if (missing(doi))
                doi <- idConverter(pmid, type = "doi")
        }
        if(missing(doi)){
            doi = ""
        }
        if(missing(pmid)){
            pmid = ""
        }
        bib <- NULL
        if (!missing(bibentry)) {
            if (is(bibentry, "bibentry")) {
                bib <- format(bibentry, style = 'html')
                if(!missing(abstract)){
                    bibentry$abstract <- abstract
                }
            }
        }else{
            bibentry <- NULL
        }
        return(
            APPconf(
                title = title[1],
                id = destinationFolder[1],
                species = species[1],
                ref = list(
                    bib = bib,
                    doi = doi,
                    pmid = pmid,
                    entry = bibentry
                ),
                type = datatype,
                markers = markers,
                keywords = keywords
            )
        )
    }

#' load data from cellRanger
#' @param outsFolder the outs folder of cellRanger
#' @importFrom Seurat CreateSeuratObject Read10X
#' @importFrom SeuratObject CreateDimReducObject
#' @importFrom utils read.csv
#' @importFrom data.table fread
#' @return An SeuratObject
createSeuFromCellRanger <- function(outsFolder) {
    analysisFolder <- file.path(outsFolder, "analysis")
    matrixFolder <-
        file.path(outsFolder, "filtered_feature_bc_matrix")
    stopifnot(
        "'analysis' folder must exits" =
            file.exists(analysisFolder))
    stopifnot(
        "'filtered_feature_bc_matrix' folder must exits" =
            file.exists(matrixFolder))
    seu <- CreateSeuratObject(Read10X(matrixFolder))
    projections <- dir(
        analysisFolder,
        "projection.csv",
        recursive = TRUE,
        full.names = TRUE)
    # projections name will be
    # analysisFolder/pca/gene_expression_x_components/projection.csv
    projs <- basename(dirname(dirname(
        sub(analysisFolder, "", projections, fixed = TRUE)
    )))
    for (i in seq_along(projections)) {
        projection <- read.csv(projections[i], row.names = 1)
        colnames(projection) <- sub("\\.", "_", colnames(projection))
        seu[[projs[i]]] <-
            CreateDimReducObject(
                embeddings = as.matrix(projection),
                assay = "RNA")
    }
    clusters <- dir(
        analysisFolder,
        "clusters.csv",
        recursive = TRUE,
        full.names = TRUE)
    clus <-
        sub("^.*gene_expression_(.*?)\\/clusters.csv",
            "\\1",
            clusters)
    for (i in seq_along(clusters)) {
        cluster <- read.csv(clusters[i], row.names = 1)
        colnames(cluster) <- clus[i]
        seu[[clus[i]]] <- as.factor(cluster[, 1])
    }
    de_tbls <- dir(
        analysisFolder,
        "differential_expression.csv",
        recursive = TRUE,
        full.names = TRUE
    )
    names(de_tbls) <- sub(
        "^.*gene_expression_(.*?)\\/differential_expression.csv",
        "\\1",
        clusters)
    misc <- lapply(de_tbls, read.csv, row.names = 1)
    ## reformat
    misc <- lapply(misc, function(.ele) {
        gene <- .ele$Feature.Name
        .ele <- .ele[, -1, drop = FALSE]
        .ele <- lapply(seq.int(ncol(.ele) / 3), function(.e) {
            .e <- .ele[, (.e - 1) * 3 + seq.int(3)]
            cluster <- sub(
                "Cluster.(.*?).Mean.Counts",
                "\\1",
                colnames(.e)[1])
            colnames(.e) <- sub("Cluster.(.*?)\\.", "", colnames(.e))
            cbind(.e, cluster, gene)
        })
        .ele <- do.call(rbind, .ele)
        .ele <- .ele[.ele[, 3] < 0.05, , drop = FALSE]
    })
    Misc(seu, "markers") <- misc
    seu
}

#' load data from a count matrix
#' @param matrix count matrix
#' @param meta cell-level meta data
#' @param genes character. gene names, will be the rownames of the matrix
#' @param cluster the cluster coordinates
#' @param ... The parameter passed to read.delim when read cluster file.
#' @importFrom data.table fread
#' @importFrom Seurat CreateSeuratObject
#' @importFrom SeuratObject CreateDimReducObject
#' @return An SeuratObject
createSeuFromMatrix <- function(
        matrix, meta, genes, cluster, ...) {
    if (missing(matrix) | missing(meta)) {
        stop('matrix and meta is requred.')
    }
    mat <- fread(matrix)
    meta <- read.delim(meta, header = TRUE)
    if (missing(genes)) {
        mat <- mat[!duplicated(mat[[1]]),]
        genes <- mat[[1]]
        mat <- mat[, -1]
    }
    stopifnot(length(genes) == nrow(matrix))
    mat <- data.frame(mat, row.names = genes)
    rownames(meta) <- colnames(mat)
    if (identical(colnames(mat), make.names(meta[, 1], unique = TRUE))) {
        meta <- meta[, -1]
    }
    getCluster <- function(cluster) {
        clusterfile <- read.delim(cluster, ...)
        stopifnot(identical(
            colnames(mat),
            make.names(clusterfile$V1, unique = TRUE)
        ))
        clusterfile <- clusterfile[, -1]
        rownames(clusterfile) <- colnames(mat)
        if (all(grepl("^V", colnames(clusterfile)), na.rm = TRUE)) {
            colnames(clusterfile)[c(1, 2)] <- c("tSNE_1", "tSNE_2")
        }
        clusterfile
    }
    if (is.list(cluster)) {
        stopifnot(length(names(cluster)) == length(cluster))
        clusters <- lapply(cluster, getCluster)
    } else{
        clusters <- list('tsne' = getCluster(cluster))
        names(clusters) <- lapply(clusters, function(.ele) {
            tolower(gsub("_\\d+$", "", colnames(.ele)[1]))
        })[[1]]
    }
    
    # create seu
    seu <- CreateSeuratObject(mat, meta.data = meta)
    if (length(clusters)) {
        for (i in seq_along(clusters)) {
            cluster <-
                CreateDimReducObject(embeddings = as.matrix(clusters[[i]]))
            seu[[names(clusters)[i]]] <- cluster
        }
    }
    seu
}

#' Add slingshot lineages to the dataset
#' @importFrom slingshot getLineages
#' @param dimred reduced dim
#' @param clusterings A named list with cluster labels
#' @noRd
addSlingshot <- function(dimred, clusterings) {
    stopifnot(length(names(clusterings)) > 0)
    lapply(clusterings, function(clustering) {
        slingshot::getLineages(data = dimred, clusterLabels = clustering)
    })
}

#' Add tricycle results to the dataset
#' tricycle is not a R/Bioconductor available package
#' DO NOT export
#' @noRd
#' @param exp A numeric matrix of **log-expression** values where rows are
#'  features and columns are cells.
#'  Alternatively, a SummarizedExperiment or SingleCellExperiment containing
#'  such a matrix.
#' @param meta metadata for the expression data
#' @param gname.type The type of gene names as in gname or rownames of exp.
#'  It can be either 'ENSEMBL' or 'SYMBOL'.
#' @param species It can be either 'mouse' or 'human'.
addTricycle <- function(exp, gname.type, species, meta) {
    species <- match.arg(species, choices = c("human", "mouse"))
    exp <- tricycle::project_cycle_space(
        exp,
        gname.type = gname.type,
        species = species)
    exp <- tricycle::estimate_cycle_position(exp)
    exp <- tricycle::estimate_Schwabe_stage(
        exp,
        gname.type = gname.type,
        species = species)
    stopifnot(identical(
        rownames(meta),
        rownames(SummarizedExperiment::colData(exp))
    ))
    tricyclePosition <-
        SummarizedExperiment::colData(exp)$tricyclePosition
    CCStage <- SummarizedExperiment::colData(exp)$CCStage
    return(list(
        tricyclePosition = tricyclePosition,
        CCStage = CCStage))
}

test_that("createDataset works not correct", {
  appconf <- createAppConfig(
    title="pbmc_small",
    destinationFolder = "pbmc_small",
    species = "Homo sapiens",
    doi="10.1038/nbt.3192",
    datatype = "scRNAseq",
    keywords = c("pbmc", "testdata"))
  app_path <- file.path(tempdir(), 'test_createDataset')
  dir.create(app_path, recursive = TRUE)
  on.exit(unlink(app_path, recursive = TRUE))
  createDataSet(appconf, pbmc_small, datafolder=app_path)
  p <- system.file("extdata", "data", "pbmc_small",
                   package = "scRNAseqApp")
  for(f in dir(p)){
    expect_true(file.exists(file.path(app_path, 'pbmc_small', f)))
  }
})

# Plot cell information on dimred
#' @noRd
#' @importFrom ggplot2 ggplot aes .data geom_point xlab ylab
#' scale_color_gradientn guides guide_colorbar scale_color_manual
#' guide_legend theme element_text coord_fixed geom_segment
#' @importFrom ggrepel geom_text_repel
#' @importFrom stats weighted.mean
#' @importFrom slingshot SlingshotDataSet slingLineages slingClusterLabels
#' slingMST slingParams
#' @importFrom data.table .SD
#' @importFrom SingleCellExperiment reducedDim
scDRcell <- function(
        inpConf,
        inpMeta,
        dimRedX,
        dimRedY,
        cellinfoID,
        cellinfoName=cellinfoID,
        subsetCellKey,
        subsetCellVal,
        pointSize,
        gradientCol,
        GeneExprDotOrd,
        labelsFontsize = 24,
        labelsFontFamily = 'Helvetica',
        plotAspectRatio,
        keepXYlables,
        inplab,
        dataset,
        geneIdMap,
        valueFilterKey,
        valueFilterCutoff,
        valueFilterCutoff2,
        hideFilterCell=FALSE,
        inpSlingshot,
        slingshotFilename,
        inpShowEdge,
        edgeFilename,
        editorStatus,
        ...) {
    subFilterColname <- 'subValue'
    subGrpColname <- 'sub'
    valColname <- 'val'
    subsetCellKey <- subsetCellKey[subsetCellKey!="N/A"]
    subsetCellVal <- namedSubsetCellVals(subsetCellKey, subsetCellVal)
    if(cellinfoName!=cellinfoID){
        if(is.na(cellinfoName)||cellinfoName==""){
            cellinfoName <- cellinfoID
        }
    }
    # Prepare ggData
    ggData <- inpMeta[, unique(c(
        inpConf[inpConf$UI == dimRedX]$ID,
        inpConf[inpConf$UI == dimRedY]$ID,
        inpConf[inpConf$UI == cellinfoID]$ID,
        inpConf[inpConf$UI %in% subsetCellKey]$ID,
        inpConf[inpConf$UI == cellinfoName]$ID)),
        with = FALSE]
    if (ncol(ggData) < 3)
        return(ggplot())
    colnames(ggData)[c(1,2)] <- c("X", "Y")
    ggData <-
        cbindFilterValues(
            ggData,
            inpConf,
            inpMeta,
            subFilterColname,
            geneIdMap,
            dataset,
            valueFilterKey,
            valueFilterCutoff,
            valueFilterCutoff2
        )
    rat <- getRatio(ggData)
    keep <- filterCells(
        ggData,
        subsetCellKey,
        subsetCellVal,
        subFilterColname,
        valueFilterCutoff,
        valueFilterCutoff2,
        inpConf)
    
    if(length(subsetCellKey)==0){
        return(ggplot())
    }
    if(subsetCellKey[1]==cellinfoID){
        subGrpColname <- valColname
        colnames(ggData)[3] <- valColname
    }else{## make the first subsetCellKey as sub
        if(ncol(ggData)<4){
            return(ggplot())
        }
        colnames(ggData)[c(3,4)] <- c(valColname, subGrpColname)
    }
    bgCells <- sum(!keep) > 0
    
    if(inpShowEdge){
        if (file.exists(edgeFilename)) {
            edges <- readRDS(edgeFilename)
            # edgeFilename must be a RDS filename. The RDS file must be a list of 
            # edges. The names of the list is the reductions name, such as umap
            if(!is.list(edges)){
                inpShowEdge <- FALSE
            }
            dimRed_prefix <- sub('.$', '', dimRedX)
            if(dimRed_prefix!=sub('.$', '', dimRedY)){
                inpShowEdge <- FALSE
            }
            if(!dimRed_prefix %in% names(edges)){
                inpShowEdge <- FALSE
            }else{
                edges <- edges[[dimRed_prefix]]
            }
            if(inpShowEdge){
                ggData$idx <- seq.int(nrow(ggData))
            }
        }else{
            inpShowEdge <- FALSE
        }
    }
    
    if (bgCells) {
        ggData2 <- ggData[!keep]
        ggData <- ggData[keep]
    }
    
    ggData <- orderGeneExpr(ggData, GeneExprDotOrd, valColname)
    
    # Do factoring if required
    ggData <- relevelData(ggData, valColname)
    ggCol <- relevelCol(inpConf, cellinfoID, ggData, valColname)
    
    if(cellinfoName!=cellinfoID){
        if(cellinfoName==subsetCellKey[1]){
            cellinfoName <- subGrpColname
        }
        tab <- unique(ggData[, c(valColname, cellinfoName), with = FALSE])
        if(length(levels(ggData[[valColname]])) != nrow(tab)){
            # showNotification(paste(
            #     'The length of cell labels are not identical',
            #     'with that of the cell info IDs'),
            #     type = "warning")
            cellinfoName <- names(ggCol)
        }else{
            cellinfoName <- tab[match(names(ggCol), tab[[valColname]]),
                                cellinfoName, with = FALSE]
            cellinfoName <- as.character(cellinfoName[[1]])
        }
    }else{
        cellinfoName <- names(ggCol)
    }
    
    # Actual ggplot
    ggOut <- ggXYplot(ggData)
    if (bgCells) {
        ggOut <- labelBackgroundCells(
            ggOut,
            ggData2,
            pointSize,
            color = "snow2",
            shape = 16,
            hide = hideFilterCell)
    }
    
    if (inpShowEdge) {
        keep <- edges[, 1] %in% ggData$idx & edges[, 2] %in% ggData$idx
        if(sum(keep)){
            edgeDf <- cbind(ggData[match(edges[keep, 1], ggData$idx), c('X', 'Y')],
                            ggData[match(edges[keep, 2], ggData$idx), c('X', 'Y')])
            colnames(edgeDf) <- c('x', 'y', 'xend', 'yend')
            ggOut <- ggOut +
                geom_segment(
                    data = edgeDf,
                    aes(
                        x = .data[["x"]],
                        y = .data[["y"]],
                        xend = .data[["xend"]],
                        yend = .data[["yend"]]
                    ),
                    inherit.aes = FALSE,
                    linewidth = 0.5,
                    colour = '#CCCCCC',
                    alpha = 0.5
                )
        }
        
    }
    
    ggOut <- pointPlot(
        ggOut = ggOut,
        pointSize = pointSize,
        fontSize = labelsFontsize,
        labelsFontFamily = labelsFontFamily,
        dimRedX = dimRedX,
        dimRedY = dimRedY,
        keepXYlables = keepXYlables)
    # slingshot
    if (inpSlingshot) {
        if (file.exists(slingshotFilename)) {
            lineages <- readRDS(slingshotFilename)
            if (cellinfoID %in% names(lineages)) {
                lineages <- lineages[[cellinfoID]]
                x <- SlingshotDataSet(lineages)
                X <- reducedDim(x)
                checkRedDimName <- function(a, b) {
                    a <- gsub('[^a-z]', '', tolower(a))
                    b <- gsub('[^a-z]', '', tolower(b))
                    a %in% b
                }
                if (checkRedDimName(dimRedX, colnames(X)) &&
                    checkRedDimName(dimRedY, colnames(X))) {
                    linInd <- seq_along(slingLineages(x))
                    clusterLabels <- slingClusterLabels(x)
                    connectivity <- slingMST(x)
                    clusters <- rownames(connectivity)
                    nclus <- nrow(connectivity)
                    if (nclus > 1) {
                        centers <- t(vapply(clusters, function(clID) {
                            w <- clusterLabels[, clID]
                            return(apply(X, 2, weighted.mean, w = w))
                        }, rep(0, ncol(
                            X
                        ))))
                        rownames(centers) <- clusters
                        X <- X[rowSums(clusterLabels) > 0, , drop = FALSE]
                        clusterLabels <-
                            clusterLabels[
                                rowSums(clusterLabels) > 0, ,
                                drop = FALSE]
                        linC <- slingParams(x)
                        clus2include <- unique(unlist(slingLineages(x)[linInd]))
                        lineDf <- data.frame()
                        for (i in seq_len(nclus - 1)) {
                            for (j in seq(i + 1, nclus)) {
                                if (connectivity[i, j] == 1 &&
                                    all(clusters[c(i, j)] %in% clus2include,
                                        na.rm = TRUE)) {
                                    lineDf <-
                                        rbind(
                                            lineDf, c(
                                                centers[
                                                    i, c(1, 2), drop = TRUE],
                                                centers[
                                                    j, c(1, 2), drop = TRUE]))
                                }
                            }
                        }
                        colnames(lineDf) <- c("x", "y", "xend", "yend")
                        pts <- centers[clusters %in% clus2include, c(1, 2)]
                        colnames(pts) <- c("x", "y")
                        pts <- cbind(as.data.frame(pts), color = 'black')
                        if (any(linC$start.given)) {
                            if (length(linC$start.clus[linC$start.given]) > 0) {
                                pts[linC$start.clus[linC$start.given],
                                    "color"] <- "green3"
                            }
                        }
                        if (any(linC$end.given)) {
                            if (length(linC$end.clus[linC$end.given]) > 0) {
                                pts[linC$end.clus[linC$end.given],
                                    "color"] <- "red2"
                            }
                        }
                        ggOut <- ggOut +
                            geom_segment(
                                data = lineDf,
                                aes(
                                    x = .data[["x"]],
                                    y = .data[["y"]],
                                    xend = .data[["xend"]],
                                    yend = .data[["yend"]]
                                ),
                                inherit.aes = FALSE
                            ) +
                            geom_point(
                                data = pts,
                                aes(
                                    x = .data[["x"]],
                                    y = .data[["y"]],
                                    color = .data[["color"]]
                                ),
                                size = pointSize * 3,
                                alpha = .5,
                                inherit.aes = FALSE
                            )
                    }
                }
            }
        }
    }
    # label
    if (is.na(inpConf[inpConf$UI == cellinfoID]$fCL)) {
        ggOut <- ggOut +
            scale_color_gradientn("", colours = availableThemes(gradientCol)) +
            guides(color = guide_colorbar(barwidth = 15))
    } else {
        ggOut <- ggOut + scale_color_manual(
            "", values = ggCol, labels=cellinfoName) +
            theme(legend.text = element_text(size = labelsFontsize,
                                             family = labelsFontFamily))
        if(length(ggCol)>50){
            ggOut <- ggOut +
                guides(color = "none")
            showNotification(paste('Too many overlapping labels.',
                                   'Not all labels can be show!',
                                   'The legend are also removed.',
                                   'Consider filter out some data points.'),
                             type = "warning")
        }else{
            ggOut <- ggOut +
                guides(color = guide_legend(
                    override.aes = list(size = 5),
                    nrow = inpConf[inpConf$UI == cellinfoID]$fRow
                ))
        }
        if (inplab) {
            ggData3 <- 
                ggData[, list(
                    X = mean(.SD$X),
                    Y = mean(.SD$Y)), by = valColname]
            lListX <-
                min(nchar(paste0(ggData3[[valColname]], collapse = "")), 200)
            lListX <- .globals$lList - (0.25 * floor(lListX / 50))
            ggOut <- ggOut +
                geom_text_repel(
                    data = ggData3,
                    aes(
                        .data[["X"]], .data[["Y"]],
                        label = .data[[valColname]]),
                    color = "grey10",
                    bg.color = "grey95",
                    bg.r = 0.15,
                    size = labelsFontsize*.3508333,
                    family = labelsFontFamily,
                    seed = 123
                )
        }
    }
    ggOut <- fixCoord(ggOut, plotAspectRatio, rat)
    return(ggOut)
}

#' @importFrom DT DTOutput
#' @importFrom magrittr %>%
coExpr3dUI <- function(id) {
    tabPanel(
        value = id,
        htmlOutput(NS(id, 'GeneExpr')),
        tabsubTitleUI(
            id,
            'GeneExpr',
            description = paste(
                "In this tab, users can visualise the coexpression of two ",
                "genes on low-dimensional representions."
            )
        ),
        fluidRow(
            column(3, dimensionReductionUI(id)),
            column(5, subsetCellByInfoUI(id)),
            column(4, graphicsControlUI(id))
        ),
        fluidRow(
            column(
                3,
                style = "border-right: 2px solid black",
                htmlOutput(NS0(id, "subPlotTitle", 1)),
                geneExprUI(id, 1),
                geneExprUI(id, 2),
                geneCoExprPlotControlUI(id, 1, plotly = TRUE)
            ),
            column(
                9,
                fluidRow(column(12, uiOutput(
                    NS0(id, "GeneExpr3Doup.ui", 1)
                ))),
                downloadButton(
                    NS(id, 'downloadExpr'),
                    "Download expression for clicked cell"
                ),
                verbatimTextOutput(NS(id, 'clicked'))
            )
        )
    )
}

#' @importFrom DT renderDT
#' @importFrom magrittr %>%
#' @importFrom plotly plotlyOutput renderPlotly event_data ggplotly
#'  event_register
coExpr3dServer <- function(id, dataSource, optCrt) {
    moduleServer(id, function(input, output, session) {
        ## title
        output$GeneExpr <-
            renderUI({
                HTML(paste("3D Gene", dataSource()$terms['coexpression']))
            })
        ## subtitle
        output$GeneExprSubTitle <-
            renderUI({
                h4(paste(
                    sub(
                        substr(dataSource()$terms['coexpression'], 1, 1),
                        toupper(substr(
                            dataSource()$terms['coexpression'],
                            1, 1)),
                        dataSource()$terms['coexpression']
                    ),
                    "of two genes on reduced dimensions"
                ))
            })
        ## input column 1
        ### Dimension Reduction
        updateDimRedSelInputPair(session, input, dataSource)
        ## input column 2
        updateSubsetCellUI(id, input, output, session, dataSource)
        
        ## plot region
        ### sub region title
        output$subPlotTitle1 <-
            renderUI({
                h4(paste("Gene", dataSource()$terms['expression']))
            })
        ### dropdown list
        updateSelectizeInput(
            session,
            "GeneName1",
            choices = sort(names(dataSource()$sc1gene)),
            server = TRUE,
            selected = dataSource()$sc1def$gene1,
            options = list(
                maxOptions = .globals$maxNumGene,
                create = TRUE,
                persist = TRUE,
                render = I(optCrt)
            )
        )
        updateSelectizeInput(
            session,
            "GeneName2",
            choices = sort(names(dataSource()$sc1gene)),
            server = TRUE,
            selected = dataSource()$sc1def$gene2,
            options = list(
                maxOptions = .globals$maxNumGene,
                create = TRUE,
                persist = TRUE,
                render = I(optCrt)
            )
        )
        ### plots
        plot3d <- reactive({
            scDRcoex(
                inpConf=dataSource()$sc1conf,
                inpMeta=dataSource()$sc1meta,
                dimRedX=input$GeneExprdrX,
                dimRedY=input$GeneExprdrY,
                gene1=input$GeneName1,
                gene2=input$GeneName2,
                subsetCellKey=input$subsetCell,
                subsetCellVal=getSubsetCellVal(input),
                dataset=dataSource()$dataset,
                geneIdMap=dataSource()$sc1gene,
                plotType="3D",
                pointSize=input$GeneExprsiz,
                GeneExprDotCol=input$CoExprcol1,
                GeneExprDotOrd=input$CoExprord1,
                labelsFontsize=input$GeneExprfsz,
                labelsFontFamily=input$GeneExprfml,
                plotAspectRatio=input$GeneExprasp,
                keepXYlables=input$GeneExprtxt,
                hideFilterCell=input$CoExprhid1
            )
        })
        output$GeneExpr3Doup1 <- renderPlotly({
            ggplotly(plot3d()) %>% event_register("plotly_click")
        })
        output$GeneExpr3Doup.ui1 <- renderUI({
            plotlyOutput(
                NS0(id, "GeneExpr3Doup", 1),
                height = .globals$pList1[input$GeneExprpsz])
        })
        output$downloadExpr <- exprDownloadHandler(
            dataSource()$sc1gene,
            dataSource()$dataset,
            dataSource()$sc1meta)
        output$clicked <- plotly3d_click(session)
    })
}

#' @importFrom shinymanager auth_ui secure_server use_language check_credentials
loginUI <- function(loginNavbarTitle, defaultDataset) {
    lan <- use_language("en")
    tabPanel(
        HTML(loginNavbarTitle),
        auth_ui(
            id = "auth",
            tags_bottom = tags$div(tags$a(
                href = paste0('?data=', defaultDataset),
                "Go to application",
                icon = icon("share")
            )),
            # set language ?
            lan = lan
        ),
        shinymanager_where("authentication"),
        shinymanager_language(lan$get_language())
    )
}
loginServer <- function(input, output, session) {
    auth <-
        secure_server(check_credentials = check_credentials(
            file.path(.globals$app_path, .globals$credential_path),
            passphrase = .globals$passphrase))
}

#' @importFrom utils getFromNamespace
#' @importFrom shinymanager use_language secure_app fab_button pwd_ui
secureUI <- function(ui0, timeout = 30) {
    function(request) {
        query <- parseQueryString(request$QUERY_STRING)
        token <- gsub('\"', "", query$token)
        .tok <- getFromNamespace(".tok", "shinymanager")
        if (.tok$is_valid(token)) {
            enable_admin <- TRUE
            head_auth <- NULL
            fab_position <- "bottom-right"
            admin <- query$admin
            language <- query$language
            if (!is.null(language)) {
                lan <- use_language(gsub('\"', "", language))
            }
            is_forced_chg_pwd <-
                getFromNamespace(
                    "is_force_chg_pwd",
                    "shinymanager")(token = token)
            if (is_forced_chg_pwd) {
                args <- list()
                args$id <- "password"
                args$lan <- lan
                pwd_ui <-
                    fluidPage(
                        theme = .globals$theme,
                        tags$head(head_auth),
                        do.call(pwd_ui, args),
                        shinymanager_where("password"),
                        shinymanager_language(lan$get_language())
                    )
                return(pwd_ui)
            }
            .tok$set_timeout(timeout)
            if (isTRUE(enable_admin) && .tok$is_admin(token) &
                identical(admin, "true") &
                !is.null(.tok$get_sqlite_path())) {
                navbarPage(
                    title = "Admin",
                    theme = theme,
                    header =
                        tagList(
                            tags$style(
                                ".navbar-header{margin-left:16.66% !important;}"
                            ),
                            fab_button(
                                position = fab_position,
                                actionButton(
                                    inputId = ".shinymanager_logout",
                                    label = lan$get("Logout"),
                                    icon = icon("sign-out")
                                ),
                                actionButton(
                                    inputId = ".shinymanager_app",
                                    label = lan$get("Go to application"),
                                    icon = icon("share")
                                )
                            ),
                            shinymanager_where("admin")
                        ),
                    tabPanel(
                        title = tagList(icon("home"), lan$get("Home")),
                        value = "home",
                        getFromNamespace(
                            "admin_ui",
                            "shinymanager")("admin", lan),
                        shinymanager_language(lan$get_language())
                    ),
                    tabPanel(
                        title = lan$get("Logs"),
                        getFromNamespace(
                            "logs_ui",
                            "shinymanager")("logs", lan),
                        shinymanager_language(lan$get_language())
                    ),
                    tabPanel(
                        title = 'SiteInfo',
                        webstatsUI('webstats')),
                    tabPanel(
                        title = 'UploadData',
                        uploadUI("upload")),
                    tabPanel(
                        title = 'EditData',
                        editUI("editdata"))
                )
            } else {
                if (isTRUE(enable_admin) && .tok$is_admin(token) &&
                    !is.null(.tok$get_sqlite_path())) {
                    menu <-
                        fab_button(
                            position = fab_position,
                            actionButton(
                                inputId = ".shinymanager_logout",
                                label = lan$get("Logout"),
                                icon = icon("sign-out")
                            ),
                            actionButton(
                                inputId = ".shinymanager_admin",
                                label = lan$get("Administrator mode"),
                                icon = icon("cogs")
                            )
                        )
                }
                else {
                    if (isTRUE(enable_admin) && .tok$is_admin(token) &&
                        is.null(.tok$get_sqlite_path())) {
                        warning(
                            "Admin mode is only available",
                            "when using a SQLite database!",
                            call. = FALSE
                        )
                    }
                    menu <- fab_button(
                        position = fab_position,
                        actionButton(
                            inputId = ".shinymanager_logout",
                            label = lan$get("Logout"),
                            icon = icon("sign-out")
                        )
                    )
                }
                getFromNamespace("save_logs", "shinymanager")(token)
                if (is.function(ui0)) {
                    ui <- ui0(request)
                }
                tagList(
                    ui,
                    menu,
                    shinymanager_where("application"),
                    shinymanager_language(lan$get_language()),
                    singleton(tags$head(
                        tags$script(src = "shinymanager/timeout.js")
                    ))
                )
            }
        } else{
            ui0(request)
        }
    }
}

#' Create a shinycell config data.table
#' This function was simplified from shinycell for package submission
#' @noRd
#' @param obj input single-cell object for Seurat (v3+)
#' @param theme color theme. default is "Paired" from ColorBrewer palettes.
#' @param meta.to.include columns to include from the single-cell metadata.
#'   Default is \code{NA}, which is to use all columns. Users can specify
#'   the columns to include, which must match one of the following:
#'   \itemize{
#'     \item{Seurat objects}: column names in \code{seu@meta.data}
#'       i.e. \code{colnames(seu@meta.data)}
#'   }
#' @param legendCols maximum number of columns allowed when displaying the
#'   legends of categorical metadata
#' @param maxLevels maximum number of levels allowed for categorical metadata.
#'   Metadata with nlevels > maxLevels will be discarded automatically
#'
#' @return config data.table
#' @importFrom SeuratObject Reductions
#' @importFrom data.table data.table
#' @importFrom grDevices colorRampPalette
#' @importFrom RColorBrewer brewer.pal
createConfig <- function(
        obj,
        theme="Paired",
        meta.to.include = NA,
        legendCols = 4,
        maxLevels = 50) {
    # Extract corresponding metadata
    drExist <- TRUE
    if (inherits(obj, "Seurat")) {
        # Seurat Object
        objMeta <- obj[[]]
        if (length(Reductions(obj)) == 0) {
            drExist <- FALSE
        }
    }  else {
        stop("Only Seurat object is accepted!")
    }
    if (!drExist) {
        stop(
            "Can not detect any dimension reduction data \n",
            "       e.g. umap / tsne. Has any analysis been performed?"
        )
    }
    if(!theme %in% availableThemes("name")){
        stop('theme must be one of ', availableThemes("namestring"))
    }
    # Checks and get list of metadata to include
    if (is.na(meta.to.include[1])) {
        meta.to.include <- colnames(objMeta)
    }
    if (length(meta.to.include) < 2) {
        stop("At least 2 metadata is required!")
    }
    
    # Start making config data.table
    scConf <- data.table()
    for (iMeta in meta.to.include) {
        tmpConf <- data.table(
            ID = iMeta,
            UI = iMeta,
            fID = NA,
            fUI = NA,
            fCL = NA,
            fRow = NA,
            default = 0,
            grp = FALSE
        )
        
        # Convert to factors if metadata contains characters
        if (is.character(objMeta[[iMeta]])) {
            objMeta[[iMeta]] <- factor(objMeta[[iMeta]])
        }
        
        # Additional preprocessing for categorical metadata
        nLevels <- nlevels(objMeta[[iMeta]])
        if (nLevels <= maxLevels) {
            if(any(grepl('\\|', levels(objMeta[[iMeta]])))){
                stop('metadata can not contain "|".')
            }
            if (nLevels >= 2) {
                tmpConf$fID <- paste0(levels(objMeta[[iMeta]]), collapse = "|")
                tmpConf$fUI <- tmpConf$fID
                tmpConf$fCL <-
                    paste0(
                        scColorRampPalette(nLevels, theme),
                        collapse = "|")
                tmpConf$fRow <- ceiling(nLevels / legendCols)
                tmpConf$grp <- TRUE
            } else if (nLevels == 1) {
                tmpConf$fID <- levels(objMeta[[iMeta]])
                tmpConf$fUI <- tmpConf$fID
                tmpConf$fCL <- "black"
                    tmpConf$fRow <- 1
            }
            scConf <- rbindlist(list(scConf, tmpConf))
        }
    }
    
    # Set defaults
    def1 <- grep("ident|library", scConf$ID, ignore.case = TRUE)[1]
    def2 <- grep("clust", scConf$ID, ignore.case = TRUE)
    def2 <- setdiff(def2, def1)[1]
    if (is.na(def1)) {
        def1 <- setdiff(c(1, 2), def2)[1]
    }
    if (is.na(def2)) {
        def2 <- setdiff(c(1, 2), def1)[1]
    }
    scConf[def1]$default <- 1
    scConf[def2]$default <- 2
    
    # STOP if there is no single multi-level covariate
    if (nrow(scConf[scConf$grp == TRUE]) == 0) {
        stop("Can not detect any multi-group cell metadata.")
    }
    
    return(scConf)
}

# Plot theme
#' @importFrom ggplot2 theme element_text element_rect element_line
#' element_blank
sctheme <-
    function(
        base_size = 24,
        family = "Helvetica",
        XYval = TRUE,
        Xang = 0,
        XjusH = 0.5) {
        oupTheme <- theme(
            text = element_text(size = base_size, family = family),
            panel.background = element_rect(fill = "white", colour = NA),
            axis.line = element_line(colour = "black"),
            axis.ticks = element_line(colour = "black", size = base_size / 20),
            axis.title = element_text(face = "bold"),
            axis.text = element_text(size = base_size, family = family),
            axis.text.x = element_text(angle = Xang, hjust = XjusH),
            legend.position = "bottom",
            legend.key = element_rect(colour = NA, fill = NA)
        )
        if (!XYval) {
            oupTheme <- oupTheme + theme(
                axis.text.x = element_blank(),
                axis.ticks.x = element_blank(),
                axis.text.y = element_blank(),
                axis.ticks.y = element_blank()
            )
        }
        return(oupTheme)
    }

plotPieDimUI <- function(id) {
    tabPanel(
        value = id,
        HTML("Sunburst"),
        h4("Gene expression scatter_pie plot"),
        "In this tab, users can visualise the gene expression patterns of ",
        "multiple genes grouped by categorical cell information ",
        "(e.g. library / cluster). ",
        "The expression of each gene will be re-scaled to 0-1. ",
        "It will only plot top 5000 events.",
        br(),
        fluidRow(
            column(3, dimensionReductionUI(id)),
            column(
                5,
                subsetCellByInfoUI(id),
                subsetCellByFilterUI(id)),
            column(4, graphicsControlUI(id, GeneExpraspSelect = 'Fixed'))
        ),
        br(),
        fluidRow(
            column(
                3,
                style = "border-right: 2px solid black",
                textAreaInput(
                    NS(id, "genelist"),
                    HTML(
                        "List of gene names <br />
                        (Max 10 genes (over 5 will be busy), <br />
                        separated by , or ; or newline):"
                    ),
                    height = "200px",
                    value = NULL
                ) %>%
                    shinyhelper::helper(
                        type = "inline",
                        size = "m",
                        fade = TRUE,
                        title = "List of genes to plot on Sunburst",
                        content = c(
                            "Input genes to plot",
                            "- Maximum 10 genes",
                            "(due to plot space limitations)",
                            "- Genes should be separated by comma,",
                            "semicolon or newline"
                        )
                    ),
                checkboxInput(
                    NS(id, "CoExpred"),
                    "Co-expressed (all>0)", value = TRUE),
                tagList(
                    actionButton(NS(id, "CoExprtog"), "Toggle plot controls"),
                    conditionalPanel(
                        condition = paste0("input.CoExprtog", " % 2 == 1"),
                        ns = NS(id),
                        checkboxInput(
                            NS(id, "CoExprCircle"),
                            "Circle Cells", value = FALSE),
                        checkboxInput(
                            NS(id, "CoExprBg"),
                            "Plot cells as Background", value = TRUE),
                        checkboxInput(
                            NS(id, "CoExprMarkGrp"), "lable groups",
                            value = FALSE),
                        radioButtons(
                            NS(id, "CoExprType"),
                            "Plot type",
                            choices = c("sunburst", "pie", 'donut', "bar", "sum", "max", "mean"),
                            selected = "sunburst"
                        ),
                        conditionalPanel(
                            condition = "['sum', 'max', 'mean'].includes(input.CoExprType)",
                            ns=NS(id),
                            radioButtons(
                                NS(id, "CoExprCol"), "Colour:",
                                inline = TRUE,
                                choices = availableThemes("sequence"),
                                selected = availableThemes("sequence")[1])),
                        sliderInput(
                            NS(id, "CoExprAlpha"),
                            "Transparency",
                            min = 0,
                            max = 1,
                            step = 0.01,
                            value = .8
                        )
                    )
                )
            ),
            column(9, h4(htmlOutput(
                NS(id, "oupTxt")
            )), geneExprDotPlotUI(id))
        )
    )
}

plotPieDimServer <- function(id, dataSource, optCrt) {
    moduleServer(id, function(input, output, session) {
        ## input column 1
        ### Dimension Reduction
        updateDimRedSelInputPair(session, input, dataSource)
        ## input column 2
        updateSubsetCellUI(id, input, output, session, dataSource)
        updateFilterCellUI(id, optCrt, input, output, session, dataSource)
        ## input column
        genelist <- dataSource()$sc1def$genes
        if (!is.null(dataSource()$genelist)) {
            if (length(dataSource()$genelist) > 1) {
                genelist <- dataSource()$genelist
            } else{
                genelist <- c(dataSource()$genelist, genelist)
            }
        }
        updateTextAreaInput(
            session, "genelist",
            value = paste0(
                genelist[seq.int(min(4, length(genelist)))],
                collapse = ", "))
        ### plots
        plot1 <- reactive({
            scPieDim(
                inpConf = dataSource()$sc1conf,
                inpMeta = dataSource()$sc1meta,
                dataset = dataSource()$dataset,
                geneIdMap = dataSource()$sc1gene,
                dimRedX = input$GeneExprdrX,
                dimRedY = input$GeneExprdrY,
                genelist = input$genelist,
                subsetCellKey = input$subsetCell,
                subsetCellVal = getSubsetCellVal(input),
                valueFilterKey = input$filterCell,
                valueFilterCutoff = input$filterCellVal,
                valueFilterCutoff2 = input$filterCellVal2,
                CoExpred = input$CoExpred,
                pointSize = input$GeneExprsiz,
                lableCircle = input$CoExprCircle,
                plotCellBg = input$CoExprBg,
                markGrp = input$CoExprMarkGrp,
                alpha = input$CoExprAlpha,
                gradientCol= input$CoExprCol,
                plotType = input$CoExprType,
                labelsFontsize = input$GeneExprfsz,
                labelsFontFamily=input$GeneExprfml,
                plotAspectRatio = input$GeneExprasp,
                keepXYlables = input$GeneExprtxt
            )
        })
        updateGeneExprDotPlotUI(
            postfix = 1,
            id,
            input,
            output,
            session,
            plot1,
            .globals$pList1[input$GeneExprpsz],
            dataSource()$dataset,
            input$GeneExprdrX,
            input$GeneExprdrY,
            'sunburst'
        )
    })
}

geneAccGeneExprUI <- function(id) {
    tabPanel(
        value = id,
        htmlOutput(NS(id, 'AccExpr')),
        tabsubTitleUI(
            id,
            'AccExpr',
            description = paste(
                "In this tab, users can visualise both accessibility and ",
                "feature information side-by-side."
            )
        ),
        fluidRow(
            column(3, dimensionReductionUI(id)),
            column(5, subsetCellByInfoUI(id, multiple=FALSE)),
            column(4, graphicsControlUI(id))
        ),
        fluidRow(
            column(
                6,
                style = "border-right: 2px solid black",
                h4("ATAC track"),
                fluidRow(
                    column(6, geneAccUI(id, 1)),
                    column(6, geneAccPlotControlUI(id, 1))),
                geneExprDotPlotUI(id, 1)
            ),
            column(
                6,
                htmlOutput(NS0(id, "subPlotTitle", 2)),
                fluidRow(
                    column(6, geneExprUI(id, 2)),
                    column(6, geneExprPlotControlUI(id, 2))),
                geneExprDotPlotUI(id, 2)
            )
        )
    )
}
#' @importFrom DT formatRound renderDT
#' @importFrom magrittr %>%
geneAccGeneExprServer <- function(id, dataSource, optCrt) {
    moduleServer(id, function(input, output, session) {
        ## title
        if(dataSource()$data_types[[dataSource()$dataset]] %in% 
           c("scMultiome", "scATACseq")){
            data_type <- dataSource()$data_types[[dataSource()$dataset]]
            output$AccExpr <- renderUI({
                HTML(paste0(
                    "GeneAcc vs ",
                    dataSource()$terms["GeneExpr"]))
            })
            ## subtitle
            output$AccExprSubTitle <-
                renderUI({
                    h4(paste(
                        "ATAC vs gene",
                        dataSource()$terms['expression'],
                        "on reduced dimensions"))
                })
            ## input column 1
            ### Dimension Reduction
            updateDimRedSelInputPair(session, input, dataSource)
            ## input column 2
            updateSubsetCellUI(id, input, output, session, dataSource)
            
            ## plot region
            ### sub region title
            output$subPlotTitle2 <-
                renderUI({
                    h4(paste("Gene", dataSource()$terms['expression']))
                })
            selectedGene <- dataSource()$sc1def$gene1
            ### geneAcc
            updateGeneAccPlot(
                postfix = 1,
                genePostfix = 2,
                optCrt,
                id,
                input,
                output,
                session,
                dataSource)
            
            ### geneExpr
            if (!is.null(dataSource()$genelist)) {
                selectedGene <- dataSource()$genelist[1]
            }
            updateGeneExprPlot(
                postfix = 2,
                selectedGene,
                optCrt,
                id,
                input,
                output,
                session,
                dataSource)
        }else{
            output$AccExpr <- renderUI({
                HTML("")
            })
        }
    })
}

#' @importFrom magrittr %>%
tabDownloader <- function() {
    tabPanel(
        value = "DownloadDataset",
        HTML("Download dataset"),
        h4("Download the dataset"),
        "In this tab, users can download the dataset.",
        br(),
        br(),
        fluidRow(
            column(
                3,
                style = "border-right: 2px solid black",
                selectInput(
                    "scDwl1inp1", "Cell information:",
                    choices = NULL) %>%
                    shinyhelper::helper(
                        type = "inline",
                        size = "m",
                        fade = TRUE,
                        title = "Cell information to group cells by",
                        content =
                            paste(
                                "Select categorical cell information",
                                "to group cells by")
                    ),
                selectInput(
                    "scDwl1inp1a", "Cell information to subset:",
                    choices = NULL) %>%
                    shinyhelper::helper(
                        type = "inline",
                        size = "m",
                        fade = TRUE,
                        title = "Cell information to subset by:",
                        content = c(
                            paste(
                                "Select categorical cell information",
                                "to sebset cells by"),
                            "- cells are shown in different subsets"
                        )
                    ),
                selectInput("scDwl1inp2", "Cell Info:", choices = NULL) %>%
                    shinyhelper::helper(
                        type = "inline",
                        size = "m",
                        fade = TRUE,
                        title = "Cell Info",
                        content = c("Select cell info  to download")
                    )
            )
        )
    )
}

# Plot heatmap for proportion
#' @importFrom ggplot2 ggplot aes .data scale_color_gradientn ylab xlab geom_tile geom_text
#' scale_fill_manual theme
#' @importFrom data.table data.table
#' @importFrom reshape2 melt
#' @importFrom stats prcomp
scCorProp <- function(
        proportions,
        testby = 'Y',
        method = c("pearson", "kendall", "spearman"),
        type=c('heatmap', 'PCA'),
        labelsFontsize=24,
        labelsFontFamily = 'Helvetica',
        inpcols='White-Red'
        ) {
    type <- match.arg(type)
    method <- match.arg(method)
    if(testby=='X'){
        proportions <- t(proportions)
    }
    ggOut <- switch(type,
                    PCA={
                        pca <- prcomp(proportions)
                        pca <- data.table(pca$rotation,
                                          keep.rownames = TRUE)
                        ggplot(pca, aes(.data[['PC1']],
                                        .data[['PC2']],
                                        label = .data[['rn']])) +
                            geom_point() +
                            geom_text(hjust=0, vjust=1) +
                            sctheme(
                                base_size = labelsFontsize,
                                family = labelsFontFamily)
                    },
                    heatmap={
                        cor <- cor(proportions, method = method)
                        cor <- reshape2::melt(data.table(cor, keep.rownames = TRUE),
                                    id.vars = c("rn"))
                        colnames(cor) <- c('Var1', 'Var2', 'value')
                        cor$Var1 <- factor(as.character(cor$Var1),
                                           levels = levels(cor$Var2))
                        ggplot(cor, aes(.data[["Var1"]], .data[["Var2"]],
                                        fill = .data[["value"]],
                                        label = round(.data[["value"]], 3))) +
                            geom_tile() +
                            geom_text() +
                            xlab("") + ylab("") +
                            sctheme(
                                base_size = labelsFontsize,
                                family = labelsFontFamily,
                                Xang = 45,
                                XjusH = 1) +
                            scale_fill_gradientn(
                                "",
                                limits = c(ifelse(any(cor$value<0), -1, 0), 1),
                                colours = availableThemes(inpcols)) +
                            theme(legend.position = "right")
                    })
    return(ggOut)
}

updateDimRedSelInput <-
    function(session, inputId, label, conf, selected) {
        updateSelectInput(
            session,
            inputId,
            label,
            choices = conf[conf$dimred == TRUE]$UI,
            selected = selected)
    }
#' @importFrom utils adist
updateDimRedSelInputPair <-
    function(session, input, dataSource) {
        updateDimRedSelInput(
            session,
            "GeneExprdrX",
            "X-axis:",
            dataSource()$sc1conf,
            dataSource()$sc1def$dimred[1]
        )
        updateDimRedSelInput(
            session,
            "GeneExprdrY",
            "Y-axis:",
            dataSource()$sc1conf,
            dataSource()$sc1def$dimred[2]
        )
        observeEvent(input$GeneExprdrX, {
            try({
                conf <- dataSource()$sc1conf
                choices <- conf[conf$dimred == TRUE]$UI
                choices <- choices[choices!=input$GeneExprdrX]
                dist <- adist(input$GeneExprdrX, choices)
                updateDimRedSelInput(
                    session, 
                    "GeneExprdrY",
                    "Y-axis:",
                    dataSource()$sc1conf,
                    choices[which.min(dist)])
            })
        })
    }
getGroupUI <- function(dataSource) {
    dataSource()$sc1conf[dataSource()$sc1conf$grp == TRUE]$UI
}
getNonGroupUI <- function(dataSource) {
    dataSource()$sc1conf[is.na(dataSource()$sc1conf$fID)]$UI
}
updateSubsetCellUI <-
    function(
        id,
        input,
        output,
        session,
        dataSource,
        addNA = FALSE,
        ABcolumns = "") {
        choices <- dataSource()$sc1conf[dataSource()$sc1conf$grp == TRUE]$UI
        if (addNA) {
            selected  <- "N/A"
            choices <- c("N/A", choices)
        } else{
            selected <- dataSource()$sc1def$grp1
        }
        subsetCell.ui <- reactiveValues()
        lapply(ABcolumns, function(ABcolumn){
            output[[paste0("subsetCellSel.ui", ABcolumn)]] <- renderUI({
                selectInput(
                    NS0(id, "subsetCell", ABcolumn),
                    "Cell information to subset:",
                    choices = choices,
                    selected = selected,
                    multiple = 
                        if(length(input[[paste0("subsetCell.multi",
                                                ABcolumn)]])>0)
                            as.logical(input[[paste0("subsetCell.multi",
                                                     ABcolumn)]]%%2)
                        else FALSE)
            })
            observeEvent(
                input[[paste0("subsetCell.multi", ABcolumn)]],
                updateActionButton(
                    session = session,
                    inputId = paste0("subsetCell.multi", ABcolumn),
                    label = ifelse(
                        as.logical(input[[paste0("subsetCell.multi",
                                                 ABcolumn)]]%%2),
                        "single", "multiple"))
            )
            
            subsetCell.ui[[paste0("uis", ABcolumn)]] <- list()
            observeEvent(input[[paste0("subsetCell", ABcolumn)]],{
                subsetCell <- input[[paste0("subsetCell", ABcolumn)]][
                    input[[paste0("subsetCell", ABcolumn)]]!="N/A"]
                sub_name <- dataSource()$sc1conf$UI %in% subsetCell
                x <- dataSource()$sc1conf[sub_name]$fID
                if(length(x)!=length(subsetCell)){
                    return(NULL)
                }
                sub <-strsplit(x, "\\|")
                names(sub) <- dataSource()$sc1conf[sub_name]$UI
                subsetCell.ui[[paste0("uis", ABcolumn)]] <- list()
                for(subid in subsetCell){
                    choices <- sub[[subid]]
                    if(!is.null(choices)){
                        subid1 <- paste0(subid, ABcolumn)
                        subsetCell.ui[[paste0("uis", ABcolumn)]][[subid]] <- 
                            tagList(
                                div(
                                    style = 
                                        paste(
                                            "max-height: 150px; display:flex;",
                                            "flex-direction: column;",
                                            "overflow-y: auto;")
                                    ,
                                    actionButton(
                                        NS0(id, 'subsetCell.uncheck', subid1),
                                        label=textOutput(
                                            NS0(id, 'subsetCell.uncheckLab',
                                                subid1),
                                            inline = TRUE)),
                                    checkboxGroupInput(
                                        NS0(id, "subsetCellVal", subid1),
                                        subid,
                                        inline = TRUE,
                                        choices = choices,
                                        selected = choices
                                    ),
                                    div(
                                        style = "visibility:hidden;",
                                        textInput(
                                            NS0(id, "subsetCellValChoices",
                                                subid1),
                                            label = NULL,
                                            value = paste(choices,
                                                          collapse = "|")
                                        )
                                    )
                                )
                            )
                        output[[paste0("subsetCell.uncheckLab", subid1)]] <-
                            renderPrint(cat("Uncheck All"))
                        observeEvent(input[[paste0("subsetCell.uncheck", 
                                                   subid1)]],
                          {
                              sub <- strsplit(
                                  input[[paste0("subsetCellValChoices",
                                                subid1)]],
                                  "\\|")[[1]]
                              if(length(input[[paste0("subsetCellVal",
                                                      subid1)]])>0){
                                  selected <- NULL
                                  uncheckLab <- 'Check All'
                              }else{
                                  selected <- sub
                                  uncheckLab <- 'Uncheck All'
                              }
                              output[[paste0("subsetCell.uncheckLab",
                                             subid1)]] <-
                                  renderPrint(cat(uncheckLab))
                              updateCheckboxGroupInput(
                                  session = session,
                                  inputId = paste0('subsetCellVal', subid1),
                                  inline = TRUE,
                                  choices = sub,
                                  selected = selected
                              )
                          })
                    }
                }
            })
            
            output[[paste0("subsetCell.ui", ABcolumn)]] <- 
                renderUI({subsetCell.ui[[paste0("uis", ABcolumn)]]})
        })
    }
updateFilterCellUI <-
    function(
        id,
        optCrt,
        input,
        output,
        session,
        dataSource) {
        updateSelectizeInput(
            session,
            "filterCell",
            server = TRUE,
            choices = c(
                getNonGroupUI(dataSource),
                sort(names(
                    dataSource()$sc1gene
                ))),
            selected = getNonGroupUI(dataSource)[1],
            options = list(
                maxOptions =
                    length(getNonGroupUI(dataSource)) + 3,
                create = TRUE,
                persist = TRUE,
                render = I(optCrt)
            )
        )
        output$filterCell.ui <- renderUI({
            if (!input$filterCell %in% dataSource()$sc1conf$UI) {
                val <- read_exprs(
                    dataSource()$dataset,
                    dataSource()$sc1gene[input$filterCell],
                    valueOnly = TRUE)
                val2 <- c()
            } else{
                val <- dataSource()$sc1meta[[
                    dataSource()$sc1conf[
                        dataSource()$sc1conf$UI == input$filterCell]$ID]]
                if(grepl('[12]$', input$filterCell) &&
                   sum(grepl(paste0('^', sub('.$', '',
                                             input$filterCell)),
                             dataSource()$sc1conf$UI))==2){
                    filterCellKey2 <- ifelse(
                        grepl('1$', input$filterCell),
                        sub('.$', '2', input$filterCell),
                        sub('.$', '1', input$filterCell)
                    )
                    val2 <- dataSource()$sc1meta[[
                        dataSource()$sc1conf[
                            dataSource()$sc1conf$UI == filterCellKey2]$ID]]
                }else{
                    val2 <- c()
                }
            }
            minv <- floor(min(val, na.rm = TRUE))
            getM <- function(val){
                val <- max(val, na.rm = TRUE)
                if (val <= 1)
                    maxv <- round(val, digits = 3)
                if (val > 1 && val <= 10)
                    maxv <- round(val, digits = 1)
                if (val > 10)
                    maxv <- ceiling(val)
                return(maxv)
            }
            maxv <- getM(val)
            if(length(val2)>1){
                minv2 <- floor(min(val2, na.rm = TRUE))
                maxv2 <- getM(val2)
                tagList(
                    sliderInput(
                        NS(id, "filterCellVal"),
                        paste("Filter the cells by", input$filterCell),
                        min = minv,
                        max = maxv,
                        value = c(minv, maxv)
                    ),
                    sliderInput(
                        NS(id, "filterCellVal2"),
                        paste("Filter the cells by", filterCellKey2),
                        min = minv2,
                        max = maxv2,
                        value = c(minv2, maxv2)
                    )
                )
            }else{
                tagList(
                    sliderInput(
                        NS(id, "filterCellVal"),
                        "Filter the cells by value",
                        min = minv,
                        max = maxv,
                        value = c(minv, maxv)
                    )
                )
            }
        })
    }

#' @importFrom colourpicker colourInput
updateGeneExprDotPlotUI <-
    function(
        postfix = 1,
        id,
        input,
        output,
        session,
        plotX,
        height,
        ...,
        handlerFUN = plotsDownloadHandler,
        isInfoPlot = FALSE,
        dataSource = NULL) {
        # Reactive values to store zoom ranges
        ranges <- reactiveValues(x = NULL, y = NULL)
        output[[paste0("GeneExproup", postfix)]] <- renderPlot({
            addLimits(darkTheme(plotX()), x=ranges$x, y=ranges$y,
                      coord=input[[paste0('coord', postfix)]],
                      id=id, postfix=postfix, input=input)
        }, background=darkTheme(returnBG=TRUE))
        
        output[[paste0("GeneExproup.ui", postfix)]] <- renderUI({
            plotOutput(
                NS0(id, "GeneExproup", postfix),
                width = ifelse(
                    input[[paste0("GeneExproup.w", postfix)]]==
                        .globals$figWidth,
                    '100%', input[[paste0("GeneExproup.w", postfix)]]*72),
                height = ifelse(
                    input[[paste0("GeneExproup.h", postfix)]]==
                        .globals$figHeight, height,
                    input[[paste0("GeneExproup.h", postfix)]]*72),
                dblclick = NS0(id, 'GeneExproup.dbl', postfix),
                click = NS0(id, 'GeneExproup.clk', postfix),
                brush = brushOpts(NS0(id, 'GeneExproup.brush', postfix),
                                  resetOnNew = TRUE))
        })
        # Handle brush (drag selection) for zooming
        observeEvent(input[[paste0("GeneExproup.brush", postfix)]], {
            brush <- input[[paste0("GeneExproup.brush", postfix)]]
            if (!is.null(brush)) {
                ranges$x <- c(brush$xmin, brush$xmax)
                ranges$y <- c(brush$ymin, brush$ymax)
            }
        })
        nearest_element <- function(e){
            if(is.null(e)) return("undefined", 'undefined')
            p <- plotX()
            ggp1 <- ggplot_build(p)
            xrg <- ggp1$layout$panel_params[[1]]$x.range
            yrg <- ggp1$layout$panel_params[[1]]$y.range
            text_layer_id <- vapply(p$layers, function(.ele){
                is(.ele$geom, 'GeomTextRepel')
            }, FUN.VALUE = logical(1L))
            text_layers <- 
                do.call(rbind, 
                        lapply(ggp1$data[text_layer_id],
                               function(.ele){
                                   .ele[, c('x', 'y', 'label', 'size')]
                               }))
            points_layers <- 
                do.call(rbind,
                        lapply(ggp1$data[!text_layer_id],
                               function(.ele){
                                   .ele[, c('x', 'y', 'colour')]
                               }))
            text_layers$width <-
                grid::convertWidth(grid::stringWidth('W'),
                                   'npc', valueOnly = TRUE)*
                (nchar(as.character(text_layers$label))+2)
            text_layers$height <-
                grid::convertHeight(grid::stringHeight('H'),
                                    'npc', valueOnly = TRUE)*2.25
            nearestLabel <- (e$x - text_layers$x)^2 + (e$y - text_layers$y)^2
            ## TODO: fix it, the text are in bottom left
            inRange <- abs((e$x - text_layers$x)/diff(xrg))<=text_layers$width &
                abs((e$y - text_layers$y)/diff(yrg))<= text_layers$height
            nearestLabel <- nearestLabel==min(nearestLabel) & inRange
            if(any(nearestLabel)){
                nearestLabel <- 
                    c('text',
                      as.character(text_layers$label[which(nearestLabel)[1]]))
            }else{
                nearestLabel <- (e$x - points_layers$x)^2 +
                    (e$y - points_layers$y)^2
                nearestLabel <- c('colour',
                                  points_layers$colour[which.min(nearestLabel)[1]])
            }
            return(nearestLabel)
        }
        if(isInfoPlot && checkPrivilege(dataSource()$auth$privilege,
                                        dataSource()$dataset)){
            ## make the duplicate button available
            # updateActionButton(session,
            #                    paste0("CellInfodup", postfix),
            #                    disabled = FALSE)
            # updateActionButton(session,
            #                    paste0("CellInforename", postfix),
            #                    disabled = FALSE)
            # updateActionButton(session,
            #                    paste0("CellInfodel", postfix),
            #                    disabled = FALSE)
            session$sendCustomMessage("toggle_div",
                                      paste0(NS0(id, "CellInfodup", postfix),
                                             'container'))
            observeEvent(input[[paste0("CellInfodup",postfix)]], {
                ## duplicated current cell info
                if(input[[paste0('CellInfodname', postfix)]]!=""){
                    updated <- updateMetaData(
                        dataset = dataSource()$dataset,
                        inpConf = dataSource()$sc1conf,
                        inpMeta = dataSource()$sc1meta,
                        privilege = dataSource()$auth$privilege,
                        info = input[[paste0('CellInfodname', postfix)]],
                        oldvalue = input[[paste0('CellInfo', postfix)]],
                        newvalue = 'duplicate')
                    if(updated){
                        session$sendCustomMessage(
                            type='updateEditorStatus',
                            message = list(id=id, postfix=postfix))
                        updateSelectInput(
                            session,
                            inputId = paste0('CellInfo', postfix),
                            choices = c(dataSource()$sc1conf$UI,
                                        input[[paste0('CellInfodname',
                                                      postfix)]]),
                            selected = input[[paste0('CellInfodname', postfix)]]
                        )
                    }else{
                        adminMsg('Something wrong! Please check the name.',
                                 "error")
                    }
                }else{
                    adminMsg('New name is not provided!',
                             "error")
                }
            })
            observeEvent(input[[paste0("CellInforename",postfix)]], {
                ## rename current cell info
                if(input[[paste0('CellInfodname', postfix)]]!=""){
                    updated <- updateMetaData(
                        dataset = dataSource()$dataset,
                        inpConf = dataSource()$sc1conf,
                        inpMeta = dataSource()$sc1meta,
                        privilege = dataSource()$auth$privilege,
                        info = input[[paste0('CellInfodname', postfix)]],
                        oldvalue = input[[paste0('CellInfo', postfix)]],
                        newvalue = 'rename')
                    if(updated){
                        session$sendCustomMessage(
                            type='updateEditorStatus',
                            message = list(id=id, postfix=postfix))
                        updateSelectInput(
                            session,
                            inputId = paste0('CellInfo', postfix),
                            choices = c(dataSource()$sc1conf$UI[
                                dataSource()$sc1conf$UI!=
                                    input[[paste0('CellInfo', postfix)]]
                            ],
                            input[[paste0('CellInfodname', postfix)]]),
                            selected = input[[paste0('CellInfodname', postfix)]]
                        )
                    }else{
                        adminMsg('Something wrong! Please check the name.',
                                 "error")
                    }
                }else{
                    adminMsg('New name is not provided!',
                             "error")
                }
            })
            observeEvent(input[[paste0("CellInfodel",postfix)]], {
                showModal(modalDialog(
                    tagList(p(
                        "Are you sure you want to delete the cell info: ",
                        input[[paste0("CellInfo",postfix)]])
                    ),
                    title = paste("Delete info",
                                  input[[paste0("CellInfo",postfix)]]),
                    footer = tagList(
                        actionButton(
                            NS0(id, "CellInfoConfirmDelete",postfix),
                            "Delete"),
                        modalButton("Cancel")
                    )
                ))
            })
            observeEvent(input[[paste0("CellInfoConfirmDelete",  postfix)]], {
                removeModal()
                ## delete current cell info
                updated <- updateMetaData(
                    dataset = dataSource()$dataset,
                    inpConf = dataSource()$sc1conf,
                    inpMeta = dataSource()$sc1meta,
                    privilege = dataSource()$auth$privilege,
                    info = 'CellInfoConfirmDelete',
                    oldvalue = input[[paste0('CellInfo', postfix)]],
                    newvalue = 'delete')
                if(updated){
                    session$sendCustomMessage(
                        type='updateEditorStatus',
                        message = list(id=id, postfix=postfix))
                    choices <- c(dataSource()$sc1conf$UI[
                        dataSource()$sc1conf$UI!=
                            input[[paste0('CellInfo', postfix)]]
                    ])
                    updateSelectInput(
                        session,
                        inputId = paste0('CellInfo', postfix),
                        choices = choices,
                        selected = choices[1]
                    )
                }else{
                    adminMsg('Something wrong!',
                             "error")
                }
            })
            observeEvent(input[[paste0("GeneExproup.dbl", postfix)]],{
                evt <- input[[paste0("GeneExproup.dbl", postfix)]]
                if(!is.null(evt)){
                    session$sendCustomMessage(
                        type='placeGeneExproupInfoEditorBox',
                        message = id)
                    output[[paste0("GeneExproup.info", postfix)]] <- renderUI({
                        val <- nearest_element(evt)
                        fluidRow(
                            column(4,
                                   if(val[1]=='colour'){
                                       colourInput(
                                           NS0(id, "GeneExproup.upd",
                                               postfix),
                                           label = NULL,
                                           value = val[2]
                                       )
                                   }else{
                                       textInput(NS0(id, "GeneExproup.upd",
                                                     postfix),
                                                 label = NULL,
                                                 value = val[2])
                                   },
                                   div(
                                       style = "visibility:hidden;",
                                       textInput(NS0(id, 'GeneExproup.vtp', 
                                                     postfix),
                                                 label = NULL,
                                                 value = val[1]),
                                       textInput(NS0(id, "GeneExproup.old",
                                                     postfix),
                                                 label = NULL,
                                                 value = val[2]))),
                            column(4, actionButton(NS0(id, "GeneExproup.submit",
                                                       postfix),
                                                   label = 'update')),
                            column(4),
                            style=paste0('position:absolute; left:',
                                         input$current_mouseX,'px; top:',
                                         input$current_mouseY, 'px;')
                        )
                    })
                }
            })
            observeEvent(input[[paste0("GeneExproup.clk", postfix)]],{
                output[[paste0("GeneExproup.info", postfix)]] <- 
                    renderUI({div()})
            })
            observeEvent(input[[paste0("GeneExproup.submit",postfix)]], {
                if(!is.null(dataSource)){
                    updated <- updateMetaData(
                        dataset = dataSource()$dataset,
                        inpConf = dataSource()$sc1conf,
                        inpMeta = dataSource()$sc1meta,
                        privilege = dataSource()$auth$privilege,
                        info = input[[paste0('CellInfo', postfix)]],
                        oldvalue = input[[paste0("GeneExproup.old",
                                                 postfix)]],
                        newvalue = input[[paste0("GeneExproup.upd",
                                                 postfix)]])
                    if(updated){
                        session$sendCustomMessage(
                            type='updateEditorStatus',
                            message = list(id=id, postfix=postfix))
                    }
                    output[[paste0("GeneExproup.info", postfix)]] <- 
                        renderUI({div()})
                }
            })
        }else{## not info
            ## zoom in for ATAC
            ## zoom in for gene expression
            # observeEvent(input[[paste0("GeneExproup.clk", postfix)]],{
            #     ranges$x <- NULL
            #     ranges$y <- NULL
            # })
            observeEvent(input[[paste0("GeneExproup.dbl", postfix)]],{
                ranges$x <- NULL
                ranges$y <- NULL
                if(!is.null(input[[paste0('GeneExpext.info', postfix)]])){
                    updateTextInput(
                        inputId = paste0('GeneExpext.info', postfix),
                        value = '')
                }
            })
        }
        
        output[[paste0("GeneExproup.dwn", postfix)]] <-
            handlerFUN(
                input = input,
                postfix = postfix,
                plotX,
                ...)
        
    }

updateCellInfoPlot <-
    function(
        postfix = 1,
        id,
        input,
        output,
        session,
        dataSource) {
        cellInfoLabel <- paste0('CellInfo', postfix)
        cellInfoName <- paste0('CellInfoname', postfix)
        observeEvent(input[[cellInfoLabel]],{
            updateSelectInput(
                session,
                cellInfoName,
                "Cell info labels",
                choices = c(dataSource()$sc1conf$UI),
                selected = input[[cellInfoLabel]]
            )
        })
        updateSelectInput(
            session,
            cellInfoLabel,
            "Cell information:",
            choices = dataSource()$sc1conf$UI,
            selected = dataSource()$sc1def[[paste0("meta", postfix)]]
        )
        plotX <- reactive({
            scDRcell(
                inpConf=dataSource()$sc1conf,
                inpMeta=dataSource()$sc1meta,
                dimRedX=input$GeneExprdrX,
                dimRedY=input$GeneExprdrY,
                cellinfoID=input[[cellInfoLabel]],
                cellinfoName=input[[cellInfoName]],
                subsetCellKey=input$subsetCell,
                subsetCellVal=getSubsetCellVal(input),
                pointSize=input$GeneExprsiz,
                gradientCol=input[[paste0("CellInfocol", postfix)]],
                GeneExprDotOrd=input[[paste0("CellInfoord", postfix)]],
                labelsFontsize=input$GeneExprfsz,
                labelsFontFamily=input$GeneExprfml,
                plotAspectRatio=input$GeneExprasp,
                keepXYlables=input$GeneExprtxt,
                inplab=input[[paste0("CellInfolab", postfix)]],
                hideFilterCell=input[[paste0("CellInfohid", postfix)]],
                inpSlingshot = input[[paste0("CellInfoslingshot", postfix)]],
                slingshotFilename = file.path(
                    .globals$datafolder,
                    dataSource()$dataset,
                    .globals$filenames[["slingshot"]]
                ),
                inpShowEdge = input[[paste0("CellInfoedge", postfix)]],
                edgeFilename = file.path(
                    .globals$datafolder,
                    dataSource()$dataset,
                    .globals$filenames[["sc1edge"]]
                ),
                editorStatus = ifelse(
                    length(input[[paste0('editorStatus', postfix)]]),
                    input[[paste0('editorStatus', postfix)]], NA)
            )
        })
        updateGeneExprDotPlotUI(
            postfix,
            id,
            input,
            output,
            session,
            plotX,
            .globals$pList1[input$GeneExprpsz],
            dataSource()$dataset,
            input$GeneExprdrX,
            input$GeneExprdrY,
            input[[cellInfoLabel]],
            isInfoPlot = TRUE,
            dataSource = dataSource
        )
    }

expandGR <- function(coor, ext){
    if(is.na(ext[1])) ext <- 0
    start(coor) <- max(1, start(coor) - ext)
    end(coor) <- min(2^31-1, max(start(coor), end(coor) + ext))
    coor
}
#' @importFrom GenomicRanges strand start end `strand<-` `start<-` `end<-`
#' @importFrom IRanges shift
getCoordByGeneSymbol <- function(symbol, genes, links){
    if(!missing(genes)){
        genes0 <- genes[genes$gene_name %in% symbol | genes$gene_id %in% symbol]
        if(length(genes0)<1) return(NULL)
        strand(genes0) <- "*"
        coor <- range(genes0)[1]
        genes0 <- c(start(genes0), end(genes0))
        links0 <- links[links$gene %in% symbol]
        if(length(links)>0){
            if(length(links0$peak)){
                peaks0 <- do.call(rbind, strsplit(links0$peak, "-"))
                peaks0 <- as.numeric(peaks0[, c(2, 3)])
            }else{
                peaks0 <- cbind(start(links0), end(links0))
            }
            
        }else{
            peaks0 <- NULL
        }
        g0 <- range(c(genes0, peaks0))
        start(coor) <- max(1, g0[1] - round(diff(g0)/5))
        end(coor) <- g0[2] + round(diff(g0)/5)
    }else{
        links0 <- links[links$gene %in% symbol]
        if(length(links0)>0){
            if(length(links$pvalue)==length(links0)){
                coor <- links0[order(links0$pvalue)]
            }else{
                coor <- links0
            }
            if(length(coor$peak)!=length(coor)){
                coor$peak <-paste(as.character(seqnames(coor)),
                                  start(coor),
                                  end(coor),
                                  sep='-')
            }
        }else{
            coor <- NULL
        }
    }
    coor
}
getGeneSymbolByCoord <- function(coor, links){
    if(length(links$peak)!=length(links$peak)){
        links$peak <- paste(as.character(seqnames(links)),
                            start(links), end(links),
                            sep='-')
    }
    links0 <- links[links$peak %in% coor]
    if(length(links0)>0){
        if(length(links0$pvalue)==length(links0)){
            return(links0[order(links0$pvalue)]$gene)
        }else{
            return(links0$gene)
        }
    }else{
        return(NULL)
    }
}

updateAccCoordInputs <- function(session, coordLabel, coor){
    if(is(coor, "GRanges")){
        updateTextInput(
            session,
            coordLabel,
            value = as(coor, "character"))
        updateSliderInput(
            session,
            'regionselector',
            value = c(start(coor), end(coor)),
            step = max(1, round(width(coor)/100)),
            min = start(coor),
            max = end(coor)
        )
    }
}

updateGeneAccPlot <-
    function(
        postfix = 1,
        genePostfix = 2,
        optCrt,
        id,
        input,
        output,
        session,
        dataSource){
        GeneNameLabel <- paste0('GeneName', genePostfix)
        coordLabel <- paste0('coord', postfix)
        genes <- readData("sc1anno", dataSource()$dataset)
        links <- readData("sc1link", dataSource()$dataset)
        observeEvent(input[[GeneNameLabel]], {
            coor <- getCoordByGeneSymbol(input[[GeneNameLabel]], genes, links)
            updateAccCoordInputs(session, coordLabel, coor)
        })
        getCoor <- function(){
            coor <- GRanges()
            tryCatch({coor <- GRanges(input[[coordLabel]])},
                     error=function(e){
                         showNotification(
                             as.character(e),
                             duration = 5,
                             type = 'warning'
                         )
                     })
            return(coor)
        }
        observeEvent(input$zoomin, {
            coor <- getCoor()
            if(length(coor)){
                updateAccCoordInputs(session, coordLabel,
                                     expandGR(coor, -width(coor)/4))
            }
            
        })
        observeEvent(input$zoomout, {
            coor <- getCoor()
            if(length(coor)){
                updateAccCoordInputs(session, coordLabel,
                                     expandGR(coor, width(coor)*2))
            }
        })
        observeEvent(input$moveleft, {
            coor <- getCoor()
            if(length(coor)){
                updateAccCoordInputs(session, coordLabel,
                                     shift(coor, -width(coor)/2))
            }
        })
        observeEvent(input$moveright, {
            coor <- getCoor()
            if(length(coor)){
                updateAccCoordInputs(session, coordLabel,
                                     shift(coor, width(coor)/2))
            }
        })
        observeEvent(input$regionsubmit, {
            if(grepl(":", input[[coordLabel]])){
                coor <- getCoor()
                change <- FALSE
                if(start(coor) != input$regionselector[1]){
                    start(coor) <- input$regionselector[1]
                    change <- TRUE
                }
                if(end(coor) != input$regionselector[2]){
                    end(coor) <- input$regionselector[2]
                    change <- TRUE
                }
                if(change){
                    updateAccCoordInputs(session, coordLabel, coor)
                }
            }
        })
        
        plotX <- reactive({
            scDRatac(
                inpConf = dataSource()$sc1conf,
                inpMeta = dataSource()$sc1meta,
                dimRedX = input$GeneExprdrX,
                dimRedY = input$GeneExprdrY,
                gene1 = input[[GeneNameLabel]],
                coord = input[[coordLabel]],
                subsetCellKey = input$subsetCell,
                subsetCellVal = 
                    input[[paste0("subsetCellVal", input$subsetCell)]],
                dataset = dataSource()$dataset,
                geneIdMap = dataSource()$sc1gene,
                pointSize = input$GeneExprsiz,
                gradientCol = input[[paste0("GeneExprcol", postfix)]],
                labelsFontsize = input$GeneExprfsz,
                labelsFontFamily=input$GeneExprfml,
                plotAspectRatio = input$GeneExprasp,
                keepXYlables = input$GeneExprtxt)
        })
        updateGeneExprDotPlotUI(
            postfix,
            id,
            input,
            output,
            session,
            plotX,
            paste0((length(input$subsetCellVal)+4)*150, "px"),
            dataSource()$dataset,
            input$GeneExprdrX,
            input$GeneExprdrY,
            input[[GeneNameLabel]]
        )
    }
        
updateGeneExprPlot <-
    function(
        postfix = 1,
        selectedGene,
        optCrt,
        id,
        input,
        output,
        session,
        dataSource) {
        GeneNameLabel <- paste0('GeneName', postfix)
        updateSelectizeInput(
            session,
            GeneNameLabel,
            choices = sort(names(dataSource()$sc1gene)),
            server = TRUE,
            selected = selectedGene,
            options = list(
                maxOptions = .globals$maxNumGene,
                create = TRUE,
                persist = TRUE,
                render = I(optCrt)
            )
        )
        observeEvent(input[[GeneNameLabel]], {
            if(isTRUE(input[[GeneNameLabel]]!=dataSource()$sc1def$gene1 &&
                      input[[GeneNameLabel]]!=dataSource()$sc1def$gene2 &&
                      input[[GeneNameLabel]]!="")){
                updateSearchTable(input[[GeneNameLabel]])
            }
        }, ignoreInit=TRUE)
        ### plots
        plotX <- reactive({
            scDRgene(
                inpConf=dataSource()$sc1conf,
                inpMeta=dataSource()$sc1meta,
                dimRedX=input$GeneExprdrX,
                dimRedY=input$GeneExprdrY,
                gene1=input[[GeneNameLabel]],
                subsetCellKey=input$subsetCell,
                subsetCellVal=getSubsetCellVal(input),
                dataset=dataSource()$dataset,
                geneIdMap=dataSource()$sc1gene,
                pointSize=input$GeneExprsiz,
                gradientCol=input[[paste0("GeneExprcol", postfix)]],
                GeneExprDotOrd=input[[paste0("GeneExprord", postfix)]],
                labelsFontsize=input$GeneExprfsz,
                labelsFontFamily=input$GeneExprfml,
                plotAspectRatio=input$GeneExprasp,
                keepXYlables=input$GeneExprtxt,
                inpPlt=input[[paste0("GeneExprtype", postfix)]],
                inpXlim=if (input[[paste0("GeneExprxlimb", postfix)]] %% 2 == 0)
                    0
                else
                    input[[paste0("GeneExprxlim", postfix)]],
                inpColRange =
                    if (input[[paste0("GeneExprrgb", postfix)]] %% 2 == 0)
                        0
                    else
                        input[[paste0("GeneExprrg", postfix)]],
                hideFilterCell = input[[paste0("GeneExprhid", postfix)]]
            )
        })
        updateGeneExprDotPlotUI(
            postfix,
            id,
            input,
            output,
            session,
            plotX,
            .globals$pList1[input$GeneExprpsz],
            dataSource()$dataset,
            input$GeneExprdrX,
            input$GeneExprdrY,
            input[[GeneNameLabel]]
        )
    }

updateSubsetGeneExprPlot <-
    function(
        postfix = 1,
        subgrp,
        optCrt,
        inpColRange,
        id,
        input,
        output,
        session,
        dataSource) {
        GeneNameLabel <- paste0("GeneExprsub", postfix, "b")
        ### sub region title
        output[[paste0("subPlotTitle", postfix)]] <-
            renderUI({
                h4(paste("Gene", dataSource()$terms['expression']))
            })
        ### select which cells to show
        output[[paste0("GeneExprgrp.ui", postfix)]] <- renderUI({
            subgrp <- subgrp(dataSource, input)
            selected <- ifelse(
                postfix == 1,
                subgrp[1],
                ifelse(length(subgrp) > 1, subgrp[2], subgrp[1]))
            checkboxGroupInput(
                NS(id, GeneNameLabel),
                "Select which cells to show",
                inline = TRUE,
                choices = subgrp,
                selected = selected
            )
        })
        ### plots
        plotX <- reactive({
            scDRgene(
                inpConf=dataSource()$sc1conf,
                inpMeta=dataSource()$sc1meta,
                dimRedX=input$GeneExprdrX,
                dimRedY=input$GeneExprdrY,
                gene1=input$GeneName,
                subsetCellKey=c(input$CellInfo, input$subsetCell),
                subsetCellVal=getSubsetCellVal(
                    input, list(input[[GeneNameLabel]]), input$CellInfo),
                dataset=dataSource()$dataset,
                geneIdMap=dataSource()$sc1gene,
                pointSize=input$GeneExprsiz,
                gradientCol=input[[paste0("GeneExprcol", postfix)]],
                GeneExprDotOrd=input[[paste0("GeneExprord", postfix)]],
                labelsFontsize=input$GeneExprfsz,
                labelsFontFamily=input$GeneExprfml,
                plotAspectRatio=input$GeneExprasp,
                keepXYlables=input$GeneExprtxt,
                inpPlt=input[[paste0("GeneExprtype", postfix)]],
                inpXlim=if (input[[paste0("GeneExprxlimb", postfix)]] %% 2 == 0)
                    0
                else
                    input[[paste0("GeneExprxlim", postfix)]],
                inpColRange = 
                    if (input[[paste0("GeneExprrgb", postfix)]] %% 2 == 0) {
                        inpColRange()
                    } else{
                        input[[paste0("GeneExprrg", postfix)]]
                    },
                valueFilterKey = input$filterCell,
                valueFilterCutoff = input$filterCellVal,
                valueFilterCutoff2 = input$filterCellVal2,
                hideFilterCell = input[[paste0("GeneExprhid", postfix)]]
            )
        })
        updateGeneExprDotPlotUI(
            postfix,
            id,
            input,
            output,
            session,
            plotX,
            .globals$pList1[input$GeneExprpsz],
            dataSource()$dataset,
            input$GeneExprdrX,
            input$GeneExprdrY,
            input$GeneName,
            input$CellInfo
        )
    }

# sub module related
updateSubModulePlotUI <-
    function(
        postfix = 1,
        pid,
        id,
        input,
        output,
        session,
        interactive,
        plotX,
        height,
        ...) {
        if (isTRUE(interactive)) {
            output[[paste0("GeneExproup", postfix)]] <-
                renderPlotly({
                    ggplotly(plotX()) %>% event_register("plotly_click")
                })
            output[[paste0("GeneExproup.ui", postfix)]] <- renderUI({
                plotlyOutput(
                    NS0(NS(pid, id), "GeneExproup", postfix),
                    height = height)
            })
        } else{
            output[[paste0("GeneExproup", postfix)]] <- renderPlot({
                plotX()
            })
            output[[paste0("GeneExproup.ui", postfix)]] <- renderUI({
                plotOutput(
                    NS0(NS(pid, id), "GeneExproup", postfix),
                    width = ifelse(
                        input[[paste0("GeneExproup.w", postfix)]]==
                            .globals$figWidth,
                        '100%', input[[paste0("GeneExproup.w", postfix)]]*72),
                    height = ifelse(
                        input[[paste0("GeneExproup.h", postfix)]]==
                            .globals$figHeight, height,
                        input[[paste0("GeneExproup.h", postfix)]]*72))
            })
            output[[paste0("GeneExproup.dwn", postfix)]] <-
                plotsDownloadHandler(
                    input = input,
                    postfix = postfix,
                    plotX,
                    ...)
        }
    }

subModuleMenuObservor <- function(
        id,
        input,
        p_session,
        dataSource,
        observeEvtList) {
    observeEvent(input$close, {
        updateTextInput(p_session, "removePlotModule", value = id)
    })
    observeEvent(input$movedown, {
        updateTextInput(p_session, "movedownPlotModule", value = id)
    })
    observeEvent(input$moveup, {
        updateTextInput(p_session, "moveupPlotModule", value = id)
    })
    observeEvent(input$resize, {
        updateTextInput(p_session, "resizePlotModule", value = id)
    })
    observeEvent(input$CellInfosubgrp1, {
        updateTextInput(p_session, "changeSubsetContext",
                        value = paste(id, input$CellInfosubgrp1, sep='___'))
    })
    #if subset group B is selected but the B have no value
    # send click message to the TogT B group
    observeEvent(input$CellInfosubgrp1, {
        if(length(p_session$input[[paste0("subsetCell",
                                          input$CellInfosubgrp1)]])==0){
            p_session$sendCustomMessage(
                'click_subset_btn',
                paste0("explorer-subsetTogT", input$CellInfosubgrp1))
        }
    })
    if (is.null(p_session$userData$defaults[[dataSource()$dataset]][[id]]))
        p_session$userData$defaults[[dataSource()$dataset]][[id]] <-
        list()
    lapply(observeEvtList, function(evt) {
        observeEvent(input[[evt]], {
            p_session$userData$defaults[[dataSource()$dataset]][[id]][[evt]] <-
                input[[evt]]
        })
    })
}

## plots related
getRatio <- function(ggData) {
    ## help function
    return((
        max(ggData$X, na.rm = TRUE) - min(ggData$X, na.rm = TRUE)) / (
            max(ggData$Y, na.rm = TRUE) - min(ggData$Y, na.rm = TRUE)))
}
orderGeneExpr <- function(ggData, GeneExprDotOrd, coln) {
    if (GeneExprDotOrd == "Max-1st") {
        ggData <- ggData[order(ggData[, coln, with = FALSE])]
    } else if (GeneExprDotOrd == "Min-1st") {
        ggData <- ggData[order(-ggData[, coln, with = FALSE])]
    } else if (GeneExprDotOrd == "Random") {
        ggData <- ggData[sample(nrow(ggData))]
    }
    return(ggData)
}
subGrp <- function(ggData, ui_key, grpVal, config) {
    for(k in ui_key){
        if (k != "N/A" && length(grpVal[[k]])) {
            ggData <- ggData[
                ggData[[config[
                    config$UI == k]$ID]] %in% grpVal[[k]],
                , drop = FALSE]
        }
    }
    return(ggData)
}
subsetData <- function(ggData, subKey, subValue) {
    if (length(subValue) != 0 & length(subValue) !=
        nlevels(ggData[, subKey, with = FALSE])) {
        ggData <- ggData[ggData[, subKey, with = FALSE] %in% subValue]
    }
    return(ggData)
}
relevelData <- function(ggData, coln) {
    ggLvl <- levels(ggData[[coln]])
    if (length(ggLvl)) {
        ggLvl <- ggLvl[ggLvl %in% unique(ggData[[coln]])]
        ggLvl <- sortLevels(ggLvl)
        ggData[[coln]] <- factor(ggData[[coln]], levels = ggLvl)
    }
    return(ggData)
}
extractGrpColor <- function(config, ui_key) {
    ggCol <- strsplit(config[config$UI == ui_key]$fCL, "\\|")[[1]]
    names(ggCol) <-
        strsplit(config[config$UI == ui_key]$fID, "\\|")[[1]]
    return(ggCol)
}
relevelCol <- function(inpConf, ui_key, ggData, coln) {
    ggCol <- NULL
    if (!is.na(inpConf[inpConf$UI == ui_key[1]]$fCL)) {
        ggCol <- extractGrpColor(inpConf, ui_key)
        ggCol <- ggCol[levels(ggData[[coln]])]
    }
    return(ggCol)
}
fixCoord <- function(ggOut, aspectRatio, ratio) {
    if (aspectRatio == "Square") {
        ggOut <- ggOut + coord_fixed(ratio = ratio)
    } else if (aspectRatio == "Fixed") {
        ggOut <- ggOut + coord_fixed()
    }
    return(ggOut)
}
labelBackgroundCells <- function(
        ggOut,
        ggData,
        pointSize,
        color = "snow2",
        shape = 16,
        hide = FALSE) {
    if(hide) return(ggOut)
    ggOut + geom_point(
        data = ggData,
        color = color,
        size = pointSize,
        shape = shape
    )
}
pointPlot <- function(
        ggOut,
        pointSize,
        fontSize = 24,
        labelsFontFamily = 'Helvetica',
        dimRedX,
        dimRedY,
        keepXYlables,
        shape = 16) {
    ggOut + geom_point(size = pointSize, shape = 16) +
        xlab(dimRedX) + ylab(dimRedY) +
        sctheme(base_size = fontSize,
                family = labelsFontFamily,
                XYval = keepXYlables)
}
ggXYplot <- function(ggData) {
    ggplot(ggData, aes(
        .data[["X"]], .data[["Y"]],
        color = .data[["val"]] #,customdata = data[["sampleID"]]
        ))
}
getTotalNumber <- function(nGrid = 16, nPad = 2) {
    return(nGrid + nPad * 2)
}
getCoexpCol <- function(
        colorPairs,
        nGrid = 16,
        nPad = 2) {
    cInp <- strsplit(colorPairs, "; ")[[1]]
    if (cInp[1] == "Red (Gene1)") {
        c10 <- c(255, 0, 0)
    } else if (cInp[1] == "Orange (Gene1)") {
        c10 <- c(255, 140, 0)
    } else {
        c10 <- c(0, 255, 0)
    }
    if (length(cInp) > 1) {
        if (cInp[2] == "Green (Gene2)") {
            c01 <- c(0, 255, 0)
        } else {
            c01 <- c(0, 0, 255)
        }
    } else{
        c01 <- c(0, 0, 255)
    }
    
    c00 <- c(217, 217, 217)
    c11 <- c10 + c01
    nTot <- getTotalNumber(nGrid, nPad)
    gg <- data.table(
        v1 = rep(0:nTot, nTot + 1),
        v2 = sort(rep(0:nTot, nTot + 1)))
    gg$vv1 <- gg$v1 - nPad
    gg[gg$vv1 < 0]$vv1 <- 0
    gg[gg$vv1 > nGrid]$vv1 <- nGrid
    gg$vv2 <- gg$v2 - nPad
    gg[gg$vv2 < 0]$vv2 <- 0
    gg[gg$vv2 > nGrid]$vv2 <- nGrid
    gg$cR <-
        bilinear(gg$vv1, gg$vv2, nGrid, c00[1], c10[1], c01[1], c11[1])
    gg$cG <-
        bilinear(gg$vv1, gg$vv2, nGrid, c00[2], c10[2], c01[2], c11[2])
    gg$cB <-
        bilinear(gg$vv1, gg$vv2, nGrid, c00[3], c10[3], c01[3], c11[3])
    gg$cMix <- rgb(gg$cR, gg$cG, gg$cB, maxColorValue = 255)
    gg <- gg[, c("v1", "v2", "cMix")]
    return(gg)
}
getCoexpVal <- function(ggData, dataset, geneIdMap, gene1, gene2) {
    ggData$val1 <- read_exprs(dataset, geneIdMap[gene1], valueOnly = TRUE)
    ggData$val2 <-
        read_exprs(dataset, geneIdMap[gene2], valueOnly = TRUE)
    ggData[ggData$val1 < 0]$val1 <- 0
    ggData[ggData$val2 < 0]$val2 <- 0
    return(ggData)
}
getFilterKey2 <- function(coln){
    paste0(coln, '2')
}
cbindFilterValues <-
    function(
        ggData,
        config,
        meta,
        coln,
        geneIdMap,
        dataset,
        valueFilterKey,
        valueFilterCutoff,
        valueFilterCutoff2) {
        if (!missing(valueFilterKey) && !missing(valueFilterCutoff)) {
            if (valueFilterKey %in% config$UI) {
                ggData <-
                    cbind(
                        ggData,
                        subValue =
                            meta[, config[
                                config$UI == valueFilterKey]$ID,
                                with = FALSE])
                colnames(ggData)[ncol(ggData)] <- coln
                if(!missing(valueFilterCutoff2)){
                    valueFilterKey2 <- ifelse(grepl('1$', valueFilterKey),
                                              sub('1$', '2', valueFilterKey),
                                              sub('2$', '1', valueFilterKey))
                    ggData <-
                        cbind(
                            ggData,
                            subValue =
                                meta[, config[
                                    config$UI == valueFilterKey2]$ID,
                                    with = FALSE])
                    colnames(ggData)[ncol(ggData)] <- getFilterKey2(coln)
                }
            } else if (valueFilterKey %in% names(geneIdMap)) {
                subValue <- read_exprs(
                    dataset,
                    geneIdMap[valueFilterKey],
                    valueOnly = TRUE)
                if (any(subValue < 0))
                    subValue[subValue < 0] <- 0
                ggData <- cbind(ggData, subValue = subValue)
                colnames(ggData)[ncol(ggData)] <- coln
            }
        }
        return(ggData)
    }
# get the subsetCellVals
getSubsetCellVal <- function(input, extralist, extralistname, group=""){
    if(!missing(extralist)) stopifnot(is.list(extralist)&&length(extralist)==1)
    subsetCell <- input[[paste0("subsetCell", group)]]
    subsetCell <- subsetCell[subsetCell!="N/A"]
    names(subsetCell) <- subsetCell
    subsetCell <- lapply(subsetCell, function(subid){
        input[[paste0("subsetCellVal", subid, group)]]
    })
    if(!missing(extralist) && !missing(extralistname)){
        names(extralist) <- extralistname
        if(extralistname %in% names(subsetCell)){
            subsetCell[[extralistname]] <- 
                intersect(subsetCell[[extralistname]], extralist[[1]])
        }else{
            subsetCell <- c(extralist, subsetCell)
        }
        
    }
    return(subsetCell)
}
# check the pairs of subsetCellKey and subsetCellVals
namedSubsetCellVals <- function(subsetCellKey, subsetCellVal){
    subsetCellKey <- subsetCellKey[subsetCellKey!="N/A"]
    if(length(subsetCellKey)==1 && !is.list(subsetCellVal)){
        subsetCellVal <- list(subsetCellVal)
        names(subsetCellVal) <- subsetCellKey
        return(subsetCellVal)
    }
    stopifnot(is.list(subsetCellVal))
    stopifnot(length(names(subsetCellVal))==length(subsetCellVal))
    subsetCellVal[unique(subsetCellKey)]
}

filterCells <- function(
        ggData,
        subsetCellKey,
        subsetCellVal,
        valueFilterKey,
        valueFilterCutoff,
        valueFilterCutoff2,
        inpConf) {
    keep <- rep(TRUE, nrow(ggData))
    if (!missing(subsetCellKey) && !missing(subsetCellVal)) {
        if(length(subsetCellKey)==1){
            if(!is.list(subsetCellVal)){
                subsetCellVal <- list(subsetCellVal)
                names(subsetCellVal) <- subsetCellKey
            }
        }
        if(length(names(subsetCellVal))==0){
            names(subsetCellVal) <- subsetCellKey
        }
        for(skey in subsetCellKey){
            sid <- inpConf[inpConf$UI == skey]$ID
            if(length(sid)==0) next
            if(length(colnames(ggData))==0) next
            if(!sid %in% colnames(ggData)) next
            if (length(subsetCellVal[[skey]]) != nlevels(ggData[[sid]])) {
                keep <- keep & ggData[[sid]] %in% subsetCellVal[[skey]]
            }
        }
    }
    if (!missing(valueFilterKey) && !missing(valueFilterCutoff)) {
        if (length(valueFilterCutoff) != 0) {
            keep <- keep & ggData[[valueFilterKey]] >= valueFilterCutoff[1]
            if(length(valueFilterCutoff)>1){
                keep <- keep & ggData[[valueFilterKey]] <= valueFilterCutoff[2]
            }
        }
        if (!missing(valueFilterCutoff2)) {
            if(length(valueFilterCutoff2)>0){
                valueFilterKey2 <- getFilterKey2(valueFilterKey)
                keep <- keep & ggData[[valueFilterKey2]] >= valueFilterCutoff2[1]
                if(length(valueFilterCutoff2)>1){
                    keep <- keep & ggData[[valueFilterKey2]] <= valueFilterCutoff2[2]
                }
            }
        }
    }
    return(keep)
}

updateRankList <- function(input, output, dataSource, uid, pid, input_id){
    observeEvent(input[[uid]], {
        output[[pid]] <- renderUI(rank_list(
            text = "Drag, drop and re-order the following items:",
            labels = sort(as.character(unlist(unique(dataSource()$sc1meta[
                , dataSource()$sc1conf[
                    dataSource()$sc1conf$UI == input[[uid]]]$ID, with = FALSE])))),
            input_id = input_id
        ))
    })
}

cellInfoCellInfoUI <- function(id) {
    tabPanel(
        value = id,
        HTML("CellInfo vs CellInfo"),
        h4("Cell information vs cell information on dimension reduction"),
        "In this tab, users can visualise two cell informations side-by-side ",
        "on low-dimensional representions.",
        br(),
        br(),
        fluidRow(
            column(3, dimensionReductionUI(id)),
            column(5, subsetCellByInfoUI(id)),
            column(4, graphicsControlUI(id))
        ),
        fluidRow(
            column(
                6,
                style = "border-right: 2px solid black",
                h4("Cell information 1"),
                fluidRow(
                    column(6, cellInfoUI(id, 1)),
                    column(6, cellInfoPlotControlUI(id, 1))),
                geneExprDotPlotUI(id, 1, editor=TRUE)
            ),
            column(
                6,
                h4("Cell information 2"),
                fluidRow(
                    column(6, cellInfoUI(id, 2)),
                    column(6, cellInfoPlotControlUI(id, 2))),
                geneExprDotPlotUI(id, 2, editor=TRUE)
            )
        )
    )
}
cellInfoCellInfoServer <- function(id, dataSource, optCrt) {
    moduleServer(id, function(input, output, session) {
        ## input column 1
        ### Dimension Reduction
        updateDimRedSelInputPair(session, input, dataSource)
        ## input column 2
        updateSubsetCellUI(id, input, output, session, dataSource)
        
        ## plot region
        ### dropdown list
        updateCellInfoPlot(1, id, input, output, session, dataSource)
        updateCellInfoPlot(2, id, input, output, session, dataSource)
    })
}

scVlnUI <- function(id, postfix = 1, subgrp=.globals$subsetgroup[1]) {
    subModuleContainerUI(
        id,
        mainSelectUI = tagList(
            selectInput(
                NS(id, "CellInfoX"),
                "X-axis:",
                choices = NULL,
                width = "100px"
            ),
            selectInput(
                NS(id, "CellInfoY"),
                "Y-axis:",
                choices = NULL,
                width = "100px"
            )
        ),
        menuUI = contextMenuViolinUI(id, group = subgrp),
        contentUI = geneExprDotPlotUI(id, postfix = postfix)
    )
}
scVlnServer <- function(
        pid,
        id,
        dataSource,
        optCrt,
        p_input,
        p_session,
        interactive,
        postfix = 1) {
    moduleServer(id, function(input, output, session) {
        if (is.null(
            p_session$userData$defaults[[dataSource()$dataset]][[id]])) {
            defaults <- list(
                CellInfoX = dataSource()$sc1def$grp2,
                CellInfoY = dataSource()$sc1def$grp1
            )
        } else{
            defaults <-
                p_session$userData$defaults[[dataSource()$dataset]][[id]]
        }
        ## title
        updateSelectInput(
            session,
            "CellInfoX",
            choices = getGroupUI(dataSource),
            selected = defaults$CellInfoX
        )
        updateSelectizeInput(
            session,
            "CellInfoY",
            server = TRUE,
            choices = c(
                getNonGroupUI(dataSource),
                sort(names(
                    dataSource()$sc1gene
                ))),
            selected =  defaults$CellInfoY,
            options = list(
                maxOptions =
                    length(getNonGroupUI(dataSource)) + 3,
                create = TRUE,
                persist = TRUE,
                render = I(optCrt)
            )
        )
        subModuleMenuObservor(
            id,
            input,
            p_session,
            dataSource,
            c("CellInfoX", "CellInfoY", "plottyp", "plotpts")
        )
        updateRankList(
            input, output, dataSource, "CellInfoX", "plotXord",
            NS(NS(pid, id), "cellinfoXorder"))
        ## plot
        plot1 <- reactive({
            scVioBox(
                inpConf = dataSource()$sc1conf,
                inpMeta = dataSource()$sc1meta,
                infoX = input$CellInfoX,
                infoY = input$CellInfoY,
                subsetCellKey=p_input[[paste0("subsetCell",
                                              input[[paste0("CellInfosubgrp",
                                                            postfix)]])]],
                subsetCellVal=
                    getSubsetCellVal(p_input,
                                     group=input[[paste0("CellInfosubgrp",
                                                         postfix)]]),
                valueFilterKey = p_input$filterCell,
                valueFilterCutoff = p_input$filterCellVal,
                valueFilterCutoff2 = p_input$filterCellVal2,
                dataset = dataSource()$dataset,
                geneIdMap = dataSource()$sc1gene,
                inptyp = input$plottyp,
                inppts = input$plotpts,
                pointSize = p_input$GeneExprsiz,
                labelsFontsize = p_input$GeneExprfsz,
                labelsFontFamily=p_input$GeneExprfml,
                reorder = input$plotord,
                orderX = input$cellinfoXorder,
                addnoise = input$addnoise
            )
        })
        updateSubModulePlotUI(
            postfix,
            pid,
            id,
            input,
            output,
            session,
            interactive,
            plot1,
            .globals$pList1[p_input$GeneExprpsz],
            dataSource()$dataset,
            input$plottyp,
            input$CellInfoX
        )
    })
}

#' @importFrom DT DTOutput
webstatsUI <- function (id) {
    ns <- NS(id)
    fluidRow(column(
        width = 10,
        offset = 1,
        textAreaInput(
            ns('description'),
            label = 'Homepage description',
            value = paste(readLines(file.path(
                .globals$app_path, "doc.txt")), collapse = "\n"),
            width="100%"
        ),
        actionButton(ns('update'), "update web description"),
        actionButton(ns('restart'), 'force restart the App'),
        fluidRow(column(width = 3, DTOutput(ns(
            "summary"
        ))),
        column(
            width = 9, plotOutput(ns("distPlot"), height = "300px")
        )),
        DTOutput(ns("issues")),
        DTOutput(ns("counter")),
        DTOutput(ns("search"))
    ))
}
#' @importFrom DT renderDT JS
#' @importFrom S4Vectors SimpleList DataFrame
#' @importFrom jsonlite parse_json
#' @importFrom data.table as.data.table .SD
#' @importFrom reshape2 melt
#' @importFrom ggplot2 geom_bar position_dodge geom_text aes labs theme_minimal
webstatsServer <- function(id) {
    moduleServer(id, function(input, output, session) {
        ## web description
        observeEvent(input$update, {
            writeLines(
                input$description,
                file.path(.globals$app_path, "doc.txt"))
        })
        
        ## force restart App
        observeEvent(input$restart, {
            writeLines(
                date(),
                file.path(.globals$app_path, "restart.txt")
            )
            adminMsg("reloading the session now!", type = "error")
            Sys.sleep(2)
            session$reload()
        })
        
        ## stats
        cache <- file.path(.globals$app_path, "www", "cache.rds")
        
        freegeoip <- function(ip) {
            if (file.exists(cache)) {
                ip_rds <- readRDS(cache)
            } else{
                ip_rds <- list()
            }
            ip_ <- ip_rds[ip[ip %in% names(ip_rds)]]
            ip <- ip[!ip %in% names(ip_rds)]
            ip_rds_addition <- lapply(ip, function(.ip) {
                url <- paste0("http://ip-api.com/json/", .ip)
                json <- tryCatch({
                    readLines(url, warn = FALSE)
                }, error = function(.e){
                    message(.e)
                    NULL
                })
                if(!is.null(json)){
                    parse_json(json)
                }else{
                    NULL
                }
            })
            names(ip_rds_addition) <- ip
            ip_rds_addition <- ip_rds_addition[!is.null(ip_rds_addition)]
            ip_rds <- c(ip_rds, ip_rds_addition)
            saveRDS(ip_rds, file = cache)
            ret <- c(ip_, ip_rds_addition)
            return(ret)
        }
        data_1 <- reactiveValues()
        observe({
            data_1$df <- reactive(({
                listVisitors()
            }))
        })
        data <- reactive(data_1$df())
        
        ips <- reactive({
            req(data())
            dat <- isolate(data())
            d <- unique(dat$ip)
            d <- d[!is.na(d)]
            d <- d[!is.null(d)]
            ip_info <- freegeoip(d)
            coln <- c(
                "status",
                "country",
                "countryCode",
                "region",
                "regionName",
                "city",
                "zip",
                "lat",
                "lon",
                "timezone",
                "isp",
                "org",
                "as",
                "query"
            )
            ip_info <- lapply(ip_info, function(.ele) {
                unlist(.ele)[coln]
            })
            ip_info <- do.call(rbind, ip_info)
            colnames(ip_info) <- coln
            merge(
                dat,
                ip_info,
                by.x = "ip",
                by.y = "query",
                all.x = TRUE
            )
        })
        summaryTable <- reactive({
            if(!tableExists(.globals$counterTableName)){
                req(ips())
                tab <- isolate(ips()[, c("ip", "date")])
                tab[, "date"] <- as.Date(tab[, "date"])
                tab[, "month"] <-
                    paste(format(tab$date, "%Y"), months(tab$date), sep = "-")
                tab <- as.data.table(tab)
                dat <- tab[, {
                    list(
                        total = length(.SD$ip),
                        uniqueIP = length(unique(.SD$ip)))
                }, by = 'month']
                dat <- reshape2::melt(dat, id.vars = "month")
            }else{
                dat = listVisitors(summary = TRUE)
                colnames(dat) <- c('month', 'total', 'uniqueIP')
                dat <- reshape2::melt(dat, id.vars = "month")
            }
        })
        output$distPlot <- renderPlot({
            req(summaryTable)
            dat <- summaryTable()
            ggplot(
                dat,
                aes(
                    x = .data[["month"]],
                    y = .data[["value"]],
                    fill = .data[["variable"]]
                )) +
                geom_bar(stat = "identity", position = position_dodge()) +
                geom_text(
                    aes(label = .data[["value"]]),
                    vjust = 1.6,
                    color = "black",
                    position = position_dodge(1 / 12)
                ) +
                #scale_fill_brewer(palette="Paired") +
                labs(
                    x = "",
                    y = "conts",
                    fill = "Legend") +
                theme_minimal()
        })
        output$summary <- renderDT({
            req(summaryTable)
            dat <- as.data.table(summaryTable())
            dat <- dat[, {
                list(total = sum(.SD$value))
            }, by = 'variable']
            levels(dat$variable) <-
                c(levels(dat$variable), "database counts")
            dat <-
                rbind(
                    dat,
                    list(
                        "database counts",
                        length(getDataSets(
                            privilege = "all"))))
            dat
        }, options = list(dom = 't'), rownames = FALSE)
        output$counter <- renderDT({
            if(!tableExists(.globals$counterTableName)){
                req(ips())
                ips()
            }else{
                listVisitors()
            }
        },
        extensions = 'Buttons',
        options = list(
            dom = 'Brtip',
            lengthMenu = list(c(10, 25, 50, -1), c(10, 25, 50, "All")),
            buttons = c('pageLength', 'copy', 'csv', 'excel', 'pdf', 'print')
        ))
        updateCommentsTable <- function(){
            output$issues <- renderDT({
                data_to_display <- 
                    listComments(page_size=.globals$totalComments,
                                 full=TRUE, all=TRUE)
                if(nrow(data_to_display)<1){
                    data_to_display
                }else{
                    data_to_display$comment <- vapply(data_to_display$comment,
                                                      FUN = markdown,
                                                      FUN.VALUE = character(1L))
                    # Add a column for delete buttons
                    # The inputId of each button will be "delete_button_ROWID"
                    delete_buttons <- paste0(
                        '<button id="delete_button_', data_to_display$id, '" ',
                        'type="button" class="btn btn-danger btn-sm delete-btn" ',
                        'onclick="Shiny.onInputChange(\'', NS(id, "delete_clicked"),
                        '\', this.id)"',
                        '>Delete</button>'
                    )
                    # Add the delete buttons as a new column to the data
                    cbind(data_to_display, Actions = delete_buttons)
                }
            },
            editable = list(target='cell', disable=list(columns = c(1, 10))), # 10 is the delete button
            escape = FALSE, # IMPORTANT: Allow HTML rendering for buttons
            extensions = 'Buttons',
            options = list(
                dom = 'Brtip',
                lengthMenu = list(c(10, 25, 50, -1), c(10, 25, 50, "All")),
                buttons = c('pageLength', 'copy', 'csv', 'excel', 'pdf', 'print')
            ))
        }
        updateCommentsTable()
        observeEvent(input$issues_cell_edit, {
            row_idx  <- input$issues_cell_edit$row
            col_idx <- input$issues_cell_edit$col
            new_value <- input$issues_cell_edit$value
            original_data <- listComments(page_size=.globals$totalComments,
                                          full=TRUE, all=TRUE)
            id_to_update <- original_data[row_idx, "id"]
            col_name_to_update <- colnames(original_data)[col_idx]
            updateComments(id=id_to_update,
                           coln=col_name_to_update,
                           val=new_value)
        })
        # Observe delete button clicks
        observeEvent(input$delete_clicked, {
            showModal(
                modalDialog(
                    title = "Confirm deletion",
                    "Are you sure you want to delete this data?",
                    footer = tagList(
                        actionButton(NS(id, "confirmDelBtn"), "Confirm"), # This is the "Yes" button
                        modalButton("Cancel") # This is the "No" button
                    )
                )
            )
        })
        observeEvent(input$confirmDelBtn, {
            removeModal()
            # Extract the row ID from the button's inputId
            # The inputId is "delete_button_ROWID"
            button_id <- input$delete_clicked
            id_to_delete <- as.numeric(gsub("delete_button_", "", button_id))
            deleteComments(id=id_to_delete)
            # refresh table
            updateCommentsTable()
        })
        
        ## search keys
        output$search <- renderDT(listSearchTable())
    })
}

#' Create a scRNAseqApp project
#' @description To run scRNAseqApp, you need to first create a directory
#' which contains the required files.
#' @param app_path path, a directory where do you want to create the app
#' @param root character(1), the user name for administrator
#' @param password character(1), the password for administrator
#' @param datafolder the folder where saved the dataset for the app
#' @param overwrite logical(1), overwrite the `app_path` if there is
#'  a project.
#' @param app_title,app_description character(1). The title and 
#' description of the home page.
#' @param passphrase A password to protect the data inside the database.
#' @return no returns. This function will copy files to `app_path`
#' @export
#' @importFrom shinymanager create_db
#' @importFrom scrypt hashPassword
#' @importFrom utils write.table
#' @examples
#' if(interactive()){
#'     scInit()
#' }

scInit <- function(
    app_path = getwd(),
    root = 'admin',
    password = 'scRNAseqApp',
    datafolder = 'data',
    overwrite = FALSE,
    app_title = 'scRNAseq Database',
    app_description = 'This database is a collection of
        single cell RNA-seq data.',
    passphrase = NULL) {
    stopifnot(is.logical(overwrite) && length(overwrite) == 1)
    .globals$app_path <- app_path
    if(!is_abs_path(datafolder)){
        .globals$datafolder <- file.path(app_path, datafolder)
    }else{
        .globals$datafolder <- datafolder
    }
    if (!dir.exists(app_path)) {
        dir.create(app_path, recursive = TRUE)
    }
    message("Now copy pbmc_small sample files")
    file.copy(
        system.file("extdata", "data", package = "scRNAseqApp"),
        to = app_path,
        recursive = TRUE,
        overwrite = overwrite
    )
    # Write the doc.txt
    writeLines(c(
        paste('<h4>', app_title, '</h4>'),
        paste('<p>', app_description, '</p>')
    ),
    file.path(app_path, "doc.txt"))
    # Write the restart.txt
    writeLines(date(), file.path(app_path, 'restart.txt'))
    
    dir.create(file.path(app_path, 'app_cache'))
    
    www <- file.path(app_path, "www")
    dir.create(www)
    # Prepare the downloader folder
    dir.create(file.path(app_path, .globals$downloadFolder))
    # Write the counter file
    visitor <-
        data.frame(
            date = Sys.time(),
            ip = '127.0.0.1',
            agent = 'Mozilla/5.0')
    write.table(
        visitor,
        file = file.path(www, 'counter.tsv'),
        quote = FALSE,
        sep = "\t",
        row.names = FALSE
    )
    # define credentials
    credentials <- data.frame(
        user = root,
        password = hashPassword(password),
        admin = TRUE,
        is_hashed_password = TRUE,
        privilege = 'all',
        stringsAsFactors = FALSE
    )
    # Init the database
    dir.create(file.path(app_path, dirname(.globals$credential_path)),
               recursive = TRUE, showWarnings = FALSE)
    create_db(
        credentials_data = credentials,
        sqlite_path = file.path(app_path, .globals$credential_path),
        passphrase = passphrase
    )
    updateConfigTable()
    touchGeneTable()
    touchGenename2Symbol()
    .globals$counterFilename <- file.path(www, 'counter.tsv')
    touchVisitorTable()
    # Write the app.R
    writeLines(c(
        "library(scRNAseqApp)",
        paste0("scRNAseqApp(datafolder='", datafolder, "')")
    ),
    file.path(app_path, "app.R"))
    return(invisible())
}

test_that("utilities works not correct", {
  app_path <- file.path(tempdir(), 'test_scRNAseqApp')
  dir.create(app_path, recursive = TRUE)
  on.exit(unlink(app_path, recursive = TRUE))
  scInit(app_path=app_path)
  expect_true(file.exists(file.path(app_path, 'doc.txt')))
  expect_true(file.exists(file.path(app_path, .globals$credential_path)))
  expect_true(file.exists(file.path(app_path, 'www', 'counter.tsv')))
  setwd(app_path)
  
  # app <- scRNAseqApp()
  # env <- environment(app$serverFuncSource)
  # testServer(env$server, {
  #     session$setInputs(selectedDatasets = 'pbmc_small')
  # })
  # created by record_test()
  # review by testthat::snapshot_review('scRNAseqApp/')
  app <- shinytest2::AppDriver$new(app_dir='.', name = "testAllFun",
                                   variant='pbmc_small',
                                   width = 1080,
                                   height = 960)
  app$set_inputs(remote_addr = '127.0.0.1', allow_no_input_binding_ = TRUE)
  ## select the default dataset
  app$set_inputs(selectedDatasets = 'pbmc_small')
  app$set_inputs(`cellInfoGeneExpr-GeneName2` = "PPBP")
  app$set_inputs(`geneExprGeneExpr-GeneName1` = "PPBP")
  app$set_inputs(`geneExprGeneExpr-GeneName2` = "IGLL5")
  app$set_inputs(`coExpr-GeneName1` = "PPBP")
  app$set_inputs(`coExpr-GeneName2` = "IGLL5")
  app$set_inputs(`coExpr3d-GeneName1` = "PPBP")
  app$set_inputs(`coExpr3d-GeneName2` = "IGLL5")
  app$set_inputs(`sunburst-filterCell` = "nCount_RNA")
  app$set_inputs(`subsetGeneExpr-GeneName` = "PPBP")
  app$set_inputs(`subsetGeneExpr-filterCell` = "nCount_RNA")
  app$set_inputs(`vioBoxPlot-filterCell` = "nCount_RNA")
  app$set_inputs(`explorer-filterCell` = "nCount_RNA")
  ## check cellInfoGeneExpr
  app$set_inputs(topnav = "cellInfoGeneExpr")
  app$set_inputs(`cellInfoGeneExpr-GeneExproup.clk2` = character(0),
                 allow_no_input_binding_ = TRUE)
  app$set_inputs(`cellInfoGeneExpr-GeneExproup.dbl2` = character(0),
                 allow_no_input_binding_ = TRUE)
  app$set_inputs(`cellInfoGeneExpr-GeneExproup.clk1` = character(0),
                 allow_no_input_binding_ = TRUE)
  app$set_inputs(`cellInfoGeneExpr-GeneExproup.dbl1` = character(0),
                 allow_no_input_binding_ = TRUE)
  app$set_inputs(`cellInfoGeneExpr-CellInfo1` = "letters")
  app$click("cellInfoGeneExpr-CellInfotog1")
  app$set_inputs(`cellInfoGeneExpr-CellInfocol1` = "Blue-Yellow-Red")
  app$set_inputs(`cellInfoGeneExpr-CellInfoname1` = "groups")
  app$set_inputs(`cellInfoGeneExpr-CellInfo1` = "groups")
  app$set_inputs(`cellInfoGeneExpr-CellInfo1` = "celltype")
  app$set_inputs(`cellInfoGeneExpr-GeneName2` = "CST3")
  app$set_inputs(`cellInfoGeneExpr-GeneExprdrX` = "PC1")
  # app$expect_screenshot()
  ## check cellInfoCellInfo
  app$set_inputs(topnav = "cellInfoCellInfo")
  app$set_inputs(`cellInfoCellInfo-GeneExproup.clk2` = character(0),
                 allow_no_input_binding_ = TRUE)
  app$set_inputs(`cellInfoCellInfo-GeneExproup.dbl2` = character(0),
                 allow_no_input_binding_ = TRUE)
  app$set_inputs(`cellInfoCellInfo-GeneExproup.clk1` = character(0),
                 allow_no_input_binding_ = TRUE)
  app$set_inputs(`cellInfoCellInfo-GeneExproup.dbl1` = character(0),
                 allow_no_input_binding_ = TRUE)
  app$set_inputs(`cellInfoCellInfo-subsetCell` = "letters")
  app$set_inputs(`cellInfoCellInfo-subsetCellValletters` = "B")
  app$set_inputs(`cellInfoCellInfo-CellInfo1` = "letters")
  app$set_inputs(`cellInfoCellInfo-CellInfo2` = "groups")
  app$click("cellInfoCellInfo-graphicTog")
  app$set_inputs(`cellInfoCellInfo-GeneExprfsz` = 12)
  # app$expect_screenshot()
  ## check subsetGeneExpr
  app$set_inputs(topnav = "subsetGeneExpr")
  app$set_inputs(`subsetGeneExpr-GeneExproup.clk2` = character(0),
                 allow_no_input_binding_ = TRUE)
  app$set_inputs(`subsetGeneExpr-GeneExproup.dbl2` = character(0),
                 allow_no_input_binding_ = TRUE)
  app$set_inputs(`subsetGeneExpr-GeneExproup.clk1` = character(0),
                 allow_no_input_binding_ = TRUE)
  app$set_inputs(`subsetGeneExpr-GeneExproup.dbl1` = character(0),
                 allow_no_input_binding_ = TRUE)
  app$set_inputs(`subsetGeneExpr-filterCell` = "tSNE1")
  app$set_inputs(`subsetGeneExpr-filterCellVal` = c(-36, 16))
  app$set_inputs(`subsetGeneExpr-filterCellVal2` = c(-7, 24))
  app$click("subsetGeneExpr-GeneExprtog1")
  app$set_inputs(`subsetGeneExpr-GeneExprhid1` = TRUE)
  app$click("subsetGeneExpr-GeneExprtog2")
  app$set_inputs(`subsetGeneExpr-GeneExprhid2` = TRUE)
  app$click("subsetGeneExpr-CellInfoTableTog1")
  app$set_inputs(`subsetGeneExpr-GeneExpr.dt1_rows_current` = c(1, 2),
                 allow_no_input_binding_ = TRUE)
  app$set_inputs(`subsetGeneExpr-GeneExpr.dt1_rows_all` = c(1, 2),
                 allow_no_input_binding_ = TRUE)
  app$set_inputs(`subsetGeneExpr-GeneExpr.dt1_state` = 
                     c(1753799453390, 0, -1, "", 
                       TRUE, FALSE, TRUE, 
                       c(TRUE, "", TRUE, FALSE, TRUE),
                       c(TRUE, "", TRUE, FALSE, TRUE),
                       c(TRUE, "", TRUE, FALSE, TRUE),
                       c(TRUE, "", TRUE, FALSE, TRUE)), 
                 allow_no_input_binding_ = TRUE)
  #app$expect_download("subsetGeneExpr-GeneExproup.dwn1")
  # app$expect_screenshot()
  ## check geneExprGeneExpr
  app$set_inputs(topnav = "geneExprGeneExpr")
  app$set_inputs(`geneExprGeneExpr-GeneExproup.clk2` = character(0),
                 allow_no_input_binding_ = TRUE)
  app$set_inputs(`geneExprGeneExpr-GeneExproup.dbl2` = character(0),
                 allow_no_input_binding_ = TRUE)
  app$set_inputs(`geneExprGeneExpr-GeneExproup.clk1` = character(0),
                 allow_no_input_binding_ = TRUE)
  app$set_inputs(`geneExprGeneExpr-GeneExproup.dbl1` = character(0),
                 allow_no_input_binding_ = TRUE)
  app$click("geneExprGeneExpr-graphicTog")
  app$click("geneExprGeneExpr-graphicTog")
  app$click("geneExprGeneExpr-GeneExprtog2")
  app$set_inputs(`geneExprGeneExpr-GeneExprtype2` = "Ridgeplot")
  app$click("geneExprGeneExpr-GeneExprtog1")
  app$click("geneExprGeneExpr-GeneExprrgb1")
  app$set_inputs(`geneExprGeneExpr-GeneExprrg1` = 8)
  # app$expect_screenshot()
  ## check coExpr
  app$set_inputs(topnav = "coExpr")
  app$set_inputs(`coExpr-coExpr.dt_rows_current` = c(1, 2, 3),
                 allow_no_input_binding_ = TRUE)
  app$set_inputs(`coExpr-coExpr.dt_rows_all` = c(1, 2, 3),
                 allow_no_input_binding_ = TRUE)
  app$set_inputs(`coExpr-coExpr.dt_state` = 
                     c(1753799510042, 0, -1, "", TRUE, FALSE, TRUE,
                       c(TRUE, "", TRUE, FALSE, TRUE),
                       c(TRUE, "", TRUE, FALSE, TRUE),
                       c(TRUE, "", TRUE, FALSE, TRUE)),
                 allow_no_input_binding_ = TRUE)
  app$set_inputs(`coExpr-GeneExproup.clk2` = character(0),
                 allow_no_input_binding_ = TRUE)
  app$set_inputs(`coExpr-GeneExproup.dbl2` = character(0),
                 allow_no_input_binding_ = TRUE)
  app$set_inputs(`coExpr-GeneExproup.clk1` = character(0),
                 allow_no_input_binding_ = TRUE)
  app$set_inputs(`coExpr-GeneExproup.dbl1` = character(0),
                 allow_no_input_binding_ = TRUE)
  app$set_inputs(`coExpr-GeneName1` = "COTL1")
  app$set_inputs(`coExpr-coExpr.dt_rows_current` = c(1, 2, 3, 4),
                 allow_no_input_binding_ = TRUE)
  app$set_inputs(`coExpr-coExpr.dt_rows_all` = c(1, 2, 3, 4),
                 allow_no_input_binding_ = TRUE)
  app$set_inputs(`coExpr-coExpr.dt_state` = 
                     c(1753799513828, 0, -1, "", TRUE, FALSE, TRUE,
                       c(TRUE, "", TRUE, FALSE, TRUE),
                       c(TRUE, "", TRUE, FALSE, TRUE),
                       c(TRUE, "", TRUE, FALSE, TRUE)),
                 allow_no_input_binding_ = TRUE)
  # app$expect_screenshot()
  ## check coExpr3d
  app$set_inputs(topnav = "coExpr3d")
  app$set_inputs(`plotly_hover-A` = character(0),
                 allow_no_input_binding_ = TRUE, 
                 priority_ = "event")
  app$set_inputs(topnav = "sunburst")
  app$set_inputs(`sunburst-GeneExproup.clk1` = character(0),
                 allow_no_input_binding_ = TRUE)
  app$set_inputs(`sunburst-GeneExproup.dbl1` = character(0),
                 allow_no_input_binding_ = TRUE)
  app$set_inputs(`sunburst-genelist` = "PPBP, IGLL5, VDAC3")
  app$set_inputs(`sunburst-CoExpred` = FALSE)
  app$set_inputs(`sunburst-filterCell` = "tSNE1")
  app$set_inputs(`sunburst-filterCellVal` = c(27, 46))
  app$set_inputs(`sunburst-filterCellVal2` = c(-3, 24))
  app$click("sunburst-graphicTog")
  app$set_inputs(`sunburst-GeneExprtxt` = TRUE)
  app$set_inputs(`sunburst-filterCellVal2` = c(6, 24))
  app$set_inputs(`sunburst-filterCellVal2` = c(9, 24))
  # app$expect_screenshot()
  ## check vioBoxPlot
  app$set_inputs(topnav = "vioBoxPlot")
  app$set_inputs(`vioBoxPlot-GeneExproup.clk1` = character(0),
                 allow_no_input_binding_ = TRUE)
  app$set_inputs(`vioBoxPlot-GeneExproup.dbl1` = character(0),
                 allow_no_input_binding_ = TRUE)
  app$set_inputs(`vioBoxPlot-CellInfoY` = "letters")
  app$set_inputs(`vioBoxPlot-subsetCell` = "celltype")
  app$set_inputs(`vioBoxPlot-filterCellVal` = c(41, 506))
  app$set_inputs(`vioBoxPlot-plottyp` = "boxplot")
  app$set_inputs(`vioBoxPlot-plotpts` = TRUE)
  app$set_inputs(`vioBoxPlot-addnoise` = FALSE)
  app$set_inputs(`vioBoxPlot-plotord` = TRUE)
  app$set_inputs(`vioBoxPlot-cellinfoXorder` = c("g1", "g2"),
                 allow_no_input_binding_ = TRUE)
  app$set_inputs(`vioBoxPlot-cellinfoXorder` = c("g2", "g1"),
                 allow_no_input_binding_ = TRUE)
  app$set_inputs(`vioBoxPlot-plotsord` = TRUE)
  app$set_inputs(`vioBoxPlot-cellinfoSorder` = c("A", "B"),
                 allow_no_input_binding_ = TRUE)
  app$set_inputs(`vioBoxPlot-cellinfoSorder` = c("B", "A"),
                 allow_no_input_binding_ = TRUE)
  # app$expect_screenshot()
  ## check proportion
  app$set_inputs(topnav = "proportion")
  app$set_inputs(`proportion-GeneExproup.clk1` = character(0),
                 allow_no_input_binding_ = TRUE)
  app$set_inputs(`proportion-GeneExproup.dbl1` = character(0),
                 allow_no_input_binding_ = TRUE)
  app$set_inputs(`proportion-CellInfoY` = "letters")
  app$set_inputs(`proportion-plottyp` = "CellNumbers")
  app$set_inputs(`proportion-plotflp` = TRUE)
  app$set_inputs(`proportion-plotord` = TRUE)
  app$set_inputs(`proportion-cellinfoYorder` = 
                     c("A", "B"), allow_no_input_binding_ = TRUE)
  app$set_inputs(`proportion-cellinfoXorder` = 
                     c("CD16 Mono", "CD14 Mono", "CD16+ NK", 
                       "CD1c DC", "CD4 Memory", "CD4 Naive", "CD8 Effector 1",
                       "CD8 Effector 2", "CD8 Memory", 
                       "CD8 Naive", "Intermediate Mono", "MAIT",
                       "Memory B", "Naive B", "Platelets"), 
                 allow_no_input_binding_ = TRUE)
  app$click("proportion-statstog")
  app$set_inputs(`proportion-proportion.dt_rows_current` = 
                     c(1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15),
                 allow_no_input_binding_ = TRUE)
  app$set_inputs(`proportion-proportion.dt_rows_all` = 
                     c(1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15),
                 allow_no_input_binding_ = TRUE)
  app$set_inputs(`proportion-proportion.dt_state` = 
                     c(1753799688415, 0, -1, "", TRUE, FALSE, TRUE,
                       c(TRUE, "", TRUE, FALSE, TRUE),
                       c(TRUE, "", TRUE, FALSE, TRUE),
                       c(TRUE, "", TRUE, FALSE, TRUE),
                       c(TRUE, "", TRUE, FALSE, TRUE),
                       c(TRUE, "", TRUE, FALSE, TRUE),
                       c(TRUE, "", TRUE, FALSE, TRUE),
                       c(TRUE, "", TRUE, FALSE, TRUE),
                       c(TRUE, "", TRUE, FALSE, TRUE)),
                 allow_no_input_binding_ = TRUE)
  # app$expect_screenshot()
  app$click("proportion-pearsontog")
  app$set_inputs(`proportion-GeneExproup.clk2` = character(0),
                 allow_no_input_binding_ = TRUE)
  app$set_inputs(`proportion-GeneExproup.dbl2` = character(0),
                 allow_no_input_binding_ = TRUE)
  # app$expect_screenshot()
  ## check bubleHeatmap
  app$set_inputs(topnav = "bubbleHeatmap")
  app$set_inputs(`bubbleHeatmap-GeneExproup.clk1` = character(0),
                 allow_no_input_binding_ = TRUE)
  app$set_inputs(`bubbleHeatmap-GeneExproup.dbl1` = character(0),
                 allow_no_input_binding_ = TRUE)
  app$click("bubbleHeatmap-plottog")
  app$set_inputs(`bubbleHeatmap-plotpsz` = "Large")
  app$set_inputs(`bubbleHeatmap-plotcols` = "Viridis")
  app$set_inputs(`bubbleHeatmap-row_dend_width` = 20.2)
  app$set_inputs(`bubbleHeatmap-legend_side` = "right")
  app$set_inputs(`bubbleHeatmap-row_dend_side` = "right")
  app$set_inputs(`bubbleHeatmap-plotsiz` = 4)
  app$set_inputs(`bubbleHeatmap-plotflp` = TRUE)
  app$set_inputs(`bubbleHeatmap-plotcol` = TRUE)
  app$set_inputs(`bubbleHeatmap-plottyp` = "Heatmap")
  app$set_inputs(`bubbleHeatmap-plottyp` = "Violin")
  app$set_inputs(`bubbleHeatmap-addnoise` = FALSE)
  app$set_inputs(`bubbleHeatmap-plottyp` = "Heatmap")
  app$set_inputs(`bubbleHeatmap-CellInfoY` = "letters")
  # app$expect_screenshot()
  ## check waffle
  app$set_inputs(topnav = "waffle")
  app$set_inputs(`waffle-GeneExproup.clk1` = character(0),
                 allow_no_input_binding_ = TRUE)
  app$set_inputs(`waffle-GeneExproup.dbl1` = character(0),
                 allow_no_input_binding_ = TRUE)
  # app$expect_screenshot()
  ## check explorer
  app$set_inputs(topnav = "explorer")
  app$click("explorer-newModule")
  app$set_inputs(`explorer-moduleName` = "gene expression")
  app$click("explorer-newModule")
  app$set_inputs(`explorer-plot_2-GeneName1` = "PPBP")
  app$set_inputs(`explorer-moduleName` = "proportion")
  app$click("explorer-newModule")
  app$set_inputs(`explorer-plot_2-GeneName1` = "PPBP")
  app$set_inputs(`explorer-moduleName` = "violin/box plot")
  app$click("explorer-newModule")
  app$set_inputs(`explorer-plot_2-GeneName1` = "PPBP")
  app$click("explorer-plot_2-moveup")
  app$set_inputs(`explorer-plot_2-GeneName1` = "PPBP")
  app$set_inputs(`plotly_hover-A` = 
                     paste0("[{\"curveNumber\":1,",
                            "\"pointNumber\":32,",
                            "\"x\":44.68867319807703,",
                            "\"y\":21.35127489929749,",
                            "\"z\":3.169925001442312,",
                            "\"customdata\":\"GTCATACTTCGCCT\"}]"), 
                 allow_no_input_binding_ = TRUE, priority_ = "event")
  # app$expect_screenshot()
  ## check issues
  .globals$commentsIntervals <- 0
  app$set_inputs(topnav = "issues")
  app$set_inputs(`issues-issues_state` = 
                     c(1755988380399, 0, 10, "", TRUE, FALSE, TRUE,
                       c(TRUE, "", TRUE, FALSE, TRUE),
                       c(TRUE, "", TRUE, FALSE, TRUE),
                       c(TRUE, "", TRUE, FALSE, TRUE),
                       c(TRUE, "", TRUE, FALSE, TRUE)),
                 allow_no_input_binding_ = TRUE)
  # Update output value
  app$set_inputs(`issues-uid` = "insert")
  app$set_inputs(`issues-email` = "insert@email.com")
  app$set_inputs(`issues-title` = "test ")
  app$set_inputs(`issues-comment` = "test comment")
  # Update output value
  app$click("issues-submit")
  app$click("issues-confirmSubmitBtn")
  # app$expect_screenshot()
})

plotVioBoxUI <- function(id) {
    tabPanel(
        value = id,
        HTML("Violinplot / Boxplot"),
        h4("Cell information / gene expression violin plot / box plot"),
        "In this tab, users can visualise the gene expression ",
        "or continuous cell information ",
        "(e.g. Number of UMIs / module score) across groups of cells ",
        "(e.g. libary / clusters).",
        br(),
        br(),
        fluidRow(
            column(
                3,
                style = "border-right: 2px solid black",
                xaxisCellInfoUI(id),
                yaxisCellInfoUI(id),
                subsetCellByInfoUI(id, mini = TRUE),
                subsetCellByFilterUI(
                    id,
                    label = "Cell Info / Gene name (Y-axis):",
                    title = "Cell Info / Gene to plot",
                    content = c(
                        "Select cell info / gene to plot on Y-axis",
                        "- Can be continuous cell information ",
                        "(e.g. nUMIs / scores)",
                        "- Can also be gene expression"
                    )
                ),
                radioButtons(
                    NS(id, "plottyp"),
                    "Plot type:",
                    choices = c("violin", "boxplot"),
                    selected = "violin",
                    inline = TRUE
                ),
                checkboxInput(
                    NS(id, "plotpts"), "Show data points",
                    value = FALSE),
                checkboxInput(
                    NS(id, 'addnoise'),
                    "Add noise", value = TRUE
                ),
                checkboxInput(
                    NS(id, "plotord"),
                    "Reorder the contents", value = FALSE
                ),
                conditionalPanel(
                    condition = "input.plotord % 2 == 1",
                    ns=NS(id),
                    uiOutput(outputId = NS(id, "plotXord"))
                ),
                conditionalPanel(
                    condition = "input.CellInfoY != ''",
                    ns = NS(id),
                    checkboxInput(
                        NS(id, "plotsord"),
                        "Reorder the spliter", value = FALSE
                    ),
                    conditionalPanel(
                        condition = "input.plotsord % 2 == 1",
                        ns=NS(id),
                        uiOutput(outputId = NS(id, "plotSord"))
                    )
                ),
                boxPlotControlUI(id)
            ),
            column(9, geneExprDotPlotUI(id),
                   br(),
                   actionButton(NS(id, "statstog"), "Toggle to show statistics"),
                   conditionalPanel(
                       condition = "input.statstog % 2 == 1",
                       ns = NS(id),
                       h4("Statistics"),
                       DTOutput(NS(id, "violin.dt"))
                   ))
        )
    )
}
#' @importFrom utils combn
#' @importFrom stats wilcox.test
plotVioBoxServer <- function(id, dataSource, optCrt) {
    moduleServer(id, function(input, output, session) {
        ## input column
        updateSelectInput(
            session,
            "CellInfoX",
            "Cell information (X-axis):",
            choices = getGroupUI(dataSource),
            selected = dataSource()$sc1def$grp1
        )
        
        updateSelectInput(
            session,
            "CellInfoY",
            "Split by:",
            choices = c("N/A", getGroupUI(dataSource)),
            selected = NA
        )
        
        updateSubsetCellUI(id, input, output, session, dataSource, addNA = TRUE)
        updateFilterCellUI(id, optCrt, input, output, session, dataSource)
        updateRankList(
            input, output, dataSource, "CellInfoX", "plotXord",
            NS(id, "cellinfoXorder"))
        updateRankList(
            input, output, dataSource, "CellInfoY", "plotSord",
            NS(id, "cellinfoSorder"))
        
        ## plot region
        ### plots
        plot1 <- reactive({
            scVioBox(
                inpConf = dataSource()$sc1conf,
                inpMeta = dataSource()$sc1meta,
                infoX = input$CellInfoX,
                infoY = input$filterCell,
                subsetCellKey = input$subsetCell,
                subsetCellVal = getSubsetCellVal(input),
                valueFilterKey = input$filterCell,
                valueFilterCutoff = input$filterCellVal,
                valueFilterCutoff2 = input$filterCellVal2,
                dataset = dataSource()$dataset,
                geneIdMap = dataSource()$sc1gene,
                inptyp = input$plottyp,
                inppts = input$plotpts,
                pointSize = input$plotsiz,
                labelsFontsize = input$plotfsz,
                labelsFontFamily = input$plotfml,
                reorder = input$plotord,
                orderX = input$cellinfoXorder,
                splitBy=input$CellInfoY,
                sreorder = input$plotsord,
                orderS = input$cellinfoSorder,
                addnoise = input$addnoise
            )
        })
        updateGeneExprDotPlotUI(
            postfix = 1,
            id,
            input,
            output,
            session,
            plot1,
            .globals$pList2[input$plotpsz],
            dataSource()$dataset,
            input$plottyp,
            input$CellInfoX,
            input$filterCell
        )
        output$violin.dt <- renderDT({
            violin <- plot1()$data
            violin <- split(violin$val, violin$X)
            cbn <- combn(names(violin), 2, simplify = FALSE)
            res <- lapply(cbn, function(.ele) {
                withCallingHandlers({
                    wt <- wilcox.test(violin[[.ele[1]]],
                                      violin[[.ele[2]]])
                }, warning = function(w){
                    showNotification(
                        as.character(w),
                        duration = 5,
                        type = 'warning'
                    )
                }, error = function(e){
                    showNotification(
                        as.character(e),
                        duration = 5,
                        type = 'error'
                    )
                    wt <- list('p.value'=NA)
                })
                
                return(as.data.frame(wt[
                    c('alternative',
                      'statistic',
                      'p.value',
                      'method')
                ]))
            })
            
            res <- cbind(do.call(rbind, cbn), do.call(rbind, res))
            colnames(res)[c(1, 2)] <- c('group 1', 'group 2')
            datatable(
                res,
                rownames = FALSE,
                extensions = "Buttons",
                options = list(
                    pageLength = -1,
                    dom = "tB",
                    buttons = c("copy", "csv", "excel")
                )
            )
        })
    })
}

homeUI <- function() {
    tabPanel(## fack home
        title = "",
        value = "home",
        icon = icon("home"))
}

#' @importFrom bibtex read.bib
#' @importFrom RefManageR GetBibEntryWithDOI PrintBibliography
aboutUI <- function(
    request, id, banner, defaultDataset, datasets, appconf, doc = "doc.txt",
    showHelpVideo=FALSE) {
        ns <- NS(id)
        if(is.character(showHelpVideo)){
            helperVideoURL <- showHelpVideo
            showHelpVideo <- TRUE
        }else{
            helperVideoURL <- "https://www.youtube.com/embed/videoseries?si=VgcP3HxCrvtp59yS&amp;list=PL0uThf-sipbnepUDzVmAHpNWUy65F4HOm&amp;showinfo=0"
        }
        query <- parseQueryString(request[["QUERY_STRING"]])
        query <- parseQuery(query, defaultDataset)
        defaultDataset <- query['defaultDataset']
        if(!defaultDataset %in% datasets){
            if(query['from']=='token'){
                ## from token
                res <- updateDatasetForToken(defaultDataset, datasets)
                datasets <- res$datasets
                appconf <- res$appconf
                data_types <- res$data_types
            }else{
                if(defaultDataset %in% names(datasets)){
                   defaultDataset <- datasets[defaultDataset]
                }else{
                    ## something wrong here
                    defaultDataset <- datasets[1] 
                }
            }
        }
        refs <- listReferences()
        species <- listSpecies()
        dbs <- listDatasets(
            key = datasets,
            privilege='all',
            named = TRUE)
        tabPanel(
            title = div(
                selectInput(
                    'selectedDatasets',
                    label = NULL,
                    choices = dbs,
                    selected = defaultDataset,
                    width = "100%"
                )
            ),
            value = 'about',
            div(id = 'about-doc', includeHTML(doc)),
            hr(),
            htmlOutput(ns("ref")),
            hr(),
            div(
                class = "about-display-container about-content about-top",
                style = 'width:100%;min-height:300px;',
                img(class = "about-image about-glass", src = banner),
                absolutePanel(
                    top = "15%",
                    left = "10%",
                    width = "80%",
                    draggable = TRUE,
                    div(
                        class =
                            "panel panel-danger about-glass",
                        div(
                            class =
                                paste(
                                    "panel-heading",
                                    "about-container",
                                    "about-padding-16"),
                            fluidRow(
                                column(9,
                                       textInput(
                                           ns('search'),
                                           label = NULL,
                                           width = "100%",
                                           placeholder = 'Search the database'
                                       )),
                                column(1, 
                                      actionButton(
                                          ns('sbtn'),
                                          label = 'Search',
                                          width = '100px'
                                      ))
                            )
                        ),
                        div(
                            class = paste(
                                "panel-body",
                                "about-bar",
                                "border-bottom-info",
                                "shadow h-100 py-2"),
                            htmlOutput(ns('search_res')),
                            div(
                                style = "visibility:hidden;",
                                checkboxInput(
                                    ns("s_res_flag"),
                                    label = NULL,
                                    value = FALSE
                                )
                            ))
                    )
                )
            ),
            hr(),
            div(
                class = "about-container",
                summaryBox(
                    "DATASETS",
                    length(dbs), #textOutput(ns('dataset_counts')),
                    width = 3,
                    icon = "database",
                    style = "success",
                    border = "bottom",
                    info = 'Public dataset list:',
                    details = paste("<ol>\n",
                                    paste(paste("<li><p><a href='?data=",
                                                dbs,
                                                "'>",
                                                names(dbs),
                                                "</a></p></li>"), collapse = "\n"),
                                    "\n</ol>"),
                    id = id
                ),
                summaryBox(
                    "REFERENCES",
                    length(refs), #textOutput(ns('reference_count')),
                    width = 3,
                    icon = "book-open",
                    style = "danger",
                    border = "bottom",
                    info = 'Full reference list:',
                    details = paste("<ol>\n",
                                    paste(paste("<li>",
                                                refs,
                                                "</li>"), collapse = "\n"),
                                    "\n</ol>")
                ),
                summaryBox(
                    "VISITORS",
                    listVisitors(ipCounter = TRUE)$uniqueIP, #textOutput(ns('visitor_count')),
                    width = 3,
                    icon = "eye",
                    style = "primary",
                    border = "bottom",
                    info = 'Vistor counts:',
                    details = ''
                ),
                summaryBox(
                    "SPECIES",
                    tags$a(length(species),
                           'data-toggle'='tooltip',
                           'data-html'="true",
                           title=paste(species,
                                       collapse = ',\u{000A}')), #textOutput(ns('species_count')),
                    width = 3,
                    icon = "fish",
                    style = "warning",
                    border = "bottom",
                    info = 'Full species list:',
                    details = paste(species,
                                    collapse = ',<br/>')
                )
            ),
            hr(),
            div(
                id = 'informationBox',
                uiOutput(ns('privateDatasets')),
                h4("Full reference list:"),
                div(id='infolist',
                    HTML(paste("<ol>\n",
                           paste(paste("<li>",
                                       refs,
                                       "</li>"), collapse = "\n"),
                           "\n</ol>")))
            ),
            if(!showHelpVideo){
                p(
                    imageOutput('total_visitor', width = "50%", height = "150px")
                )
            }else{
                fluidRow(
                    column(
                        6,
                        imageOutput('total_visitor',
                                    width = "100%", height = "240px")),
                    column(
                        6,
                        tags$iframe(
                            width="100%",
                            height='240px',
                            src=helperVideoURL,
                            frameborder="0",
                            allow="accelerometer; autoplay; clipboard-write; encrypted-media; gyroscope; picture-in-picture; web-share;",
                            referrerpolicy="strict-origin-when-cross-origin",
                            allowfullscreen=NA)
                    )
                )
            },
            div(
                style = "display:none;",
                textInput(
                    "remote_agent", "remote_agent",
                    value = request[["HTTP_USER_AGENT"]]),
                visitorDependencies()
            ),
            hr(),
            p(
                em("This webpage was modified from "),
                a(
                    "ShinyCell",
                    href = "https://github.com/SGDDNB/ShinyCell",
                    target = "_blank")
            )
        )
    }
aboutServer <- function(id, dataSource, optCrt) {
    moduleServer(id, function(input, output, session) {
        bibentry <- function(key){
            getRef(dataSource()$dataset, key, dataSource()$appconf)
        }
        global <- reactiveValues(search_results = list(), evt = list())
        output$ref <- renderUI(
            if(!is.na(bibentry('ref_title'))){
                tagList(
                    h5(bibentry('ref_title')),
                    if (!is.na(bibentry("ref_abstract"))) {
                        if(!grepl(
                    "^The abstract of a scientific paper represents a concise",
                            bibentry("ref_abstract")
                        )){
                            HTML(bibentry("ref_abstract"))
                        }
                    },
                    h5("Reference for current data"),
                    HTML(
                        bibentry("ref_bib")
                    )
                )
            }else{
                h5('No reference for current data.')
            })
        output$statsDATASETS <- 
            renderText({length(dataSource()$available_datasets)})
        observeEvent(dataSource()$available_datasets, {
            updateSelectInput(inputId = 'selectedDatasets',
                    label = NULL,
                    choices = dataSource()$available_datasets,
                    selected = dataSource()$dataset
            )
            output$statsDATASETS <- 
                renderText({length(dataSource()$available_datasets)})
            ## private datasets
            publicDbs <- checkAvailableDataSets(privilege = NULL)
            privateDbs <- setdiff(dataSource()$available_datasets, publicDbs)
            if(length(privateDbs)){
                output$privateDatasets <- renderUI({
                    tagList(
                        h5('Private dataset list:'),
                        tags$ul(
                            lapply(privateDbs, function(.ele){
                                # add token if logged
                                token <- getLogToken(session = session)
                                if(!is.null(token)){
                                    token <- paste0('&token=', token)
                                }
                                return(tags$li(
                                    tags$a(
                                        href=paste0('?data=', .ele, token),
                                        .ele,
                                        onclick = 
                                            paste0('event.preventDefault();',
                                                   'Shiny.onInputChange(',
                                                   '\"search_results_select_button\",  \"',
                                                   .ele,'\");window.scrollTo(0, 0);'))
                                ))
                            })
                        ) 
                    )
                })
            }
        })
        observeEvent(input$sbtn, {
            if (input$search != '' && input$search != "Type key words here") {
                showNotification(
                    paste('Searching', input$search, 'in database.'),
                    duration = 3,
                    type = 'message'
                )
                updateSearch(
                    input$search,
                    dataSource()$available_datasets,
                    auth = dataSource()$auth,
                    global = reactive({
                        global
                    }),
                    id = id,
                    input = input,
                    output = output,
                    session = session
                )
            }else{
                output$search_res <- renderUI(tags$div())
            }
        })
    })
}

#' Generate data files required for shiny app
#'
#' Copied from ShinyCell to generate the data files.
#' @noRd
#' @param obj input single-cell object for Seurat (v3+)
#' @param scConf config data.table
#' @param assayName assay in single-cell data object to use for plotting
#'   gene expression, which must match one of the following:
#'   \itemize{
#'     \item{Seurat objects}: "RNA" or "integrated" assay,
#'       default is "RNA"
#'   }
#' @param gexSlot layer in single-cell gex assay to plot.
#' Default is to use the "data" layer
#' @param atacAssayName assay in single-cell data object to use for plotting
#' open chromatin.
#' @param atacSlot layer in single-cell atac assay to plot.
#' Default is to use the "data" layer
#' @param appDir specify directory to create the shiny app in
#' @param defaultGene1 specify primary default gene to show
#' @param defaultGene2 specify secondary default gene to show
#' @param default.multigene character vector specifying default genes to
#'   show in bubbleplot / heatmap
#' @param default.dimred character vector specifying the two default dimension
#'   reductions. Default is to use UMAP if not TSNE embeddings
#' @param default.symbol character(1L) specifying the default rownames to be used. If use default, the gene symbols will be the row names of the assay. If one column name of the meta.feature of the assay is supplied, the function will try to extract the symbols from the meta.feature slot of the assay. 
#' @param chunkSize number of genes written to h5file at any one time. Lower
#'   this number to reduce memory consumption. Should not be less than 10
#' @param binSize number of bps for each bin for ATAC fragment coverage. Used
#' to reduce the file size of bigwig.
#' @param fragmentNameMapList list of named character vector. 
#' The name map list must be the same order as the fragment list in the object.
#' For each element of the list,
#' the names of the vector are the name of the fragment and
#' the vector contains the cell names (column names of the assay). 
#' You can try \link{extractFragmentNameMapList}.
#' @return data files required for shiny app
#' @importFrom IRanges tile Views viewMeans ranges nearest
#' @importFrom SeuratObject GetAssayData VariableFeatures Embeddings Reductions
#' @importFrom data.table data.table as.data.table
#' @importFrom rhdf5 h5createFile h5createGroup h5createDataset h5write
#' @importFrom Rsamtools TabixFile seqnamesTabix scanTabix
#' @importFrom GenomeInfoDb keepSeqlevels seqinfo seqnames seqlevelsStyle
#' `seqlevelsStyle<-`
#' @importFrom GenomicRanges GRanges width coverage GRangesList
#' @importFrom rtracklayer export
#' @importFrom utils read.table
#' @importFrom fs path_sanitize
makeShinyFiles <- function(
        obj,
        scConf,
        assayName,
        gexSlot = c("data", "scale.data", "counts"),
        atacAssayName,
        atacSlot = c("data", "scale.data", "counts"),
        appDir = "data",
        defaultGene1 = NA,
        defaultGene2 = NA,
        default.multigene = NA,
        default.dimred = NA,
        default.symbol = 'rownames',
        chunkSize = 500,
        binSize = 1,
        fragmentNameMapList) {
    stopifnot(is.numeric(binSize))
    ### Preprocessing and checks
    # Generate defaults for assayName / slot
    stopifnot(is(obj, "Seurat"))
    # Seurat Object
    if (missing(assayName)) {
        assayName <- "RNA"
    } else{
        assayName <- assayName[1]
    }
    gexSlot <- match.arg(gexSlot)
    atacSlot <- match.arg(atacSlot)
    gexAsy <- extAssayData(obj, assay = assayName, slot = gexSlot)
    gex.matdim <- dim(gexAsy)
    gex.rownm <- rownames(gexAsy)
    gex.colnm <- colnames(gexAsy)
    useRownames <- TRUE
    if(default.symbol!='rownames'){
        featureMeta <- obj[[assayName]][[]]
        if(default.symbol %in% colnames(featureMeta)){
            ## rename the gexAsy
            newRN <- featureMeta[, default.symbol[1]]
            newRN[is.na(newRN)] <- gex.rownm[is.na(newRN)]
            newRN[newRN==''] <- gex.rownm[newRN=='']
            if(any(duplicated(newRN))){
                dup_RN <- newRN %in% newRN[duplicated(newRN)]
                newRN[dup_RN] <- 
                    paste(newRN[dup_RN], gex.rownm[dup_RN], sep='-')
            }
            oldRN <- gex.rownm
            gex.rownm <- newRN
            rm(newRN)
            useRownames <- FALSE
        }
    }
    defGenes <- VariableFeatures(obj)
    if(length(defGenes)>10) defGenes <- defGenes[seq.int(10)]
    if (is.na(defGenes[1])) {
        warning(
            "Variable genes for seurat object not found! Have you ",
            "ran `FindVariableFeatures` or `SCTransform`?"
        )
        defGenes <- gex.rownm
        if(length(defGenes)>10) defGenes <- defGenes[seq.int(10)]
    }
    if(!useRownames){
        if(any(oldRN %in% gex.rownm)){
            defGenes <- gex.rownm[match(defGenes, oldRN)]
            defGenes <- defGenes[!is.na(defGenes)]
            if(length(defGenes)==0){
                defGenes <- gex.rownm
                if(length(defGenes)>10) defGenes <- defGenes[seq.int(10)]
            }
        }
    }
    sc1meta <- data.table(sampleID = rownames(obj[[]]), obj[[]])
    
    geneMap <- gex.rownm
    names(geneMap) <- gex.rownm    # Basically no mapping
    
    defGenes <- geneMap[defGenes]
    
    # Check defaultGene1 / defaultGene2 / default.multigene
    defaultGene1 <- defaultGene1[1]
    defaultGene2 <- defaultGene2[1]
    if(!useRownames){
        if(!defaultGene1 %in% gex.rownm){
            defaultGene1 <- gex.rownm[match(defaultGene1, oldRN)]
        }
        if(!defaultGene2 %in% gex.rownm){
            defaultGene2 <- gex.rownm[match(defaultGene2, oldRN)]
        }
        if(any(!default.multigene %in% gex.rownm)){
            default.multigene <- gex.rownm[match(default.multigene, oldRN)]
        }
    }
    if (is.na(defaultGene1)) {
        defaultGene1 <- defGenes[1]
    }
    if (is.na(defaultGene2)) {
        defaultGene2 <- defGenes[2]
    }
    if (is.na(default.multigene[1])) {
        default.multigene <- defGenes
    }
    if (defaultGene1 %in% geneMap) {
        defaultGene1 <- defaultGene1
    } else {
        warning(
            "defaultGene1 doesn't exist in gene expression, using defaults...")
        defaultGene1 <- defGenes[1]
    }
    if (defaultGene2 %in% geneMap) {
        defaultGene2 <- defaultGene2
    } else {
        warning(
            "defaultGene2 doesn't exist in gene expression, using defaults...")
        defaultGene2 <- defGenes[2]
    }
    if (all(default.multigene %in% geneMap)) {
        default.multigene <- default.multigene
    } else {
        warning(
            "default.multigene doesn't exist in gene expression, ",
            "using defaults...")
        default.multigene <- defGenes
    }
    
    ## check for ATAC
    if (!missing(atacAssayName)) {
        if (atacAssayName %in% Assays(obj)) {
            peaks <- obj[[atacAssayName]]
            ## links, the links between peaks and gene symbol,
            ## used to create the matrix table
            links <- peaks@links #GetAssayData(obj, layer = "links")
            ## annotations, used to plot gene model
            annotations <- peaks@annotation #GetAssayData(obj, layer = "annotation")
            if (length(annotations) < 1) {
                stop("scATAC annotation data are not available.")
            }
            if(any(!c('gene_name', 'tx_id') %in% colnames(mcols(annotations)))){
                stop('gene_name and tx_id must exist as metadata for scATAC ',
                     'annotations. Check it by Annotation(obj[["',
                     atacAssayName, '"]])')
            }
            if (length(links) < 1) {
                warning("scATAC links data are not available.")
            }
            regions <- seqinfo(annotations)
            tryCatch({
                regions <- as(regions, "GRanges")
            }, error=function(.e){
                stop("Cannot get genomic informations from annotations.",
                     "Check it by as(seqinfo(Annotation(obj[['",
                     atacAssayName, "']])), 'GRanges')")
            })
            ## get fragments for each cell and group
            fragments <- peaks@fragments #GetAssayData(obj, layer = "fragments")
            if (missing(fragmentNameMapList)){
                fragmentNameMapList <- sc1meta$sampleID
                names(fragmentNameMapList) <- sc1meta$sampleID
                fragmentNameMapList <- rep(list(fragmentNameMapList),
                                           length(fragments))
            }
            stopifnot(is.list(fragmentNameMapList))
            stopifnot(
                'The length of fragmentNameMapList is not same as the fragment'
                      =length(fragmentNameMapList)==length(fragments))
            null <- lapply(seq_along(fragmentNameMapList), function(k){
                if(sum(sc1meta$sampleID  %in% fragmentNameMapList[[k]])==0){
                    stop('fragmentNameMapList[[',
                         k, ']] does not contain the cell names')
                }
            })
            ## check reads name
            for(k in seq_along(fragments)){
                fragment.path <- fragments[[k]]@path
                if(file.exists(fragment.path)){
                    tbx <- open(TabixFile(fragment.path, yieldSize=100))
                    on.exit({close(tbx)})
                    if(length(reads <- scanTabix(tbx)[[1]])){
                        reads <- read.table(text = reads)
                        colnames(reads) <- 
                            c("seqnames", "start", "end", "name", "score")
                        if(length(intersect(
                            reads$name,
                            names(fragmentNameMapList[[k]])))==0){
                            reads_name <- paste(head(reads$name, n=5),
                                                collapse=', ')
                            cells_name <- paste(head(
                                names(fragmentNameMapList[[k]]),
                                                     n=5),
                                                collapse=', ')
                            stop("The fragment ", k,
                                 " names are not same format as",
                                 " the cell names or the map vector.\n",
                                 "The top 5 fragment names: ",
                                 reads_name,
                                 "\n",
                                 "The top 5 cell names: ",
                                 cells_name,
                                 '\nPlease use the parameter ',
                                 'fragmentNameMapList',
                                 ' to map the correct name.')
                        }
                    }
                    close(tbx)
                    on.exit()
                }else{
                    stop('fragments file are missing for ', fragment.path,
                         '. Please check it by ?Fragment and ?UpdatePath')
                }
            }
        }
    }
    
    # save data
    sc1conf <- scConf
    sc1conf$dimred <- FALSE
    sc1meta <-
        sc1meta[, c("sampleID", as.character(sc1conf$ID)), with = FALSE]
    # Factor metadata again
    for (i in as.character(sc1conf[!is.na(sc1conf$fID)]$ID)) {
        sc1meta[[i]] <- factor(
            sc1meta[[i]],
            levels =
                strsplit(sc1conf[sc1conf$ID == i]$fID, "\\|")[[1]])
        levels(sc1meta[[i]]) <-
            strsplit(sc1conf[sc1conf$ID == i]$fUI, "\\|")[[1]]
        sc1conf[sc1conf$ID == i]$fID <- sc1conf[sc1conf$ID == i]$fUI
    }
    # Extract dimred and append to both XXXmeta.rds and XXXconf.rds...
    for (iDR in Reductions(obj)) {
        drMat <- Embeddings(obj[[iDR]])
        if (ncol(drMat) > 5) {
            drMat <- drMat[, seq.int(5)]
        }  # Take first 5 components only
        drMat <- drMat[sc1meta$sampleID,]          # Ensure ordering
        drMat <- as.data.table(drMat)
        sc1meta <- cbind(sc1meta, drMat)
        
        # Update sc1conf accordingly
        tmp <- data.table(
            ID = colnames(drMat),
            UI = colnames(drMat),
            fID = NA,
            fUI = NA,
            fCL = NA,
            fRow = NA,
            default = 0,
            grp = FALSE,
            dimred = TRUE
        )
        tmp$UI <- gsub("_", "", tmp$UI)
        sc1conf <- rbindlist(list(sc1conf, tmp))
    }
    sc1conf$ID <- as.character(sc1conf$ID)     # Remove levels
    
    # Make XXXgexpr.h5
    if (!dir.exists(appDir)) {
        dir.create(appDir, recursive = TRUE)
    }
    
    filename <- file.path(appDir, .globals$filenames$sc1gexpr)
    if(h5createFile(filename)){
        if(h5createGroup(filename, .globals$h5fGrpPrefix)){
            if(h5createDataset(
                filename,
                dataset = .globals$h5fGrp,
                dims = gex.matdim,
                maxdims = gex.matdim,
                H5type = "H5T_NATIVE_FLOAT", #storage.mode(gexAsy[1]),
                chunk = c(1, gex.matdim[2]),
                filter = 'GZIP',
                level = 6)){
                chk <- chunkSize
                while (chk > (gex.matdim[1] - 8)) {
                    # Account for cases where nGene < chunkSize
                    chk <-
                        floor(chk / 2)
                }
                for (i in seq.int(floor((gex.matdim[1] - 8) / chk))) {
                    h5write(as.matrix(gexAsy[((i - 1) * chk + 1):(i * chk), ]),
                            file = filename,
                            name = .globals$h5fGrp,
                            index = list(((i - 1) * chk + 1):(i * chk), NULL))
                }
                h5write(as.matrix(gexAsy[(i * chk + 1):gex.matdim[1], ]),
                        file = filename,
                        name = .globals$h5fGrp,
                        index = list((i * chk + 1):gex.matdim[1], NULL))
            }else{
                stop("can not create dataset:", .globals$h5fGrp)
            }
        }else{
            stop("can not create group:", .globals$h5fGrpPrefix)
        }
    }else{
        stop("can not create file:", filename)
    }
    
    if (!isTRUE(all.equal(sc1meta$sampleID, gex.colnm))) {
        sc1meta$sampleID <- factor(sc1meta$sampleID, levels = gex.colnm)
        sc1meta <- sc1meta[order(sc1meta$sampleID)]
        sc1meta$sampleID <- as.character(sc1meta$sampleID)
    }
    
    # Make XXXgenes.rds
    sc1gene <- seq(gex.matdim[1])
    names(geneMap) <- NULL
    names(sc1gene) <- geneMap
    sc1gene <- sc1gene[order(names(sc1gene))]
    sc1gene <- sc1gene[order(nchar(names(sc1gene)))]
    
    # Make XXXdef.rds (list of defaults)
    if (all(default.dimred %in% sc1conf[sc1conf$dimred == TRUE]$ID)) {
        default.dimred[1] <- sc1conf[sc1conf$ID == default.dimred[1]]$UI
        default.dimred[2] <- sc1conf[sc1conf$ID == default.dimred[2]]$UI
    } else if (all(default.dimred %in% sc1conf[sc1conf$dimred == TRUE]$UI)) {
        default.dimred <- default.dimred    # Nothing happens
    } else {
        warn <- TRUE
        if (is.na(default.dimred[1])) {
            default.dimred <- "umap"
            warn <- FALSE
        }
        # Try to guess... and give a warning
        guess <- gsub("[0-9]", "", default.dimred[1])
        if (length(
            grep(
                guess, sc1conf[sc1conf$dimred == TRUE]$UI,
                ignore.case = TRUE)) >= 2) {
            default.dimred <-
                sc1conf[sc1conf$dimred == TRUE]$UI[
                    grep(
                        guess, sc1conf[sc1conf$dimred == TRUE]$UI,
                        ignore.case = TRUE)[c(1, 2)]]
        } else {
            nDR <- length(sc1conf[sc1conf$dimred == TRUE]$UI)
            default.dimred <-
                sc1conf[sc1conf$dimred == TRUE]$UI[(nDR - 1):nDR]
        }
        if (warn) {
            warning(
                "default.dimred not found, switching to ",
                default.dimred[1],
                " and ",
                default.dimred[1]
            )
        } # Warn if user-supplied default.dimred is not found
    }
    # Note that we stored the display name here
    sc1def <- list()
    sc1def$meta1 <-
        sc1conf[sc1conf$default == 1]$UI   # Use display name
    sc1def$meta2 <-
        sc1conf[sc1conf$default == 2]$UI   # Use display name
    sc1def$gene1 <- defaultGene1              # Actual == Display name
    sc1def$gene2 <- defaultGene2              # Actual == Display name
    sc1def$genes <-
        default.multigene          # Actual == Display name
    sc1def$dimred <- default.dimred            # Use display name
    tmp <- nrow(sc1conf[sc1conf$default != 0 & sc1conf$grp == TRUE])
    if (tmp == 2) {
        sc1def$grp1 <- sc1def$meta1
        sc1def$grp2 <- sc1def$meta2
    } else if (tmp == 1) {
        sc1def$grp1 <-
            sc1conf[sc1conf$default != 0 & sc1conf$grp == TRUE]$UI
        if (nrow(
            sc1conf[sc1conf$default == 0 & sc1conf$grp == TRUE]) == 0) {
            sc1def$grp2 <- sc1def$grp1
        } else {
            sc1def$grp2 <-
                sc1conf[sc1conf$default == 0 & sc1conf$grp == TRUE]$UI[1]
        }
    } else {
        sc1def$grp1 <-
            sc1conf[sc1conf$default == 0 & sc1conf$grp == TRUE]$UI[1]
        if (nrow(
            sc1conf[sc1conf$default == 0 & sc1conf$grp == TRUE]) < 2) {
            sc1def$grp2 <- sc1def$grp1
        } else {
            sc1def$grp2 <-
                sc1conf[sc1conf$default == 0 & sc1conf$grp == TRUE]$UI[2]
        }
    }
    sc1conf <- sc1conf[, -c("fUI", "default"), with = FALSE]
    
    ### Saving objects
    saveRDS(sc1conf, file = file.path(appDir, .globals$filenames$sc1conf))
    saveRDS(sc1meta, file = file.path(appDir, .globals$filenames$sc1meta))
    saveRDS(sc1gene, file = file.path(appDir, .globals$filenames$sc1gene))
    saveRDS(sc1def,  file = file.path(appDir, .globals$filenames$sc1def))
    
    ### save ATAC objects
    if (!missing(atacAssayName)) {
        if (atacAssayName %in% Assays(obj)) {
            rm(gexAsy)
            DefaultAssay(obj) <- atacAssayName
            peaks <- obj[[atacAssayName]]
            ## links, the links between peaks and gene symbol,
            ## used to create the matrix table
            links <- peaks@links #GetAssayData(obj, layer = "links")
            if (length(links) < 1) {
                warning("scATAC links data are not available.")
            }
            ## annotations, used to plot gene model
            annotations <- peaks@annotation #GetAssayData(obj, layer = "annotation")
            if (length(annotations) < 1) {
                stop("scATAC annotation data are not available.")
            }
            if(any(!c('gene_name', 'tx_id') %in% colnames(mcols(annotations)))){
                stop('gene_name and tx_id must exist as metadata for scATAC ',
                     'annotations. Check it by Annotation(obj[["',
                     atacAssayName, '"]])')
            }
            ## get fragments for each cell and group
            fragments <- peaks@fragments #GetAssayData(obj, layer = "fragments")
            regions <- seqinfo(annotations)
            tryCatch({
                regions <- as(regions, "GRanges")
            }, error=function(.e){
                warning("Cannot get genomic informations")
            })
            grp <- sc1conf[sc1conf$grp, ]$ID
            if(is(regions, "GRanges")){
                message("The following steps will cost memories.")
                res <- list()
                for(k in seq_along(fragments)){
                    fragment.path <- fragments[[k]]@path
                    if(file.exists(fragment.path)){
                        tabix.file <- TabixFile(fragment.path)
                        open(con = tabix.file)
                        on.exit(close(tabix.file))
                        region <- regions
                        seq_x <- as.character(seqnames(x = region))
                        seq_y <- seqnamesTabix(file = tabix.file)
                        seq_x_style <- seqlevelsStyle(seq_x)
                        seq_y_style <- seqlevelsStyle(seq_y)
                        if(length(intersect(seq_x_style, seq_y_style))==0){
                            seqlevelsStyle(region)<-seq_y_style[1]
                        }
                        seqnames.in.both <- intersect(
                            x = seqnames(x = region),
                            y = seqnamesTabix(file = tabix.file))
                        region <- keepSeqlevels(
                            x = region,
                            value = seqnames.in.both,
                            pruning.mode = "coarse")
                        message('Creating coverage for ', fragment.path)
                        coverage <- lapply(seq_along(region), function(i){
                            reads <- scanTabix(
                                file = tabix.file,
                                param = region[i])
                            reads <- read.table(text = reads[[1]])
                            colnames(reads) <- 
                                c("seqnames", "start", "end", "name", "score")
                            reads <- GRanges(reads)
                            tryCatch({
                                seqlevelsStyle(reads)<-seq_x_style[1]
                            }, error = function(e){
                                message(e, '\nCannot convert the seqstyle to ',
                                        seq_x_style[1],' for ', region[i])
                            })
                            reads.grp <- lapply(grp, function(.grp){
                                lapply(split(
                                    reads,
                                    sc1meta[[.grp]][
                                        match(fragmentNameMapList[[k]][
                                            reads$name],
                                            sc1meta$sampleID
                                              )]),
                                    function(.e){
                                        coverage(.e, weight = .e$score)
                                    })
                            })
                            rm(reads)
                            names(reads.grp) <- grp
                            reads.grp
                        })
                        seqlevelsStyle(region)<-seq_x_style[1]
                        names(coverage) <- as.character(seqnames(region))
                        ## coverage is 3 level list,
                        ## level 1, chromosome
                        ## level 2, group
                        ## level 3, factors in group
                        if(length(coverage)){
                            res[[k]] <- list()
                            for(i in names(coverage[[1]])){
                                res[[k]][[i]] <- list()
                                for(j in names(coverage[[1]][[i]])){
                                    res[[k]][[i]][[j]] <-
                                        Reduce(c, lapply(coverage,
                                                         function(.cvg){
                                            .cvg[[i]][[j]]
                                        }))
                                }
                            }
                        }
                        
                        close(tabix.file)
                        on.exit()
                    }
                }
                if(length(res)>0){
                    message('Exporting coverage to bigwig files.')
                    ## accumulate the signals
                    if(length(res)>1){
                        for(i in seq_along(res)[-1]){
                            N_grp <- names(res[[1]])
                            names(N_grp) <- N_grp
                            res[[1]] <- lapply(N_grp, function(.grp){
                                N_fac <- names(res[[1]][[.grp]])
                                names(N_fac) <- N_fac
                                lapply(N_fac, function(.fac){
                                    if(sum(lengths(
                                        res[[1]][[.grp]][[.fac]]))==0){
                                        return(res[[i]][[.grp]][[.fac]])
                                    }
                                    if(sum(lengths(
                                        res[[i]][[.grp]][[.fac]]))==0){
                                        return(res[[1]][[.grp]][[.fac]])
                                    }
                                    res[[1]][[.grp]][[.fac]] +
                                        res[[i]][[.grp]][[.fac]]
                                })
                            })
                        }
                    }
                    ## normalization
                    res <- lapply(res[[1]], function(.grp) {
                        mapply(
                            .grp,
                            rep(binSize, length(.grp))[seq_along(.grp)],
                            FUN=function(.fac, bs){
                                if(bs>1){
                                    # summarize by bin
                                    .fac_gr <- GRanges(.fac)
                                    wid <- width(.fac_gr)
                                    k <- wid<bs
                                    if(any(k)){
                                        .fac_0 <- .fac_gr[!k]
                                        .fac_1 <- .fac_gr[k]
                                        ## sumarize signal by bin size
                                        .fac_1.rd <- reduce(.fac_1,
                                                            with.revmap=TRUE)
                                        .fac_1_tile <- tile(.fac_1.rd, width=bs)
                                        .fac_1 <- unlist(.fac_1_tile)
                                        .fac_0 <- c(.fac_0, .fac_1)
                                        .fac_gr <- split(.fac_0,
                                                         seqnames(.fac_0))
                                        .fac <- 
                                            mapply(
                                                .fac, .fac_gr[names(.fac)],
                                                FUN=function(.rle, .gr){
                                                    .ir <- ranges(.gr)
                                                    .gr$score <-
                                                        viewMeans(Views(.rle,
                                                                        .ir),
                                                                  na.rm=TRUE)
                                                    .gr
                                                })
                                        rm(.fac_0, .fac_1,
                                           .fac_1.rd, .fac_1_tile, .fac_gr)
                                        .fac <- unlist(GRangesList(.fac))
                                    }else{
                                        .fac <- .fac_gr
                                    }
                                }else{
                                    .fac <- GRanges(.fac)
                                }
                                .fac <- .fac[.fac$score!=0]
                                ## normalize by FPKM
                                .s <- .fac$score * width(.fac)/1e3
                                .fac$score <- 1e6*.fac$score/sum(.s)
                                .fac
                            }, SIMPLIFY = FALSE)
                    })
                    ## export
                    mapply(res, names(res), FUN=function(.grp, .grpname){
                        .grp <- .grp[lengths(.grp)>0]
                        mapply(.grp, names(.grp), FUN=function(.fac, .facname){
                            .facname <- path_sanitize(.facname)
                            pf <- file.path(
                                appDir, .globals$filenames$bwspath, .grpname)
                            dir.create(pf,
                                recursive = TRUE, showWarnings=FALSE)
                            export(.fac, file.path(
                                pf,
                                paste0(.facname, ".bigwig")),
                                format = "BigWig")
                        })
                    })
                }
            }
            # asy used to create coverage files,
            # Note this is different from fragment signals
            # it just show the counts in each called peaks
            acAsy <- extAssayData(obj, assay = atacAssayName, slot = atacSlot)
            acAsy <- acAsy[, sc1meta$sampleID]
            peaks <- do.call(rbind, strsplit(rownames(acAsy), "-"))
            peaks <- as.data.frame(peaks)
            if(ncol(peaks)==3){
                writeATACdata(acAsy, appDir)
                colnames(peaks) <- c("seqnames", "start", "end")
                mode(peaks[, 2]) <- "numeric"
                mode(peaks[, 3]) <- "numeric"
                saveRDS(peaks, file = file.path(
                    appDir, .globals$filenames$sc1peak
                ))
                if(length(links)<1){
                    message(" Linking peak by nearest TSS.")
                    links <- GRanges(peaks)
                    n <- nearest(links, annotations)
                    links$gene <- annotations$gene_name[n]
                }
            }else{
                warning("Ask rownames of ATAC assays format is chr-start-end!")
            }
            saveRDS(links, file = file.path(
                appDir, .globals$filenames$sc1link
            ))
            saveRDS(annotations, file = file.path(
                appDir, .globals$filenames$sc1anno
            ))
        }
    }
    return(sc1conf)
}

#' Extract fragment name map
#' @description
#' Try to extract cell name information. It will be used to map the fragment cell
#' name to the object cell name.
#' @param obj input single-cell object for Seurat (v3+)
#' @param atacAssayName assay in single-cell data object to use for plotting
#' open chromatin.
#' @return A list of fragment name map.
#' @importFrom methods slotNames
#' @export
#' @examples
#' library(Seurat)
#' #library(Signac)
#' #fnm <- extractFragmentNameMapList(atac_small, 'peaks')
#' 

extractFragmentNameMapList <- function(obj, atacAssayName){
    stopifnot(is(obj, "Seurat"))
    fragments <- obj[[atacAssayName]]@fragments #GetAssayData(obj, layer = "fragments")
    if(length(fragments)==0) stop('No fragments available!')
    fragmentNameMapList <- lapply(fragments, function(.ele) {
        if(!'cells' %in% slotNames(.ele)) 
            stop('cells is not a slotName for the fragments!')
        n=names(.ele@cells)
        names(n) <- .ele@cells
        n
    })
}

#' waffle plot for single cell expressions
#' @noRd
#' @importFrom stats as.formula
#' @importFrom ggplot2 coord_equal facet_grid unit scale_fill_gradientn .data
#' @importFrom grDevices hcl.colors
#' @param expr expression table
#' @param groupCol The column name for `facet_grid`
#' @param gradientCol Color sets
#' @param xyMaxRatio The ratio of x,y. The parameter is used to avoid long plots
#' Not used yet.
scWafflePlot <- function(
        expr,
        groupCol = 'splitBy',
        gradientCol = rev(hcl.colors(20, "RdYlGn")),
        xyMaxRatio = 5) {
    if (length(gradientCol) == 1) {
        gradientCol <- availableThemes(gradientCol)
    }
    if (groupCol[1] %in% colnames(expr)) {
        if (!all(
            as.character(expr[[groupCol]]) ==
            as.character(expr$grpBy), na.rm = TRUE)) {
            expr$geneName <- paste(
                as.character(expr$geneName),
                as.character(expr[[groupCol]]))
        }
    }
    data <- expr[, c("geneName", "grpBy", "val"), with = FALSE]
    data <- data[order(expr$geneName, expr$grpBy, -1 * expr$val),]
    data <- data[!is.na(data$val),]
    groupMax <- data[, {
        list(N = max(.SD[, list(N = .N), by = 'grpBy']$N))
    }, by = 'geneName']
    data <- merge(data, groupMax)
    ggData <- data[, {
        this_N <- max(100, .SD$N[1])
        .SD[, {
            list(idx <- as.numeric(
                cut(seq.int(this_N), 100))[seq.int(.N)])
            .SD[, {
                list(exprs = mean(.SD$val, na.rm = TRUE))
            }, by = 'idx']
        }, by = 'grpBy']
    }, by = 'geneName']
    
    ggData$x <- (ggData$idx - 1) %% 10 + 1
    ggData$y <- floor((ggData$idx - 1) / 10) + 1
    
    # len_cols <- length(unique(data$grpBy))
    # len_rows <- length(unique(data$geneName))
    # xyRatio <- len_rows / len_cols
    # if (xyRatio < 1 / xyMaxRatio) {
    #     ## plot is too wider
    #     ## change the geneName
    #     ## add blank after geneName by grpBy group
    # } else{
    #     if (xyRatio > xyMaxRatio) {
    #         ## plot is too long
    #         ## change the grpBy
    #         ## add blank after grpBy
    #     }
    # }
    
    p <- ggplot(ggData, aes(
        .data[["x"]], .data[["y"]], fill = .data[["exprs"]])) +
        geom_tile(color = 'white') + coord_equal() +
        facet_grid(as.formula("geneName ~ grpBy")) +
        theme_minimal() +
        theme(
            axis.text = element_blank(),
            axis.ticks = element_blank()) +
        guides(fill = guide_colorbar(barwidth = unit(3, "mm"))) +
        xlab("") + ylab("") +
        scale_fill_gradientn(colors = gradientCol)
    return(p)
}


scDRwafflePlot <- function(
        dataset,
        geneIdMap,
        inpConf,
        inpMeta,
        gene,
        groupBy,
        splitBy,
        gradientCol,
        grpKey,
        grpVal) {
    if (isQuote(gene)) {
        genenames <- geneIdMap[
            names(geneIdMap) %in% removeQuote(gene)]
    } else{
        if (isAsterisk(gene)) {
            gene <- isAsterisk(gene, transform = TRUE)
            genenames <-
                geneIdMap[grepl(gene, names(geneIdMap), ignore.case = TRUE)]
        } else{
            geneList <- scGeneList(gene, geneIdMap)
            geneList <- geneList[geneList$present == TRUE]
            genenames <- geneList$gene
        }
    }
    if (length(genenames) > 0) {
        #waffle plot
        if (missing(groupBy)) {
            groupBy <- getCelltypeCol(
                inpConf,
                celltypePattern =
                    .globals$groupColPattern)
        }
        if (missing(splitBy)) {
            splitBy <- NA
        }
        ggData <-
            read_exprs(
                dataset,
                geneIdMap[genenames],
                inpMeta,
                inpConf,
                groupBy,
                splitBy,
                valueOnly = FALSE)
        ggData[ggData$val < 0]$val <- 0
        
        ggData <- subGrp(ggData, grpKey, grpVal, inpConf)
        if("filter" %in% colnames(ggData)){
            ggData$filter <- NULL
        }
        return(scWafflePlot(
            ggData, 'splitBy', gradientCol))
    } else{
        return(ggplot())
    }
}

scExprUI <- function(id, postfix = 1, subgrp=.globals$subsetgroup[1]) {
    subModuleContainerUI(
        id,
        mainSelectUI = selectInput(
            NS0(id, "GeneName", postfix),
            "Gene name:", choices =
                NULL),
        menuUI = contextMenuGeneExprUI(id, postfix, group = subgrp),
        contentUI = geneExprDotPlotUI(id, postfix)
    )
}
scExprServer <- function(
        pid,
        id,
        dataSource,
        optCrt,
        p_input,
        p_session,
        interactive,
        postfix = 1) {
    moduleServer(id, function(input, output, session) {
        ## title
        GeneNameLabel <- paste0('GeneName', postfix)
        if (is.null(
            p_session$userData$defaults[[dataSource()$dataset]][[id]])) {
            defaults <- list()
            defaults[[GeneNameLabel]] <- dataSource()$sc1def$gene1
        } else{
            defaults <-
                p_session$userData$defaults[[dataSource()$dataset]][[id]]
        }
        updateSelectizeInput(
            session,
            GeneNameLabel,
            choices = sort(names(dataSource()$sc1gene)),
            server = TRUE,
            selected = defaults[[GeneNameLabel]],
            options = list(
                maxOptions = .globals$maxNumGene,
                create = TRUE,
                persist = TRUE,
                render = I(optCrt)
            )
        )
        subModuleMenuObservor(
            id,
            input,
            p_session,
            dataSource,
            c(
                GeneNameLabel,
                paste0("GeneExprcol", postfix),
                paste0("GeneExprord", postfix),
                paste0("GeneExprtype", postfix),
                paste0("GeneExprxlim", postfix),
                paste0("GeneExprrg", postfix)
            )
        )
        ### plots
        plotX <- reactive({
            scDRgene(
                inpConf=dataSource()$sc1conf,
                inpMeta=dataSource()$sc1meta,
                dimRedX=p_input$GeneExprdrX,
                dimRedY=p_input$GeneExprdrY,
                gene1=input[[GeneNameLabel]],
                subsetCellKey=p_input[[paste0("subsetCell",
                                              input[[paste0("CellInfosubgrp",
                                                            postfix)]])]],
                subsetCellVal=
                    getSubsetCellVal(p_input,
                                     group=input[[paste0("CellInfosubgrp",
                                                         postfix)]]),
                dataset=dataSource()$dataset,
                geneIdMap=dataSource()$sc1gene,
                pointSize=p_input$GeneExprsiz,
                gradientCol=input[[paste0("GeneExprcol", postfix)]],
                GeneExprDotOrd=input[[paste0("GeneExprord", postfix)]],
                labelsFontsize=p_input$GeneExprfsz,
                labelsFontFamily=p_input$GeneExprfml,
                plotAspectRatio=p_input$GeneExprasp,
                keepXYlables=p_input$GeneExprtxt,
                inpPlt=input[[paste0("GeneExprtype", postfix)]],
                inpXlim=if (input[[paste0("GeneExprxlimb", postfix)]] %% 2 == 0)
                    0
                else
                    input[[paste0("GeneExprxlim", postfix)]],
                inpColRange =
                    if (input[[paste0("GeneExprrgb", postfix)]] %% 2 == 0)
                        0
                    else
                        input[[paste0("GeneExprrg", postfix)]],
                valueFilterKey = p_input$filterCell,
                valueFilterCutoff = p_input$filterCellVal,
                valueFilterCutoff2 = p_input$filterCellVal2,
                hideFilterCell = input[[paste0("GeneExprhid", postfix)]]
            )
        })
        updateSubModulePlotUI(
            postfix,
            pid,
            id,
            input,
            output,
            session,
            interactive,
            plotX,
            .globals$pList1[p_input$GeneExprpsz],
            dataSource()$dataset,
            p_input$GeneExprdrX,
            p_input$GeneExprdrY,
            input[[GeneNameLabel]]
        )
    })
}

plotProportionUI <- function(id) {
    tabPanel(
        value = id,
        HTML("Proportion plot"),
        h4("Proportion / cell numbers across different cell information"),
        "In this tab, users can visualise the composition of single cells ",
        "based on one discrete ",
        "cell information across another discrete cell information. ",
        "Usage examples include the library ",
        "or cellcycle composition across clusters.",
        br(),
        br(),
        fluidRow(
            column(
                3,
                style = "border-right: 2px solid black",
                xaxisCellInfoUI(id),
                subsetCellByInfoUI(id, mini = TRUE),
                selectInput(
                    NS(id, "CellInfoY"),
                    "Cell information to group / colour by:",
                    choices = NULL
                ) %>%
                    shinyhelper::helper(
                        type = "inline",
                        size = "m",
                        fade = TRUE,
                        title = "Cell information to group / colour cells by",
                        content = c(
                            "Select categorical cell information to group / ",
                            "colour cells by",
                            "- Proportion / cell numbers are shown ",
                            "in different colours"
                        )
                    ),
                radioButtons(
                    NS(id, "plottyp"),
                    "Plot value:",
                    choices = c("Proportion", "CellNumbers"),
                    selected = "Proportion",
                    inline = TRUE
                ),
                checkboxInput(
                    NS(id, "plotflp"),
                    "Flip X/Y", value = FALSE),
                checkboxInput(
                    NS(id, "plotord"),
                    "Reorder the contents", value = FALSE
                ),
                conditionalPanel(
                    condition = "input.plotord % 2 == 1",
                    ns=NS(id),
                    uiOutput(outputId = NS(id, "plotXord")),
                    uiOutput(outputId = NS(id, "plotYord"))
                ),
                boxPlotControlUI(id, withPoints = FALSE)
            ),
            column(
                9,
                geneExprDotPlotUI(id),
                br(),
                actionButton(NS(id, "statstog"), "Toggle to show Chi-squared statistics"),
                conditionalPanel(
                    condition = "input.statstog % 2 == 1",
                    ns = NS(id),
                    DTOutput(NS(id, "proportion.dt"))
                ),
                actionButton(NS(id, "pearsontog"),
                             "Toggle to show correlation plot"),
                conditionalPanel(
                    condition = "input.pearsontog % 2 == 1",
                    ns = NS(id),
                    radioButtons(
                        NS(id, "testby"),
                        "Group by:",
                        choices = c("X", "Y"),
                        selected = "Y",
                        inline = TRUE
                    ),
                    radioButtons(
                        NS(id, "testmethod"),
                        "Method:",
                        choices = c("spearman", "kendall", "pearson"),
                        selected = "spearman",
                        inline = TRUE
                    ),
                    radioButtons(
                        NS(id, "cortyp"),
                        "Method:",
                        choices = c("heatmap", "PCA"),
                        selected = "heatmap",
                        inline = TRUE
                    ),
                    radioButtons(
                        NS(id, "plotcols"), "Colour scheme:",
                        choices = availableThemes("sequence"),
                        selected = availableThemes("sequence")[2]),
                    geneExprDotPlotUI(id, postfix=2)
                )
            )
        )
    )
}
#' @importFrom DT datatable
#' @importFrom sortable rank_list
#' @importFrom data.table dcast.data.table
#' @importFrom stats chisq.test
plotProportionServer <- function(id, dataSource, optCrt) {
    moduleServer(id, function(input, output, session) {
        ## input column
        updateSelectInput(
            session,
            "CellInfoX",
            "Cell information (X-axis):",
            choices = getGroupUI(dataSource),
            selected = dataSource()$sc1def$grp2
        )
        
        updateSubsetCellUI(id, input, output, session, dataSource, addNA = TRUE)
        
        updateSelectInput(
            session,
            "CellInfoY",
            "Cell information to group / colour by:",
            choices = getGroupUI(dataSource),
            selected = dataSource()$sc1def$grp1
        )
        
        updateRankList(
            input, output, dataSource, "CellInfoX", "plotXord",
            NS(id, "cellinfoXorder"))
        updateRankList(
            input, output, dataSource, "CellInfoY", "plotYord",
            NS(id, "cellinfoYorder"))
        
        ## plot region
        ### plots
        plot1 <- reactive({
            scProp(
                inpConf = dataSource()$sc1conf,
                inpMeta = dataSource()$sc1meta,
                infoX = input$CellInfoX,
                infoY = input$CellInfoY,
                subsetCellKey = input$subsetCell,
                subsetCellVal = getSubsetCellVal(input),
                inptyp = input$plottyp,
                flipXY = input$plotflp,
                labelsFontsize = input$plotfsz,
                labelsFontFamily = input$plotfml,
                reorder = input$plotord,
                orderX = input$cellinfoXorder,
                orderY = input$cellinfoYorder
            )
        })
        updateGeneExprDotPlotUI(
            postfix = 1,
            id,
            input,
            output,
            session,
            plot1,
            .globals$pList2[input$plotpsz],
            dataSource()$dataset,
            input$plottyp,
            input$CellInfoX,
            input$CellInfoY
        )
        
        getProportion <- function(plot1){
            proportion <- plot1()$data
            proportion <- dcast.data.table(proportion, as.formula('X~grp'),
                                           value.var = 'nCells')
            proportion[is.na(proportion)] <- 0
            proportion <- as.data.frame(proportion)
            rownames(proportion) <- proportion$X
            proportion$X <- NULL
            proportion
        }
        output$proportion.dt <- renderDT({
            proportion <- getProportion(plot1)
            withCallingHandlers({
                chisq <- chisq.test(proportion)
            }, warning = function(w){
                showNotification(
                    as.character(w),
                    duration = 5,
                    type = 'warning'
                )
            }, error = function(e){
                showNotification(
                    as.character(e),
                    duration = 5,
                    type = 'error'
                )
                chisq <- list('observed'=data.frame(nodata=NA),
                              'expected'=data.frame(nodata=NA),
                              'residuals'=data.frame(nodata=NA),
                              'p.value'=NA)
            })
            
            chi_out <- lapply(c('observed', 'expected', 'residuals'),
                              function(.ele){
                                  colnames(chisq[[.ele]]) <-
                                      paste(.ele, colnames(chisq[[.ele]]),
                                            sep='.')
                                  chisq[[.ele]]
                              })
            chi_out <- do.call(cbind, chi_out)
            chi_out <- as.data.frame(chi_out)
            chi_out <- cbind(chisq.test.p.value=
                                 c(chisq$p.value,
                                   rep(NA, nrow(chi_out)-1)),
                             chi_out)
            datatable(
                chi_out,
                rownames = TRUE,
                extensions = "Buttons",
                options = list(
                    pageLength = -1,
                    dom = "tB",
                    buttons = c("copy", "csv", "excel")
                )
            )
        })
        
        plot2 <- reactive({
            scCorProp(
                getProportion(plot1),
                testby = input$testby,
                method = input$testmethod,
                type = input$cortyp,
                labelsFontsize = input$plotfsz,
                labelsFontFamily = input$plotfml,
                inpcols = input$plotcols
            )
        })
        updateGeneExprDotPlotUI(
            postfix = 2,
            id,
            input,
            output,
            session,
            plot2,
            .globals$pList2[input$plotpsz],
            dataSource()$dataset,
            input$plottyp,
            input$CellInfoX,
            input$CellInfoY
        )
    })
}

# rename the function to compatable with old functions
getDataSets <- function(datasets, appconf, privilege=NULL) {
    datasets <- listDatasets(key = datasets,
                             privilege=privilege,
                             named=missing(appconf))
    return(datasets)
}
# rename the old getDataSets function to checkAvailableDataSets
# to check the datasets by list the folders
checkAvailableDataSets <- function(privilege='all', token=NULL) {
    datasets <- list.dirs(
        .globals$datafolder,
        full.names = FALSE,
        recursive = FALSE)
    tokenList <- list()
    if(!is.null(token)){
        tokenList <- getTokenList()
    }
    datasets <- datasets[vapply(
        datasets,
        FUN = checkFiles,
        FUN.VALUE = logical(1L),
        privilege = privilege,
        token = token,
        tokenList = tokenList)]
    return(datasets)
}

getNamedDataSets <- function(privilege=NULL) {
    datasets <- checkAvailableDataSets(privilege = privilege)
    names(datasets) <- datasets
    appconf <- getAppConfObj(datasets=datasets, privilege=privilege)
    if(all(datasets==names(appconf))){
        n <- vapply(appconf, function(.ele)
            .ele$title,
            FUN.VALUE = character(1))
        if (length(n) == length(datasets)) {
            names(datasets) <- n
        }
    }
    datasets <- datasets[order(names(datasets))]
}


# check if all the required files are available
checkFiles <- function(folder, privilege, token, tokenList) {
    stopifnot(length(folder)==1)
    if(checkLocker(folder)){
        if(!checkPrivilege(privilege=privilege, datasetname = folder) &&
           !checkToken(tokenList=tokenList, token = token, dataset = folder)){
            return(FALSE)
        }
    }
    if(!checkAccessibility(folder, bool=TRUE)){
        return(FALSE)
    }
    all(
        c(
            .globals$filenames$appconf,
            .globals$filenames$sc1conf,
            .globals$filenames$sc1def,
            .globals$filenames$sc1gene,
            .globals$filenames$sc1gexpr,
            .globals$filenames$sc1meta
        ) %in%
            list.files(
                file.path(.globals$datafolder, folder),
                full.names = FALSE)
    )
    
}

getDefaultDataset <- function(
        defaultDataset = "pbmc_small", datasets, privilege=NULL) {
    if(missing(datasets)) datasets <- getDataSets(privilege = privilege)
    if (!defaultDataset %in% datasets) {
        defaultDataset <- datasets[1]
    }
    defaultDataset
}

getAppConf <- function(datasets, privilege=NULL) {
    if(missing(datasets)) datasets <- getDataSets(privilege = privilege)
    appconf <- getConfigTable()
    appconf <- appconf[match(datasets, appconf$id), , drop=FALSE]
    keep <- !is.na(appconf$id)
    appconf <- appconf[keep, , drop=FALSE]
    appconf <- apply(appconf, 1, as.list)
    names(appconf) <- datasets[keep]
    return(appconf)
}

getAppConfObj <- function(datasets, privilege=NULL) {
    if(missing(datasets)){
        datasets <- checkAvailableDataSets(privilege = privilege)
    }
    appconf <- lapply(datasets, function(.ele) {
        conf <- readData("appconf", .ele)
        stopifnot(
            'id is not identical with the folder name' =
                .ele == conf$id)
        conf
    })
    names(appconf) <- datasets
    return(appconf)
}

getDataType <- function(appconf) {
    stopifnot(!missing(appconf))
    data_types <- vapply(appconf, function(.ele)
        .ele$type,
        FUN.VALUE = character(1))
}

updateDatasetForToken <- function(defaultDataset, datasets){
    datasets1 <- getDataSets(privilege = 'all')
    datasets <- datasets1[datasets1 %in% c(datasets, defaultDataset)]
    appconf <- getAppConf(datasets = datasets, privilege = 'all')
    datasets <- listDatasets(key = datasets,
                             privilege='all',
                             named=TRUE)
    data_types <- getDataType(appconf)
    return(list(datasets=datasets, appconf=appconf, data_types=data_types))
}

updateSymbolDict <- function(datasets, privilege=NULL, updateDB = FALSE) {
    if(missing(datasets)) datasets <- getDataSets(privilege = privilege)
    touchGeneTable(updateDB)
    symbols <- listGeneSymbols(datasets = datasets)
    symbols$symbol
}

getRef <- function(dataset, key, appconf) {
    stopifnot(!missing(appconf))
    stopifnot(!missing(dataset))
    stopifnot(!missing(key))
    if (key %in% names(appconf[[dataset]])) {
        appconf[[dataset]][[key]]
    } else{
        NA
    }
}

trimBib <- function(bib) {
    return(sub("^(<.*?>)\\[\\d+\\]", "\\1", gsub("^\\s+", "", bib)))
}
magicQuote <- function(s) {
    return(gsub('"', "'", s, fixed = TRUE))
}
get_full_ref_list <- function(appconf, returnLen = FALSE) {
    ref <- lapply(appconf, function(.ele) {
        .ele <- .ele$ref
        if (!is.null(.ele$entry)) {
            bib <- format(.ele$entry, style = "html")
            bib <- trimBib(bib)
            return(list(TRUE, bib))
        }
        if (!is.null(.ele$bib)) {
            bib <- trimBib(.ele$bib)
            if (bib != "" && !is.na(bib)) {
                if (!is.null(.ele$doi)) {
                    bib <- paste(
                        bib,
                        paste0(
                            "<a href='https://doi.org/",
                            .ele$doi,
                            "'>",
                            .ele$doi,
                            "</a>"
                        )
                    )
                }
                if (!is.null(.ele$pmid)) {
                    bib <- paste(
                        bib,
                        paste0(
                            "<a href='https://www.ncbi.nlm.nih.gov/pubmed",
                            .ele$pmid,
                            "'>PMID:",
                            .ele$pmid,
                            "</a>"
                        )
                    )
                }
                return(list(TRUE, bib))
            }
        }
        return(FALSE)
    })
    keep <- vapply(ref, function(.ele)
        .ele[[1]], logical(1L))
    ref <- ref[keep]
    ref <- unlist(lapply(ref, function(.ele)
        .ele[[2]]))
    ref <- sort(ref)
    ref <- unique(ref)
    if (returnLen)
        return(length(ref))
    ref <- paste("<li>", ref, "</li>")
    ref <- paste("<ol>\n", paste(ref, collapse = "\n"), "\n</ol>")
    HTML(ref)
}
checkAccessibility <- function(f, bool=FALSE){
    fa <- file.access(file.path(.globals$datafolder, f), mode = 4) != -1 # file not readable
    if(bool) return(fa)
    if(any(fa)){
        showNotification('Some file not accessable.', type='error')
        error_message <- paste("File'", 
                               paste(f[fa], collapse = "','"),
                                "'is not readable. Please check permissions.")
        warning(error_message)
        datasets <- paste(basename(dirname(f[fa])),
                          collapse = ', ')
        insertComments(uid='system',
                       email='system@sys.account',
                       title='file not accessible',
                       comment=paste('File not readable for', datasets),
                       dataset=datasets)
        f <- f[!fa]
    }
    return(f)
}
getTokenList <- function() {
    token <- dir(
        .globals$datafolder,
        .globals$filenames$token,
        recursive = TRUE,
        full.names = TRUE
    )
    token <- checkAccessibility(token)
    token_n <- lapply(token, readLines)
    token <- rep(basename(dirname(token)), lengths(token_n))
    names(token) <- unlist(token_n)
    token <- as.list(token)
}

checkLocker <- function(datasetname) {
    file.exists(file.path(
        .globals$datafolder,
        datasetname,
        .globals$filenames$locker
    ))
}

checkToken <- function(tokenList, token, dataset) {
    if(length(token)==0){
        return(FALSE)
    }
    if (token %in% names(tokenList)) {
        return(tokenList[[token]] == dataset)
    } else{
        return(FALSE)
    }
}

checkPrivilege <- function(privilege, datasetname) {
    if (is.null(privilege))
        return(FALSE)
    privilege == "all" || grepl(datasetname, privilege, fixed = TRUE)
}

checkMisc <- function(slot, folder) {
    file.exists(file.path(
        .globals$datafolder,
        folder,
        .globals$filenames[[slot]]))
}

scInfoUI <- function(id, postfix = 1, subgrp=.globals$subsetgroup[1]) {
    subModuleContainerUI(
        id,
        mainSelectUI = selectInput(
            NS0(id, "CellInfo", postfix),
            "Cell info:",
            choices = NULL),
        menuUI = contextMenuCellInfoUI(id, postfix, group=subgrp),
        contentUI = geneExprDotPlotUI(id, postfix)
    )
}
scInfoServer <- function(
        pid,
        id,
        dataSource,
        optCrt,
        p_input,
        p_session,
        interactive,
        postfix = 1) {
    moduleServer(id, function(input, output, session) {
        ## title
        cellInfoLabel <- paste0('CellInfo', postfix)
        if (is.null(
            p_session$userData$defaults[[dataSource()$dataset]][[id]])) {
            defaults <- list()
            defaults[[cellInfoLabel]] <- dataSource()$sc1def$meta1
        } else{
            defaults <-
                p_session$userData$defaults[[dataSource()$dataset]][[id]]
        }
        updateSelectInput(
            session,
            cellInfoLabel,
            "Cell info:",
            choices = dataSource()$sc1conf$UI,
            selected = defaults[[cellInfoLabel]]
        )
        subModuleMenuObservor(
            id,
            input,
            p_session,
            dataSource,
            c(
                cellInfoLabel,
                paste0("CellInfocol", postfix),
                paste0("CellInfoord", postfix),
                paste0("CellInfolab", postfix),
                paste0("CellInfoslingshot", postfix)
            )
        )
        ## plot
        plotX <- reactive({
            scDRcell(
                inpConf=dataSource()$sc1conf,
                inpMeta=dataSource()$sc1meta,
                dimRedX=p_input$GeneExprdrX,
                dimRedY=p_input$GeneExprdrY,
                cellinfoID=input[[cellInfoLabel]],
                subsetCellKey=p_input[[paste0("subsetCell",
                                              input[[paste0("CellInfosubgrp",
                                                            postfix)]])]],
                subsetCellVal=
                    getSubsetCellVal(p_input,
                                     group=input[[paste0("CellInfosubgrp",
                                                         postfix)]]),
                pointSize=p_input$GeneExprsiz,
                gradientCol=input[[paste0("CellInfocol", postfix)]],
                GeneExprDotOrd=input[[paste0("CellInfoord", postfix)]],
                labelsFontsize=p_input$GeneExprfsz,
                labelsFontFamily=p_input$GeneExprfml,
                plotAspectRatio=p_input$GeneExprasp,
                keepXYlables=p_input$GeneExprtxt,
                inplab=input[[paste0("CellInfolab", postfix)]],
                dataset = dataSource()$dataset,
                geneIdMap = dataSource()$sc1gene,
                valueFilterKey = p_input$filterCell,
                valueFilterCutoff = p_input$filterCellVal,
                valueFilterCutoff2 = p_input$filterCellVal2,
                hideFilterCell=input[[paste0("CellInfolab", postfix)]],
                inpSlingshot = input[[paste0("CellInfoslingshot", postfix)]],
                slingshotFilename = file.path(
                    .globals$datafolder,
                    dataSource()$dataset,
                    .globals$filenames[["slingshot"]]
                ),
                inpShowEdge = input[[paste0("CellInfoedge", postfix)]],
                edgeFilename = file.path(
                    .globals$datafolder,
                    dataSource()$dataset,
                    .globals$filenames[["sc1edge"]]
                ),
                interactive = interactive
            )
        })
        updateSubModulePlotUI(
            postfix,
            pid,
            id,
            input,
            output,
            session,
            interactive,
            plotX,
            .globals$pList1[p_input$GeneExprpsz],
            dataSource()$dataset,
            p_input$GeneExprdrX,
            p_input$GeneExprdrY,
            input[[cellInfoLabel]]
        )
    })
}

#' @importFrom DT DTOutput
#' @importFrom magrittr %>%
coExprUI <- function(id) {
    tabPanel(
        value = id,
        htmlOutput(NS(id, 'GeneExpr')),
        tabsubTitleUI(
            id,
            'GeneExpr',
            description = paste(
                "In this tab, users can visualise the coexpression of two ",
                "genes on low-dimensional representions."
            )
        ),
        fluidRow(
            column(3, dimensionReductionUI(id)),
            column(5, subsetCellByInfoUI(id)),
            column(4, graphicsControlUI(id))
        ),
        fluidRow(
            column(
                3,
                style = "border-right: 2px solid black",
                htmlOutput(NS0(id, "subPlotTitle", 1)),
                geneExprUI(id, 1),
                geneExprUI(id, 2),
                geneCoExprPlotControlUI(id, 1)
            ),
            column(
                9,
                fluidRow(
                    column(
                        8, style = "border-right: 2px solid black",
                        geneExprDotPlotUI(id, 1)),
                    column(
                        4,
                        geneExprDotPlotUI(id, 2),
                        br(),
                        h4("Cell numbers"),
                        DTOutput(NS(id, "coExpr.dt"))
                    )
                ))
        )
    )
}
#' @importFrom DT renderDT
#' @importFrom magrittr %>%
#' @importFrom plotly plotlyOutput renderPlotly
coExprServer <- function(id, dataSource, optCrt) {
    moduleServer(id, function(input, output, session) {
        ## title
        output$GeneExpr <-
            renderUI({
                HTML(paste("Gene", dataSource()$terms['coexpression']))
            })
        ## subtitle
        output$GeneExprSubTitle <-
            renderUI({
                h4(paste(
                    sub(
                        substr(dataSource()$terms['coexpression'], 1, 1),
                        toupper(substr(
                            dataSource()$terms['coexpression'], 1, 1)),
                        dataSource()$terms['coexpression']
                    ),
                    "of two genes on reduced dimensions"
                ))
            })
        ## input column 1
        ### Dimension Reduction
        updateDimRedSelInputPair(session, input, dataSource)
        ## input column 2
        updateSubsetCellUI(id, input, output, session, dataSource)
        
        ## plot region
        ### sub region title
        output$subPlotTitle1 <-
            renderUI({
                h4(paste("Gene", dataSource()$terms['expression']))
            })
        ### dropdown list
        updateSelectizeInput(
            session,
            "GeneName1",
            choices = sort(names(dataSource()$sc1gene)),
            server = TRUE,
            selected = dataSource()$sc1def$gene1,
            options = list(
                maxOptions = .globals$maxNumGene,
                create = TRUE,
                persist = TRUE,
                render = I(optCrt)
            )
        )
        updateSelectizeInput(
            session,
            "GeneName2",
            choices = sort(names(dataSource()$sc1gene)),
            server = TRUE,
            selected = dataSource()$sc1def$gene2,
            options = list(
                maxOptions = .globals$maxNumGene,
                create = TRUE,
                persist = TRUE,
                render = I(optCrt)
            )
        )
        ### plots
        plot1 <- reactive({
            scDRcoex(
                inpConf=dataSource()$sc1conf,
                inpMeta=dataSource()$sc1meta,
                dimRedX=input$GeneExprdrX,
                dimRedY=input$GeneExprdrY,
                gene1=input$GeneName1,
                gene2=input$GeneName2,
                subsetCellKey=input$subsetCell,
                subsetCellVal=getSubsetCellVal(input),
                dataset=dataSource()$dataset,
                geneIdMap=dataSource()$sc1gene,
                plotType="2D",
                pointSize=input$GeneExprsiz,
                GeneExprDotCol=input$CoExprcol1,
                GeneExprDotOrd=input$CoExprord1,
                labelsFontsize=input$GeneExprfsz,
                labelsFontFamily=input$GeneExprfml,
                plotAspectRatio=input$GeneExprasp,
                keepXYlables=input$GeneExprtxt,
                hideFilterCell=input$CoExprhid1
            )
        })
        updateGeneExprDotPlotUI(
            postfix = 1,
            id,
            input,
            output,
            session,
            plot1,
            .globals$pList1[input$GeneExprpsz],
            dataSource()$dataset,
            input$GeneExprdrX,
            input$GeneExprdrY,
            input$GeneName1,
            input$GeneName2
        )
        
        plot2 <- reactive({
            scDRcoexLeg(input$GeneName1,
                        input$GeneName2,
                        colorPairs = input$CoExprcol1,
                        labelsFontsize = input$GeneExprfsz,
                        labelsFontFamily=input$GeneExprfml)
        })
        updateGeneExprDotPlotUI(
            postfix = 2,
            id,
            input,
            output,
            session,
            plot2,
            300,
            dataSource()$dataset,
            input$GeneExprdrX,
            input$GeneExprdrY,
            input$GeneName1,
            input$GeneName2
        )
        
        output$coExpr.dt <- renderDT({
            ggData <- scDRcoexNum(
                dataSource()$sc1conf,
                dataSource()$sc1meta,
                input$GeneName1,
                input$GeneName2,
                input$subsetCell,
                getSubsetCellVal(input),
                dataSource()$dataset,
                dataSource()$sc1gene
            )
            datatable(
                ggData,
                rownames = FALSE,
                extensions = "Buttons",
                options = list(
                    pageLength = -1,
                    dom = "tB",
                    buttons = c("copy", "csv", "excel")
                )
            ) %>%
                formatRound(columns = c("percent"),
                            digits = 2)
        })
    })
}

downloaderUI <- function(id) {
    files <- dir(file.path(
        .globals$app_path,
        .globals$downloadFolder),
                 full.names = TRUE,
                 recursive = TRUE)
    if(length(files)){
        files <- sub(file.path(
            .globals$app_path,
            .globals$downloadFolder), "", files, fixed = TRUE)
        folders <- dirname(files)
        files <- split(files, folders)
        fr <- mapply(files, names(files), FUN=function(fn, fd){
            readme <- file.path(fd, 'readme')
            fluidRow(
                div(strong(sub('^\\/', '', fd)),
                    if(readme %in% fn){
                        includeText(file.path(.globals$app_path,
                                              .globals$downloadFolder,
                                              readme))
                    }),
                div(class='inline-float',
                    lapply(fn, function(.fn){
                        downloadButton(NS(id, make.names(.fn)),
                                       basename(.fn))}))
            )
        }, SIMPLIFY = FALSE)
        tabPanel(
            value = id,
            HTML("Download"),
            h4("Download data"),
            "In this tab, users can view and download data.",
            br(),
            br(),
            fr
        )
    }
}
downloaderServer <- function(id) {
    moduleServer(id, function(input, output, session) {
        files <- dir(file.path(
            .globals$app_path,
            .globals$downloadFolder),
            full.names = TRUE,
            recursive = TRUE)
        if(length(files)){
            files <- sub(file.path(
                .globals$app_path,
                .globals$downloadFolder), "", files, fixed = TRUE)
            folders <- dirname(files)
            files <- split(files, folders)
            observe({
                lapply(seq_along(files), function(i){
                    lapply(seq_along(files[[i]]), function(j){
                        output[[make.names(files[[i]][j])]] <- downloadHandler(
                            filename = basename(files[[i]][j]),
                            content = function(file){
                                file.copy(file.path(.globals$app_path,
                                                    .globals$downloadFolder,
                                                    files[[i]][j]),
                                          file)
                            }
                        )
                    })
                })
            })
        }
    })
}

plotMonocleUI <- function(id) {
    ns <- NS(id)
    tabPanel(
        value = id,
        HTML("Monocle"),
        h4("Monocle plot cell"),
        "In this tab, users can visualise the cells",
        "along with their trajectories.",
        br(),
        br(),
        fluidRow(
            column(
                3,
                style = "border-right: 2px solid black",
                selectInput(
                    ns("group"),
                    "select the data group",
                    choices = c(),
                    selected = NULL
                ),
                selectInput(
                    ns("root"),
                    "select the root point",
                    choices = c(),
                    selected = NULL
                ),
                checkboxInput(
                    ns("show_trajectory_graph"),
                    "show trajectory graph",
                    value = TRUE
                ),
                checkboxInput(
                    ns("label_principal_points"),
                    "label principal points",
                    value = TRUE
                ),
                checkboxInput(
                    ns("label_leaves"),
                    "label leaves",
                    value = TRUE),
                checkboxInput(
                    ns("label_roots"),
                    "label roots",
                    value = TRUE),
                boxPlotControlUI(id, withPoints = FALSE, withColor = FALSE)
            ),
            column(9, h4(htmlOutput(
                NS(id, "oupTxt")
            )), geneExprDotPlotUI(id))
        )
    )
}
plotMonocleServer <- function(id, dataSource, optCrt) {
    if (checkMisc("monocle", dataSource()$dataset)) {
        monocle <- readData("monocle", dataSource()$dataset)
        reduction_method <- monocle$reduction_method
        if (is.null(reduction_method))
            reduction_method <- "UMAP"
        monocle$reduction_method <- NULL
    } else{
        monocle <- NULL
    }
    moduleServer(id, function(input, output, session) {
        updateSelectInput(
            session,
            "group",
            choices = c(),
            selected = NULL)
        updateSelectInput(
            session,
            "root",
            choices = c(),
            selected = NULL)
        has_monocle <- FALSE
        if (length(monocle)) {
            has_monocle <- TRUE
            monocle <- lapply(monocle, function(.ele) {
                n <- names(.ele)
                if (is.null(n)) {
                    names(.ele) <-  paste0("node", seq_along(.ele))
                }
                .ele
            })
            ## input column
            updateSelectInput(
                session,
                "group",
                choices = names(monocle),
                selected = names(monocle)[1]
            )
            
            observeEvent(input$group, {
                if (!is.null(monocle[[input$group]])) {
                    n <- names(monocle[[input$group]])
                    
                    updateSelectInput(
                        session,
                        "root",
                        choices = n,
                        selected = n[1])
                }
            })
            ## plot region
            ### plots
            plot1 <- reactive({
                scDRmonocle(
                    data = monocle[[input$group]][[input$root]],
                    reduction_method = reduction_method,
                    alpha = 1,
                    cell_size = .globals$fList[input$plotpsz],
                    cell_stroke = .globals$fList[input$plotpsz] / 4,
                    trajectory_graph_segment_size = .75,
                    trajectory_graph_color = 'grey28',
                    graph_label_size = input$plotfsz / 6,
                    font_size = input$plotfsz,
                    labelsFontFamily = input$plotfml,
                    show_trajectory_graph = input$show_trajectory_graph,
                    label_principal_points = input$label_principal_points,
                    label_leaves = input$label_leaves,
                    label_roots = input$label_roots
                )
            })
            updateGeneExprDotPlotUI(
                postfix = 1,
                id,
                input,
                output,
                session,
                plot1,
                .globals$pList2[input$plotpsz],
                dataSource()$dataset,
                input$group,
                input$node
            )
        }
        # update the ui
        output$oupTxt <- renderUI({
            if (!has_monocle) {
                HTML("No data available!")
            } else {
                HTML("")
            }
        })
    })
}

setClassUnion("character_OR_NULL", c("character", "NULL"))

#' Class \code{"APPconf"}
#' @description An object of class \code{"APPconf"} represents
#'  the metadata for a dataset.
#' @aliases APPconf
#' @rdname APPconf-class
#' @slot title character(1). Title of the data
#' @slot id character(1). Folder name of the data
#' @slot species character(1). species
#' @slot ref Reference information in a list with element bib, doi, pmid
#' and entry. Entry must be an object of \link[utils]{bibentry}
#' @slot type character(1). Type of the data, scRNAseq or scATACseq.
#' @slot markers list. A list of data.frame represents cell
#'  markers.
#' @slot keywords character. A vector of characters represents the
#'  keywords of the study.
#' @slot groupCol character. The key group column name to separate
#'  the cells.
#' @importFrom methods setClass representation prototype setMethod
#'  new `slot<-` setClassUnion
#' @export
#' @examples
#' appconf <- readRDS(system.file("extdata", "data",
#'     "pbmc_small", "appconf.rds", package="scRNAseqApp"))
#' appconf

setClass(
    "APPconf",
    representation = representation(
        title = "character",
        id = "character",
        species = "character",
        ref = "list",
        type = "character",
        markers = "list",
        keywords = "character",
        groupCol = "character_OR_NULL"
    ),
    prototype = prototype(
        title = "scRNAseqApp",
        id = 'sample_data',
        ref = list(),
        type = "scRNAseqApp",
        markers = list(),
        keywords = "",
        groupCol = NULL
    ),
    validity = function(object) {
        if (length(object@title) != 1) {
            return("title must be character(1L)")
        }
        if (length(object@id) != 1) {
            return("id must be character(1L)")
        }
        if (make.names(object@id) != object@id) {
            return("id must be a safe name.")
        }
        if (!object@type %in% c("scRNAseq", "scATACseq", "scMultiome")) {
            return("type must be scRNAseq, scATACseq or scMultiome")
        }
        if (length(object@ref$entry)) {
            if (!is(object@ref$entry, "bibentry")) {
                return("ref$entry must be an object of bibentry.")
            }
        }
        if (length(object@markers)) {
            out <- lapply(object@markers, function(.ele) {
                if (!is.data.frame(.ele)) {
                    return("markers must be a list of data.frame")
                }
                if (length(rownames(.ele)) != nrow(.ele)) {
                    return(
                        "markers must be a list of data.frame with rownames
                        and the rownames show be the marker name"
                    )
                }
                return(NULL)
            })
            out <- out[lengths(out) > 0]
            if (length(out)) {
                return(out)
            }
        }
        return(TRUE)
    }
)

#' @rdname APPconf-class
#' @param \dots Each argument in \dots becomes an slot in the new
#' \code{"APPconf"}-class.
#' @return A APPconf object.
#' @export
APPconf <- function(...) {
    new("APPconf", ...)
}

#' @noRd
#' @importFrom shinyhelper helper
helper1 <- function(...,
                    category,
                    title = NULL,
                    content = NULL) {
    if (missing(category))
        category <- "geneName"
    if (is.null(title)) {
        titles <- c(
            "geneName" = "Gene information to colour cells by",
            "cellInfo" = "Cell information to colour cells by",
            "cellInfoX" = "Cell information to group cells by",
            "cellInfoY" = "Cell Info / Gene to plot",
            "subsetCellInfo" = "Cell information to subset by"
        )
        title <- titles[category]
    }
    if (is.null(content)) {
        contents <-
            list(
                "geneName" = c(
                    paste(
                        "Select gene to colour cells by gene",
                        "expression/accessibility"),
                    paste(
                        "- Gene expression/accessibility are coloured in a",
                        "White-Red colour scheme which can be",
                        "changed in the plot controls"
                    ),
                    paste(
                        "- Gene name support autocomplete.",
                        "Try to input gene name in the input box."
                    )
                ),
                "cellInfo" = c(
                    "Select cell information to colour cells",
                    "- Categorical covariates have a fixed colour palette",
                    paste(
                        "- Continuous covariates are coloured in a",
                        "Blue-Yellow-Red colour scheme, which can be",
                        "changed in the plot controls"
                    )
                ),
                "cellInfoX" = c(
                    "Select categorical cell information to group cells by",
                    "- Single cells are grouped by this categorical covariate",
                    paste(
                        "- Plotted as the X-axis of the violin plot /",
                        "box plot / bar plot")
                ),
                "cellInfoY" = c(
                    "Select cell info to plot for group"
                ),
                "subsetCellInfo" = c(
                    "Select categorical cell information to subset cells by",
                    "- cells are shown in different subsets"
                )
            )
        
        content <- contents[[category]]
    }
    shinyhelper::helper(
        ...,
        type = "inline",
        size = "m",
        fade = TRUE,
        title = title,
        content = content
    )
}

#' scRNAseqApp main function
#' @description create a scRNAseqApp once the initialization is done.
#' @param app_path path, a directory where do you want to create the app
#' @param datafolder the folder where saved the dataset for the app
#' @param defaultDataset default dataset for the app.
#' @param windowTitle The title that should be displayed by the browser window.
#' @param favicon The favicon for the page.
#' @param banner The banner image.
#' @param footer The footer html contents.
#' @param maxRequestSize Maximal upload file size. Default is 1G.
#' @param timeout Timeout session (minutes) before logout if sleeping.
#'  Default to 30. 0 to disable.
#' @param theme A theme.
#' @param use_bs_themer logical(1). Used to determine the theme.
#' @param showHelpVideo logical(1) or character(1). 
#' Show help videos in homepage or not. If an url is provided, the
#' url will be embedded as a iframe element.
#' @param ... parameters can be passed to shinyApp except ui and server.
#' @import shiny
#' @importFrom utils packageVersion packageDescription read.delim
#' @importFrom xfun base64_uri is_abs_path
#' @importFrom shinyhelper observe_helpers
#' @importFrom ggplot2 ggplot aes geom_bar theme_minimal xlab ylab
#' @importFrom bslib bs_theme bs_themer
#' @export
#' @return An object that represents the app.
#' @examples
#' if(interactive()){
#'     app_path=tempdir()
#'     scInit(app_path=app_path)
#'     setwd(app_path)
#'     scRNAseqApp()
#' }

scRNAseqApp <- function(
        app_path = getwd(),
        datafolder = "data",
        defaultDataset = "pbmc_small",
        windowTitle = "scRNAseq/scATACseq database",
        favicon = system.file(
            'assets', 'img', 'favicon.ico',
            package = 'scRNAseqApp'),
        banner = system.file(
            'assets', 'img', 'banner.png',
            package = 'scRNAseqApp'),
        footer = tagList(
            HTML("&copy;"),
            "2024 -",
            format(Sys.Date(), "%Y"),
            "jianhong@morgridge"
        ),
        maxRequestSize = 1073741824,
        timeout = 30,
        theme = bs_theme(bootswatch = 'lumen'),
        use_bs_themer = FALSE,
        showHelpVideo = FALSE,
        ...) {
    stopifnot(is(theme, "bs_theme"))
    .globals$theme <- theme
    .globals$app_path <- app_path
    .globals$counterFilename <- file.path(
        app_path, "www/counter.tsv")
    ## config file
    .globals <- loadConfigFile(.globals, app_path)
    if(!is_abs_path(datafolder)){
        .globals$datafolder <- file.path(app_path, datafolder)
    }else{
        .globals$datafolder <- datafolder
    }
    stopifnot(file.exists(.globals$datafolder))
    # check if the db is encrypted
    if(isEncrypted()){
        .globals$passphrase <-
            readline('Please input the passphrase for the user database:')
    }
    ## load banner
    banner <- base64_uri(banner)
    ## load default parameters
    loginNavbarTitle <- "Switch User"
    if(!tableExists(.globals$configTableName)){
        updateConfigTable()
    }
    if(!tableExists(.globals$geneSymbolTableName)){
        touchGeneTable()
    }
    datasets <- listDatasets(privilege='all', named=TRUE)
    avdb <- sort(checkAvailableDataSets(privilege = 'all'))
    if(!identical(avdb, sort(unname(datasets)))){
        updateConfigTable()
        touchGeneTable(updateDB=TRUE)
    }
    datasets <- listDatasets(named=TRUE)
    defaultDataset <- getDefaultDataset(
        defaultDataset = defaultDataset,
        datasets = datasets)
    appconf <- getAppConf(datasets = datasets)
    data_types <- getDataType(appconf = appconf)

    ui0 <- function(req) {
        fluidPage(
            ### theme
            theme = .globals$theme,
            navbarPage(
                title = tags$head(
                    tags$link(rel="icon", 
                              href=base64_uri(favicon), 
                              type="image/x-icon")),
                windowTitle = windowTitle,
                id = "topnav",
                footer = div(
                    p(
                        em(
                            windowTitle,
                            " (Version:",
                            a(as.character(packageVersion("scRNAseqApp")),
                              href = packageDescription("scRNAseqApp",
                                                        fields='URL'),
                              target = "_blank"),
                            ")"
                        ),
                        footer,
                        style = 'text-align:right;'
                    ),
                    class = "about-right border-top-info"
                ),
                ### Tab: change dataset
                aboutUI(
                    req,
                    "about",
                    banner,
                    defaultDataset,
                    datasets,
                    appconf,
                    doc = file.path(.globals$app_path, "doc.txt"),
                    showHelpVideo = showHelpVideo),
                ## fake home
                homeUI(),
                navbarMenu(
                    "CellInfo/GeneExpr",
                    ### Tab: cellInfo vs geneExpr on dimRed
                    cellInfoGeneExprUI("cellInfoGeneExpr"),
                    ### Tab: cellInfo vs cellInfo on dimRed
                    cellInfoCellInfoUI("cellInfoCellInfo"),
                    ### Tab: subset gene expr
                    subsetGeneExprUI("subsetGeneExpr"),
                    ### Tab: ATAC vs Gene expr
                    geneAccGeneExprUI("ATACvsExpr")
                ),
                navbarMenu(
                    "Co-expression",
                    ### Tab: geneExpr vs geneExpr on dimRed
                    geneExprGeneExprUI("geneExprGeneExpr"),
                    ### Tab: Gene coexpression plot
                    coExprUI("coExpr"),
                    ### Tab: 3d Gene coexpression plot
                    coExpr3dUI("coExpr3d"),
                    ### Tab: sunburst plot
                    plotPieDimUI('sunburst')
                ),
                navbarMenu(
                    "Stats",
                    ### Tab: violinplot / boxplot
                    plotVioBoxUI("vioBoxPlot"),
                    ### Tab: Proportion plot
                    plotProportionUI("proportion"),
                    ### Tab: Multiple gene expr
                    plotBubbleHeatmapUI("bubbleHeatmap"),
                    ### Tab: monocle
                    #plotMonocleUI("monocle"),
                    ### Tab: waffle
                    plotWaffleUI("waffle")
                ),
                subsetPlotsUI('explorer'),
                ### Tab: downloader
                downloaderUI('downloader'),
                ### Tab: comments
                issueUI('issues'),
                ### Tab: Login form
                #tabLogin(),
                loginUI(loginNavbarTitle, defaultDataset)
            )
        )
    }
    ## security_login
    ui <- secureUI(ui0, timeout)
    
    ### Start server code
    server <- function(input, output, session) {
        ## load local storage
        session$sendCustomMessage("load_key", "defaultDataset")
        ## set theme
        if (is.null(getShinyOption("bootstrapTheme"))) {
            shinyOptions("bootstrapTheme" = .globals$theme)
        }
        if (use_bs_themer &&
            is(getShinyOption("bootstrapTheme"), "bs_theme")) {
            bs_themer()
        }
        ### resize the max upload file size for admin
        options(shiny.maxRequestSize = maxRequestSize) # 1G
        ### For all tags and Server-side selectize
        observe_helpers()
        optCrt <- "{ option_create: function(data,escape) {
    return('<div class=\"create\"><strong>' + '</strong></div>');
    } }"
        touchGenename2Symbol()
        dataSource <- reactiveValues(
            # all available datasets
            available_datasets = datasets,
            # current dataset
            dataset = defaultDataset,
            # all data configs, used for search
            appconf = appconf,
            # data type
            data_types = data_types,
            cell = NULL,
            # genelist for cellInfoGeneExpr or heatmap plot from query string
            genelist = NULL,
            # active tab from query string
            activetab = NULL,
            # config for current sc data
            sc1conf = NULL,
            # def for current sc data
            sc1def = NULL,
            # gene for current sc data
            sc1gene = NULL,
            # meta for current sc data
            sc1meta = NULL,
            #search_cache
            search_results = NULL,
            # user logged
            Logged = FALSE,
            # tab UI for scRNAseq/scATACseq
            terms = .globals$terms[["scRNAseq"]],
            # authority
            auth = NULL,
            # username
            Username = "",
            # passward
            Password = "",
            # toekn
            token = ""
        ) 
        
        updateDataSource <- function(datasets, selected){
            if(missing(datasets)) datasets <- 
                    listDatasets(privilege = dataSource$auth$privilege,
                                 named=TRUE)
            appconf <- getAppConf(
                datasets = datasets,
                privilege = 'all'
            )
            if(any(!datasets %in% names(appconf))){
                warning('Some dataset does not contain appconf!')
            }
            dataSource$appconf <- appconf
            dataSource$available_datasets <- datasets
            dataSource$dataset <- selected
            dataSource$data_types <- getDataType(appconf = appconf)
            updateSelectInput(
                session,
                "selectedDatasets",
                choices = dataSource$available_datasets,
                selected = dataSource$dataset
            )
        }
        
        ### check available data
        compareAvailableDatasets <- reactivePoll(
            1000,session,
            checkFunc = function(){
                sort(checkAvailableDataSets(
                    privilege = dataSource$auth$privilege))
                },
            valueFunc = function(){
                sort(unname(isolate(dataSource$available_datasets)))
            })
        observeEvent(
            compareAvailableDatasets(),
            ignoreNULL = TRUE,
            ignoreInit = TRUE,
            {
                updateConfigTable()
                touchGeneTable(updateDB=TRUE)
                ad <- listDatasets(privilege = dataSource$auth$privilege,
                                   named=TRUE)
                if (!all(
                    dataSource$available_datasets %in% ad,
                    na.rm = TRUE) ||
                    !all(
                        ad %in% dataSource$available_datasets,
                        na.rm = TRUE)) {
                    updateDataSource(ad, input$selectedDatasets)
                }
            })
        ## login
        dataSource$auth <- loginServer(input, output, session)
        observe({
            if (!is.null(isolate(dataSource$auth$admin))) {
                if (isolate(dataSource$auth$admin)) {
                    webstatsServer("webstats")
                    uploadServer("upload")
                    editServer("editdata")
                }
            }
        })
        ## update visitor stats
        updateVisitorTable(input, output, session)
        ## parse query strings, and load old session
        observe({
            query <- parseQueryString(session$clientData$url_search)
            query_has_results <- FALSE
            if (!is.null(query[['gene']])) {
                genes <- strsplit(query[['gene']], ";")[[1]]
                dataSource$genelist <- genes
            } else{
                dataSource$genelist <- NULL
            }
            if (!is.null(query[['tab']])) {
                dataSource$activetab <- query[['tab']]
            } else {
                dataSource$activetab <- NULL
            }
            if (!is.null(query[['cell']])) {
                ## not used yet
                cell <- strsplit(query[['cell']], ";")[[1]]
                dataSource$cell <- cell
            } else{
                dataSource$cell <- NULL
            }
            p_query <- parseQuery(query, defaultDataset)
            if(p_query['from']=='data'){
                session$userData[["defaultdata_init"]] <- TRUE
                query_has_results <- TRUE
                if(p_query["defaultDataset"] %in% dataSource$available_datasets){
                    defaultDataset <- p_query["defaultDataset"]
                    updateSelectInput(
                        session,
                        'selectedDatasets',
                        choices = dataSource$available_datasets,
                        selected = p_query["defaultDataset"]
                    )
                }
            }else{
                if(p_query['from']=='token'){
                    dataSource$token <- query[["token"]]
                    defaultDataset <- p_query["defaultDataset"]
                    session$userData[["defaultdata_init"]] <- TRUE
                    query_has_results <- TRUE
                    showNotification(
                        "Datasets from token!",
                        duration = 3,
                        type = "message")
                    updateSelectInput(
                        session,
                        'selectedDatasets',
                        choices = c(dataSource$available_datasets,
                                    p_query["defaultDataset"]),
                        selected = p_query["defaultDataset"])
                }
            }
            
            if (!use_bs_themer &
                !query_has_results) {
                ## not work when load themer
                ## load remote user data
                observeEvent(input$default_defaultDataset, {
                    if (isTRUE(isolate(
                        session$userData[["defaultdata_init"]]))) {
                            return()
                    } else {
                        session$userData[["defaultdata_init"]] <- TRUE
                        dd <- isolate(input$default_defaultDataset)
                        if (!is.null(dd)) {
                            if (dd %in% dataSource$available_datasets &&
                                dd != dataSource$dataset) {
                                showNotification(
                                    "Resuming your last session!",
                                    duration = 3,
                                    type = "message")
                                updateSelectInput(
                                    session,
                                    'selectedDatasets',
                                    choices = dataSource$available_datasets,
                                    selected = dd)
                            }
                        }
                    }
                })
            }
        })
        ## when click the search results
        observeEvent(input$search_results_select_button, {
            updateSelectInput(session, 'selectedDatasets',
                              selected=input$search_results_select_button)
        })
        ## change dataset
        observeEvent(input$selectedDatasets, {
            ## in case the data is deleted, refresh the datasets
            availabledb <- 
                checkAvailableDataSets(privilege = dataSource$auth$privilege,
                                       token = dataSource$token)
            if (input$selectedDatasets %in% availabledb) {
                # in available datasets for current user
                if(dataSource$token != ""){#with token
                    if (checkToken(
                        getTokenList(),
                        dataSource$token,
                        input$selectedDatasets)) {
                        dataSource$Logged <- TRUE
                        dataSource$auth$privilege <- 
                            unique(c(dataSource$auth$privilege,
                                     input$selectedDatasets))
                        res <- updateDatasetForToken(
                            input$selectedDatasets,
                            dataSource$available_datasets)
                        dataSource$available_datasets <- res$datasets
                        dataSource$appconf <- res$appconf
                        dataSource$data_types <- res$data_types
                        dataSource$dataset <- input$selectedDatasets
                    }
                    updateDataSource(
                        datasets = listDatasets(availabledb,
                                                privilege = 'all',
                                                named = TRUE),
                        selected = input$selectedDatasets)

                }
                redirectionData()
            } else{
                if (checkLocker(input$selectedDatasets)) {
                    # dataset is not open to current user
                    dataSource$Logged <- FALSE
                    if (dataSource$token != "") {} else{
                        if (!is.null(dataSource$auth)) {
                            if (checkPrivilege(
                                dataSource$auth$privilege,
                                input$selectedDatasets)) {
                                dataSource$Logged <- TRUE
                                dataSource$available_datasets <- 
                                    listDatasets(dataSource$auth$privilege,
                                                 named=TRUE)
                                updateDataSource(
                                    datasets = dataSource$available_datasets,
                                    selected = input$selectedDatasets)
                            }
                        }
                    }
                    if (!dataSource$Logged) {
                        updateTabsetPanel(
                            session,
                            "topnav",
                            selected = loginNavbarTitle)
                    }else{
                        redirectionData()
                    }
                }else{
                    ##data was deleted
                    updateDataSource(datasets = dataSource$available_datasets,
                                     selected = availabledb[1])
                }
            }
            session$sendCustomMessage(
                "save_key",
                paste(
                    "defaultDataset",
                    isolate(input$selectedDatasets),
                    sep = "|"
                ))
        })
        ## refresh data when change dataset
        refreshData <- function(input, output, session) {
            if (dataSource$dataset %in% names(dataSource$data_types)) {
                dataSource$terms <-
                    .globals$terms[[
                        dataSource$data_types[[
                            dataSource$dataset]]]]
            }
            dataSource <- loadData(dataSource)
            output$dataTitle <- renderUI({
                HTML(names(datasets)[datasets == input$selectedDatasets])
            })
            
            aboutServer("about", reactive({
                dataSource
            }), optCrt)
            ### Plots for tab cell info vs gene expression
            cellInfoGeneExprServer(
                "cellInfoGeneExpr",
                reactive({
                    dataSource
                }),
                optCrt)
            
            ### Plots for tab cell info vs cell info
            cellInfoCellInfoServer(
                "cellInfoCellInfo",
                reactive({
                    dataSource
                }),
                optCrt)
            
            ### Plots for tab gene expression vs gene expression
            geneExprGeneExprServer(
                "geneExprGeneExpr",
                reactive({
                    dataSource
                }),
                optCrt)
            
            ### ATAC vs Expr
            geneAccGeneExprServer(
                "ATACvsExpr",
                reactive({
                    dataSource
                }),
                optCrt)
            
            ### Plots for tab co-expression
            coExprServer(
                "coExpr",
                reactive({
                    dataSource
                }),
                optCrt)
            
            ### Plots for tab 3d co-expression
            coExpr3dServer(
                "coExpr3d",
                reactive({
                    dataSource
                }),
                optCrt)
            
            ### Plots for tab PieDim
            plotPieDimServer(
                'sunburst',
                reactive({
                    dataSource
                }),
                optCrt)
            
            ### Plots for tab subset
            subsetGeneExprServer(
                "subsetGeneExpr",
                reactive({
                    dataSource
                }),
                optCrt)
            
            ### Plots for tab violion
            plotVioBoxServer(
                "vioBoxPlot",
                reactive({
                    dataSource
                }),
                optCrt)
            
            
            ### Plots for tab proportion
            plotProportionServer(
                "proportion",
                reactive({
                    dataSource
                }),
                optCrt)
            
            
            ### Plots for tab bubble heatmap
            plotBubbleHeatmapServer(
                "bubbleHeatmap",
                reactive({
                    dataSource
                }),
                optCrt)
            
            ### Plots for waffle
            plotWaffleServer(
                "waffle",
                reactive({
                    dataSource
                }),
                optCrt)
            
            subsetPlotsServer(
                'explorer',
                reactive({
                    dataSource
                }),
                optCrt)
        }
        ## update UI and server
        redirectionData <- function(){
            dataSource$dataset <- input$selectedDatasets
            refreshData(input, output, session)
            if (length(dataSource$genelist) == 1) {
                updateTabsetPanel(
                    session,
                    'topnav',
                    selected = ifelse(is.null(dataSource$activetab),
                                      "cellInfoGeneExpr",
                                      dataSource$activetab))
            } else{
                if (length(dataSource$genelist) > 1) {
                    updateTabsetPanel(
                        session,
                        'topnav',
                        selected = "bubbleHeatmap")
                }
            }
        }
        ## comments server
        issueServer('issues',
                    reactive({
                        dataSource
                    }))
        ## download server
        downloaderServer("downloader")
        ## observe event for editor
        reLoadData <- function(){
            dataSource <- loadData(dataSource)
        }
        observeEvent(input$editorStatus, {
            reLoadData()
        })
    }
    shinyApp(ui = ui, server = server, ...)
}

# Plot gene coexpression on dimred
bilinear <- function(x, y, xy, Q11, Q21, Q12, Q22) {
    oup <- (xy - x) * (xy - y) * Q11 + x * (xy - y) *
        Q21 + (xy - x) * y * Q12 + x *
        y * Q22
    oup <- oup / (xy * xy)
    return(oup)
}
#' @importFrom grDevices rgb
#' @importFrom data.table data.table
#' @importFrom plotly plot_ly layout
#' @importFrom ggplot2 ggplot aes .data geom_point xlab ylab
#' scale_color_gradientn guides guide_colorbar coord_fixed
scDRcoex <- function(
        inpConf,
        inpMeta,
        dimRedX,
        dimRedY,
        gene1,
        gene2,
        subsetCellKey,
        subsetCellVal,
        dataset,
        geneIdMap,
        plotType,
        pointSize,
        GeneExprDotCol,
        GeneExprDotOrd,
        labelsFontsize = 24,
        labelsFontFamily = 'Helvetica',
        plotAspectRatio,
        keepXYlables,
        valueFilterKey,
        valueFilterCutoff,
        valueFilterCutoff2,
        hideFilterCell = FALSE) {
    if (is.null(gene1) || is.null(gene2) || gene1 == "" || gene2 == "") {
        return(NULL)
    }
    subFilterColname <- 'subValue'
    subGrpColname <- 'sub'
    subsetCellKey <- subsetCellKey[subsetCellKey!="N/A"]
    subsetCellVal <- namedSubsetCellVals(subsetCellKey, subsetCellVal)
    # Prepare ggData
    ggData <- inpMeta[, unique(c(
        inpConf[inpConf$UI == dimRedX]$ID,
        inpConf[inpConf$UI == dimRedY]$ID,
        inpConf[inpConf$UI %in% subsetCellKey]$ID)),
        with = FALSE]
    if (nrow(ggData) == 0)
        return(NULL)
    if(ncol(ggData)<3){
        return(NULL)
    }
    colnames(ggData)[c(1, 2)] <- c("X", "Y")
    if (plotType == "3D") {
        ggData$sampleID <- inpMeta$sampleID
    }
    ggData <-
        cbindFilterValues(
            ggData,
            inpConf,
            inpMeta,
            subFilterColname,
            geneIdMap,
            dataset,
            valueFilterKey,
            valueFilterCutoff,
            valueFilterCutoff2
        )
    rat <- getRatio(ggData)
    
    ggData <- getCoexpVal(ggData, dataset, geneIdMap, gene1, gene2)
    keep <- filterCells(
        ggData,
        subsetCellKey,
        subsetCellVal,
        subFilterColname,
        valueFilterCutoff,
        valueFilterCutoff2,
        inpConf)
    
    colnames(ggData)[3] <- subGrpColname ## make the first subsetCellKey as sub
    bgCells <- sum(!keep) > 0
    if (bgCells) {
        ggData2 <- ggData[!keep]
        ggData <- ggData[keep]
    }
    ## color for group
    # Do factoring if required
    ggData <- relevelData(ggData, subGrpColname)
    if (plotType == "3D") {
        ggCol <- relevelCol(inpConf, subsetCellKey[1], ggData, subGrpColname)
    }
    
    # Generate coex color palette
    cInp <- strsplit(GeneExprDotCol, "; ")[[1]]
    
    nTot <- getTotalNumber(nGrid = 16, nPad = 2)
    gg <- getCoexpCol(GeneExprDotCol, nGrid = 16, nPad = 2)
    
    # Map colours
    mv1 <- max(ggData$val1, na.rm = TRUE)
    mv2 <- max(ggData$val2, na.rm = TRUE)
    ggData$v1 <-
        round(nTot * ggData$val1 / mv1)
    ggData$v2 <-
        round(nTot * ggData$val2 / mv2)
    ggData$v0 <- ggData$v1 + ggData$v2
    ggData <- gg[ggData, on = c("v1", "v2")]
    ggData <- orderGeneExpr(ggData, GeneExprDotOrd, 'v0')
    ggData <- ggData[!is.na(ggData$X) & !is.na(ggData$Y), ]
    if(nrow(ggData)==0) return(ggplot())
    ggData$cMix[is.na(ggData$cMix)] <- "snow2"
    
    # Actual ggplot
    if (plotType == "3D") {
        nTot <- max(c(ggData$val1, ggData$val2), na.rm = TRUE)
        ggData$norm1 <-
            round(nTot * ggData$val1 / mv1)
        ggData$norm2 <-
            round(nTot * ggData$val2 / mv2)
        ggData$Z <- log2(ggData$norm1 + 1) - log2(ggData$norm2 + 1)
        return(layout(
            plot_ly(
                x = ggData$X,
                y = ggData$Y,
                z = ggData$Z,
                customdata = ggData$sampleID,
                type = "scatter3d",
                mode = "markers",
                color = if (GeneExprDotCol == "Default") {
                    ggData[[subGrpColname]]
                } else{
                    ggData$Z
                },
                colors = if (GeneExprDotCol == "Default")
                    ggCol
                else
                    availableThemes(GeneExprDotCol),
                text = paste0(
                    gene1, ": ", ggData$val1, "\n",
                    gene2, ": ", ggData$val2),
                size = 1
            ),
            scene =
                list(
                    xaxis = list(title = dimRedX),
                    yaxis = list(title = dimRedY),
                    zaxis = list(title = paste0(
                        'Log2 Fold Change (',
                        gene1, '/', gene2, ')'
                    ))
                )
        ))
    }
    ggOut <- ggplot(ggData, aes(x = .data[["X"]], y = .data[["Y"]]))
    if (bgCells) {
        ggOut <- labelBackgroundCells(
            ggOut,
            ggData2,
            pointSize,
            color = "snow2",
            shape = 16,
            hide = hideFilterCell)
    }
    ggOut <- ggOut +
        geom_point(
            size = pointSize,
            shape = 16,
            color = ggData$cMix) +
        xlab(dimRedX) + ylab(dimRedY) +
        sctheme(
            base_size = labelsFontsize,
            family = labelsFontFamily,
            XYval = keepXYlables)
    
    ggOut <- fixCoord(ggOut, plotAspectRatio, rat)
    return(ggOut)
}

scAccUI <- function(id, postfix = 1, subgrp=.globals$subsetgroup[1]) {
    subModuleContainerUI(
        id,
        mainSelectUI = tagList(
            selectInput(
                NS0(id, "coord", postfix),
                "Coordinates:",
                choices = NULL),
            selectInput(
                NS0(id, "GeneName", postfix),
                "Linked gene name:",
                choices = NULL)
        ),
        menuUI = contextMenuGeneExprUI(id, postfix, group = subgrp),
        contentUI = geneExprDotPlotUI(id, postfix)
    )
}
scAccServer <- function(
        pid,
        id,
        dataSource,
        optCrt,
        p_input,
        p_session,
        interactive,
        postfix = 1) {
    moduleServer(id, function(input, output, session) {
        ## title
        CoordLabel <- paste0('coord', postfix)
        GeneNameLabel <- paste0('GeneName', postfix)
        links <- readData("sc1link", dataSource()$dataset)
        if(length(links$peak)!=length(links)){
            links$peak <- paste(as.character(seqnames(links)),
                                start(links), end(links),
                                sep='-')
        }
        availableTargets <- sort(names(dataSource()$sc1gene))
        if(length(links$gene)){
            availableTargets <- availableTargets[availableTargets %in% 
                                                     links$gene]
        }
        if(length(links$peak)){
            availablePeaks <- sort(links$peak)
        }else{
            if(length(links)){
                availablePeaks <- paste(as.character(seqnames(links)),
                                        start(links), end(links),
                                        sep='-')
            }else{
                availablePeaks <-
                    apply(readData("sc1peak", dataSource()$dataset),
                          1, paste, collapse='-')
            }
        }
        
        if (is.null(
            p_session$userData$defaults[[dataSource()$dataset]][[id]])) {
            defaults <- list()
            defaults[[CoordLabel]] <-
                getCoordByGeneSymbol(dataSource()$sc1def$gene1, links=links)
            defaults[[GeneNameLabel]] <- dataSource()$sc1def$gene1
        } else{
            defaults <-
                p_session$userData$defaults[[dataSource()$dataset]][[id]]
        }
        if(is.null(defaults[[CoordLabel]])){
            defaults[[CoordLabel]] <-
                getCoordByGeneSymbol(dataSource()$sc1def$gene1, links=links)
        }
        if(!is.character(defaults[[CoordLabel]])){
            defaults[[CoordLabel]] <- defaults[[CoordLabel]]$peak[1]
        }
        programChange <- reactiveValues(from='')
        observeEvent(input[[GeneNameLabel]], {
            if(input[[GeneNameLabel]]!='' && programChange$from!='coor'){
                coor <- 
                    getCoordByGeneSymbol(input[[GeneNameLabel]], links=links)
                if(is(coor, 'GRanges')){
                    coor <- coor[coor$peak!='']
                    if(length(coor$peak)){
                        if(length(defaults[[CoordLabel]])){
                            updateCoord <- !defaults[[CoordLabel]] %in%
                                coor$peak
                        }else{
                            updateCoord <- TRUE
                        }
                        if(updateCoord){
                            defaults[[CoordLabel]] <- coor$peak[1]
                            if(defaults[[CoordLabel]]!=input[[CoordLabel]]){
                                programChange$from <- 'gene'
                                updateSelectizeInput(
                                    session,
                                    CoordLabel,
                                    choices = unique(c(coor$peak,
                                                       availablePeaks)),
                                    server = TRUE,
                                    selected = defaults[[CoordLabel]],
                                    options = list(
                                        maxOptions = min(.globals$maxNumGene,
                                                         length(coor)),
                                        create = TRUE,
                                        persist = TRUE,
                                        render = I(optCrt)
                                    )
                                )
                            }
                        }
                        
                    }
                }else{
                    showNotification(
                        paste("No linked peak by givening gene: ",
                              input[[GeneNameLabel]]),
                        duration = 5,
                        closeButton = TRUE,
                        type = "error"
                    )
                }
            }
            if(input[[GeneNameLabel]]!='' && programChange$from=='coor'){
                programChange$from <- ''
            }
        }, ignoreInit = TRUE)
        observeEvent(input[[CoordLabel]], {
            if(input[[CoordLabel]]!='' && programChange$from!='gene'){
                gene <- 
                    getGeneSymbolByCoord(input[[CoordLabel]], links=links)
                if(length(gene)){
                    if(any(gene %in% names(dataSource()$sc1gene))){
                        if(length(defaults[[GeneNameLabel]])){
                            updateGeneSymbol <- !defaults[[GeneNameLabel]] %in%
                                gene
                        }else{
                            updateGeneSymbol <- TRUE
                        }
                        if(updateGeneSymbol){
                            defaults[[GeneNameLabel]] <- 
                                gene[gene %in% names(dataSource()$sc1gene)][1]
                            if(defaults[[GeneNameLabel]]!=input[[GeneNameLabel]]){
                                programChange$from <- 'coor'
                                updateSelectizeInput(
                                    session,
                                    GeneNameLabel,
                                    choices = availableTargets,
                                    server = TRUE,
                                    selected = defaults[[GeneNameLabel]],
                                    options = list(
                                        maxOptions = .globals$maxNumGene,
                                        create = TRUE,
                                        persist = TRUE,
                                        render = I(optCrt)
                                    )
                                )
                            }
                        }
                    }
                }else{
                    showNotification(
                        paste("No linked gene by givening coordinate",
                              input[[CoordLabel]]),
                        duration = 5,
                        closeButton = TRUE,
                        type = "error"
                    )
                }
            }
            if(input[[CoordLabel]]!='' && programChange$from=='gene'){
                programChange$from <- ''
            }
        }, ignoreInit = TRUE)
        updateSelectizeInput(
            session,
            GeneNameLabel,
            choices = availableTargets,
            server = TRUE,
            selected = defaults[[GeneNameLabel]],
            options = list(
                maxOptions = .globals$maxNumGene,
                create = TRUE,
                persist = TRUE,
                render = I(optCrt)
            )
        )
        updateSelectizeInput(
            session,
            CoordLabel,
            choices = availablePeaks,
            server = TRUE,
            selected = if(length(defaults[[CoordLabel]])) defaults[[CoordLabel]] else availablePeaks[1],
            options = list(
                maxOptions = .globals$maxNumGene,
                create = TRUE,
                persist = TRUE,
                render = I(optCrt)
            )
        )
        subModuleMenuObservor(
            id,
            input,
            p_session,
            dataSource,
            c(
                CoordLabel,
                GeneNameLabel,
                paste0("GeneExprcol", postfix),
                paste0("GeneExprord", postfix),
                paste0("GeneExprtype", postfix),
                paste0("GeneExprxlim", postfix),
                paste0("GeneExprrg", postfix)
            )
        )
        ### plots
        plotX <- reactive({
            scDRgene(
                inpConf=dataSource()$sc1conf,
                inpMeta=dataSource()$sc1meta,
                dimRedX=p_input$GeneExprdrX,
                dimRedY=p_input$GeneExprdrY,
                gene1=input[[CoordLabel]],
                subsetCellKey=p_input[[paste0("subsetCell",
                                              input[[paste0("CellInfosubgrp",
                                                            postfix)]])]],
                subsetCellVal=
                    getSubsetCellVal(p_input,
                                     group=input[[paste0("CellInfosubgrp",
                                                         postfix)]]),
                dataset=dataSource()$dataset,
                geneIdMap=dataSource()$sc1gene,
                pointSize=p_input$GeneExprsiz,
                gradientCol=input[[paste0("GeneExprcol", postfix)]],
                GeneExprDotOrd=input[[paste0("GeneExprord", postfix)]],
                labelsFontsize=p_input$GeneExprfsz,
                labelsFontFamily=p_input$GeneExprfml,
                plotAspectRatio=p_input$GeneExprasp,
                keepXYlables=p_input$GeneExprtxt,
                inpPlt=input[[paste0("GeneExprtype", postfix)]],
                inpXlim=if (input[[paste0("GeneExprxlimb", postfix)]] %% 2 == 0)
                    0
                else
                    input[[paste0("GeneExprxlim", postfix)]],
                inpColRange =
                    if (input[[paste0("GeneExprrgb", postfix)]] %% 2 == 0)
                        0
                else
                    input[[paste0("GeneExprrg", postfix)]],
                valueFilterKey = p_input$filterCell,
                valueFilterCutoff = p_input$filterCellVal,
                valueFilterCutoff2 = p_input$filterCellVal2,
                hideFilterCell = input[[paste0("GeneExprhid", postfix)]],
                geneType = 'coor'
            )
        })
        updateSubModulePlotUI(
            postfix,
            pid,
            id,
            input,
            output,
            session,
            interactive,
            plotX,
            .globals$pList1[p_input$GeneExprpsz],
            dataSource()$dataset,
            p_input$GeneExprdrX,
            p_input$GeneExprdrY,
            input[[CoordLabel]]
        )
    })
}

plotWaffleUI <- function(id) {
    tabPanel(
        value = id,
        HTML("Waffle Plot"),
        h4("Gene expression waffle plot"),
        "In this tab, users can visualise the gene expression by waffle plot",
        "across groups of cells (e.g. libary / clusters).",
        "Each cell in waffle plot represents the expression mean value of 1% ",
        "number of the maximal group cells",
        br(),
        br(),
        fluidRow(
            column(
                3,
                style = "border-right: 2px solid black",
                textAreaInput(
                    NS(id, "genelist"),
                    HTML(
                        "List of gene names <br />
                            (Max 3 genes (over 3 will be very slow), <br />
                            separated by , or ; or newline):"
                    ),
                    height = "100px",
                    value = NULL
                ),
                xaxisCellInfoUI(id),
                yaxisCellInfoUI(id),
                subsetCellByInfoUI(id, mini = TRUE),
                boxPlotControlUI(
                    id, withPoints = FALSE, withColor = TRUE,
                    withFontSize = FALSE)
            ),
            column(9, geneExprDotPlotUI(id))
        )
    )
}
plotWaffleServer <- function(id, dataSource, optCrt, postfix = 1) {
    moduleServer(id, function(input, output, session) {
        ## input column
        genelist <- dataSource()$sc1def$gene1
        if (!is.null(dataSource()$genelist)) {
            if (length(dataSource()$genelist) > 1) {
                genelist <- dataSource()$genelist[1]
            } else{
                genelist <- c(dataSource()$genelist, genelist)
            }
        }
        updateTextAreaInput(
            session, "genelist",
            value = paste0(genelist, collapse = ", "))
        updateSelectInput(
            session,
            "CellInfoX",
            "Group by:",
            choices = getGroupUI(dataSource),
            selected = dataSource()$sc1def$grp1
        )
        updateSelectInput(
            session,
            "CellInfoY",
            "Split by:",
            choices = getGroupUI(dataSource),
            selected = dataSource()$sc1def$grp2
        )
        updateSubsetCellUI(id, input, output, session, dataSource, addNA = TRUE)
        updateFilterCellUI(id, optCrt, input, output, session, dataSource)
        ### plots
        plotX <- reactive({
            scDRwafflePlot(
                dataSource()$dataset,
                dataSource()$sc1gene,
                dataSource()$sc1conf,
                dataSource()$sc1meta,
                input$genelist,
                input$CellInfoX,
                input$CellInfoY,
                input$plotcols,
                input$subsetCell,
                getSubsetCellVal(input)
            )
        })
        updateGeneExprDotPlotUI(
            postfix,
            id,
            input,
            output,
            session,
            plotX,
            .globals$pList3[input$plotpsz],
            dataSource()$dataset,
            make.names(input$genelist)
        )
    })
}

# Plot violin / boxplot
#' @importFrom stats rnorm
#' @importFrom ggplot2 ggplot aes .data geom_violin geom_boxplot
#' geom_jitter xlab ylab scale_fill_manual theme
scVioBox <- function(
        inpConf,
        inpMeta,
        infoX,
        infoY,
        subsetCellKey,
        subsetCellVal,
        valueFilterKey,
        valueFilterCutoff,
        valueFilterCutoff2,
        dataset,
        geneIdMap,
        inptyp,
        inppts,
        pointSize,
        labelsFontsize = 24,
        labelsFontFamily = 'Helvetica',
        reorder=FALSE,
        orderX,
        splitBy,
        sreorder=FALSE,
        orderS,
        addnoise=TRUE) {
    # Prepare ggData
    ggData <- inpMeta[, c(
        inpConf[inpConf$UI == infoX]$ID,
        inpConf[inpConf$grp == TRUE]$ID), ## will include all subsetCellKey
        with = FALSE]
    colnames(ggData)[1] <- c("X")
    # Load in either cell meta or gene expr
    if (infoY %in% inpConf$UI) {
        ggData$val <- inpMeta[[inpConf[inpConf$UI == infoY]$ID]]
    } else {
        ggData$val <- read_exprs(dataset, geneIdMap[infoY], valueOnly = TRUE)
        ggData[ggData$val < 0]$val <- 0
        if(addnoise){
            tmpNoise <-
                rnorm(length(ggData$val)) * diff(range(ggData$val)) / 1000
            ggData$val <- ggData$val + tmpNoise
        }
    }
    # Load splitBy
    if(!missing(splitBy)){
        if(splitBy!=""){
            if(splitBy %in% inpConf$UI){
                ggData$splitBy <- inpMeta[[inpConf[inpConf$UI == splitBy]$ID]]
                if(sreorder){
                    ggData$splitBy <- factor(as.character(ggData$splitBy),
                                             levels = orderS)
                }
            }
        }
    }
    # filter the cell
    subFilterColname <- 'filter'
    ggData <-
        cbindFilterValues(
            ggData,
            inpConf,
            inpMeta,
            subFilterColname,
            geneIdMap,
            dataset,
            valueFilterKey,
            valueFilterCutoff,
            valueFilterCutoff2
        )
    subsetCellKey <- subsetCellKey[subsetCellKey!="N/A"]
    subsetCellVal <- namedSubsetCellVals(subsetCellKey, subsetCellVal)
    keep <- filterCells(
        ggData,
        subsetCellKey,
        subsetCellVal,
        subFilterColname,
        valueFilterCutoff,
        valueFilterCutoff2,
        inpConf)
    ggData <- ggData[keep]
    
    # Do factoring
    if(reorder){
        ggData$X <- factor(ggData$X, levels=orderX)
    }else{
        ggData <- relevelData(ggData, "X")
    }
    if('splitBy' %in% colnames(ggData)){
        ggCol <- relevelCol(inpConf, splitBy, ggData, "splitBy")
    }else{
        ggCol <- relevelCol(inpConf, infoX, ggData, "X")
    }
    
    
    # Actual ggplot
    showLegend <- FALSE
    if (inptyp == "violin") {
        if('splitBy' %in% colnames(ggData)){
            if(length(unique(ggData$splitBy))==2){
                ggOut <- ggplot(ggData, aes(
                    .data[["X"]], .data[["val"]], fill = .data[["splitBy"]])) +
                    geom_split_violin(scale = "width")
            }else{
                ggOut <- ggplot(ggData, aes(
                    .data[["X"]], .data[["val"]], fill = .data[["splitBy"]])) +
                    geom_violin(scale = "width")
            }
            showLegend <- TRUE
        }else{
            ggOut <- ggplot(ggData, aes(
                .data[["X"]], .data[["val"]], fill = .data[["X"]])) +
                geom_violin(scale = "width")
        }
    } else {
        if('splitBy' %in% colnames(ggData)){
            ggOut <- ggplot(ggData, aes(
                .data[["X"]], .data[["val"]], fill = .data[["splitBy"]])) +
                geom_boxplot()
            showLegend <- TRUE
        }else{
            ggOut <- ggplot(ggData, aes(
                .data[["X"]], .data[["val"]], fill = .data[["X"]])) +
                geom_boxplot()
        }
        
    }
    if (inppts) {
        ggOut <- ggOut + geom_jitter(size = pointSize, shape = 16)
    }
    ggOut <- ggOut + xlab(infoX) + ylab(infoY) +
        sctheme(
            base_size = labelsFontsize,
            family = labelsFontFamily,
            Xang = 45,
            XjusH = 1) +
        scale_fill_manual("", values = ggCol)
    if(showLegend){
        ggOut <- ggOut +
            theme(legend.position = "bottom")
    }else{
        ggOut <- ggOut +
            theme(legend.position = "none")
    }
    return(ggOut)
}

# Plot ATAC track on dimred
#' @importFrom ggplot2 ggplot aes .data geom_line xlab ylab guides
#' scale_x_continuous xlim vars
#' @importFrom gridExtra grid.arrange
#' @importFrom rtracklayer import
#' @importFrom GenomicRanges GRanges strand seqnames gaps
#' @importFrom patchwork wrap_plots
#' @importFrom IRanges subsetByOverlaps
#' @importFrom shiny Progress
scDRatac <- function(
        inpConf,
        inpMeta,
        dimRedX,
        dimRedY,
        gene1,
        coord,
        subsetCellKey,
        subsetCellVal, # not support multiple key and val pairs
        dataset,
        geneIdMap,
        pointSize,
        gradientCol,
        labelsFontsize=24,
        labelsFontFamily="Helvetica",
        plotAspectRatio,
        keepXYlables,
        ...) {
    if (gene1[1] == "") {
        return(ggplot())
    }
    if (coord[1] == "") {
        return(ggplot())
    }
    if (is.na(geneIdMap[gene1])) {
        return(ggplot())
    }
    if (is.null(geneIdMap[gene1])) {
        return(ggplot())
    }
    # Prepare ggData
    pf <- file.path(
        .globals$datafolder, dataset,
        .globals$filenames$bwspath,
        subsetCellKey)
    bws <- dir(pf, ".bigwig$")
    if(length(bws)==0){
        warning('no bws detected. Please check the folder privileges for ', pf)
        return(ggplot())
    }
    gr <- GRanges(coord)
    if(width(gr)[1]>10000000){
        showNotification("The region is greater than 10M! Please decrease the plot region.",
                         type = "error")
        return(ggplot())
    }
    progress <- Progress$new()
    progress$set(message = 'Plot ATAC signals', value =0)
    on.exit(progress$close())
    updateProgress <- function(value = NULL, n = 1, detail=NULL) {
        if (is.null(value)) {
            value <- progress$getValue()
            value <- value + (progress$getMax() - value) / n
        }
        progress$set(value = value, detail=detail)
    }
    atac_sig <- lapply(subsetCellVal, function(.ele){
        sample <- .ele
        f <- file.path(pf, paste0(path_sanitize(.ele), ".bigwig"))
        if(file.exists(f)){
            sig <- import(f, format="BigWig", which=gr, as="GRanges")
        }else{
            sig <- GRanges()
        }
        updateProgress(n=length(subsetCellVal)+5, detail=paste('loading', sample))
        return(sig)
    })
    names(atac_sig) <- subsetCellVal
    atac_sig <- lapply(atac_sig, function(.ele){
        .ele <- sort(c(.ele, gaps(.ele)))
        .ele <- .ele[strand(.ele)=="*"]
        .ele <- subsetByOverlaps(.ele, gr)
        .ele$score[is.na(.ele$score)] <- 0
        start(.ele)[start(.ele)<start(gr)[1]] <- start(gr)[1]
        end(.ele)[end(.ele)>end(gr)[1]] <- end(gr)[1]
        .ele
    })
    updateProgress(n=5)
    ggData <- lapply(atac_sig, function(.ele){
        .ele <- as.data.frame(.ele)
        .ele <- .ele[, c("start", "end", "score")]
        colnames(.ele) <- NULL
        .ele <- data.frame(
            x=c(.ele[, 1], .ele[, 2]),
            y=c(.ele[, 3], .ele[, 3]))
        .ele[order(.ele[, 1]), , drop=FALSE]
    })
    updateProgress(n=4)
    LN <- length(ggData)
    gp <- rep(names(ggData), vapply(ggData, FUN=nrow, FUN.VALUE = integer(1L)))
    ggData <- do.call(rbind, ggData)
    if(length(dim(ggData))==0){
        return(ggplot())
    }
    if(nrow(ggData)<1){
        return(ggplot())
    }
    if(ncol(ggData)!=2){
        return(ggplot())
    }
    colnames(ggData) <- c("x", "y")
    ggData <- as.data.frame(ggData)
    # Actual ggplot
    ggData$group <- factor(
        gp,
        levels = rev(sortLevels(as.character(
            unique(gp)
        ))))
    ggOut <- ggplot(
        ggData,
        aes(
            x = .data[["x"]],
            y = .data[["y"]],
            fill = .data[["group"]]
        )) +
        geom_line() +
        scale_x_continuous(expand = c(0, 0)) +
        facet_grid(rows = vars(ggData$group)) +
        ylab("Groups") +
        xlab("Acc Level") +
        scale_y_continuous(breaks = scales::breaks_extended(n = 2)) +
        theme_classic() +
        theme(
            strip.text.y.right = element_text(angle = 0, hjust=0, 
                                              size = labelsFontsize,
                                              family = labelsFontFamily),
            strip.background = element_blank() )
    updateProgress(n=3)
    linkp <- LinkPlot(dataset, gr)
    anno <- AnnotationPlot(dataset, gr, ylimt0 = ifelse(length(linkp)>0, -1, -0.6)) + 
        PeakPlot(dataset, gr) +
        linkp
    updateProgress(n=2)
    return(wrap_plots(
        ggOut, anno, 
        nrow=2, ncol=1,
        heights=c(LN, ifelse(length(linkp)>0, 2.5, 1.5))))
}

#' Plot peaks in a genomic region
#'
#' @noRd
#'
#' @param dataset dataset folder
#' @param region A genomic region to plot
#' @param color Color to use.
#' 
#' @return Returns a \code{\link[ggplot2]{ggplot}} object
#' @importFrom GenomicRanges start end
#' @importFrom IRanges subsetByOverlaps
#' @importFrom GenomeInfoDb seqnames
#' @importFrom ggplot2 ggplot aes geom_segment theme_classic element_blank
#' theme xlab ylab scale_color_manual .data
PeakPlot <- function(
        dataset,
        region,
        color = "dimgrey"
) {
    peaks <- readRDS(file.path(
        .globals$datafolder,
        dataset,
        .globals$filenames$sc1peak
    ))
    peaks <- GRanges(peaks)
    # subset to covered range
    peak.intersect <- subsetByOverlaps(x = peaks, ranges = region)
    peak.df <- as.data.frame(x = peak.intersect)
    start.pos <- start(x = region)
    end.pos <- end(x = region)
    chromosome <- seqnames(x = region)
    
    if (nrow(x = peak.df) > 0) {
        peak.df$start[peak.df$start < start.pos] <- start.pos
        peak.df$end[peak.df$end > end.pos] <- end.pos
        peak.plot <- geom_segment(
                data = peak.df,
                aes(
                    x = .data$start,
                    y = .1,
                    xend = .data$end,
                    yend = .1),
                size = 2,
                color = color,
                inherit.aes = FALSE)
    } else {
        # no peaks present in region, make empty panel
        peak.plot <- NULL
    }
    return(peak.plot)
}

#' Plot linked genomic elements
#'
#' Display links between pairs of genomic elements within a given region of the
#' genome.
#' @noRd
#' @param dataset dataset folder
#' @param region A genomic region to plot#' 
#'
#' @return Returns a \code{\link[ggplot2]{ggplot}} object
#' @importFrom IRanges subsetByOverlaps
#' @importFrom GenomicRanges start end
#' @importFrom GenomeInfoDb seqnames
#' @importFrom ggplot2 ggplot geom_hline aes theme_classic xlim
#' ylab theme element_blank scale_color_gradient2
#' @importFrom ggforce geom_bezier
LinkPlot <- function(
        dataset,
        region
) {
    chromosome <- seqnames(x = region)
    
    # extract link information
    links <- readRDS(file.path(
        .globals$datafolder,
        dataset,
        .globals$filenames$sc1link))
    
    # if links not set, return NULL
    if (length(x = links) == 0) {
        return(NULL)
    }
    if (!'score' %in% colnames(mcols(links))){
        return(NULL)
    }
    ## re-scale scores to fit the color bar (.1-.9)
    min.color <- min(0, min(links$score))
    oldRange <- range(
        c(min.color, links$score[is.finite(links$score)]), na.rm = TRUE)
    oldRange <- diff(oldRange)
    if(oldRange==0) oldRange <- 1
    links$score <- (links$score - min.color) * .8 / oldRange + .1
    # subset to those in region
    links.keep <- subsetByOverlaps(x = links, ranges = region, type='within')
    
    link.df <- as.data.frame(x = links.keep)
    # plot
    if (nrow(x = link.df) > 0) {
        # convert to format for geom_bezier
        df <- data.frame(
            x = c(link.df$start,
                  (link.df$start + link.df$end) / 2,
                  link.df$end),
            y = c(rep(x = 0, nrow(x = link.df)),
                  rep(x = -1, nrow(x = link.df)),
                  rep(x = 0, nrow(x = link.df))),
            score = rep(link.df$score, 3),
            group = rep(seq_along(links.keep), 3)
        )
        p <- geom_bezier(data = df, aes(
            x = .data$x, y = .data$y, color = .data$score, group=.data$group
        ), inherit.aes = FALSE)
    } else {
        p <- NULL
    }
    return(p)
}

#' Plot gene annotations
#'
#' Display gene annotations in a given region of the genome.
#' @noRd
#' @param dataset dataset folder
#' @param region A genomic region to plot
#' @param mode Display mode. Choose either "gene" or "transcript" to determine
#' whether genes or transcripts are plotted.
#' @param ylimt0 ylim0
#' @return Returns a \code{\link[ggplot2]{ggplot}} object
#' @importFrom IRanges subsetByOverlaps
#' @importFrom GenomicRanges start end
#' @importFrom GenomeInfoDb seqnames
#' @importFrom ggplot2 theme_classic ylim xlim ylab xlab
#' geom_segment geom_text aes scale_color_manual .data
#' @importFrom grid arrow
#' @importFrom S4Vectors split
AnnotationPlot <- function(
        dataset,
        region,
        mode = c("gene", "transcript"),
        ylimt0 = -0.6
) {
    mode <- match.arg(mode)
    if(mode == "gene") {
        collapse_transcript <- TRUE
        label <- "gene_name"
    } else if (mode == "transcript") {
        collapse_transcript <- FALSE
        label <- "tx_id"
    }
    annotation <- readRDS(
        file.path(
            .globals$datafolder,
            dataset,
            .globals$filenames$sc1anno))
    if (is.null(x = annotation)) {
        return(NULL)
    }
    # get names of genes that overlap region, then subset to include only those
    # genes. This avoids truncating the gene if it runs outside the region
    annotation.subset <- subsetByOverlaps(x = annotation, ranges = region)
    if (mode == "gene") {
        genes.keep <- unique(x = annotation.subset$gene_name)
        annotation.subset <- annotation[
            match(
                x = annotation$gene_name,
                table = genes.keep,
                nomatch = 0L) > 0L
        ]
    } else {
        tx.keep <- unique(x = annotation.subset$tx_id)
        annotation.subset <- annotation[
            match(x = annotation$tx_id, table = tx.keep, nomatch = 0L) > 0L
        ]
    }
    
    start.pos <- start(x = region)
    end.pos <- end(x = region)
    if (length(x = annotation.subset) == 0) {
        # make empty plot
        p <- ggplot(data = data.frame())
        y_limit <- c(ylimt0, 1)
    } else {
        annotation_df_list <- reformat_annotations(
            annotation = annotation.subset,
            start.pos = start.pos,
            end.pos = end.pos,
            collapse_transcript = collapse_transcript
        )
        colorFun <- function(strand){
            ifelse(strand=="-", 0, 1)
        }
        p <- ggplot() +
            # exons
            geom_segment(
                data = annotation_df_list$exons,
                mapping = aes(
                    x = .data$start,
                    y = .data$dodge,
                    xend = .data$end,
                    yend = .data$dodge,
                    color = colorFun(.data$strand)
                ),
                show.legend = FALSE,
                size = 3
            ) +
            # gene body
            geom_segment(
                data = annotation_df_list$labels,
                mapping = aes(
                    x = .data$start,
                    y = .data$dodge,
                    xend = .data$end,
                    yend = .data$dodge,
                    color = colorFun(.data$strand)
                ),
                show.legend = FALSE,
                size = 1/2
            )
        if (nrow(x = annotation_df_list$plus) > 0) {
            # forward strand arrows
            p <- p + geom_segment(
                data = annotation_df_list$plus,
                mapping = aes(
                    x = .data$start,
                    y = .data$dodge,
                    xend = .data$end,
                    yend = .data$dodge,
                    color = colorFun(.data$strand)
                ),
                arrow = arrow(
                    ends = "last",
                    type = "open",
                    angle = 45,
                    length = unit(x = 0.04, units = "inches")
                ),
                show.legend = FALSE,
                size = 1/2
            )
        }
        if (nrow(x = annotation_df_list$minus) > 0) {
            # reverse strand arrows
            p <- p + geom_segment(
                data = annotation_df_list$minus,
                mapping = aes(
                    x = .data$start,
                    y = .data$dodge,
                    xend = .data$end,
                    yend = .data$dodge,
                    color = colorFun(.data$strand)
                ),
                arrow = arrow(
                    ends = "first",
                    type = "open",
                    angle = 45,
                    length = unit(x = 0.04, units = "inches")
                ),
                show.legend = FALSE,
                size = 1/2
            )
        }
        # label genes
        n_stack <- max(annotation_df_list$labels$dodge)
        annotation_df_list$labels$dodge <- annotation_df_list$labels$dodge + 0.2
        p <- p + geom_text(
            data = annotation_df_list$labels,
            mapping = aes(
                x = .data$position,
                y = .data$dodge,
                label = .data[[label]]),
            size = 2.5
        )
        y_limit <- c(ylimt0, n_stack + 0.4)
    }
    color_breaks <- c(-.05, seq(.1, .9, length.out=9), 1.05)
    colors <- c(
        "darkgreen",
        '#007FFF', '#4CC3FF', '#99EDFF', '#CCFFFF',
        '#FFFFCC', '#FFEE99', '#FFC34C', '#FF7F00',
        "darkblue")
    p <- p +
        theme_classic() +
        ylab("Genes/peaks") +
        xlab(label = paste0(seqnames(region), " position (bp)")) +
        xlim(start.pos, end.pos) +
        ylim(y_limit) +
        theme(
            axis.ticks.y = element_blank(),
            axis.text.y = element_blank()
        ) +
        scale_color_gradientn(
            colors = colors, breaks = color_breaks,
            name = 'score')
    return(p)
}

# copied from Signac
split_body <- function(df, width = 1000) {
wd <- df$end - df$start
nbreak <- wd / width
if (nbreak > 1) {
    steps <- 0:(nbreak)
    starts <- (width * steps) + df$start
    starts[starts > df$end] <- NULL
} else {
    starts <- df$end
}
breaks <- data.frame(
    seqnames = df$seqnames[[1]],
    start = starts,
    end = starts + 1,
    strand = df$strand[[1]],
    tx_id = df$tx_id[[1]],
    gene_name = df$gene_name[[1]],
    gene_biotype = df$gene_biotype[[1]],
    type = "arrow"
)
return(breaks)
}
#' @importFrom GenomicRanges makeGRangesFromDataFrame reduce
record_overlapping <- function(
        annotation,
        min.gapwidth = 1000,
        collapse_transcript = TRUE
) {
    # convert back to granges
    annotation$strand <- "*"
    gr <- makeGRangesFromDataFrame(
        df = annotation[annotation$type == "body", ], keep.extra.columns = TRUE
    )
    # work out which ranges overlap
    collapsed <- reduce(
        x = gr, with.revmap = TRUE, min.gapwidth = min.gapwidth
    )$revmap
    idx <- seq_along(gr)
    for (i in seq_along(collapsed)) {
        mrg <- collapsed[[i]]
        for (j in seq_along(mrg)) {
            idx[[mrg[[j]]]] <- j
        }
    }
    if (collapse_transcript) {
        names(x = idx) <- gr$gene_name
    } else {
        names(x = idx) <- gr$tx_id
    }
    return(idx)
}

#' @importFrom S4Vectors mcols
reformat_annotations <- function(
        annotation,
        start.pos,
        end.pos,
        collapse_transcript = TRUE
) {
    total.width <- end.pos - start.pos
    tick.freq <- total.width / 50
    for(i in c('tx_id', 'gene_name')){
        if(!i %in% colnames(mcols(annotation))){
            stop(i, ' is missing from annotation.',
                 ' Please reprepare the annotations.')
        }
    }
    if(length(annotation$gene_biotype)!=length(annotation)){
        if(length(annotation)){
            annotation$gene_biotype <- NA
        }
    }
    if('type' %in% colnames(mcols(annotation))){
        annotation <- annotation[annotation$type == "exon"]
    }
    annotation <- unname(annotation)
    exons <- as.data.frame(x = annotation)
    if (collapse_transcript) {
        annotation <- split(
            x = annotation,
            f = annotation$gene_name
        )
    } else {
        annotation <- split(
            x = annotation,
            f = annotation$tx_id
        )
    }
    annotation <- lapply(X = annotation, FUN = as.data.frame)
    
    # add gene total start / end
    gene_bodies <- list()
    for (i in seq_along(annotation)) {
        df <- data.frame(
            seqnames = annotation[[i]]$seqnames[[1]],
            start = min(annotation[[i]]$start),
            end = max(annotation[[i]]$end),
            strand = annotation[[i]]$strand[[1]],
            tx_id = annotation[[i]]$tx_id[[1]],
            gene_name = annotation[[i]]$gene_name[[1]],
            gene_biotype = annotation[[i]]$gene_biotype[[1]],
            type = "body"
        )
        # trim any that extend beyond region
        df$start <- ifelse(
            test = df$start < start.pos,
            yes = start.pos,
            no = df$start
        )
        df$end <- ifelse(
            test = df$end > end.pos,
            yes = end.pos,
            no = df$end
        )
        breaks <- split_body(df = df, width = tick.freq)
        df <- rbind(df, breaks)
        gene_bodies[[i]] <- df
    }
    gene_bodies <- do.call(what = rbind, args = gene_bodies)
    
    # record if genes overlap
    overlap_idx <- record_overlapping(
        annotation = gene_bodies,
        min.gapwidth = 1000,
        collapse_transcript = collapse_transcript
    )
    # overlap_idx <- overlap_idx
    if (collapse_transcript) {
        gene_bodies$dodge <- overlap_idx[gene_bodies$gene_name]
        exons$dodge <- overlap_idx[exons$gene_name]
    } else {
        gene_bodies$dodge <- overlap_idx[gene_bodies$tx_id]
        exons$dodge <- overlap_idx[exons$tx_id]
    }
    
    label_df <- gene_bodies[gene_bodies$type == "body", ]
    label_df$width <- label_df$end - label_df$start
    label_df$position <- label_df$start + (label_df$width / 2)
    
    onplus <- gene_bodies[gene_bodies$strand %in% c("*", "+"), ]
    onminus <- gene_bodies[gene_bodies$strand == "-", ]
    
    return(
        list(
            "labels" = label_df,
            "exons" = exons,
            "plus" = onplus,
            "minus" = onminus
        )
    )
}

require("scRNAseqApp") || stop("unable to load Package:scRNAseqApp")
require("Seurat") || stop("unable to load Package::Seurat")
require("testthat") || stop("unable to load testthat")
require('shinytest2') || stop("unable to load shinytest2")
test_check("scRNAseqApp")

#' @importFrom htmltools htmlDependency
#' @importFrom utils packageVersion
#' @importFrom methods getPackageName
visitorDependencies <- function(){
    htmlDependency(
        name = "scRNAseqApp-assets",
        version = packageVersion(getPackageName()),
        package = "scRNAseqApp",
        src = "assets",
        script = c("js/script.js"),
        stylesheet = c("css/style.css")
    )
}
# parse querystring
parseQuery <- function(query, defaultDataset){
    stopifnot(is.list(query))
    from <- 'default'
    if (!is.null(query[['token']])) {
        token <- getTokenList()
        if (query[["token"]] %in% names(token)) {
            defaultDataset <- token[[query[['token']]]]
            from <- 'token'
        }else{
            if (!is.null(query[['data']])) {
                defaultDataset <- query[['data']]
                from <- 'data'
            }
        }
    } else {
        if (!is.null(query[['data']])) {
            defaultDataset <- query[['data']]
            from <- 'data'
        }
    }
    return(c(defaultDataset=defaultDataset, from=from))
}
# summary box for home page
summaryBox <- function(
        title, value,
        width = 4,
        icon = "fas fa-chart-bar",
        style = "info",
        border = "left",
        info = '',
        details = '',
        id = NULL) {
    div(
        class = c(paste0("col-md-", width), "about-left-border"),
        div(
            class = paste0(
                "card border-",
                border, "-",
                style,
                " shadow h-100 py-2"),
            div(
                class = "card-body",
                div(
                    class = "row no-gutters align-items-center",
                    div(
                        class = "col mr-2",
                        div(
                            class = paste0(
                                "text-xs font-weight-bold text-",
                                style, " text-uppercase mb-1"),
                            title
                        ),
                        div(
                            class = "h5 mb-0 font-weight-bold text-gray-800",
                            if(!is.null(id)){
                                textOutput(NS(id, id=paste0('stats', title)),
                                       inline = TRUE)
                            }else{
                                value
                            }
                        )
                    ),
                    div(
                        class = "col-auto about-large-icon about-right",
                        icon(class=paste0("about-", style), icon)
                    )
                )
            )
        ),
        info = info,
        details = details
    )
}

plotLoader <- function(ui, bufferStr='loading...'){
    id <- removeQuote(gsub("^.*?id=(.*?)\\s+.*$", "\\1", ui))
    tagList(
        div(
            class = 'ploader-container',
            div(
                class = 'ploader',
                id=paste0("scRNAseqAppLoader-", id),
                bufferStr
            ),
            ui
        )
    )
}

#' Function to extract legend
#' @noRd
#' @importFrom ggplot2 ggplot_gtable ggplot_build
g_legend <- function(a.gplot){
    tmp <- ggplot_gtable(ggplot_build(a.gplot))
    leg <- which(vapply(tmp$grobs,
                        function(x) x$name, character(1L)) == "guide-box")
    legend <- tmp$grobs[[leg]]
    legend
}

getLogToken <- function(session){
    query <- getQueryString(session = session)
    if(!is.null(query$token)){
        gsub('\"', "", query$token)
    } else {
        NULL
    }
}
# update search results
#' @importFrom utils combn
updateSearch <- function(
        key_words, datasets,
        auth, global, page=1,
        id, input, output, session){
    updateCheckboxInput(session=session,
                        inputId='s_res_flag',
                        value = FALSE)
    on.exit({
        updateCheckboxInput(session=session,
                            inputId='s_res_flag',
                            value = TRUE)
    })
    if(key_words %in% names(getTokenList())){## is token
        token <- key_words
        dataset <- getTokenList()[[token]]
        datasets <- listDatasets(privilege='all', named=TRUE)
        output$search_res <- renderUI(
            tags$ul(
                class='about-ul',
                tags$li(
                    tags$a(
                        href=paste0('?token=', token),
                        names(datasets)[datasets==dataset])
                )
            )
        )
        return(invisible())
    }
    updateSearchTable(key_words)##record the search key
    if( isGene(
        key_words,
        datasets,
        maxEvent = .globals$maxNumGene) ||
        mappingToSymbols(key_words)){## check if it is a gene
        search_res <- checkGene(
            key_words, datasets=datasets,
            auth = auth,
            global = global, page=page,
            id=id, input=input, output=output, session=session)
        output$search_res <-
            renderUI(search_res$UI)
        for(i in seq_along(search_res$PLOT)){
            local({
                output[[names(search_res$PLOT)[i]]] <- search_res$PLOT[[i]]
            })
        }
    }else{
        key_words <- strsplit(key_words, "\\s+")[[1]]
        key_words <- gsub("[^a-zA-Z0-9._'\"*-]+", "", key_words)
        key_words <- unique(key_words)
        if(length(key_words)>1){
            key_words <- lapply(seq.int(length(key_words)), function(m){
                cmb <- combn(key_words, m, simplify = FALSE)
                lapply(cmb, paste, collapse = " ")
            })
            key_words <- unlist(key_words)
        }
        
        res_data <- lapply(getAppConf(privilege = auth$privilege), function(.ele){
            x <- paste(as.character(.ele), collapse = " ")
            m <- vapply(key_words, grepl, logical(1L), x = x, ignore.case=TRUE)
            m <- sum(m)
            return(c(m, .ele$id, .ele$title))
        })
        ## update search_res
        res_data <- do.call(rbind, res_data)
        res_data <- res_data[res_data[, 1]>0, , drop=FALSE]
        res_data <- res_data[
            order(res_data[, 1], decreasing = TRUE), -1, drop=FALSE]
        if(nrow(res_data)>0){
            output$search_res <- renderUI(
                tags$ul(
                    class='about-ul',
                    apply(res_data, 1, function(.ele){
                        # add token if logged
                        token <- getLogToken(session = session)
                        if(!is.null(token)){
                            token <- paste0('&token=', token)
                        }
                        return(tags$li(
                            tags$a(
                                href=paste0('?data=', .ele[1], token),
                                .ele[2],
                                onclick = 
                                    paste0('event.preventDefault();',
                                    'Shiny.onInputChange(',
                                    '\"search_results_select_button\",  \"',
                                    .ele[1],'\");window.scrollTo(0, 0);'))
                            ))
                    })
                )
            )
        }else{
            output$search_res <- renderUI(tags$div(
                "Sorry, I can do nothing with this. Try different one.
                If you see this sentence is fading out,
                I'm working hard on your request.
                Have a coffee."))
        }
    }
}

#' convert search keys to gene symbols
#' @noRd
#' @param x the key words to be converted
#' @param transform convert the x to gene symbol
mappingToSymbols <- function(x, transform=FALSE){
    # quote all gene symbols
    # convert names in database to gene symbols
    rs <- unique(mapGeneSymbols(x)$gene)
    if(!transform){
        return(length(rs)>0)
    }else{
        if(length(rs)>0){
            if(length(rs)==1){
                return(paste0('"', rs, '"'))
            }else{
                return(rs)
            }
        }else{
            return(x)
        }
    }
}

#' check if a symbol is a gene
#' @noRd
#' @param symbol the character to be checked
#' @param datasets the available datasets
isGene <- function(symbol, datasets, maxEvent=3){
    if(isQuote(symbol)){
        symbol <- removeQuote(symbol)
        g <- listGeneSymbols(genes = symbol, datasets = datasets)
        return(length(g)>0)
    }
    if(isAsterisk(symbol)){
        symbol <- isAsterisk(symbol, transform = TRUE, to = '%')
        maxEvent <- .globals$maxNumGene
        like <- TRUE
    }else{
        like <- FALSE
    }
    g <- listGeneSymbols(genes = symbol, datasets = datasets, like=like)
    length(g) > 0 && length(g) < maxEvent
}
#' check if a symbol is quoted
#' @noRd
#' @param symbol the character to be checked
isQuote <- function(symbol){
    grepl("'|\"", symbol)
}
removeQuote <- function(symbol){
    gsub("'|\"", "", symbol)
}
#' check if a symbol contain '*'
#' @noRd
#' @param symbol the character to be checked
#' @param transform change the '*' to '.*'
#' @param to The character to be changed to.
isAsterisk <- function(symbol, transform=FALSE, to='.*'){
    symbol <- symbol[1]
    isT <- grepl("*", symbol, fixed = TRUE)
    if(transform){
        if(isT){
            symbol <- gsub("*", to, symbol, fixed = TRUE)
        }
        return(symbol)
    }
    return(isT)
}
#' get cell type column name
#' @noRd
#' @param config config data table
#' @param celltypePattern the pattern of cell type column name
getCelltypeCol <- function(config, celltypePattern='celltype'){
    groupName <- config[config$grp]$ID
    ad <- adist(celltypePattern, groupName, ignore.case = TRUE)[1, ]
    groupName[which.min(ad)][1]
}

#' waffle plot
#' @noRd
#' @param expres expression table returned by `read_exprs`
#' @param id module id
#' @param plotname the name of plotUI
#' @param numGene number of gene
#' @param groupCol group column name, used to compare two groups in the plot
#' @importFrom utils adist
wafflePlot <- function(
        expres, id, plotname, numGene,
        groupCol="treatment"){
    groupValue <- expres[[groupCol]]
    if(all(
        as.character(groupValue)==
        as.character(expres$grpBy), na.rm=TRUE)){
        groupValue <- 1
    }
    list(
        UI = plotLoader(
            plotOutput(
                NS(id, plotname),
                width = '100%',
                height = paste0(min(numGene*max(
                    1, length(unique(groupValue))),
                    max(length(unique(expres$grpBy)), 6)),
                    '00px'))),
        PLOT = renderPlot(scWafflePlot(expres, groupCol))
    )
}
#' get search result by gene name
#' @noRd
#' @param key_words character(1L), gene name
#' @param datafolder the data folder
#' @param auth for locked data
#' @param id namespace
#' @return Html tags for search results
checkGene <- function(
        key_words, datasets,
        auth, global, page=1,
        id, input, output, session){
    exprs <- NULL
    limit <- 5 # return 5 record
    gene <- mappingToSymbols(key_words, transform = TRUE)
    getGeneNamesByKeyword <- function(){
        appconfs <- getAppConf(privilege = auth$privilege)
        gn <- lapply(appconfs, function(.ele){
            if(checkLocker(.ele$id)){
                if(!checkPrivilege(auth$privilege, .ele$id)){
                    return(NULL)
                }
            }
            geneIds <- readData("sc1gene", .ele$id)
            if(any(isQuote(gene))){
                genenames <- geneIds[
                    tolower(names(geneIds)) %in%
                        tolower(removeQuote(gene))]
            }else{
                if(isAsterisk(gene)){
                    gene <- isAsterisk(gene, transform = TRUE)
                    genenames <- geneIds[grepl(
                        gene, names(geneIds), ignore.case = TRUE)]
                    if(length(genenames)>.globals$limitNumGene){
                        genenames <- geneIds[grepl(
                            paste0("^",gene),
                            names(geneIds),
                            ignore.case = TRUE)]
                    }
                }else{
                    genenames <- 
                        geneIds[tolower(names(geneIds)) %in% tolower(gene)]
                }
            }
            if(length(genenames)>0){
                lgs <- listGeneSymbols(genes = names(genenames),
                                       datasets = .ele$id,
                                       checkExpr=TRUE)
                genenames <- genenames[names(genenames) %in% lgs]
            }
            return(genenames)
        })
        keep <- lengths(gn)>0
        return(list(appconfs=appconfs[keep], genenames=gn[keep]))
    }
    tryCatch({
        global <- global()
        if(is.null(global$search_results[[key_words]])){
            gn <- getGeneNamesByKeyword()
            if(length(gn$genenames)>0){
                showNotification(
                    paste('Plotting expression data for multiple datasets.',
                          'It will take a while. Please be patient.'),
                    duration = 5,
                    type = 'message'
                )
            }
            global$search_results[[key_words]] <-
                list(
                    genenames= gn$genenames,
                    appconfs = gn$appconfs,
                    page=1,
                    total=ceiling(length(gn$genenames)/limit))
        }
        global$search_results[[key_words]]$page <- page
        from <- limit*(page-1)+1
        to <- min(from+limit-1, length(global$search_results[[key_words]]$genenames))
        if(to>=from){
            genenames <- global$search_results[[key_words]]$genenames[from:to]
            appconfs <- global$search_results[[key_words]]$appconfs[from:to]
            
            exprs <- mapply(function(.appconfs, .genenames){
                .genenames <- .genenames[seq.int(min(
                    length(.genenames),
                    .globals$limitNumGene))]
                config <- readData("sc1conf", .appconfs$id)
                groupCol <- getCelltypeCol(
                    config,
                    celltypePattern =
                        .globals$groupColPattern)
                if(!is.null(.appconfs$groupCol)){
                    if(.appconfs$groupCol!=""){
                        groupCol <- .appconfs$groupCol
                    }
                }
                if(is.null(groupCol)){
                    return(NULL)
                }
                if(groupCol==""){
                    return(NULL)
                }
                ggData <-
                    read_exprs(
                        .appconfs$id,
                        .genenames,
                        readData("sc1meta", .appconfs$id),
                        config, groupCol, valueOnly=FALSE)
                ggData[ggData$val < 0]$val <- 0
                #waffle plot
                plotname <- paste0('search-plot', .appconfs$id)
                wp <- wafflePlot(
                    ggData, id,
                    plotname,
                    length(.genenames),
                    groupCol = groupCol)
                list(
                    UI = tags$li(
                        tags$a(href=paste0(
                            '?data=', .appconfs$id, '&gene=',
                            paste(names(.genenames), collapse=";")),
                            .appconfs$title, 
                            onclick = 
                                paste0('event.preventDefault();',
                                       'Shiny.onInputChange(',
                                       '\"search_results_select_button\",  \"',
                                       .appconfs$id,
                                       '\");window.scrollTo(0, 0);')),
                        wp$UI
                    ),
                    PLOT = wp$PLOT,
                    NAME = plotname
                )
            }, appconfs, genenames, SIMPLIFY = FALSE)
        }
    },
    error = function(e){
        message(e)
    })
    exprs <- exprs[lengths(exprs)>0]
    if(length(exprs)==0){
        return(list(UI=tagList(), PLOT=NULL))
    }else{
        plots <- lapply(exprs, function(.ele){
            .ele$PLOT
        })
        names(plots) <- vapply(exprs, function(.ele) .ele$NAME, character(1L))
        exprs <- lapply(exprs, function(.ele){
            .ele$UI
        })
        if(length(global$evt))
            lapply(global$evt, function(.ele){
                if(!is.null(.ele)){
                    .ele$destroy()
                }
            })
        if(!is.null(global$search_results[[key_words]])){
            if(global$search_results[[key_words]]$total>1){
                lapply(
                    seq.int(global$search_results[[key_words]]$total),
                    function(.id){
                        if(.id!=global$search_results[[key_words]]$page){
                            global$evt[[local({.id})]] <-
                                observeEvent(input[[paste0("page", .id)]],{
                                    updateSearch(
                                        key_words,
                                        datasets=datasets,
                                        auth=auth,
                                        global=reactive({global}),
                                        page = local({.id}),
                                        id=id,
                                        input=input,
                                        output=output,
                                        session=session)
                                }, ignoreInit = TRUE, once = TRUE)
                        }
                    })
            }
        }
        return(
            list(
                UI=tagList(
                    ##pagination
                    div(style="padding-left:2rem;",
                        if(global$search_results[[key_words]]$total>1){
                            lapply(
                                seq.int(global$search_results[[key_words]]$total),
                                function(.id){
                                    if(.id!=global$search_results[[key_words]]$page){
                                        tags$span(actionLink(
                                            NS(id, paste0("page", .id)),
                                            paste("page", .id),
                                            `data-value`=.id))
                                    }else{
                                        tags$em(paste("page", .id))
                                    }
                            })
                        }),
                    tags$ul(exprs)),
                PLOT=plots
            )
        )
    }
}

# vistor plots
updateVisitor <- function(input, output, session){
    updateVisitorTable(input, output, session)
    # conterFilename <- .globals$counterFilename
    # ## update visitor stats
    # update_visitor <- function(){
    #     req(input$remote_addr)
    #     counter <- read.delim(conterFilename, header = TRUE)
    #     ips <- counter$ip
    #     counter <- as.Date(counter$date)
    #     visitors <- paste(format(counter, "%d/%m/%y %H"), ips)
    #     current <- Sys.time()
    #     ip <- isolate(input$remote_addr)
    #     agent <- isolate(input$remote_agent)
    #     if(!paste(format(current, "%d/%m/%y %H"), ip) %in% visitors){
    #         write(
    #             paste(as.character(current), ip, agent, sep="\t"),
    #             conterFilename, append = TRUE)
    #     }
    # }
    # observeEvent(input$remote_addr, update_visitor())
    # output$total_visitor <- renderPlot({
    #     counter <- read.delim(conterFilename, header = TRUE)
    #     counter <- as.Date(counter$date)
    #     counter <- counter[as.numeric(difftime(
    #         as.Date(Sys.time()),
    #         counter,
    #         units = 'days'))<730]
    #     counter <- table(format(counter, "%y-%m"))
    #     counter <- as.data.frame(counter)
    #     ggplot(counter, aes(x=.data[["Var1"]], y=.data[["Freq"]])) +
    #         geom_bar(stat = "identity", fill="darkorchid4") +
    #         theme_minimal() + xlab("") + ylab("visitor counts") +
    #         theme(axis.text.x = element_text(angle = 90, vjust = 0.5, hjust=1))
    # })
}

# used to avoid suppressWarnings(as.numeric)
char2numeric <- function(x, keep="[., 0-9eE+-]"){
    keep <- grepl(paste0("^", keep, "+$"), x)
    x[!keep] <- NA
    x <- gsub('[ ,]+', '', x)
    x <- gsub('eminus', 'e-', gsub('-', '', gsub('e-', 'eminus', x)))
    x <- as.numeric(x)
    x
}

# get assay data by SeuratObject version
#' @importFrom utils packageVersion
#' @importFrom SeuratObject GetAssayData
extAssayData <- function(object, slot, ...){
    seuObjVersion <- packageVersion('SeuratObject')
    if(seuObjVersion<'5.0.0'){
        GetAssayData(object=object, slot = slot, ...)
    }else{
        GetAssayData(object=object, layer = slot, ...)
    }
}

# load config file
loadConfigFile <- function(.globals, app_path){
    configFile <- file.path(app_path, 'www/config.dcf')
    if(file.exists(configFile)){
        configs <- read.dcf(configFile)
        cn <- intersect(colnames(configs), names(.globals))
        for(i in cn){
            if(i=='theme'){
                theme <- eval(parse(text=paste0("bs_theme(", configs[1, i], ")")))
                .globals$theme <- theme
            }else{
                if(!is.list(.globals[[i]])){
                    param <- strsplit(configs[1, i], ",\\s*")[[1]]
                    mode(param) <- mode(.globals[[i]])
                    .globals[[i]] <- param
                }
            }
        }
    }else{
        configFile <- file.path(app_path, 'www/config.rds')
        ## no safety check here, need to reconsider.
        if(file.exists(configFile)){
            configs <- readRDS(configFile)
            cn <- intersect(names(configs), names(.globals))
            for(i in cn){
                if(is.list(.globals[[i]])){
                    cn2 <- intersect(names(.globals[[i]]), names(configs[[i]]))
                    for(j in cn2){
                        .globals[[i]][[j]] <- configs[[i]][[j]]
                    }
                }else{
                    .globals[[i]] <- configs[[i]]
                }
            }
        }
    }
    return(.globals)
}

#' @importFrom grDevices col2rgb
#' @importFrom ggplot2 theme
is_dark <- function(color) {
    rgb <- col2rgb(color)
    luminance <- 0.299 * rgb[1] + 0.587 * rgb[2] + 0.114 * rgb[3]
    return(luminance < 128)
}
#' @importFrom bslib bs_get_variables bs_current_theme
darkTheme <- function(p, returnBG=FALSE, bg_color){
    # Extract the background color
    if(missing(bg_color)||is.null(bg_color)){
        bg_color <- bslib::bs_get_variables(bs_current_theme(), "body-bg")
    }
    if(returnBG) return(bg_color)
    if(isTRUE(is_dark(bg_color))&&is(p, 'ggplot')){
        p <- p + 
            theme(
                plot.background = element_rect(fill = bg_color, color = NA),
                panel.background = element_rect(fill = "gray20", color = NA),
                panel.border = element_blank(),
                panel.grid = element_line(color = "gray30"),
                axis.text = element_text(color = "white"),
                axis.title = element_text(color = "white"),
                plot.title = element_text(color = "white"),
                plot.subtitle = element_text(color = "white"),
                legend.background = element_rect(fill = bg_color, color = NA),
                legend.text = element_text(color = "white"),
                legend.title = element_text(color = "white"),
                legend.key = element_rect(fill = bg_color, color = NA),
                strip.background = element_rect(fill = "gray20", color = NA),
                strip.text = element_text(color = "white")
            )
    }
    return(p)
}

#' @importFrom ggplot2 xlim ylim ggplot_build ggplot_gtable layer_scales
#' @importFrom grid convertWidth
addLimits <- function(p, x=NULL, y=NULL, coord=NULL, id, postfix, input){
    notify <- FALSE
    if(is(p, 'ggplot')){
        if(!is.null(coord)){
            ## ATAC plot
            if(!is.null(x)){
                try({
                    gr0 <- GRanges(coord)
                    gr <- c(start(gr0), end(gr0))
                    infoLabel <- paste0('GeneExpext.info', postfix)
                    savedCoord <- isolate({
                        input[[infoLabel]]
                    })
                    if(savedCoord!=''){
                        gr <- as.numeric(strsplit(savedCoord, '-')[[1]])
                    }
                    gt <- ggplot_gtable(ggplot_build(p[[1]]))
                    w <- convertWidth(gt$widths, 'npc', valueOnly=TRUE)
                    ## the plot region is in grid 7
                    left <- sum(w[seq.int(6)])
                    right <- sum(w[-seq.int(7)])
                    ## remove all the left and right part from the xlim
                    x <- (x-left)/(1-right-left)
                    x[x<0] <- 0
                    x[x>1] <- 1
                    x <- x*(gr[2]-gr[1]) + gr[1]
                    x <- round(x)
                    p[[1]] <- p[[1]] + xlim(x) ## tracks
                    p[[2]] <- p[[2]] + xlim(x) ## annotations
                    updateTextInput(inputId = infoLabel,
                                    value = paste(x, collapse = '-'))
                    notify <- TRUE
                })
            }
        }else{
            if(!is.null(x) && is(layer_scales(p)$x, 'ScaleContinuousPosition')){
                p <- p + xlim(x)
                notify <- TRUE
            }
            if(!is.null(y) && is(layer_scales(p)$y, 'ScaleContinuousPosition')){
                p <- p + ylim(y)
                notify <- TRUE
            }
        }
    }
    if(notify){
        showNotification("Double click to cancel zoom in.",
                         type = 'message')
    }
    return(p)
}

# copied internal functions from https://github.com/datastorm-open/shinymanager
shinymanager_where <- function(where) {
    tags$div(
        style = "display: none;",
        selectInput(
            inputId = "shinymanager_where",
            label = NULL,
            choices = where,
            selected = where,
            multiple = FALSE
        )
    )
}
shinymanager_language <- function(lan) {
    tags$div(
        style = "display: none;",
        selectInput(
            inputId = "shinymanager_language",
            label = NULL,
            choices = lan,
            selected = lan,
            multiple = FALSE
        )
    )
}

# A split violin plot geom
#
#' @importFrom scales zero_range
#' @importFrom ggplot2 GeomPolygon ggproto GeomViolin
#' @importFrom grid grobTree grobName
#
# @author jan-glx on StackOverflow
# @references \url{https://stackoverflow.com/questions/35717353/split-violin-plot-with-ggplot2}
# @seealso \code{\link[ggplot2]{geom_violin}}
#
GeomSplitViolin <- ggproto(
    "GeomSplitViolin",
    GeomViolin,
    # setup_data = function(data, params) {
    #   data$width <- data$width %||% params$width %||% (resolution(data$x, FALSE) * 0.9)
    #   data <- plyr::ddply(data, "group", transform, xmin = x - width/2, xmax = x + width/2)
    #   e <- globalenv()
    #   name <- paste(sample(x = letters, size = 5), collapse = '')
    #   message("Saving initial data to ", name)
    #   e[[name]] <- data
    #   return(data)
    # },
    draw_group = function(self, data, ..., draw_quantiles = NULL) {
        data$xminv <- data$x - data$violinwidth * (data$x - data$xmin)
        data$xmaxv <- data$x + data$violinwidth * (data$xmax - data$x)
        grp <- data[1, 'group']
        if (grp %% 2 == 1) {
            data$x <- data$xminv
            data.order <- data$y
        } else {
            data$x <- data$xmaxv
            data.order <- -data$y
        }
        newdata <- data[order(data.order), , drop = FALSE]
        newdata <- rbind(
            newdata[1, ],
            newdata,
            newdata[nrow(x = newdata), ],
            newdata[1, ]
        )
        newdata[c(1, nrow(x = newdata) - 1, nrow(x = newdata)), 'x'] <- round(x = newdata[1, 'x'])
        grob <- if (length(x = draw_quantiles) > 0 & !zero_range(x = range(data$y))) {
            stopifnot(all(draw_quantiles >= 0), all(draw_quantiles <= 1))
            quantiles <- QuantileSegments(data = data, draw.quantiles = draw_quantiles)
            aesthetics <- data[rep.int(x = 1, times = nrow(x = quantiles)), setdiff(x = names(x = data), y = c("x", "y")), drop = FALSE]
            aesthetics$alpha <- rep.int(x = 1, nrow(x = quantiles))
            both <- cbind(quantiles, aesthetics)
            quantile.grob <- GeomPath$draw_panel(both, ...)
            grobTree(GeomPolygon$draw_panel(newdata, ...), name = quantile.grob)
        }
        else {
            GeomPolygon$draw_panel(newdata, ...)
        }
        grob$name <- grobName(grob = grob, prefix = 'geom_split_violin')
        return(grob)
    }
)

# Create a split violin plot geom
#
# @inheritParams ggplot2::geom_violin
#
#' @importFrom ggplot2 layer
#
# @author jan-glx on StackOverflow
# @references \url{https://stackoverflow.com/questions/35717353/split-violin-plot-with-ggplot2}
# @seealso \code{\link[ggplot2]{geom_violin}}
#
geom_split_violin <- function(
        mapping = NULL,
        data = NULL,
        stat = 'ydensity',
        position = 'identity',
        ...,
        draw_quantiles = NULL,
        trim = TRUE,
        scale = 'area',
        na.rm = FALSE,
        show.legend = NA,
        inherit.aes = TRUE
) {
    return(layer(
        data = data,
        mapping = mapping,
        stat = stat,
        geom = GeomSplitViolin,
        position = position,
        show.legend = show.legend,
        inherit.aes = inherit.aes,
        params = list(
            trim = trim,
            scale = scale,
            draw_quantiles = draw_quantiles,
            na.rm = na.rm,
            ...
        )
    ))
}

#' @importFrom data.table .N
scDRnum <- function(
        inpConf,
        inpMeta,
        inpCellInfo,
        gene,
        inpsubName,
        inpsubValue,
        dataset,
        geneIdMap,
        inpsplt) {
    if (length(inpCellInfo) > 1) {
        inpCellInfoSubGroup <- inpCellInfo[-1]
        inpCellInfo <- inpCellInfo[1]
    } else{
        inpCellInfoSubGroup <- NULL
    }
    # Prepare ggData
    ggData <- inpMeta[, c(
        inpConf[inpConf$UI == inpCellInfo]$ID,
        inpConf[inpConf$UI == inpsubName]$ID),
        with = FALSE, drop = FALSE]
    if (nrow(ggData) < 1) {
        dt <- data.frame(
            "group" = numeric(),
            "nCells" = numeric(),
            "nExpress" = numeric(),
            "pctExpress" = numeric()
        )
        return(dt)
    }
    colnames(ggData) <- c("group", "sub")
    ggData$val2 <-
        read_exprs(dataset, geneIdMap[gene], valueOnly = TRUE)
    ggData[ggData$val2 < 0]$val2 <- 0
    if (length(inpsubValue) != 0 &
        length(inpsubValue) != nlevels(ggData$sub)) {
        if(is.list(inpsubValue)) inpsubValue <- unlist(inpsubValue)
        ggData <- ggData[ggData$sub %in% inpsubValue]
    }
    if (length(inpCellInfoSubGroup) != 0 &
        length(inpCellInfoSubGroup) != nlevels(ggData$group)) {
        if(is.list(inpCellInfoSubGroup)) inpCellInfoSubGroup <-
                unlist(inpCellInfoSubGroup)
        ggData <- ggData[ggData$group %in% inpCellInfoSubGroup]
    }
    
    # Split inpCellInfo if necessary
    if (is.na(inpConf[inpConf$UI == inpCellInfo]$fCL)) {
        if (inpsplt == "Quartile") {
            nBk <- 4
        }
        if (inpsplt == "Decile") {
            nBk <- 10
        }
        ggData$group <- cut(ggData$group, breaks = nBk)
    }
    
    # Actual data.table
    ggData$express <- FALSE
    ggData[ggData$val2 > 0]$express <- TRUE
    ggData1 <-
        ggData[ggData$express == TRUE, list(nExpress = .N), by = "group"]
    ggData <- ggData[, list(nCells = .N), by = "group"]
    ggData <- ggData1[ggData, on = "group"]
    ggData <- ggData[, c("group", "nCells", "nExpress"), with = FALSE]
    ggData[is.na(ggData$nExpress)]$nExpress <- 0
    ggData$pctExpress <- 100 * ggData$nExpress / ggData$nCells
    ggData <- ggData[order(ggData$group)]
    colnames(ggData)[3] <-
        paste0(colnames(ggData)[3], "_", gene)
    return(ggData)
}

adminMsg <- function(
        msg,
        type,
        duration = 5,
        close = TRUE) {
    showNotification(
        toString(msg)[1],
        duration = duration,
        closeButton = close,
        type = type
    )
}

conditionHandler <- function(cond) {
    type <- 
        vapply(c("warning", "message", "error"), function(x)
            is(cond, x),
            logical(1L))
    type <- c("warning", "message", "error")[type]
    if (length(type) == 0)
        type <- 'warning'
    if (type == "message") {
        progress <- get("progress_value", envir = .globals)
        if (progress < .9) {
            progress <- progress + .05
        }
        assign("progress_value", progress, envir = .globals)
        get("progress",
            envir = .globals)$set(message = cond$message, value = progress)
    } else{
        adminMsg(
            cond$message,
            type = type[1],
            duration = 5,
            close = TRUE
        )
    }
}

#' @noRd
#' @param startMsg,endMsg messages for start and end of the progress
#' @param expr expression for the process
adminProcess <- function(expr, startMsg, endMsg, duration=5) {
    # Create a Progress object
    assign("progress", shiny::Progress$new(), envir = .globals)
    assign("progress_value", 0, envir = .globals)
    on.exit(get("progress", envir = .globals)$close())
    get("progress", envir = .globals)$set(
        message = startMsg, value = 0)
    withCallingHandlers(
        withRestarts({
            expr
        }, muffleStop = function()
            NULL),
        message = conditionHandler,
        warning = conditionHandler,
        error = function(e) {
            conditionHandler(e)
            invokeRestart("muffleStop")
        }
    )
    get("progress", envir = .globals)$close()
    on.exit()
    if (!missing(endMsg)) {
        adminMsg(endMsg, "message", duration = duration)
    }
}

askNamespace <- function(...) {
    pkgs <- list(...)
    lapply(pkgs, function(pkg) {
        if (!requireNamespace(pkg)) {
            adminMsg(paste(
                "The", pkg, "package is required for this function!"),
                type = "error")
            return()
        }
    })
}

redirectOutput <- function(
        input,
        output,
        session,
        open = TRUE,
        tmpf = tempfile()) {
    if (open) {
        sink(tmpf)
        autoInvalidate <- reactiveTimer(1000)
        observe({
            autoInvalidate()
            output$consoleOutput <-
                renderText(readLines(tmpf), sep = "\n")
        })
        
    } else{
        sink()
        autoInvalidate <- reactiveTimer(1000000)
    }
}

#' @noRd
#' @importFrom xml2 read_xml as_list
#' @importFrom utils download.file
#' @param id id to be convert
#' @param type target id type
#' @param url the service url
idConverter <- function(
        id,
        type = c("doi", "pmid"),
        url = 'https://www.ncbi.nlm.nih.gov/pmc/utils/idconv/v1.0/') {
    type <- match.arg(type)
    url0 <- paste0(url, "?format=xml&ids=", id)
    res <- tryCatch({
        read_xml(url0)
    }, error = function(e){
        message(e, ' Try it again.')
        tmp <- tempfile()
        download.file(url0, destfile = tmp)
        read_xml(tmp)
    })
    res <- as_list(res)
    if (!is.null(res$pmcids$record)) {
        return(attr(res$pmcids$record, type))
    }
    return(NULL)
}
updateRefIDs <- function(element, input, output, session) {
    o_ele <- ifelse(element == "doi", "pmid", "doi")
    updateID <- FALSE
    updateID <- is.null(input[[o_ele]])
    if (!updateID) {
        updateID <- gsub("\\s+", '', input[[o_ele]]) == ""
    }
    
    if (updateID) {
        res <- idConverter(input[[element]], type = o_ele)
        if (!is.null(res)) {
            updateTextInput(session,
                            o_ele,
                            value = res)
        } else{
            adminMsg(
                paste('Please update the correct', toupper(o_ele)),
                'warning')
        }
    }
}

#' @importFrom methods is
updateRefById <-
    function(id, element, FUN, input, output, session) {
        if (!is.null(input[[element]])) {
            if (input[[element]] != "") {
                tryCatch({
                    bibentry <- FUN(input[[element]])
                    if (is(bibentry, "bibentry")) {
                        updateTextAreaInput(
                            session,
                            id,
                            value = format(bibentry, style = 'html'))
                    }
                    updateRefIDs(element, input, output, session)
                    return(bibentry)
                },
                error = function(e) {
                    adminMsg(e, 'error')
                },
                warning = function(w) {
                    adminMsg(w, 'warning')
                })
                
            }
        }
    }
getGrpIDs <- function(config) {
    x <- config$ID[config$grp]
    names(x) <- x
    x
}
filterGrpIDs <- function(grp_ids, meta) {
    keep <- vapply(
        grp_ids,
        FUN = function(.ele) {
            all(is.na(char2numeric(as.character(
                unique(meta[[.ele]])
            ))))
        },
        FUN.VALUE = logical(1L)
    )
    grp_ids[keep]
}

#' @importFrom utils bibentry
updateAppConf <- function(input, global) {
    markers <- global()$markers
    if (is.character(markers)) {
        markers <- markers[!is.na(markers)]
        markers <- markers[markers != ""]
        markers <- t(t(markers))
        rownames(markers) <- markers
        markers <- list(markers = as.data.frame(markers))
    }
    if(is.list(markers) && !is.data.frame(markers)){
        if(!all(vapply(markers, is.data.frame, logical(1L)))){
            markers <- as.data.frame(markers)
        }
    }
    if(is.data.frame(markers)){
        markers <- list(markers)
    }
    if (input$species2 != "" && input$species2 != "NA") {
        species <- input$species2
    } else{
        species <- input$species
    }
    entry <- global()$ref
    if(!is(entry, 'bibentry')){
        author <- 'NA'
        journal <- 'unpublished'
        year <- format(Sys.time(), '%Y')
        entry <- bibentry(
            bibtype = 'Article',
            title = input$title,
            doi = input$doi,
            pmid = input$pmid,
            author = author,
            journal = journal,
            year = year,
            abstract = input$abstract
        )
    }
    appconf <- APPconf(
        title = input$title,
        id = input$dir,
        species = species,
        ref = list(
            bib = input$reference,
            doi = input$doi,
            pmid = input$pmid,
            entry = entry
        ),
        type = input$datatype,
        markers = markers,
        keywords = input$keywords
    )
    if (!is.null(input$dir)) {
        if (input$dir != "") {
            saveAppConf(appconf)
            updateConfigTable(appconf)
        }
    }
}
updateDef <- function(input) {
    sc1def <- list(
        meta1 = input$meta1,
        meta2 = input$meta2,
        gene1 = input$gene1,
        gene2 = input$gene2,
        genes = input$multigene,
        dimred = c(input$dimred1, input$dimred2),
        grp1 = input$grp1,
        grp2 = input$grp2
    )
    if (!is.null(input$dir)) {
        if (input$dir != "") {
            saveData(sc1def, input$dir, "sc1def")
        }
    }
}
formatfID_CL <- function(x, rev = FALSE) {
    if (rev) {
        x[!is.na(x)] <- gsub(";\\s+", "|", x[!is.na(x)])
        return(x)
    }
    x[!is.na(x)] <- gsub("\\|", "; ", x[!is.na(x)])
    x
}
#' read expression matrix from sc1gexpr.h5
#' @noRd
#' @param h5filename Parent foldername of h5 file
#' @param rown,coln rownames and colnames for the expression file.
#' rownames = names(readRDS('sc1gene.rds'));
#' colnames = readRDS('sc1meta.rds')$sampleID
#' @importFrom rhdf5 h5read
readDataMatrix <- function(h5filename, rown, coln) {
    expr <- h5read(
        file.path(
            .globals$datafolder,
            h5filename,
            .globals$filenames$sc1gexpr
        ),
        .globals$h5fGrp
    )
    rownames(expr) <- rown
    colnames(expr) <- coln
    return(expr)
}

getReductionMethod <-
    function(reduction_method, forMonocole = TRUE) {
        if ("umap" %in% reduction_method) {
            reduction_method <- "UMAP"
        } else{
            if ("tsne" %in% reduction_method) {
                reduction_method <- "tSNE"
            } else{
                if ("pca" %in% reduction_method) {
                    reduction_method <- "PCA"
                } else{
                    reduction_method <- "UMAP"
                }
            }
        }
        if (!forMonocole)
            reduction_method <- tolower(reduction_method)
        reduction_method
    }

# updateMetaData by double click the info plot
# if it is color, change color
# if it is label, change label
updateMetaData <- function(dataset, inpConf, inpMeta, privilege,
                           info, oldvalue, newvalue){
    check <- FALSE
    if(checkPrivilege(privilege, dataset)){
        if(!info %in% inpConf$UI){
            if(info=='sampleID'){
                adminMsg(
                    'sampleID is not a proper info name.',
                    type = 'warning',
                    duration = 5,
                    close = TRUE
                )
            }else if(newvalue=='duplicate'){
                newvalue <- inpConf[inpConf$UI==oldvalue]
                id <- newvalue$ID
                newvalue$UI <- info
                newvalue$ID <- make.names(info)
                inpConf <- rbind(inpConf, newvalue)
                inpMeta[[newvalue$ID]] <- inpMeta[[id]]
                saveData(inpMeta, dataset, "sc1meta")
                saveData(inpConf, dataset, "sc1conf")
                check <- TRUE
            }else if(newvalue=='rename'){
                newvalue <- inpConf[inpConf$UI==oldvalue]
                id <- newvalue$ID
                newvalue$UI <- info
                newvalue$ID <- make.names(info)
                inpConf[inpConf$UI==oldvalue] <- newvalue
                colnames(inpMeta)[colnames(inpMeta)==id] <- newvalue$ID
                saveData(inpMeta, dataset, "sc1meta")
                saveData(inpConf, dataset, "sc1conf")
                check <- TRUE
            }else if(newvalue=='delete'){
                newvalue <- inpConf[inpConf$UI==oldvalue]
                #inpMeta[, (newvalue$ID):=NULL]
                inpConf <- inpConf[inpConf$UI!=oldvalue]
                saveData(inpMeta, dataset, "sc1meta")
                saveData(inpConf, dataset, "sc1conf")
                check <- TRUE
            }
        }else{
            id <- inpConf[inpConf$UI==info]$ID
            fID <- strsplit(inpConf[inpConf$UI==info]$fID, '\\|')[[1]]
            fCL <- strsplit(inpConf[inpConf$UI==info]$fCL, '\\|')[[1]]
            if(oldvalue %in% fID){
                if(grepl('Merging to:', newvalue)){
                    newvalue <- sub('Merging to:', '', newvalue)
                    newvalue <- sub('^ *', '', newvalue)
                    nCL <- fCL[which(fID==newvalue)]
                    idx <- which(fID==oldvalue)
                    fCL <- fCL[-idx]
                    fID <- fID[-idx]
                    lfactor <- is.factor(inpMeta[[info]])
                    l <- inpMeta[[info]]
                    if(lfactor){
                        l <- as.character(l)
                    }
                    l[which(l==oldvalue)] <- newvalue
                    if(lfactor){
                        inpMeta[[info]] <- factor(l, levels = fID)
                    }
                    fID <- paste(fID, collapse = '|')
                    inpConf[inpConf$UI==info, 'fID'] <- fID
                    fCL <- paste(fCL, collapse = '|')
                    inpConf[inpConf$UI==info, 'fCL'] <- fCL
                    saveData(inpMeta, dataset, "sc1meta")
                    saveData(inpConf, dataset, "sc1conf")
                    check <- TRUE
                }else if(!newvalue %in% fID){
                    fID[which(fID==oldvalue)] <- newvalue
                    lfactor <- is.factor(inpMeta[[info]])
                    l <- inpMeta[[info]]
                    if(lfactor){
                        l <- as.character(l)
                    }
                    l[which(l==oldvalue)] <- newvalue
                    if(lfactor) inpMeta[[info]] <- factor(l, levels = fID)
                    fID <- paste(fID, collapse = '|')
                    inpConf[inpConf$UI==info, 'fID'] <- fID
                    saveData(inpMeta, dataset, "sc1meta")
                    saveData(inpConf, dataset, "sc1conf")
                    check <- TRUE
                }else{
                    adminMsg(
                        'duplicated value. If you want to merge the cell info, please add "Merging to:" to the label.',
                        type = 'warning',
                        duration = 5,
                        close = TRUE
                    )
                }
            }else{
                if(oldvalue %in% fCL){
                    if(!newvalue %in% fCL){
                        fCL[which(fCL==oldvalue)] <- newvalue
                        fCL <- paste(fCL, collapse = '|')
                        inpConf[inpConf$UI==info, 'fCL'] <- fCL
                        saveData(inpConf, dataset, "sc1conf")
                        check <- TRUE
                    }else{
                        adminMsg(
                            'duplicated value',
                            type = 'warning',
                            duration = 5,
                            close = TRUE
                        )
                    }
                }
            }
        }
    }
    return(check)
}

#' @importFrom grDevices rgb
#' @importFrom data.table data.table
#' @importFrom ggplot2 ggplot aes .data xlab ylab coord_fixed
#' scale_x_continuous scale_y_continuous geom_tile
scDRcoexLeg <- function(gene1, gene2, colorPairs, labelsFontsize = 24,
                        labelsFontFamily = 'Helvetica') {
    # Generate coex color palette
    nTot <- getTotalNumber(nGrid = 16, nPad = 2)
    gg <- getCoexpCol(colorPairs, nGrid = 16, nPad = 2)
    
    # Actual ggplot
    ggOut <- ggplot(gg, aes(.data[["v1"]], .data[["v2"]])) +
        geom_tile(fill = gg$cMix) +
        xlab(gene1) + ylab(gene2) + coord_fixed(ratio = 1) +
        scale_x_continuous(
            breaks = c(0, nTot),
            labels = c("low", "high")) +
        scale_y_continuous(
            breaks = c(0, nTot),
            labels = c("low", "high")) +
        sctheme(base_size = labelsFontsize,
                family = labelsFontFamily, XYval = TRUE)
    return(ggOut)
}

cellInfoGeneExprUI <- function(id) {
    tabPanel(
        value = id,
        htmlOutput(NS(id, 'GeneExpr')),
        tabsubTitleUI(
            id,
            'GeneExpr',
            description = paste(
                "In this tab, users can visualise both cell and feature",
                "information side-by-side on low-dimensional representions."
            )
        ),
        fluidRow(
            column(3, dimensionReductionUI(id)),
            column(5, subsetCellByInfoUI(id)),
            column(4, graphicsControlUI(id))
        ),
        fluidRow(
            column(
                6,
                style = "border-right: 2px solid black",
                h4("Cell information"),
                fluidRow(
                    column(6, cellInfoUI(id, 1)),
                    column(6, cellInfoPlotControlUI(id, 1))),
                geneExprDotPlotUI(id, 1, editor=TRUE),
                br(),
                cellInfoTblUI(id, 1)
            ),
            column(
                6,
                htmlOutput(NS0(id, "subPlotTitle", 2)),
                fluidRow(
                    column(6, geneExprUI(id, 2)),
                    column(6, geneExprPlotControlUI(id, 2))),
                geneExprDotPlotUI(id, 2)
            )
        )
    )
}
#' @importFrom DT formatRound renderDT
#' @importFrom magrittr %>%
cellInfoGeneExprServer <- function(id, dataSource, optCrt) {
    moduleServer(id, function(input, output, session) {
        ## title
        output$GeneExpr <-
            renderUI({
                HTML(paste("CellInfo vs", dataSource()$terms["GeneExpr"]))
            })
        ## subtitle
        output$GeneExprSubTitle <-
            renderUI({
                h4(
                    paste(
                        "Cell information vs gene",
                        dataSource()$terms['expression'],
                        "on reduced dimensions"
                    )
                )
            })
        ## input column 1
        ### Dimension Reduction
        updateDimRedSelInputPair(session, input, dataSource)
        ## input column 2
        updateSubsetCellUI(id, input, output, session, dataSource)
        
        ## plot region
        ### sub region title
        output$subPlotTitle2 <-
            renderUI({
                h4(paste("Gene", dataSource()$terms['expression']))
            })
        ### cellInfo
        updateCellInfoPlot(1, id, input, output, session, dataSource)
        ### expression stats table
        output$GeneExpr.dt1 <- renderDT({
            ggData <- scDRnum(
                inpConf=dataSource()$sc1conf,
                inpMeta=dataSource()$sc1meta,
                inpCellInfo=input$CellInfo1,
                gene=input$GeneName2,
                inpsubName=input$subsetCell,
                inpsubValue=getSubsetCellVal(input),
                dataset=dataSource()$dataset,
                geneIdMap=dataSource()$sc1gene,
                inpsplt=input$GeneExprsplt1
            )
            datatable(
                ggData,
                rownames = FALSE,
                extensions = "Buttons",
                options = list(
                    pageLength = -1,
                    dom = "tB",
                    buttons = c("copy", "csv", "excel")
                )
            ) %>%
                formatRound(
                    columns = c("pctExpress"),
                    digits = 2)
        })
        
        ### geneExpr
        selectedGene <- dataSource()$sc1def$gene1
        if (!is.null(dataSource()$genelist)) {
            selectedGene <- dataSource()$genelist[1]
        }
        updateGeneExprPlot(
            postfix = 2,
            selectedGene,
            optCrt,
            id,
            input,
            output,
            session,
            dataSource)
    })
}

outputFileName <- function(ext, ...) {
    paste0(paste(
        ...,
        format(Sys.time(), "%Y%m%d.%M%S"),
        sep = "_"), ".", tolower(ext))
}
#' @noRd
#' @importFrom ggplot2 ggsave
#' @importFrom shiny downloadHandler
plotsDownloadHandler <- function(input, postfix, plot, ...) {
    downloadHandler(
        filename = function() {
            outputFileName(input[[paste0("GeneExproup.fmt", postfix)]], ...)
        },
        content = function(file) {
            if (input[[paste0("GeneExproup.fmt", postfix)]] == "PDF") {
                ggsave(
                    file,
                    device = 
                        tolower(input[[paste0("GeneExproup.fmt", postfix)]]),
                    height = input[[paste0("GeneExproup.h", postfix)]],
                    width = input[[paste0("GeneExproup.w", postfix)]],
                    useDingbats = FALSE,
                    plot = darkTheme(plot())
                )
            } else if(input[[paste0("GeneExproup.fmt", postfix)]] == "CSV"){
                plot <- plot()
                write.csv(plot$data, file = file)
            } else{
                ggsave(
                    file,
                    device =
                        tolower(input[[paste0("GeneExproup.fmt", postfix)]]),
                    height = input[[paste0("GeneExproup.h", postfix)]],
                    width = input[[paste0("GeneExproup.w", postfix)]],
                    plot = darkTheme(plot())
                )
            }
        }
    )
}
#' @importFrom grDevices dev.off pdf
heatmapDownloadHandler <-
    function(input, postfix, plot, ...) {
        downloadHandler(
            filename = function() {
                outputFileName(input[[paste0("GeneExproup.fmt", postfix)]], ...)
            },
            content = function(file) {
                if(input[[paste0("GeneExproup.fmt", postfix)]] == "CSV"){
                    plot <- plot()
                    env <- environment(
                        plot@ht_list$expression@matrix_param$layer_fun)
                    if('ggProp' %in% names(env)){
                        dat <- cbind(env$ggProp, env$ggMat)
                        colnames(dat) <- c(
                            paste(colnames(env$ggProp), 'prop',
                                  sep='.'),
                            paste(colnames(env$ggMat), 'rescaled.expr',
                                  sep='.'))
                    }else{
                        dat <- plot@ht_list$expression@matrix
                    }
                    write.csv(dat, file = file)
                }else{
                    if (input[[paste0("GeneExproup.fmt", postfix)]] == "PDF") {
                        pdf(
                            file,
                            height = input[[paste0("GeneExproup.h", postfix)]],
                            width = input[[paste0("GeneExproup.w", postfix)]],
                            useDingbats = FALSE
                        )
                    } else{
                        get(tolower(input[[paste0("GeneExproup.fmt", postfix)]]))(
                            file,
                            height = 
                                input[[paste0("GeneExproup.h",
                                              postfix)]] * 300,
                            width = 
                                input[[paste0("GeneExproup.w",
                                              postfix)]] * 300,
                            res = 300)
                    }
                    on.exit({
                        dev.off()
                    })
                    draw(plot()) ## for complexheatmap
                }
            }
        )
    }
#' @importFrom plotly event_data
#' @importFrom shiny getDefaultReactiveDomain
#' @importFrom utils write.csv
click_event_data <- function(session = shiny::getDefaultReactiveDomain(), ...) {
    if('plotly_click' %in% session$userData$plotlyShinyEventIDs){
        event_data(event = "plotly_click", ...)
    }else{
        return(NULL)
    }
}
lasso_select_data <- function(session = shiny::getDefaultReactiveDomain(), ...) {
    if('plotly_click' %in% session$userData$plotlyShinyEventIDs){
        event_data(event = "plotly_selected", ...)
    }else{
        return(NULL)
    }
}
exprDownloadHandler <- function(geneIdMap, dataset, meta) {
    downloadHandler(
        filename = function() {
            paste0('exprdata-', dataset, '.csv')
        },
        content = function(file) {
            d <- click_event_data()
            expr <- NULL
            if (!is.null(d)) {
                cell <- which(meta$sampleID == d$customdata)
                if (!is.null(cell)) {
                    expr <- read_exprs(
                        h5f = dataset,
                        valueOnly = TRUE,
                        cell = cell)
                    names(expr) <- names(geneIdMap[order(geneIdMap)])
                    expr <- t(t(expr))
                    colnames(expr) <- d$customdata
                }
            }
            write.csv(expr, file)
        }
    )
}

#' @importFrom plotly event_data
plotly3d_click <- function(session) {
    renderPrint({
        d <- click_event_data(session = session)
        if (is.null(d))
            "Click cell appear here"
        else {
            paste("Clicked cell", d$customdata)
        }
    })
}
plotly_selected <- function(session){
    renderPrint({
        d <- lasso_select_data(session = session)
        if (is.null(d))
            "Selected cell appear here"
        else {
            paste("Selected cell", d$customdata)
        }
    })
}

subsetPlotsUI <- function(id) {
    ns <- NS(id)
    tabPanel(
        value = id,
        HTML("Explorer"),
        tabsubTitleUI(
            id,
            'Plots',
            description = paste(
                "In this tab, users can visualise gene expression and",
                "cell information by adding new module."
            )
        ),
        fluidRow(
            column(
                3,
                dimensionReductionUI(id)),
            column(
                5,
                subsetCellByInfoUI(id, ABcolumn=.globals$subsetgroup[1]),
                subsetCellByInfoUI(id, ABcolumn=.globals$subsetgroup[2])
                ),
            column(
                4,
                graphicsControlUI(id),
                subsetCellByFilterUI(id))
        ),
        fluidRow(
            column(
                2,
                selectInput(
                    ns('moduleName'),
                    "Module type",
                    choices = c(
                        'cell info',
                        'gene expression',
                        'peak expression',
                        'proportion',
                        'violin/box plot',
                        'co-expression',
                        'co-expression 3d'
                    ),
                    selected = 'cell info'
                )
            ),
            column(
                2,
                selectInput(
                    ns('moduleWidth'),
                    "Module width",
                    choices = c('half', 'full'),
                    selected = 'half'
                )
            ),
            column(
                2,
                selectInput(
                    ns('interactive'),
                    label = "Interactive",
                    choices = c('No', 'Yes'),
                    selected = 'No'
                )
            ),
            column(
                6,
                div(
                    actionButton(
                        ns("newModule"),
                        "New Module",
                        icon = icon('plus'),
                        class = "align-action-button"
                    ),
                    actionButton(
                        ns("destroyModule"),
                        "Remove All",
                        icon = icon('minus'),
                        class = "align-action-button"
                    ),
                    div(
                        style = 'display: none;',
                        textInput(
                            ns("removePlotModule"),
                            '',
                            value = '',
                            width = 0
                        )),
                    div(
                        style = 'display: none;',
                        textInput(
                            ns("moveupPlotModule"),
                            '',
                            value = '',
                            width = 0
                        )),
                    div(
                        style = 'display: none;',
                        textInput(
                            ns("movedownPlotModule"),
                            '',
                            value = '',
                            width = 0
                        )),
                    div(
                        style = 'display: none;',
                        textInput(
                            ns("resizePlotModule"),
                            '',
                            value = '',
                            width = 0
                        )),
                    div(
                        style = 'display: none;',
                        textInput(
                            ns("changeSubsetContext"),
                            '',
                            value = '',
                            width = 0
                        ))
                ))
        ),
        fluidRow(div(
            id = ns('container'),
            htmlOutput(ns(
                'subplot'
            ))))
    )
}
subsetPlotsServer <- function(id, dataSource, optCrt) {
    moduleServer(id, function(input, output, session) {
        ## subtitle
        output$PlotsSubTitle <-
            renderUI({
                h4(paste(
                    "Explorer",
                    dataSource()$terms['expression'],
                    "on dimension reduction"
                ))
            })
        ## input column 1
        ### Dimension Reduction
        updateDimRedSelInputPair(session, input, dataSource)
        ### Information to show
        
        ## input column 2
        updateSubsetCellUI(id, input, output, session, dataSource,
                           ABcolumns=.globals$subsetgroup)
        
        ## input column 3
        updateFilterCellUI(id, optCrt, input, output, session, dataSource)
        
        ## plot region
        globals <- reactiveValues(
            containerIds = c(),
            containerInteractive = list(),
            containerUIs = list(),
            containerUIsFUN = list(),
            containerWidth = list(),
            containerServers = list(),
            containerSubsetGrp = c(),
            Idpointer = 0
        )
        if (is.null(session$userData$defaults))
            session$userData$defaults <- list()
        if (is.null(session$userData$defaults[[dataSource()$dataset]]))
            session$userData$defaults[[dataSource()$dataset]] <- list()
        uiGrid <- function(ns0, subgrp, uiFUN, width) {
            column(width = width, uiFUN(NS(id, ns0), subgrp=subgrp))
        }
        getRown <- function(w) {
            base <- 0
            currentLine <- 1
            l <- rep(0, length(w))
            for (i in seq_along(w)) {
                if (w[i] == 12) {
                    if (i == 1) {
                        l[i] <- 1
                    } else{
                        l[i] <- l[i - 1] + 1
                    }
                    currentLine <- l[i] + 1
                    base <- 0
                } else{
                    base <- base + w[i]
                    l[i] <- currentLine
                    if (base == 12) {
                        currentLine <- currentLine + 1
                        base <- 0
                    }
                }
            }
            return(l)
        }
        updatePlotModules <- function() {
            coln <- 
                ifelse(
                    globals$containerWidth[globals$containerIds] == "half",
                    6, 12)
            uis <- mapply(
                globals$containerIds,
                globals$containerSubsetGrp[globals$containerIds],
                globals$containerUIsFUN[globals$containerIds],
                coln,
                FUN = uiGrid,
                SIMPLIFY = FALSE)
            names(uis) <- globals$containerIds
            globals$containerUIs <- lapply(split(uis, getRown(coln)), fluidRow)
            output$subplot <- renderUI(globals$containerUIs)
            lapply(globals$containerIds, function(cid) {
                globals$containerServers[[cid]](
                    id,
                    cid,
                    dataSource,
                    optCrt,
                    input,
                    session,
                    globals$containerInteractive[[cid]]
                )
            })
        }
        observeEvent(input$newModule, {
            if(input$moduleName=='peak expression'){
                if(!file.exists(file.path(
                    .globals$datafolder,
                    dataSource()$dataset,
                    .globals$filenames$sc1atac))){
                    showNotification(
                        "Peak expression data not available",
                        duration = 5,
                        closeButton = TRUE,
                        type = "error"
                    )
                    return()
                }
            }
                
            if (length(globals$containerIds) >= 8) {
                showNotification(
                    "Reached the limitation.",
                    duration = 5,
                    closeButton = TRUE,
                    type = "warning"
                )
            } else{
                globals$Idpointer <- globals$Idpointer + 1
                ns0 <- paste0('plot_', globals$Idpointer)
                globals$containerIds <- c(globals$containerIds, ns0)
                globals$containerInteractive[[ns0]] <-
                    isolate(input$interactive == 'Yes')
                globals$containerWidth[[ns0]] <- isolate(input$moduleWidth)
                globals$containerSubsetGrp[ns0] <- .globals$subsetgroup[1]
                globals$containerUIsFUN[[ns0]] <- switch(
                    input$moduleName,
                    "cell info" = scInfoUI,
                    'gene expression' = scExprUI,
                    'proportion' = scPropUI,
                    'violin/box plot' = scVlnUI,
                    'co-expression' = scCoexpUI,
                    'co-expression 3d' = scCoexp3dUI,
                    'peak expression' = scAccUI,
                    scInfoUI
                )
                globals$containerServers[[ns0]] <- switch(
                    input$moduleName,
                    "cell info" = scInfoServer,
                    'gene expression' = scExprServer,
                    'proportion' = scPropServer,
                    'violin/box plot' = scVlnServer,
                    'co-expression' = scCoexpServer,
                    'co-expression 3d' = scCoexp3dServer,
                    'peak expression' = scAccServer,
                    scInfoServer
                )
                updatePlotModules()
            }
        })
        observeEvent(input$destroyModule, {
            lapply(globals$containerUIs, function(ui) {
                removeUI(ui, immediate = TRUE, session = session)
            })
            globals$containerIds <- c()
            globals$containerInteractive <- list()
            globals$containerUIsFUN <- list()
            globals$containerUIs <- list()
            globals$containerWidth <- list()
            globals$containerServers <- list()
            updatePlotModules()
        })
        observeEvent(input$removePlotModule, {
            if (input$removePlotModule != "") {
                globals$containerIds <-
                    globals$containerIds[
                        globals$containerIds !=
                            input$removePlotModule]
                globals$containerInteractive[[input$removePlotModule]] <-
                    NULL
                globals$containerWidth[[input$removePlotModule]] <- NULL
                globals$containerUIsFUN[[input$removePlotModule]] <- NULL
                globals$containerUIs[[input$removePlotModule]] <- NULL
                updatePlotModules()
            }
        })
        observeEvent(input$movedownPlotModule, {
            if (input$movedownPlotModule != "") {
                id <- which(globals$containerIds == input$movedownPlotModule)
                if (length(id) == 1) {
                    if (id != length(globals$containerIds)) {
                        tmp <- globals$containerIds[id + 1]
                        globals$containerIds[id + 1] <-
                            globals$containerIds[id]
                        globals$containerIds[id] <- tmp
                    }
                    updatePlotModules()
                }
            }
        })
        observeEvent(input$moveupPlotModule, {
            if (input$moveupPlotModule != "") {
                id <- which(globals$containerIds == input$moveupPlotModule)
                if (length(id) == 1) {
                    if (id != 1) {
                        tmp <- globals$containerIds[id - 1]
                        globals$containerIds[id - 1] <-
                            globals$containerIds[id]
                        globals$containerIds[id] <- tmp
                    }
                    updatePlotModules()
                }
            }
        })
        observeEvent(input$resizePlotModule, {
            if (input$resizePlotModule != "") {
                globals$containerWidth[[input$resizePlotModule]] <- 
                    ifelse(
                        globals$containerWidth[[input$resizePlotModule]] ==
                            "half",
                        "full",
                        "half")
                updatePlotModules()
            }
        })
        observeEvent(input$changeSubsetContext, {
            if (input$changeSubsetContext != "") {
                globals$containerSubsetGrp[[
                    sub('___.*$', '', input$changeSubsetContext)
                ]] <- sub('^.*___', '', input$changeSubsetContext)
            }
        })
    })
}

#' @importFrom data.table .N
scDRcoexNum <- function(
        inpConf,
        inpMeta,
        gene1,
        gene2,
        subsetCellKey,
        subsetCellVal,
        dataset,
        geneIdMap) {
    # Prepare ggData
    ggData <-
        inpMeta[, c(inpConf[inpConf$UI %in% subsetCellKey]$ID), with = FALSE]
    ggData <- getCoexpVal(ggData, dataset, geneIdMap, gene1, gene2)
    
    keep <- filterCells(
        ggData,
        subsetCellKey,
        subsetCellVal,
        inpConf=inpConf)
    ggData <- ggData[keep]
    
    # Actual data.table
    ggData$express <- "none"
    ggData[ggData$val1 > 0]$express <- gene1
    ggData[ggData$val2 > 0]$express <- gene2
    ggData[ggData$val1 > 0 & ggData$val2 > 0]$express <- "both"
    ggData$express <- factor(
        ggData$express,
        levels = unique(c("both", gene1, gene2, "none")))
    ggData <- ggData[, list(nCells = .N), by = "express"]
    ggData$percent <- 100 * ggData$nCells / sum(ggData$nCells)
    ggData <- ggData[order(ggData$express)]
    colnames(ggData)[1] <- "expression > 0"
    return(ggData)
}

# Plot gene expression scatter_pie
#' @importFrom data.table rbindlist
#' @importFrom scales rescale
#' @importFrom ggforce geom_arc_bar geom_circle geom_mark_hull
#' @importFrom ggplot2 geom_rect .data
scPieDim <- function(
        inpConf,
        inpMeta,
        dataset,
        geneIdMap,
        dimRedX,
        dimRedY,
        genelist,
        subsetCellKey,
        subsetCellVal,
        valueFilterKey,
        valueFilterCutoff,
        valueFilterCutoff2,
        CoExpred,
        pointSize,
        lableCircle,
        plotCellBg,
        markGrp,
        alpha,
        gradientCol="White-Red",
        plotType,
        labelsFontsize = 24,
        labelsFontFamily = 'Helvetica',
        plotAspectRatio,
        keepXYlables) {
    subFilterColname <- 'subValue'
    subGrpColname <- 'sub'
    geneList <- scGeneList(genelist, geneIdMap)
    geneList <- geneList[geneList$present == TRUE]
    shiny::validate(need(
        nrow(geneList) <= 10,
        "More than 10 genes to plot! Please reduce the gene list!"
    ))
    shiny::validate(need(
        nrow(geneList) > 1, 
        "Please input at least 2 genes to plot!"))
    
    # Prepare ggData
    ggData <- inpMeta[, c(
        "sampleID",
        inpConf[inpConf$UI == dimRedX]$ID,
        inpConf[inpConf$UI == dimRedY]$ID,
        inpConf[inpConf$UI %in% subsetCellKey]$ID),
        with = FALSE]
    if (nrow(ggData) == 0)
        return(NULL)
    if(ncol(ggData)<4){
        return(NULL)
    }
    cnid <- if(ncol(ggData)>3) 4 else 0
    colnames(ggData)[seq.int(3)] <- c("sampleID", "X", "Y")
    ggData <-
        cbindFilterValues(
            ggData,
            inpConf,
            inpMeta,
            subFilterColname,
            geneIdMap,
            dataset,
            valueFilterKey,
            valueFilterCutoff,
            valueFilterCutoff2
        )
    rat <- getRatio(ggData)
    expr <-
        lapply(geneIdMap[geneList$gene][
            seq.int(min(c(10, length(geneList$gene))))],
            read_exprs,
            h5f = dataset,
            valueOnly = TRUE)
    expr <- do.call(cbind, expr)
    expr[expr < 0] <- 0
    expr_keep <- rowSums(expr) > 0
    if (CoExpred)
        expr_keep <- expr_keep & rowSums(expr > 0) == ncol(expr)
    keep <- filterCells(
        ggData,
        subsetCellKey,
        subsetCellVal,
        subFilterColname,
        valueFilterCutoff,
        valueFilterCutoff2,
        inpConf=inpConf)
    
    if(cnid>3) colnames(ggData)[cnid] <- subGrpColname
    
    if (sum(keep & expr_keep) * nrow(geneList) > 5000) {
        ## filter the expression event, otherwise too slow
        expr_keep <- rowSums(expr)
        expr_keep <-
            expr_keep >=
            sort(expr_keep, decreasing = TRUE)[floor(1000 / nrow(geneList))]
    }
    ggData_ <- ggData[keep & expr_keep]
    expr <- expr[keep & expr_keep, , drop = FALSE]
    ggData <- ggData[keep]
    size <- diff(range(ggData$X)) / 60 * pointSize
    expr <- apply(expr, 2, rescale, to = c(size / 4, size))
    expr <- as.list(as.data.frame(expr))
    expr <- mapply(function(d, n) {
        cbind(ggData_, geneName = n, "val" = d)
    }, expr, names(expr), SIMPLIFY = FALSE)
    expr <- rbindlist(expr)
    
    expr$X0 <-
        expr$X + (as.numeric(factor(as.character(expr$geneName))) - 1) * size /
        2
    ggOut <- ggplot(data = expr, aes(x0 = .data[["X"]], y0 = .data[["Y"]]))
    if (markGrp) {
        ggOut <- ggOut +
            geom_mark_hull(
                data = ggData,
                aes(
                    x = .data[["X"]],
                    y = .data[["Y"]],
                    fill = .data[[subGrpColname]],
                    label = .data[[subGrpColname]]
                ),
                inherit.aes = FALSE,
                show.legend = FALSE
            )
    }
    if (plotCellBg) {
        ggOut <- ggOut +
            geom_point(
                data = ggData,
                aes(x = .data$X, y = .data$Y),
                color = 'snow2',
                shape = 16,
                inherit.aes = FALSE
            )
    }
    ggOut <- ggOut +
        switch(
            plotType,
            sunburst =
                geom_arc_bar(
                    aes(
                        amount = 1,
                        r0 = 0,
                        r = .data$val,
                        fill = .data$geneName
                    ),
                    stat = 'pie',
                    color = NA,
                    alpha = alpha
                ),
            pie =
                geom_arc_bar(
                    aes(
                        amount = .data$val,
                        r0 = 0,
                        r = size,
                        fill = .data$geneName
                    ),
                    stat = 'pie',
                    color = NA,
                    alpha = alpha
                ),
            donut =
                geom_arc_bar(
                    aes(
                        amount = .data$val,
                        r0 = size / 2,
                        r = size,
                        fill = .data$geneName
                    ),
                    stat = 'pie',
                    color = NA,
                    alpha = alpha
                ),
            bar =
                geom_rect(
                    aes(
                        xmin = .data$X0 - .02 * size,
                        ymin = .data$Y,
                        xmax = .data$X0 + .46 * size,
                        ymax = .data$Y + .data$val,
                        fill = .data$geneName
                    ),
                    position = "identity",
                    alpha = alpha,
                    color = NA
                ),
            sum = {
                dt_sum <- expr[, {list(val_sum=sum(.SD$val, na.rm = TRUE))},
                               by = c("sampleID", "X", "Y")]
                geom_point(
                    aes(
                        x = .data$X, y = .data$Y,
                        color = .data$val_sum
                    ),
                    data = dt_sum,
                    size = pointSize,
                    shape = 16,
                    alpha = alpha,
                    inherit.aes = FALSE
                )
            },
            max = {
                dt_max <- expr[, {list(val_max=max(.SD$val, na.rm = TRUE))},
                               by = c("sampleID", "X", "Y")]
                geom_point(
                    aes(
                        x = .data$X, y = .data$Y,
                        color = .data$val_max
                    ),
                    data = dt_max,
                    size = pointSize,
                    shape = 16,
                    alpha = alpha,
                    inherit.aes = FALSE
                )
            },
            mean = {
                dt_mean <- expr[, {list(val_mean=mean(.SD$val, na.rm = TRUE))},
                               by = c("sampleID", "X", "Y")]
                geom_point(
                    aes(
                        x = .data$X, y = .data$Y,
                        color = .data$val_mean
                    ),
                    data = dt_mean,
                    size = pointSize,
                    shape = 16,
                    alpha = alpha,
                    inherit.aes = FALSE
                )
            }
        )
    
    if(plotType %in% c("sum", "max", "mean")){
        ggOut <- ggOut + scale_color_gradientn(
            colours = availableThemes(gradientCol))
    }else{
        ggOut <- ggOut +
            guides(
                fill = guide_legend(title = 'genename'),
                alpha = "none")
        if (lableCircle) {
            if (plotType != "bar") {
                ggOut <- ggOut +
                    geom_circle(
                        aes(color = .data$sub, r = size),
                        fill = NA,
                        lwd = .5) +
                    guides(color = guide_legend(title = subsetCellKey))
            } else{
                ggOut <- ggOut +
                    geom_rect(
                        aes(
                            xmin = .data$X - .025 * size,
                            ymin = .data$Y - .05,
                            xmax = .data$X + (nrow(geneList) + .05) * size / 2,
                            ymax = .data$Y + size * 1.05,
                            color = .data$sub
                        ),
                        fill = NA,
                        linewidth = .5
                    ) +
                    guides(color = guide_legend(title = subsetCellKey))
            }
        }
    }
    ggOut <- ggOut +
        xlab(dimRedX) + ylab(dimRedY) +
        sctheme(
            base_size = labelsFontsize,
            family = labelsFontFamily, XYval = keepXYlables)
    ggOut <- fixCoord(ggOut, plotAspectRatio, rat)
    return(ggOut)
}

# Plot gene expression bubbleplot / heatmap
#' @importFrom stats as.dendrogram dist hclust quantile as.formula
#' @importFrom ComplexHeatmap merge_dendrogram Heatmap draw rowAnnotation
#' HeatmapAnnotation Legend pindex
#' @importFrom circlize colorRamp2
#' @importFrom grid gpar grid.circle unit.c unit grid.xaxis grid.yaxis 
#'  seekViewport current.viewport grid.lines
#' @importFrom rhdf5 h5read
#' @importFrom data.table rbindlist dcast.data.table data.table := uniqueN
#' @importFrom ggdendro dendro_data
#' @importFrom gridExtra arrangeGrob grid.arrange
#' @importFrom ggplot2 .data
scBubbHeat <- function(
        inpConf,
        inpMeta,
        inp,
        inpGrp,
        grpKey,
        grpVal,
        inpGrp1c,
        inpPlt,
        dataset,
        geneIdMap,
        inpScl,
        inpRow,
        row_dend_side = "left",
        row_dend_width = unit(10, "mm"),
        inpCol,
        column_dend_side = "top",
        column_dend_height = unit(10, "mm"),
        legend_side = 'bottom',
        inpcols,
        pointSize,
        labelsFontsize = 24,
        labelsFontFamily = 'Helvetica',
        flipXY,
        plotAllCells = FALSE,
        save = FALSE,
        colorBreaks,
        legendTitle = "expression",
        returnColorRange = FALSE,
        reorder=FALSE,
        orderX,
        splitBy,
        sreorder=FALSE,
        orderS,
        addnoise=TRUE) {
    # Identify genes that are in our dataset
    if (missing(inpGrp1c))
        inpGrp1c <- 0
    if (inpPlt != "Heatmap") {
        plotAllCells <- FALSE
    }
    geneList <- scGeneList(inp, geneIdMap)
    geneList <- geneList[geneList$present == TRUE]
    shiny::validate(need(
        nrow(geneList) <= .globals$maxHeatmapGene,
        paste("More than", .globals$maxHeatmapGene,
              "genes to plot! Please reduce the gene list!")
    ))
    if(missing(splitBy)){
        splitBy <- ""
    }else{
        if(is.na(splitBy)||splitBy=='N/A'){
            splitBy <- ""
        }
    }
    if(plotAllCells){
        shiny::validate(
            need(splitBy=="",
                 'Plotting all cells data does not support splitBy.')
        )
    }
    if(nrow(geneList)==1){
        shiny::validate(need(
            splitBy!="",
            "single gene ask setting of split by!"
        ))
        singleGene <- TRUE
    }else{
        singleGene <- FALSE
    }
    shiny::validate(need(
        nrow(geneList) > 0,
        "Please input at least 1 genes to plot!"))
    if(nrow(geneList)>.globals$maxNumGene){
        showNotification(
            paste('Plotting expression data for too many genes.',
                  'It will take a while. Please be patient.'),
            duration = 5,
            type = 'message'
        )
    }
    
    # Prepare ggData
    ggData <- data.table()
    vals <- h5read(file.path(
        .globals$datafolder,
        dataset,
        .globals$filenames$sc1gexpr
    ), .globals$h5fGrp, index=list(geneIdMap[geneList$gene], NULL))
    rownames(vals) <- geneList$gene
    tmp <- inpMeta[, c(
        "sampleID",
        inpConf[inpConf$grp == TRUE]$ID),
        with = FALSE]
    tmp$grpBy <- inpMeta[[inpConf[inpConf$UI == inpGrp]$ID]]
    # Load splitBy
    if(splitBy!=""){
        if(splitBy %in% inpConf$UI){
            tmp$splitBy <- inpMeta[[inpConf[inpConf$UI == splitBy]$ID]]
            if(sreorder){
                tmp$splitBy <- factor(as.character(tmp$splitBy),
                                      levels=orderS)
            }
        }
    }
    for (iGene in geneList$gene) {
        tmp$geneName <- iGene
        tmp$val <- vals[iGene, ]
        ggData <- rbindlist(list(ggData, tmp))
    }
    ggData <- subGrp(ggData, grpKey, grpVal, inpConf)
    
    if (inpPlt == "Violin") {
        ggData$geneName <- factor(as.character(ggData$geneName), levels = geneList$gene)
        if(reorder){
            ggData$grpBy <- factor(ggData$grpBy, levels=orderX)
        }
        showLegend <- FALSE
        if(addnoise){
            tmpNoise <-
                rnorm(length(ggData$val)) * diff(range(ggData$val)) / 1000
            ggData$val <- ggData$val + tmpNoise
        }
        if('splitBy' %in% colnames(ggData)){
            showLegend <- TRUE
            ggCol <- relevelCol(inpConf, splitBy, ggData, "splitBy")
            if(length(unique(ggData$splitBy))==2){
                ggOut <- ggplot(ggData, aes(
                    .data[["grpBy"]], .data[["val"]], fill = .data[["splitBy"]])) +
                    geom_split_violin(scale = "width") +
                    facet_grid(rows=as.formula("geneName ~ ."), scales = "free")
            }else{
                ggOut <- ggplot(ggData, aes(
                    .data[["grpBy"]], .data[["val"]], fill = .data[["splitBy"]])) +
                    geom_violin(scale = "width") +
                    facet_grid(rows=as.formula("geneName ~ ."), scales = "free")
            }
        }else{
            ggCol <- relevelCol(inpConf, inpGrp, ggData, "grpBy")
            ggOut <- ggplot(ggData, aes(
                .data[["grpBy"]], .data[["val"]], fill = .data[["grpBy"]])) +
                geom_violin(scale = "width") +
                facet_grid(rows=as.formula("geneName ~ ."), scales = "free")
        }
        
        if(pointSize>0) ggOut <- ggOut +
            geom_jitter(size = pointSize, shape = 16)
        ggOut <- ggOut + 
            xlab(inpGrp) + ylab("expression") +
            sctheme(
                base_size = labelsFontsize,
                family = labelsFontFamily,
                Xang = 45,
                XjusH = 1) +
            scale_fill_manual("", values = ggCol) +
            theme(legend.position = ifelse(
                showLegend, "bottom", "none"))
        return(plot(ggOut))
    }
    
    # Aggregate
    ggData$val <- expm1(ggData$val)
    ggData$val[is.infinite(ggData$val)] <-
        max(ggData$val[!is.infinite(ggData$val)], na.rm = TRUE)
    cNames <- c("geneName", "grpBy", "splitBy")[
        c("geneName", "grpBy", "splitBy") %in% 
            colnames(ggData)
    ]
    if (!plotAllCells) {
        ggData <- ggData[, list(
            val = mean(.SD$val[.SD$val >= inpGrp1c]),
            prop = sum(.SD$val > 0) / length(.SD$sampleID)
        ),
        by = cNames]
        ggDataAvg <- NULL
    } else{
        ggDataAvg <- ggData[, list(
            val = mean(.SD$val[.SD$val >= inpGrp1c]),
            prop = sum(.SD$val > 0) / length(.SD$sampleID)
        ),
        by = cNames]
        ggDataAvg$val <- log1p(ggDataAvg$val)
    }
    ggData$val <- log1p(ggData$val)
    
    # Scale if required
    colRange <- range(ggData$val, na.rm = TRUE)
    colRange1 <- quantile(
        ggData$val,
        probs = c(0, .01, .5, .99, 1),
        na.rm = TRUE,
        names = FALSE
    )
    if (inpScl) {
        ggData[, "val" := if (length(unique(.SD$val)) == 1) {
            0
        } else {
            scale(.SD$val)
        }, keyby = "geneName"]
        colRange <- range(ggData$val, na.rm = TRUE)
        if (colRange[1] < 0) {
            colRange <- c(-max(abs(
                range(ggData$val, na.rm = TRUE)
            )),
            max(abs(
                range(ggData$val, na.rm = TRUE)
            )))
        } else{
            colRange <- c(0, max(abs(
                range(ggData$val, na.rm = TRUE)
            )))
        }
        colRange1 <- quantile(
            ggData$val,
            probs = c(0, .01, .5, .99, 1),
            na.rm = TRUE,
            names = FALSE
        )
        colRange1 <- c(colRange[1], colRange1[-c(1, 5)], colRange[2])
    }
    if (returnColorRange) {
        colRange1 <- colRange1[c(-3)]
        return(colRange1)
    }
    
    if (!is.na(colorBreaks[1])) {
        if (colorBreaks[1] < colRange[1])
            colorBreaks[1] <- colRange[1]
        if (colorBreaks[2] > colRange[2])
            colorBreaks[2] <- colRange[2]
        ggData$val[ggData$val < colorBreaks[1]] <- colorBreaks[1]
        ggData$val[ggData$val > colorBreaks[2]] <- colorBreaks[2]
        if (colorBreaks[1] == colorBreaks[2]) {
            col_fun <- availableThemes(inpcols)[1]
        } else{
            col_fun <- colorRamp2(
                breaks = seq(
                    colorBreaks[1],
                    colorBreaks[2],
                    length.out = length(availableThemes(inpcols))
                ),
                colors = availableThemes(inpcols)
            )
        }
    } else{
        if (colRange[1] == colRange[2]) {
            col_fun <- availableThemes(inpcols)[1]
        } else{
            col_fun <- colorRamp2(
                breaks = seq(
                    colRange[1],
                    colRange[2],
                    length.out = length(availableThemes(inpcols))
                ),
                colors = availableThemes(inpcols)
            )
        }
    }
    # reshape the data to matrix
    if (plotAllCells) {
        ggData$grpBy <- paste(ggData$grpBy, ggData$sampleID, sep = "__")
    }else{
        if(reorder){
            ggData$grpBy <- factor(as.character(ggData$grpBy), levels=orderX)
        }
    }
    if(sreorder){
        ggData$splitBy <- factor(as.character(ggData$splitBy), levels =orderS)
    }
    
    reshapeMat <- function(value.var) {
        if(singleGene){
            ggMatrix <- dcast.data.table(
                ggData,
                as.formula("splitBy ~ grpBy"),
                value.var = value.var,
                fun.aggregate = mean
            )
            tmp <- ggMatrix$splitBy
        }else{
            ggMatrix <- dcast.data.table(
                ggData,
                as.formula(ifelse('splitBy' %in% colnames(ggData),
                                  "geneName~grpBy+splitBy", "geneName~grpBy")),
                value.var = value.var,
                fun.aggregate = mean,
                sep = '___')
            tmp <- ggMatrix$geneName
        }
        ggMatrix <- as.matrix(ggMatrix[, -1])
        ggMatrix[is.na(ggMatrix)] <- 0
        ggMatrix[is.infinite(ggMatrix)] <-
            .Machine$integer.max * sign(ggMatrix[is.infinite(ggMatrix)])
        rownames(ggMatrix) <- tmp
        return(ggMatrix)
    }
    ggMat <- reshapeMat(value.var = "val")
    if(!plotAllCells) ggProp <- reshapeMat(value.var = 'prop')
    
    if(reorder&&length(orderX)>0&&all(orderX %in% colnames(ggMat))){
        if(singleGene){
            ggMat[, orderX, drop=FALSE]
            if(!plotAllCells) ggProp <- ggProp[, orderX, drop=FALSE]
        }else{
            ggMat <- ggMat[geneList$gene[geneList$gene %in% rownames(ggMat)],
                           orderX, drop=FALSE]
            if(!plotAllCells) ggProp <- 
                    ggProp[geneList$gene[geneList$gene %in% rownames(ggProp)],
                           orderX, drop=FALSE]
        }
    }else{
        if(singleGene){
            ggMat[, colnames(ggMat), drop=FALSE]
            if(!plotAllCells) ggProp <- ggProp[, colnames(ggProp), drop=FALSE]
        }else{
            ggMatSplit <- sub('^.*___', '', colnames(ggMat))
            ggMat <- ggMat[geneList$gene[geneList$gene %in% rownames(ggMat)],
                           order(ggMatSplit), drop=FALSE]
            if(!plotAllCells) ggProp <- 
                ggProp[geneList$gene[geneList$gene %in% rownames(ggProp)],
                       order(ggMatSplit), drop=FALSE]
        }
    }
    
    ggMatSplit <- sub('^.*___', '', colnames(ggMat))
    ggMatGrp <- sub('___.*$', '', colnames(ggMat))
    colnames(ggMat) <- sub('___', '_', colnames(ggMat))
    if(!plotAllCells) colnames(ggProp) <- sub('___', '_', colnames(ggProp))
    cluster_rows <- inpRow
    if (inpRow) {
        cluster_rows <- as.dendrogram(hclust(dist(ggMat)))
    }
    cluster_columns <- inpCol
    if (inpCol) {
        cluster_columns <- as.dendrogram(hclust(dist(t(ggMat))))
    }
    
    layer_fun <- NULL
    rect_gp <- gpar(col = NA)
    font_gp <- gpar(fontsize = labelsFontsize, fontfamily = labelsFontFamily)
    title_position <- ifelse(legend_side=='bottom',
                             "lefttop","topleft")
    legend_direction <- ifelse(legend_side=='bottom',
                               "horizontal","vertical")
    if (inpPlt == "Bubbleplot") {
        if(pointSize>0){
            dotsize <- unit(pointSize*3.2, 'mm')
        }else{
            dotsize <- unit(4, "mm")
        }
        layer_fun <- function(j, i, x, y, w, h, fill){
            grid.circle(x=x,y=y,
                        r= sqrt(pindex(ggProp, i, j)) * dotsize,
                        gp = gpar(fill = col_fun(pindex(ggMat, i, j)),
                                  col = NA))
        }
        lgd_list = list(
            Legend( labels = c(0, 10, 25, 50, 75), title = "percentage",
                    title_position = title_position,
                    graphics = list(
                        function(x, y, w, h) {
                            grid.circle(x = x, y = y, r = 0  * dotsize,
                                        gp = gpar(fill = "black"))},
                        function(x, y, w, h) {
                            grid.circle(x = x, y = y, r = sqrt(0.1)  * dotsize,
                                        gp = gpar(fill = "black"))},
                        function(x, y, w, h) {
                            grid.circle(x = x, y = y, r = sqrt(0.25) * dotsize,
                                        gp = gpar(fill = "black"))},
                        function(x, y, w, h) {
                            grid.circle(x = x, y = y, r = sqrt(0.5) * dotsize,
                                        gp = gpar(fill = "black"))},
                        function(x, y, w, h) {
                            grid.circle(x = x, y = y, r = sqrt(0.75) * dotsize,
                                        gp = gpar(fill = "black"))}
                    ),
                    grid_width = 2*dotsize,
                    grid_height = 2*dotsize,
                    ncol = ifelse(legend_side=='bottom',
                                  5, 1),
                    direction = legend_direction
            ))
        rect_gp <- gpar(type = 'none')
        ## add gap for rownames and colnames
        if(flipXY){
            if(row_dend_side == 'bottom'){
                rownames(ggMat) <- rownames(ggProp) <-
                    paste0(" ", rownames(ggMat))
            }else{
                rownames(ggMat) <- rownames(ggProp) <-
                    paste0(rownames(ggMat), " ")
            }
            if(column_dend_side == "left"){
                colnames(ggMat) <- colnames(ggProp) <-
                    paste0(" ", colnames(ggMat))
            }else{
                colnames(ggMat) <- colnames(ggProp) <-
                    paste0(colnames(ggMat), " ")
            }
        }else{
            if(row_dend_side == 'left'){
                rownames(ggMat) <- rownames(ggProp) <-
                    paste0(" ", rownames(ggMat))
            }else{
                rownames(ggMat) <- rownames(ggProp) <-
                    paste0(rownames(ggMat), " ")
            }
            if(column_dend_side == 'bottom'){
                colnames(ggMat) <- colnames(ggProp) <-
                    paste0(" ", colnames(ggMat))
            }else{
                colnames(ggMat) <- colnames(ggProp) <-
                    paste0(colnames(ggMat), " ")
            }
        }
    }
    if (flipXY) {
        ggMat <- t(ggMat)
        if(!plotAllCells) ggProp <- t(ggProp)
        
        if (plotAllCells) {
            group <- ggData[[inpGrp]][match(rownames(ggMat), ggData$grpBy)]
            anno <- rowAnnotation(
                group = group,
                show_legend = FALSE,
                show_annotation_name = FALSE
            )
            ht_list <- Heatmap(
                ggMat,
                name = legendTitle,
                col = col_fun,
                heatmap_legend_param =
                    list(
                        title = legendTitle,
                        direction = legend_direction,
                        title_position = title_position
                    ),
                cluster_rows = TRUE,
                cluster_row_slices = inpCol,
                cluster_columns = cluster_rows,
                show_row_names = !plotAllCells,
                row_split = group,
                right_annotation = anno,
                row_title_rot = 0,
                column_names_rot = 45,
                layer_fun = layer_fun,
                rect_gp = rect_gp,
                use_raster = TRUE,
                row_title_gp = font_gp,
                column_title_gp = font_gp,
                row_names_gp = font_gp,
                column_names_gp = font_gp,
                row_dend_side = ifelse(column_dend_side == "right",
                                       "right", "left"),
                row_dend_width = column_dend_height,
                column_dend_side = ifelse(row_dend_side=="top",
                                          "top", "bottom"),
                column_dend_height = row_dend_width,
                row_names_side = ifelse(column_dend_side == "right",
                                        "left", "right"),
                row_title_side = ifelse(column_dend_side == "right",
                                        "left", "right"),
                column_title_side = ifelse(row_dend_side=="top",
                                           "bottom", "top")
            )
        } else{
            ht_list <- Heatmap(
                ggMat,
                name = legendTitle,
                col = col_fun,
                heatmap_legend_param =
                    list(
                        title = legendTitle,
                        direction = legend_direction,
                        title_position = title_position
                    ),
                cluster_rows = cluster_columns,
                cluster_columns = cluster_rows,
                show_row_names = !plotAllCells,
                column_names_rot = 45,
                layer_fun = layer_fun,
                rect_gp = rect_gp,
                row_title_gp = font_gp,
                column_title_gp = font_gp,
                row_names_gp = font_gp,
                column_names_gp = font_gp,
                row_dend_side = ifelse(column_dend_side == "right",
                                       "right", "left"),
                row_dend_width = column_dend_height,
                column_dend_side = ifelse(row_dend_side=="top",
                                          "top", "bottom"),
                column_dend_height = row_dend_width,
                row_names_side = ifelse(column_dend_side == "right",
                                        "left", "right"),
                column_names_side = ifelse(row_dend_side=="top",
                                           "bottom", "top")
            )
        }
    } else{
        if (plotAllCells) {
            group <- ggData[[inpGrp]][match(colnames(ggMat), ggData$grpBy)]
            anno <- HeatmapAnnotation(
                group = group,
                show_legend = FALSE,
                show_annotation_name = FALSE
            )
            ht_list <- Heatmap(
                ggMat,
                name = legendTitle,
                col = col_fun,
                heatmap_legend_param =
                    list(
                        title = legendTitle,
                        direction = legend_direction,
                        title_position = title_position
                    ),
                cluster_rows = cluster_rows,
                cluster_columns = TRUE,
                cluster_column_slices = inpCol,
                show_column_names = !plotAllCells,
                column_split = group,
                bottom_annotation = anno,
                layer_fun = layer_fun,
                rect_gp = rect_gp,
                use_raster = TRUE,
                row_title_gp = font_gp,
                column_title_gp = font_gp,
                row_names_gp = font_gp,
                column_names_gp = font_gp,
                row_dend_side = ifelse(row_dend_side == "right",
                                       "right", "left"),
                row_dend_width = row_dend_width,
                column_dend_side = ifelse(column_dend_side=="top",
                                          "top", "bottom"),,
                column_dend_height = column_dend_height,
                row_names_side = ifelse(row_dend_side == "right",
                                        "left", "right"),
                column_title_side = ifelse(column_dend_side=="top",
                                           "bottom", "top"),
                column_title_rot = 45
            )
        } else{
            ht_list <- Heatmap(
                ggMat,
                name = legendTitle,
                col = col_fun,
                heatmap_legend_param =
                    list(
                        title = legendTitle,
                        direction = legend_direction,
                        title_position = title_position
                    ),
                cluster_rows = cluster_rows,
                cluster_columns = cluster_columns,
                show_column_names = !plotAllCells,
                layer_fun = layer_fun,
                rect_gp = rect_gp,
                row_title_gp = font_gp,
                column_title_gp = font_gp,
                row_names_gp = font_gp,
                column_names_gp = font_gp,
                row_dend_side = ifelse(row_dend_side == "right",
                                       "right", "left"),
                row_dend_width = row_dend_width,
                column_dend_side = ifelse(column_dend_side=="top",
                                          "top", "bottom"),
                column_dend_height = column_dend_height,
                row_names_side = ifelse(row_dend_side == "right",
                                        "left", "right"),
                column_names_side = ifelse(column_dend_side=="top",
                                           "bottom", "top")
            )
        }
    }
    plot_axis <- function(object){
        current_vp <- current.viewport()$name
        if (current_vp == "ROOT") {
            current_vp <- "global"
        }
        vp_name <- paste0(legendTitle, "_heatmap_body_wrap")
        seekViewport(vp_name)
        xmain <- ifelse(flipXY,
                        row_dend_side=='top', 
                        column_dend_side == "top")
        ymain <- ifelse(flipXY,
                        column_dend_side == "right",
                        row_dend_side == "right")
        grid.lines(c(0, 1), as.numeric(!xmain))
        grid.xaxis(at = (seq.int(ncol(ggMat))-.5)/ncol(ggMat),
                   label = FALSE,
                   main = xmain)
        grid.lines(as.numeric(!ymain), c(0, 1))
        grid.yaxis(at = (seq.int(nrow(ggMat))-.5)/nrow(ggMat),
                   label = FALSE,
                   main = ymain)
        seekViewport(current_vp)
    }
    if (inpPlt == "Bubbleplot") {
        return(
            draw(
            ht_list,
            heatmap_legend_side = legend_side,
            annotation_legend_side = legend_side,
            annotation_legend_list = lgd_list,
            post_fun = plot_axis
            )
        )
    }else{
        return(
            draw(
            ht_list,
            heatmap_legend_side = legend_side,
            annotation_legend_side = legend_side
            )
        )
    }
}

#' The methods for \link{APPconf-class}
#' @description The assessment and replacement methods for \link{APPconf-class}
#' @name APPconf-methods
#' @family APPconf
#' @rdname APPconf-methods
#' @param object an object of APPconf
#' @importFrom methods show
#' @export
#' @aliases show,APPconf-method
#' @examples
#' appconf <- readRDS(system.file("extdata", "data",
#'     "pbmc_small", "appconf.rds", package="scRNAseqApp"))
#' appconf
setMethod("show", "APPconf", function(object) {
    cat("This is an object of APPconf \n")
    cat(
        "describe the data ",
        object@title,
        " saved in ",
        object@id,
        "\n")
    if (length(object@ref$bib)) {
        cat("Please cite the data via ", object@ref$bib, "\n")
    } else{
        if (length(object@ref$entry)) {
            cat("Please cite the data via ", object@ref$entry, "\n")
        }
    }
    cat(
        "First 5 markers for current data: ",
        head(markers(object), n = 5L),
        "\n")
    cat(
        "Keywords for current data: ",
        head(object@keywords, n = 20L),
        "\n")
})

#' @rdname APPconf-methods
#' @export
#' @param x APPconf object.
setMethod("$", "APPconf", function(x, name)
    slot(x, name))
#' @rdname APPconf-methods
#' @param name A literal character string or a name (possibly backtick quoted).
#' @param value value to replace.
#' @export
#' @examples
#' appconf$title
setReplaceMethod(
    "$",
    "APPconf",
    function(x, name, value) {
        slot(x, name, check = TRUE) <- value
        x
    })


#' @rdname APPconf-methods
#' @export
#' @param i,j indices specifying elements to extract or replace.
#' @param \dots Named or unnamed arguments to form a signature.
#' @param exact see \link[base]{Extract}
setMethod("[[", "APPconf", function(x, i, j, ..., exact = TRUE)
    slot(x, i))
#' @rdname APPconf-methods
#' @export
#' @examples
#' appconf[["title"]]
setReplaceMethod(
    "[[",
    "APPconf",
    function(x, i, ..., value) {
        slot(x, i, check = TRUE) <- value
        x
    })
#' @rdname APPconf-methods
#' @export
#' @param drop see \link[base]{drop}
#' @importFrom methods as
setMethod(
    "[",
    "APPconf",
    function(x, i, j, ..., drop = TRUE) {
        as(x, "list")[i]
    })
if (length(getGeneric("as.list")) == 0) {
    setGeneric("as.list", function(x, ...)
        standardGeneric("as.list"))
}
#' @rdname APPconf-methods
#' @export
#' @importFrom methods slotNames setAs
#' @examples
#' as.list(appconf)
setMethod(
    "as.list",
    "APPconf",
    function(x, ...) {
        ii <- slotNames(x)
        names(ii) <- ii
        lapply(ii, function(i)
            x[[i]])
    })
setAs(
    from = "APPconf",
    to = "list",
    function(from) {
        ii <- slotNames(from)
        names(ii) <- ii
        lapply(ii, function(i)
            from[[i]])
    })
if (length(getGeneric("as.character")) == 0) {
    setGeneric("as.character", function(x, ...)
        standardGeneric("as.character"))
}
#' @rdname APPconf-methods
#' @export
#' @importFrom methods slotNames setAs
#' @examples
#' as.character(appconf)
setMethod(
    "as.character",
    "APPconf",
    function(x, ...) {
        ii <- slotNames(x)
        c(markers(x), unlist(x[ii != 'markers']))
    })
setAs(
    from = "APPconf",
    to = "character",
    function(from) {
        ii <- slotNames(from)
        c(markers(from), unlist(from[ii != 'markers']))
    })

setGeneric("markers", function(x)
    standardGeneric("markers"))
#' @rdname APPconf-methods
#' @aliases markers
#' @export
#' @examples
#' markers(appconf)
setMethod("markers", "APPconf", function(x) {
    unique(unlist(lapply(x[["markers"]], rownames)))
})

if (length(getGeneric("lapply")) == 0) {
    setGeneric("lapply", function(X, FUN, ...)
        standardGeneric("lapply"))
}
#' @rdname APPconf-methods
#' @export
#' @param X an APPconf object.
#' @param FUN function used by `lapply`
#' @importFrom methods slotNames
#' @examples
#' lapply(appconf, print)
setMethod(
    "lapply",
    "APPconf",
    function(X, FUN, ...) {
        FUN <- match.fun(FUN)
        ii <- slotNames(X)
        names(ii) <- ii
        lapply(ii, function(i)
            FUN(X[[i]], ...))
    })
#' @rdname APPconf-methods
#' @export
#' @param recursive,use.names function used by \link[base:unlist]{unlist}
#' @return A named character vector.
#' @examples
#' unlist(appconf)
setMethod(
    "unlist",
    "APPconf",
    function(
        x,
        recursive = TRUE,
        use.names = TRUE) {
        unlist(as.list(x), recursive = recursive, use.names = use.names)
    })

#' @importFrom ggplot2 geom_text
scDRmonocle <- function(
        data,
        reduction_method = "UMAP",
        alpha = 1,
        cell_size = .35,
        cell_stroke = .1,
        trajectory_graph_segment_size = .75,
        trajectory_graph_color = 'grey28',
        graph_label_size = 2,
        font_size = 24,
        labelsFontFamily = 'Helvetica',
        show_trajectory_graph = TRUE,
        label_principal_points = TRUE,
        label_leaves = TRUE,
        label_roots = TRUE) {
    if (is.null(data)) {
        return(NULL)
    }
    data_df <- data$meta_data
    edge_df <- data$segments_layer_data
    princ_point_df <- data$principal_points_data
    leaf_df <- data$mst_leaf_nodes
    root_df <- data$mst_root_nodes
    ggOut <-
        ggplot(data = data_df, aes(
            x = .data[["data_dim_1"]], y = .data[["data_dim_2"]])) +
        geom_point(
            color = I("black"),
            size = 1.25 * cell_size,
            stroke = I(cell_stroke),
            na.rm = TRUE,
            alpha = I(alpha)
        ) +
        geom_point(
            aes(color = .data[["cell_color"]]),
            size = I(cell_size),
            stroke = I(cell_stroke),
            na.rm = TRUE,
            alpha = alpha
        ) +
        guides(color = guide_legend(
            title = "cluster",
            override.aes = list(size = 4)))
    if (show_trajectory_graph) {
        ggOut <- ggOut +
            geom_segment(
                aes(
                    x = .data[["source_prin_graph_dim_1"]],
                    y = .data[["source_prin_graph_dim_2"]],
                    xend = .data[["target_prin_graph_dim_1"]],
                    yend = .data[["target_prin_graph_dim_2"]]
                ),
                size = trajectory_graph_segment_size,
                color = I(trajectory_graph_color),
                linetype = "solid",
                na.rm = TRUE,
                data = edge_df
            )
    }
    if (label_principal_points) {
        ggOut <- ggOut +
            geom_point(
                aes(
                    x = .data[["prin_graph_dim_1"]],
                    y = .data[["prin_graph_dim_2"]]),
                shape = 21,
                stroke = I(trajectory_graph_segment_size),
                color = "white",
                fill = "black",
                size = I(graph_label_size * 1.5),
                na.rm = TRUE,
                princ_point_df
            )
    }
    if (label_leaves) {
        ggOut <- ggOut +
            geom_point(
                aes(
                    x = .data[["prin_graph_dim_1"]],
                    y = .data[["prin_graph_dim_2"]]),
                shape = 21,
                stroke = I(trajectory_graph_segment_size),
                color = "black",
                fill = "lightgray",
                size = I(graph_label_size * 1.5),
                na.rm = TRUE,
                leaf_df
            ) +
            geom_text(
                aes(
                    x = .data[["prin_graph_dim_1"]],
                    y = .data[["prin_graph_dim_2"]],
                    label = .data[["leaf_idx"]]
                ),
                size = I(graph_label_size),
                color = "black",
                na.rm = TRUE,
                leaf_df,
                inherit.aes = FALSE
            )
    }
    if (label_roots) {
        ggOut <- ggOut +
            geom_point(
                aes(
                    x = .data[["prin_graph_dim_1"]],
                    y = .data[["prin_graph_dim_2"]]),
                shape = 21,
                stroke = I(trajectory_graph_segment_size),
                color = "black",
                fill = "white",
                size = I(graph_label_size * 1.5),
                na.rm = TRUE,
                root_df
            ) +
            geom_text(
                aes(
                    x = .data[["prin_graph_dim_1"]],
                    y = .data[["prin_graph_dim_2"]],
                    label = .data[["root_idx"]]
                ),
                size = I(graph_label_size),
                color = "black",
                na.rm = TRUE,
                root_df
            )
    }
    ggOut <- ggOut +
        xlab(paste(reduction_method, "_1")) +
        ylab(paste(reduction_method, "_2")) +
        sctheme(base_size = font_size,
                family = labelsFontFamily)
    return(ggOut)
}

# global variables
.globals <- new.env(parent = emptyenv())
.globals$theme <- bs_theme(bootswatch = 'lumen')
# filenames
.globals$app_path <- "."
.globals$datafolder <- "data"
.globals$filenames <- list(
    appconf = "appconf.rds",
    sc1conf = "sc1conf.rds",
    sc1def = "sc1def.rds",
    sc1gene = "sc1gene.rds",
    sc1gexpr = "sc1gexpr.h5",
    sc1meta = "sc1meta.rds",
    sc1atac = "sc1atac.h5",
    sc1peak = 'sc1peak.rds',
    sc1link = 'sc1link.rds',
    sc1anno = 'sc1anno.rds',
    sc1edge = 'sc1edge.rds',
    coor_revmap = 'coorrevmap.rds',
    bwspath = 'bws',
    token = "token",
    locker = "LOCKER",
    cellchat = "cellchat.rds",
    monocle = "monocle3_pseudotime.rds",
    slingshot = "slingshot.rds"
)
.globals$tokenMinLen <- 8 # minimal token length
.globals$h5fGrpPrefix <- 'grp' #gene expression table prefix
.globals$h5fDataPrefix <- 'data'
.globals$h5fGrp <- paste(.globals$h5fGrpPrefix, .globals$h5fDataPrefix, sep="/") 
.globals$h5fATACcell <- 'cell' # ATAC cell level signals table prefix 
.globals$h5fATACcoor <- 'coor' # ATAC coor level signals table prefix
.globals$downloadFolder <- file.path("www", "download")
.globals$counterFilename <- file.path("www", "counter.tsv")
.globals$IPlocationFilename <- file.path("www", "iptable.rds")
.globals$credential_path <- file.path("www", "database.sqlite")
.globals$credentialTableName <- 'credentials'
.globals$passphrase <- NULL
.globals$configTableName <- 'config'
.globals$configTableSep <- ';'
.globals$counterTableName <- 'counter'
.globals$IPlocationTablename <- "iptable"
.globals$geneSymbolTableName <- 'gene'
.globals$gn2symTableName <- 'gn2sym'
.globals$commentsTableName <- 'issues'
.globals$totalComments <- 2500
.globals$commentsIntervals <- 10
.globals$searchTableName <- 'searchlog'
# for home search page, the groupColPattern
.globals$groupColPattern <- 'celltype'
# Colour palette
.globals$coExpColor <- c(
    "Red (Gene1); Blue (Gene2)",
    "Orange (Gene1); Blue (Gene2)",
    "Red (Gene1); Green (Gene2)",
    "Green (Gene1); Blue (Gene2)"
)
# Panel sizes
.globals$pList1 <- c(
    "Small" = "400px",
    "Medium" = "600px",
    "Large" = "800px")
.globals$pList2 <- c(
    "Small" = "500px",
    "Medium" = "700px",
    "Large" = "900px")
.globals$pList3 <- c(
    "Small" = "600px",
    "Medium" = "800px",
    "Large" = "1000px")
.globals$lList <- c(#ggrepel font size in scDRcell
    "Small" = 5,
    "Medium" = 6,
    "Large" = 7)
.globals$fList <- c(#monocle plot point size
    "Small" = .5,
    "Medium" = 1,
    "Large" = 2)

.globals$figWidth = 8.01
.globals$figHeight = 6.01
.globals$figFormats = c('PDF', 'PNG', 'TIFF', 'JPEG', 'BMP', 'CSV')

# supported organisms
.globals$supported_organisms <- c(
    "Bos taurus",
    "Caenorhabditis elegans",
    "Drosophila melanogaster",
    "Danio rerio",
    "Gallus gallus",
    "Homo sapiens",
    "Mus musculus",
    "Macaca mulatta",
    "Pan troglodytes",
    "Rattus norvegicus",
    "Saccharomyces cerevisiae",
    "Sus scrofa"
)
# tab terms
.globals$terms <- list(
    scRNAseq = c(
        GeneExpr = "GeneExpr",
        coexpression = "coexpression",
        expression = "expression"
    ),
    scATACseq = c(
        GeneExpr = "GeneScore",
        coexpression = "co-genescore",
        expression = "score"
    ),
    scMultiome = c(
        GeneExpr = "GeneExpr",
        coexpression = "coexpression",
        expression = "expression"
    )
)
# number of gene by search
.globals$maxHeatmapGene <- 100
.globals$maxNumGene <- 50
.globals$limitNumGene <- 3

# subset cell setting group
.globals$subsetgroup <- c('A', 'B')

geneExprGeneExprUI <- function(id) {
    tabPanel(
        value = id,
        htmlOutput(NS(id, 'GeneExpr')),
        tabsubTitleUI(
            id,
            'GeneExpr',
            description = paste(
                "In this tab, users can visualise two feature informations ",
                "side-by-side on low-dimensional representions."
            )
        ),
        fluidRow(
            column(3, dimensionReductionUI(id)),
            column(5, subsetCellByInfoUI(id)),
            column(4, graphicsControlUI(id))
        ),
        fluidRow(
            column(
                6,
                style = "border-right: 2px solid black",
                htmlOutput(NS0(id, "subPlotTitle", 1)),
                fluidRow(
                    column(6, geneExprUI(id, 1)),
                    column(6, geneExprPlotControlUI(id, 1))),
                geneExprDotPlotUI(id, 1)
            ),
            column(
                6,
                htmlOutput(NS0(id, "subPlotTitle", 2)),
                fluidRow(
                    column(6, geneExprUI(id, 2)),
                    column(6, geneExprPlotControlUI(id, 2))),
                geneExprDotPlotUI(id, 2)
            )
        )
    )
}
geneExprGeneExprServer <- function(id, dataSource, optCrt) {
    moduleServer(id, function(input, output, session) {
        ## title
        output$GeneExpr <-
            renderUI({
                HTML(paste(
                    dataSource()$terms["GeneExpr"],
                    "vs",
                    dataSource()$terms["GeneExpr"]))
            })
        ## subtitle
        output$GeneExprSubTitle <-
            renderUI({
                h4(
                    paste(
                        "Gene",
                        dataSource()$terms['expression'],
                        "vs gene",
                        dataSource()$terms['expression'],
                        "on dimension reduction"
                    )
                )
            })
        ## input column 1
        ### Dimension Reduction
        updateDimRedSelInputPair(session, input, dataSource)
        ## input column 2
        updateSubsetCellUI(id, input, output, session, dataSource)
        
        ## plot region
        ### sub region title
        output$subPlotTitle1 <-
            renderUI({
                h4(paste("Gene", dataSource()$terms['expression'], "1"))
            })
        output$subPlotTitle2 <-
            renderUI({
                h4(paste("Gene", dataSource()$terms['expression'], "2"))
            })
        ### gene expressions
        updateGeneExprPlot(
            postfix = 1,
            dataSource()$sc1def$gene1,
            optCrt,
            id,
            input,
            output,
            session,
            dataSource
        )
        updateGeneExprPlot(
            postfix = 2,
            dataSource()$sc1def$gene2,
            optCrt,
            id,
            input,
            output,
            session,
            dataSource
        )
    })
}

sortLevels <- function(
        lev,
        controls = c(
            "control", "wildtype",
            "wt", "cont", "contr",
            "vehicle", "sham")) {
    if (length(lev) < 2)
        return(lev)
    l0 <- lev[tolower(lev) %in% controls]
    l1 <- lev[!tolower(lev) %in% controls]
    if (length(l1) > 1) {
        ## find the same part of the levels
        l1.sub <- strsplit(l1, split = "")
        ml <- max(lengths(l1.sub))
        l1.sub <-
            lapply(l1.sub, function(.ele)
                c(.ele, rep("", ml))[seq.int(ml)])
        l1.sub <- do.call(rbind, l1.sub)
        l1.sub <- apply(l1.sub, 2, unique, simplify = FALSE)
        ml <- lengths(l1.sub)
        ml <- min(which(ml != 1), na.rm = TRUE)
        if (ml > 1) {
            l1.sub <- paste(unlist(l1.sub)[seq.int(ml - 1)], collapse = "")
            l1.sub0 <- sub(l1.sub, "", l1)
        } else{
            l1.sub0 <- l1
        }
        ## postfix
        l1.sub <- strsplit(l1.sub0, split = "")
        l1.sub <- Reduce(intersect, l1.sub)
        l1.sub <- paste(l1.sub, collapse = "")
        l1.sub <- sub(l1.sub, "", l1.sub0)
        l1.sub <- char2numeric(l1.sub)
        if (all(!is.na(l1.sub))) {
            l1 <- l1[order(l1.sub)]
        }
    }
    c(l0, l1)
}

readData <- function(slot, folder) {
    fs <- file.path(.globals$datafolder, folder, .globals$filenames[[slot]])
    if(file.exists(fs)){
        return(readRDS(fs))
    }
    return(NULL)
}
loadData <- function(dataSource) {
    for (i in c("sc1conf", "sc1def", "sc1gene", "sc1meta")) {
        dataSource[[i]] <- readData(i, dataSource$dataset)
    }
    return(dataSource)
}

saveAppConf <- function(appconf) {
    pf <- file.path(.globals$datafolder, appconf$id)
    if (!file.exists(pf))
        dir.create(pf, recursive = TRUE, showWarnings = FALSE)
    for (i in c("markers", "keywords")) {
        appconf[[i]] <- appconf[[i]][!is.na(appconf[[i]])]
        appconf[[i]] <- appconf[[i]][appconf[[i]] != ""]
    }
    saveData(appconf, appconf$id, "appconf")
}

saveData <- function(data, folder, prefix) {
    filename <- .globals$filenames[[prefix]]
    if (is.null(filename))
        filename <- paste0(prefix, ".rds")
    saveRDS(data,
            file.path(.globals$datafolder, folder, filename))
}

setLocker <- function(folder) {
    writeLines(
        character(0),
        file.path(.globals$datafolder, folder, .globals$filenames$locker)
    )
    updateConfigTblLocker(folder, TRUE)
}
removeLocker <- function(folder) {
    unlink(file.path(.globals$datafolder, folder, .globals$filenames$locker))
    updateConfigTblLocker(folder, FALSE)
}

generateToken <- function(){
    tokens <- getTokenList()
    token <- paste(sample(c(letters, LETTERS, seq(0, 9)),
                          size=12, replace = TRUE),
                   collapse = '')
    while (token %in% names(tokens)) {
        token <- paste(sample(c(letters, LETTERS, seq(0, 9)),
                              size=12, replace = TRUE),
                       collapse = '')
    }
    return(token)
}
setToken <- function(folder, token){
    if(is.null(token)){
        token <- ''
    }
    if(nchar(token)<.globals$tokenMinLen){
        token <- generateToken()
    }
    writeLines(
        token,
        file.path(.globals$datafolder, folder, .globals$filenames$token)
    )
    adminMsg(paste0("new token is: ", token,
                   "; use link: /?token=", token),
             "message", duration = 30)
}
removeToken <- function(folder){
    unlink(file.path(.globals$datafolder, folder, .globals$filenames$token))
}

writeMisc <- function(misc, folder, slot) {
    if (!is.null(misc)) {
        saveData(misc, folder, slot)
    }
}
#' read expression from h5 file
#' @noRd
#' @param h5f Parent folder name of h5 file
#' @param genesID genes IDs retrieved from sc1gene.rds
#' @param meta meta data by loading sc1meta.rds
#' @param config configs by loading sc1conf.rds
#' @param groupName The group name in the metadata colnames
#' @param valueOnly return the values of first gene
#' @param cell barcode/sampleID pos retrieved from sc1meta.rds
#' @return If valueOnly is TRUE, return expression values for first gene.
#' Otherwise, return a data.table with expressions and group information.
#' @importFrom rhdf5 h5read
#'
read_exprs <- function(
        h5f,
        genesID,
        meta,
        config,
        groupName,
        splitName,
        valueOnly = FALSE,
        cell) {
    fs <- file.path(
        .globals$datafolder,
        h5f,
        .globals$filenames$sc1gexpr)
    if(!file.exists(fs)){
        stop("No expression data available. Data may be removed.")
    }
    if (valueOnly) {
        if (!missing(cell)) {
            expr <- h5read(fs,
                           .globals$h5fGrp,
                           index=list(NULL, cell[1]))[, 1]
        } else{
            if (!is.na(genesID)) {
                expr <- h5read(fs,
                               .globals$h5fGrp,
                               index=list(genesID[1], NULL))[1, ]
            } else{
                expr <- 0
            }
        }
        return(expr)
    }
    exprs <- data.table()
    vals <- h5read(fs, .globals$h5fGrp, index=list(genesID, NULL))
    for (idx in seq_along(genesID)) {
        tmp <- meta[, c("sampleID",
                        config[config$grp == TRUE]$ID),
                    with = FALSE]
        if (!missing(groupName)) {
            tmp$grpBy <- meta[[config[config$UI == groupName]$ID]]
        }
        if (!missing(splitName)) {
            if(!is.na(splitName)) {
                if(splitName %in% config$UI){
                    tmp$splitBy <- meta[[config[config$UI == splitName]$ID]]
                }
            }
        }
        tmp$geneName <- names(genesID)[idx]
        tmp$val <- vals[idx, , drop=TRUE]
        if(all(tmp$val==0) && length(tmp$val)>0){
            setGeneExprForData(symbol = names(genesID)[idx],
                               dataset = h5f,
                               expr = 0)
        }
        exprs <- rbindlist(list(exprs, tmp))
    }
    exprs
}

#' read ATAC counts in peaks
#' The data was write as
#' /cell/cell-name/matrix
#' matrix is sparse matrix, the first column is the index number start from 1
#' of the peak; the second column is the count number.
#' @noRd
#' @importFrom rhdf5 h5createFile h5createGroup h5write
#' @importMethodsFrom Matrix summary
encodeCoord <- function(coorStr, split='-', min.gapwidth=1e5){
    stopifnot(is.character(coorStr))
    coorStr <- strsplit(coorStr, split = split)
    coorStr <- do.call(rbind, coorStr)
    coor <- GRanges(coorStr[, 1], IRanges(as.numeric(coorStr[, 2]), 
                                          as.numeric(coorStr[, 3])))
    gr <- reduce(coor, min.gapwidth=min.gapwidth, with.revmap=TRUE)
    revmap <- rep(seq_along(gr), lengths(gr$revmap))
    names(revmap) <- unlist(gr$revmap)
    return(revmap)
}
#' @importFrom IRanges IRanges nearest
decodeCoord <- function(index, revmap){
    revmap[as.character(index)]
}
writeATACdata <- function(acAsy, appDir){
    filename <- file.path(appDir, .globals$filenames$sc1atac)
    if(h5createFile(filename)){
        if(is(acAsy, 'sparseMatrix')){
            x <- summary(acAsy, sparse=TRUE)
            # search by coordinates
            revmap <- encodeCoord(rownames(acAsy))
            saveRDS(revmap, file.path(appDir, .globals$filenames$coor_revmap))
            x_coors <- split(x[, c("i", "j", "x")],
                             revmap[as.character(x[, "i"])])
            if(h5createGroup(filename, .globals$h5fATACcoor)){
                null <- mapply(function(.x, .n){
                    ij <- paste0(.globals$h5fATACcoor, '/ij', .n)
                    if(h5createDataset(filename, ij, dims = c(nrow(.x), 2),
                                       storage.mode = "integer",
                                       level=7)){
                        h5write(unname(as.matrix(.x[, c('i', 'j')])), file=filename,
                                name=ij)
                    }
                    v <- paste0(.globals$h5fATACcoor, '/v', .n)
                    if(h5createDataset(filename, v, dims = c(nrow(.x), 1),
                                       storage.mode = "double",
                                       level=7)){
                        h5write(.x[, c('x')], file=filename,
                                name=v)
                    }
                }, x_coors, names(x_coors))
            }
        }else{
            warning('The ATAC matrix must be a sparseMatrix.')
        }
    }
}
#' @importFrom rhdf5 h5read H5Lexists H5Fopen H5Fclose
readATACdata <- function(h5f, index){
    fs <- file.path(
        .globals$datafolder,
        h5f,
        .globals$filenames$sc1atac)
    h5f <- H5Fopen(fs)
    on.exit(H5Fclose(h5f))
    stopifnot(length(index)==1)
    if(H5Lexists(h5f, paste0(.globals$h5fATACcoor, '/ij', index)) &&
       H5Lexists(h5f, paste0(.globals$h5fATACcoor, '/v', index))){
        cnts <- cbind(
            h5read(h5f, paste0(.globals$h5fATACcoor, '/ij', index)),
            h5read(h5f, paste0(.globals$h5fATACcoor, '/v', index)))
        colnames(cnts) <- c('i', 'j', 'x')
    }else{
        cnts <- matrix(nrow = 0, ncol=3,
                       dimnames = list(c(), c('i', 'j', 'x')))
    }
    H5Fclose(h5f)
    on.exit()
    cnts
}

#' @importFrom rhdf5 h5ls
#' @importFrom Matrix sparseMatrix
readATACdataByCoor <- function(h5f, coord, cells, revmap, peaks){
    stopifnot(is.character(cells))
    stopifnot(is.list(coord))
    stopifnot(all(c("seqnames", "start", "end") %in% names(coord)))
    if(missing(peaks)) peaks <- readData("sc1peak", h5f)
    if(missing(revmap)) revmap <- readData('coor_revmap', h5f)
    sel <- which(peaks$seqnames %in% coord$seqnames &
                     peaks$start <= coord$end &
                     peaks$end >= coord$start)
    if(length(sel)==0){
        return(data.frame())
    }
    peaks <- paste(peaks$seqnames, peaks$start, peaks$end, sep='-')
    selIndex <- decodeCoord(sel, revmap)
    cnts <- lapply(selIndex, function(i){
        readATACdata(h5f, index=i)
    })
    cnts <- do.call(rbind, cnts)
    cnts <- cnts[cnts[, 'i'] %in% sel, , drop=FALSE]
    cnts <- sparseMatrix(i=cnts[, 'j'],
                         j=match(cnts[, 'i'], sel),
                         x=cnts[, 'x'],
                         dims=c(length(cells), length(sel)),
                         dimnames = list(
                             cells,
                             peaks[sel]
                         ))
    cnts[is.na(cnts)] <- 0
    cnts <- as.data.frame(cnts)
}

tabsubTitleUI <- function(id, title, description){
    tagList(
        htmlOutput(NS(id, paste0(title, "SubTitle"))),
        description,
        br(),br()
    )
}

#' @importFrom grDevices pdfFonts
fontUI <- function(id, fontsizePrefix='plot'){
    pdfs <- names(pdfFonts())
    if(capabilities()[["X11"]]){
        x11s <- names(grDevices::X11Fonts())
    } else {
        x11s <- c()
    }
    if(length(pdfs)>0){
        if(length(x11s)>0){
            family <- intersect(pdfs, x11s)
        }else{
            family <- pdfs
        }
    }else{
        if(length(x11s)>0){
            family <- x11s
        }else{
            family <- c('Helvetica', 'serif', 'mono')
        }
    }
    if('Helvetica' %in% family){
        selected <- 'Helvetica'
    }else{
        selected <- family[1]
    }
    tagList(
        selectInput(NS(id, paste0(fontsizePrefix, 'fml')),
                    "Font family:",
                    choices = family, selected = selected),
        numericInput(
            NS(id, paste0(fontsizePrefix, "fsz")), "Font size:",
            value = 24, min=3, max = 72, step = .5)
    )
}

graphicsControlUI <- function(id, GeneExpraspSelect="Square"){
    tagList(
        actionButton(
            NS(id, "graphicTog"),
            "Toggle graphics controls"),
        conditionalPanel(
            condition = "input.graphicTog % 2 == 1",
            ns = NS(id),
            column(
                6,
                sliderInput(
                    NS(id, "GeneExprsiz"),
                    "Point size:",
                    min = 0, max = 4, value = 1.25, step = 0.25),
                radioButtons(
                    NS(id, "GeneExprpsz"),
                    "Plot size:",
                    choices = c("Small", "Medium", "Large"),
                    selected = "Medium", inline = TRUE),
                fontUI(id, "GeneExpr")
            ),
            column(
                6, radioButtons(
                    NS(id, "GeneExprasp"), "Aspect ratio:",
                    choices = c("Square", "Fixed", "Free"),
                    selected = GeneExpraspSelect, inline = TRUE),
                checkboxInput(
                    NS(id, "GeneExprtxt"), "Show axis text", value = FALSE)
            )
        )
    )
}
NS0 <- function(namespace, id, postfix){
    NS(namespace, id=paste0(id, postfix))
}
geneAccPlotControlUI <- function(
        id, postfix=1,
        colorNames=availableThemes("sequence")){
    tagList(
        actionButton(
            NS0(id, "GeneExprtog", postfix), "Toggle plot controls"),
        conditionalPanel(
            condition = paste0("input.GeneExprtog", postfix, " % 2 == 1"),
            ns=NS(id),
            radioButtons(
                NS0(id, "GeneExprcol", postfix), "Colour:",
                inline = TRUE,
                choices = colorNames,
                selected = colorNames[1]),
            sliderInput(#region selector
                NS(id, 'regionselector'), label = NULL,
                min=0, max = 100,
                step = 1,
                value = c(0, 100),
                ticks = FALSE,
                width = "100%"),
            actionButton(
                NS(id, 'regionsubmit'),
                label = "change region",
                width = "100%"
            )
        )
    )
}
geneExprPlotControlUI <- function(
        id, postfix=1,
        colorNames=availableThemes("sequence")){
    tagList(
        actionButton(
            NS0(id, "GeneExprtog", postfix), "Toggle plot controls"),
        conditionalPanel(
            condition = paste0("input.GeneExprtog", postfix, " % 2 == 1"),
            ns=NS(id),
            radioButtons(
                NS0(id, "GeneExprtype", postfix), "Plot type",
                choices = c("Dotplot", "Ridgeplot"),
                selected = "Dotplot"),
            conditionalPanel(
                condition = paste0(
                    "input.GeneExprtype", postfix, " == 'Dotplot'"),
                ns=NS(id),
                radioButtons(
                    NS0(id, "GeneExprcol", postfix), "Colour:",
                    inline = TRUE,
                    choices = colorNames,
                    selected = colorNames[1]),
                radioButtons(
                    NS0(id, "GeneExprord", postfix), "Plot order:",
                    choices = c("Max-1st", "Min-1st",
                                "Original", "Random"),
                    selected = "Max-1st", inline = TRUE),
                checkboxInput(
                    NS0(id, "GeneExprhid", postfix),
                    "Hide filtered cells", value = FALSE),
                actionButton(
                    NS0(id, "GeneExprrgb", postfix),
                    "Manually set max color value",
                    inline = TRUE),
                conditionalPanel(
                    condition =
                        paste0("input.GeneExprrgb", postfix, " % 2 ==1"),
                    ns=NS(id),
                    numericInput(
                        NS0(id, "GeneExprrg", postfix), "Max value:",
                        value = 100))
            ),
            conditionalPanel(
                condition = paste0(
                    "input.GeneExprtype", postfix, " == 'Ridgeplot'"),
                ns=NS(id),
                actionButton(
                    NS0(id, "GeneExprxlimb", postfix),
                    "Manually set x axis", inline = TRUE),
                conditionalPanel(
                    condition = paste0(
                        "input.GeneExprxlimb", postfix, " % 2 ==1"),
                    ns=NS(id),
                    sliderInput(
                        NS0(id, "GeneExprxlim", postfix), "Xlim range:",
                        min = -10, max = 100, value = c(-1.5, 10),
                        step = 0.1))
            )
        )
    )
}
cellInfoPlotControlUI <- function(
        id, postfix=1,
        colorNames=availableThemes("sequence")){
    tagList(
        actionButton(
            NS0(id, "CellInfotog", postfix), "Toggle plot controls"),
        conditionalPanel(
            condition = paste0(
                "input.CellInfotog", postfix, " % 2 == 1"), ns=NS(id),
            radioButtons(
                NS0(id, "CellInfocol", postfix), "Colour (Continuous data):",
                inline = TRUE,
                choices = colorNames,
                selected = colorNames[1]),
            radioButtons(
                NS0(id, "CellInfoord", postfix), "Plot order:",
                choices = c("Max-1st", "Min-1st",
                            "Original", "Random"),
                selected = "Original", inline = TRUE),
            checkboxInput(
                NS0(id, "CellInfolab", postfix),
                "Show cell info labels", value = TRUE),
            selectInput(
                NS0(id, 'CellInfoname', postfix),
                "Cell info labels",
                choices = NULL
            ),
            checkboxInput(
                NS0(id, "CellInfohid", postfix),
                "Hide filtered cells", value = FALSE),
            checkboxInput(
                NS0(id, "CellInfoslingshot", postfix),
                "Show lineages", value = TRUE),
            checkboxInput(
                NS0(id, "CellInfoedge", postfix),
                "Show cell edges", value = TRUE)
        ),
        div(style = "visibility:hidden;",
            id = paste0(NS0(id, "CellInfodup", postfix), 'container'),
            div(style="display:inline-block",
            textInput(NS0(id, "CellInfodname", postfix), "",
                      placeholder = "new name", width = "100px")),
            actionButton(NS0(id, "CellInfodup", postfix),
                         "Dup"),
            actionButton(NS0(id, "CellInforename", postfix),
                         "Ren"),
            actionButton(NS0(id, "CellInfodel", postfix),
                         "Del")
        )
    )
}
geneCoExprPlotControlUI <- function(id, postfix=1, plotly=FALSE){
    choices <- .globals$coExpColor
    if(plotly){
        choices <- c("Default", availableThemes("sequence"))
    }
    tagList(
        actionButton(NS0(id, "CoExprtog", postfix), "Toggle plot controls"),
        conditionalPanel(
            condition = paste0(
                "input.CoExprtog", postfix, " % 2 == 1"), ns=NS(id),
            radioButtons(
                NS0(id, "CoExprcol", postfix), "Colour:",
                inline = TRUE,
                choices = choices,
                selected = choices[1]),
            radioButtons(
                NS0(id, "CoExprord", postfix), "Plot order:",
                choices = c("Max-1st", "Min-1st", "Original", "Random"),
                selected = "Max-1st", inline = TRUE),
            checkboxInput(
                NS0(id, "CoExprhid", postfix),
                "Hide filtered cells", value = FALSE)
        )
    )
}

boxPlotControlUI <- function(
        id, withPoints=TRUE, withColor=FALSE,
        withFontSize=TRUE,
        colorNames=availableThemes("sequence")){
    tagList(
        actionButton(
            NS(id, "plottog"), "Toggle graphics controls"),
        conditionalPanel(
            condition = "input.plottog % 2 == 1",
            ns=NS(id),
            if(withPoints) {
                sliderInput(
                    NS(id, "plotsiz"), "Data point size:",
                    min = 0, max = 4, value = 1.25, step = 0.25)
            }else{
                span()
            },
            if(withColor){
                radioButtons(
                    NS(id, "plotcols"), "Colour scheme:",
                    inline = TRUE,
                    choices = colorNames,
                    selected = colorNames[2])
            }else{
                span()
            },
            radioButtons(
                NS(id, "plotpsz"), "Plot size:",
                choices = c("Small", "Medium", "Large"),
                selected = "Medium", inline = TRUE),
            if(withFontSize){
                fontUI(id)
            }else{
                span()
            }
        )
    )
}

dimensionReductionUI <- function(id){
    tagList(
        h4("Dimension Reduction"),
        fluidRow(
            column(
                12,
                selectInput(
                    NS(id, "GeneExprdrX"),
                    "X-axis:",
                    choices = NULL),
                selectInput(
                    NS(id, "GeneExprdrY"),
                    "Y-axis:",
                    choices = NULL))
        )
    )
}
#' @importFrom magrittr %>%
subsetCellByInfoUI <- function(id, mini=FALSE, multiple=TRUE, ABcolumn){
    if(mini){
        tagList(
            uiOutput(NS(id, "subsetCellSel.ui")) %>%
                helper1(category="subsetCellInfo"),
            if(multiple) actionButton(
                NS(id, 'subsetCell.multi'),
                label="multiple") else tags$span(),
            uiOutput(NS(id, "subsetCell.ui"))
        )
    }else{
        if(missing(ABcolumn)){
            tagList(
                actionButton(NS(id, "subsetTogT"), "Toggle to subset cells"),
                conditionalPanel(
                    condition = "input.subsetTogT % 2 == 0",
                    ns = NS(id),
                    fluidRow(
                        column(9,
                               uiOutput(NS(id, "subsetCellSel.ui")) %>%
                                   helper1(category="subsetCellInfo")),
                        column(3,
                               if(multiple) actionButton(
                                   NS(id, 'subsetCell.multi'),
                                   label="multiple",
                                   class = "align-action-button")
                               else tags$span())),
                    uiOutput(NS(id, "subsetCell.ui"))
                )
            )
        }else{
            tagList(
                actionButton(NS0(id, "subsetTogT", ABcolumn),
                             paste("Toggle to subset cells setting",
                                   ABcolumn)),
                conditionalPanel(
                    condition = paste0("input.subsetTogT", ABcolumn,
                                       " % 2 == ",
                                       ifelse(ABcolumn==.globals$subsetgroup[1],
                                              0, 1)),
                    ns = NS(id),
                    fluidRow(
                        column(
                            7,
                            uiOutput(NS0(id, "subsetCellSel.ui", ABcolumn))
                        ),
                        column(
                            3,
                            if(multiple) actionButton(
                                NS0(id, 'subsetCell.multi', ABcolumn),
                                label="multiple", class = "align-action-button")
                            else tags$span()
                        )
                    ),
                    
                    uiOutput(NS0(id, "subsetCell.ui", ABcolumn))
                )
            )
        }
    }
}
#' @importFrom magrittr %>%
subsetCellByFilterUI <- function(
        id,
        label="Cell Info/Gene name to subset:",
        title=NULL,
        content=NULL){
    tagList(
        selectInput(
            NS(id, "filterCell"),
            label=label,
            choices = NULL) %>%
            helper1(category="subsetCellInfo", title=title, content=content),
        uiOutput(NS(id, "filterCell.ui"))
    )
}

geneExprDotPlotUI <- function(id, postfix=1, editor=FALSE){
    tagList(
        fluidRow(column(12, uiOutput(NS0(id, "GeneExproup.ui", postfix)))),
        div(style="display:inline-block",
            selectInput(
                NS0(id, "GeneExproup.fmt", postfix),
                "Format:", width = "75px",
                choices = .globals$figFormats,
                selected = .globals$figFormats[1])),
        div(style="display:inline-block",
            numericInput(
                NS0(id, "GeneExproup.h", postfix),
                "height:", width = "60px",
                min = 2, max = 20, value = .globals$figHeight, step = 0.5)),
        div(style="display:inline-block",
            numericInput(
                NS0(id, "GeneExproup.w", postfix),
                "width:", width = "60px",
                min = 2, max = 20, value = .globals$figWidth, step = 0.5)),
        downloadButton(NS0(id, "GeneExproup.dwn", postfix), "download"),
        if(editor){
            tagList(
                uiOutput(NS0(id, 'GeneExproup.info', postfix))
            )
        }else{
            div(style = "visibility:hidden;",
                textInput(NS0(id, 'GeneExpext.info', postfix),
                          label='', value = '')
            )
        }
    )
}

#' @importFrom magrittr %>%
cellInfoUI <- function(id, postfix=1){
    tagList(
        selectInput(
            NS0(id, "CellInfo", postfix), "Cell information:",
            choices = NULL) %>%
            helper1(category="cellInfo")
    )
}
#' @importFrom DT DTOutput
cellInfoTblUI <- function(id, postfix=1){
    tagList(
        actionButton(
            NS0(id, "CellInfoTableTog", postfix),
            "Toggle to show cell numbers / statistics"),
        conditionalPanel(
            condition = paste0("input.CellInfoTableTog", postfix, " % 2 == 1"),
            ns=NS(id),
            h4("Cell numbers / statistics"),
            radioButtons(
                NS0(id, "GeneExprsplt", postfix),
                "Split continuous cell info into:",
                choices = c("Quartile", "Decile"),
                selected = "Decile", inline = TRUE),
            DTOutput(NS0(id, "GeneExpr.dt", postfix))
        )
    )
}

#' @importFrom magrittr %>%
geneExprUI <- function(id, postfix=1){
    tagList(
        selectInput(
            NS0(id, "GeneName", postfix),
            "Gene name:", choices=NULL) %>%
            helper1(category="geneName")
    )
}

geneAccUI <- function(id, postfix=1){
    tagList(
        textInput(
            NS0(id, "coord", postfix),
            "Coordinates:", value=NULL),
        div(
            class = "acccontroler",
            actionButton(#zoom in
                NS(id, 'zoomin'), label = '', title="Zoom In",
                icon = icon('plus'),
                class = "submodule-dot-btn submodule-icon",
                style = "background: #ED594A;"),
            actionButton(#zoom out
                NS(id, 'zoomout'), label = '', title="Zoom Out",
                icon = icon('minus'),
                class = "submodule-dot-btn submodule-icon",
                style = "background: #FDD800;"),
            actionButton(#move left
                NS(id, 'moveleft'), label = '', title="Move Left",
                icon = icon('angle-left'),
                class = "submodule-dot-btn submodule-icon",
                style = "background: #006EF4;"),
            actionButton(#move right
                NS(id, 'moveright'), label = '', title="Move Right",
                icon = icon('angle-right'),
                class = "submodule-dot-btn submodule-icon",
                style = "background: #5AC05A;")
        )
    )
}

#' @importFrom magrittr %>%
xaxisCellInfoUI <- function(id){
    tagList(
        selectInput(
            NS(id, "CellInfoX"), "Cell information (X-axis):",
            choices = NULL) %>%
            helper1(category="cellInfoX")
    )
}

#' @importFrom magrittr %>%
yaxisCellInfoUI <- function(id){
    tagList(
        selectInput(
            NS(id, "CellInfoY"), "Cell Info / Gene name (Y-axis):",
            choices=NULL) %>%
            helper1(category="cellInfoY")
    )
}
# subMOduleUIs
subModuleContainerUI <- function(id, mainSelectUI, menuUI, contentUI){
    tagList(
        div(
            class="submodule-container",
            div(
                class="submodule-row",
                div(
                    class="submodule-column submodule-left",
                    actionButton(
                        NS(id, 'close'), label = '',
                        icon = icon('close'),
                        class = "submodule-dot-btn submodule-icon",
                        style = "background: #ED594A;"),
                    actionButton(
                        NS(id, 'movedown'), label = '',
                        icon = icon('angle-down'),
                        class = "submodule-dot-btn submodule-icon",
                        style = "background: #FDD800;"),
                    actionButton(
                        NS(id, 'moveup'), label = '',
                        icon = icon('angle-up'),
                        class = "submodule-dot-btn submodule-icon",
                        style = "background: #006EF4;"),
                    actionButton(
                        NS(id, 'resize'), label = '',
                        icon = icon('arrows-left-right'),
                        class = "submodule-dot-btn submodule-icon",
                        style = "background: #5AC05A;")
                ),
                div(
                    class="submodule-column submodule-middle",
                    mainSelectUI
                ),
                div(
                    class="submodule-column submodule-right",
                    menuUI
                )
            ),
            div(
                class="submodule-content",
                div(
                    contentUI
                )
            )
        )
    )
}

subsetGrpRadioButton <- function(id, label, selected, inline=TRUE){
    if(is.logical(selected)){
        if(selected){
            selected <- .globals$subsetgroup[1]
        }
    }
    if(selected %in% .globals$subsetgroup){
        radioButtons(
            inputId = id,
            label = label,
            choices = .globals$subsetgroup,
            selected = selected,
            inline = inline
        )
    }
}

contextMenuCellInfoUI <- function(
        id, postfix=1,
        colorNames=availableThemes("sequence"),
        group=FALSE){
    tagList(
        actionButton(
            NS0(id, "CellInfotog", postfix), "",
            icon = icon("bars"),
            class = "submodule-icon"),
        div(
            class="submodule-contextmenu",
            conditionalPanel(
                condition = paste0("input.CellInfotog", postfix, " % 2 == 1"),
                ns=NS(id),
                subsetGrpRadioButton(
                    id = NS0(id, 'CellInfosubgrp', postfix),
                    label = "Subset setting group:",
                    selected = group,
                    inline=TRUE),
                radioButtons(
                    NS0(id, "CellInfocol", postfix),
                    "Colour (Continuous data):",
                    inline = TRUE,
                    choices = colorNames,
                    selected = colorNames[1]),
                radioButtons(
                    NS0(id, "CellInfoord", postfix), "Plot order:",
                    choices = c("Max-1st", "Min-1st",
                                "Original", "Random"),
                    selected = "Original", inline = TRUE),
                checkboxInput(
                    NS0(id, "CellInfolab", postfix),
                    "Show cell info labels", value = TRUE),
                checkboxInput(
                    NS0(id, "CellInfoslingshot", postfix),
                    "Show lineages", value = TRUE),
                checkboxInput(
                    NS0(id, "CellInfoedge", postfix),
                    "Show cell edges", value = TRUE),
                checkboxInput(
                    NS0(id, "CellInfohid", postfix),
                    "Hide filtered cells", value = FALSE)
            )
        )
    )
}
contextMenuGeneExprUI <- function(
        id, postfix=1,
        colorNames=availableThemes("sequence"),
        group = FALSE){
    tagList(
        actionButton(
            NS0(id, "GeneExprtog", postfix), "",
            icon = icon("bars"),
            class = "submodule-icon"),
        div(
            class="submodule-contextmenu",
            conditionalPanel(
                condition = paste0("input.GeneExprtog", postfix, " % 2 == 1"),
                ns=NS(id),
                subsetGrpRadioButton(
                    id = NS0(id, 'CellInfosubgrp', postfix),
                    label = "Subset setting group:",
                    selected = group,
                    inline=TRUE),
                radioButtons(
                    NS0(id, "GeneExprtype", postfix), "Plot type",
                    choices = c("Dotplot", "Ridgeplot"),
                    selected = "Dotplot"),
                conditionalPanel(
                    condition = paste0(
                        "input.GeneExprtype",
                        postfix, " == 'Dotplot'"),
                    ns=NS(id),
                    radioButtons(
                        NS0(id, "GeneExprcol", postfix), "Colour:",
                        inline = TRUE,
                        choices = colorNames,
                        selected = colorNames[1]),
                    radioButtons(
                        NS0(id, "GeneExprord", postfix), "Plot order:",
                        choices = c("Max-1st", "Min-1st",
                                    "Original", "Random"),
                        selected = "Max-1st", inline = TRUE),
                    checkboxInput(
                        NS0(id, "GeneExprhid", postfix),
                        "Hide filtered cells", value = FALSE),
                    actionButton(
                        NS0(id, "GeneExprrgb", postfix),
                        "Manually set max color value",
                        inline = TRUE),
                    conditionalPanel(
                        condition = paste0(
                            "input.GeneExprrgb",
                            postfix, " % 2 ==1"),
                        ns=NS(id),
                        numericInput(
                            NS0(id, "GeneExprrg", postfix), "Max value:",
                            value = 100))
                ),
                conditionalPanel(
                    condition = paste0(
                        "input.GeneExprtype", postfix,
                        " == 'Ridgeplot'"),
                    ns=NS(id),
                    actionButton(
                        NS0(id, "GeneExprxlimb", postfix),
                        "Manually set x axis", inline = TRUE),
                    conditionalPanel(
                        condition = paste0(
                            "input.GeneExprxlimb",
                            postfix, " % 2 ==1"),
                        ns=NS(id),
                        sliderInput(
                            NS0(id, "GeneExprxlim", postfix), "Xlim range:",
                            min = -10, max = 100, value = c(-1.5, 10),
                            step = 0.1))
                )
            )
        )
    )
}
contextMenuCoExprUI <- function(
        id, postfix=1,
        colorNames=availableThemes("sequence"),
        plotly = FALSE,
        group = FALSE){
    choices <- .globals$coExpColor
    if(plotly){
        choices <- c("Default", availableThemes("sequence"))
    }
    tagList(
        actionButton(
            NS0(id, "CoExprtog", postfix), "",
            icon = icon("bars"),
            class = "submodule-icon"),
        div(
            class="submodule-contextmenu",
            conditionalPanel(
                condition = paste0("input.CoExprtog", postfix, " % 2 == 1"),
                ns=NS(id),
                subsetGrpRadioButton(
                    id = NS0(id, 'CellInfosubgrp', postfix),
                    label = "Subset setting group:",
                    selected = group,
                    inline=TRUE),
                radioButtons(
                    NS0(id, "CoExprcol", postfix), "Colour:",
                    choices = choices,
                    selected = choices[1]),
                radioButtons(
                    NS0(id, "CoExprord", postfix), "Plot order:",
                    choices = c("Max-1st", "Min-1st", "Original", "Random"),
                    selected = "Max-1st", inline = TRUE),
                checkboxInput(
                    NS0(id, "CoExprhid", postfix),
                    "Hide filtered cells", value = FALSE)
            )
        )
    )
}
contextMenuPropUI <- function(id, postfix=1, group = FALSE){
    tagList(
        actionButton(
            NS(id, "Proptog"), "",
            icon = icon("bars"),
            class = "submodule-icon"),
        div(
            class="submodule-contextmenu",
            conditionalPanel(
                condition = paste0("input.Proptog", " % 2 == 1"), ns=NS(id),
                subsetGrpRadioButton(
                    id = NS0(id, 'CellInfosubgrp', postfix),
                    label = "Subset setting group:",
                    selected = group,
                    inline=TRUE),
                radioButtons(
                    NS(id, "plottyp"),
                    "Plot value:",
                    choices = c("Proportion", "CellNumbers"),
                    selected = "Proportion", inline = TRUE),
                checkboxInput(
                    NS(id, "plotflp"),
                    "Flip X/Y", value = FALSE),
                checkboxInput(
                    NS(id, "plotord"),
                    "Reorder the contents", value = FALSE
                ),
                conditionalPanel(
                    condition = "input.plotord % 2 == 1",
                    ns=NS(id),
                    uiOutput(outputId = NS(id, "plotXord")),
                    uiOutput(outputId = NS(id, "plotYord"))
                )
            )
        )
    )
}
contextMenuViolinUI <- function(id, postfix=1, group = FALSE){
    tagList(
        actionButton(
            NS(id, "Propviolin"), "",
            icon = icon("bars"),
            class = "submodule-icon"),
        div(
            class="submodule-contextmenu",
            conditionalPanel(
                condition = paste0("input.Propviolin", " % 2 == 1"), ns=NS(id),
                subsetGrpRadioButton(
                    id = NS0(id, 'CellInfosubgrp', postfix),
                    label = "Subset setting group:",
                    selected = group,
                    inline=TRUE),
                radioButtons(
                    NS(id, "plottyp"), "Plot type:",
                    choices = c("violin", "boxplot"),
                    selected = "violin", inline = TRUE),
                checkboxInput(
                    NS(id, "plotpts"),
                    "Show data points",
                    value = FALSE),
                checkboxInput(
                    NS(id, 'addnoise'),
                    "Add noise", value = TRUE
                ),
                checkboxInput(
                    NS(id, "plotord"),
                    "Reorder the contents", value = FALSE
                ),
                conditionalPanel(
                    condition = "input.plotord % 2 == 1",
                    ns=NS(id),
                    uiOutput(outputId = NS(id, "plotXord"))
                )
            )
        )
    )
}

#' @importFrom DT DTOutput
#' @importFrom magrittr %>%
issueUI <- function(id) {
    ns <- NS(id)
    tabPanel(
        value = id,
        HTML("Issues"),
        tabsubTitleUI(
            id,
            'Issues',
            description = paste(
                "There is a solution to every problem:",
                "simple, quick, and wrong. ",
                "--Henry Louis Mencken"
            )
        ),
        fluidRow(
            ## new comments here
            div(class = "newcomment-form-container",
                # Header
                div(class = "newcomment-form-header",
                    h3(class = "newcomment-form-title",
                       tags$i(class = "fas fa-comment-alt",
                              style = "color: #5f6368;"),
                       "Add New Comment"
                    )
                ),
                
                # User info row
                div(class = "newcomment-form-row",
                    div(class = "newcomment-form-group-modern has-icon",
                        tags$i(class = "newcomment-input-icon fas fa-user"),
                        tags$label("for" = ns('uid'),
                                   class = "required-field", "Your Name"),
                        div(class = "newcomment-form-control-wrapper",
                            textInput(ns('uid'), 
                                      label = NULL,
                                      placeholder = "Enter your name")
                        ),
                        div(class = "newcomment-form-info",
                            tags$i(class = "fas fa-info-circle"),
                            "We'll email you only if clarification is needed.")
                    ),
                    div(class = "newcomment-form-group-modern has-icon",
                        tags$i(class = "newcomment-input-icon fas fa-envelope"),
                        tags$label("for" = ns('email'),
                                   class = "required-field", 
                                   "Email Address"),
                        div(class = "newcomment-form-control-wrapper",
                            textInput(ns('email'), 
                                      label = NULL,
                                      placeholder = "your.email@example.com")
                        ),
                        div(class = "newcomment-form-info",
                            tags$i(class = "fas fa-info-circle"),
                            "Email acts as edit password.")
                    )
                ),
                
                # Title input
                div(class = "newcomment-title-input-container",
                    div(class = "newcomment-form-group-modern",
                        tags$label("for" = ns('title'),
                                   class = "required-field",
                                   "Comment Title"),
                        div(class = "newcomment-form-control-wrapper",
                            textInput(ns('title'), 
                                      label = NULL,
                                      placeholder = "What's this comment about?",
                                      width = "100%")
                        )
                    )
                ),
                
                # Comment textarea
                div(class = "newcomment-input-container",
                    tags$label("for" = ns('comment'),
                               class = "required-field",
                               "Your Comment"),
                    div(class = "form-control-wrapper",
                        textAreaInput(ns('comment'),
                                      label = NULL,
                                      placeholder = "Share your thoughts, ask questions, or provide feedback...",
                                      rows = 6,
                                      width = "100%")
                    )
                ),
                
                # Hidden token field
                div(style = "display: none;",
                    textInput(ns('token'), 
                              label = NULL,
                              value = as.numeric(Sys.time()))
                ),
                
                # Form actions
                div(class = "newcomment-form-actions",
                    div(class = "newcomment-form-info",
                        tags$i(class = "fas fa-info-circle"),
                        "Markdown is supported in comments.",
                        span(style = "margin-left: 16px;",
                             tags$i(class = "fas fa-lock"),
                             "  We'll never share your email."
                        )
                    ),
                    actionButton(ns('submit'),
                                 "Post Comment",
                                 class = "newcomment-submit-btn",
                                 icon = icon("paper-plane"))
                ),
                
                # Preview section
                div(class = "newcomment-preview-container",
                    div(class = "newcomment-preview-header",
                        tags$i(class = "fas fa-eye"),
                        "Markdown Comment Preview"
                    ),
                    div(class = "newcomment-preview-content",
                        uiOutput(ns('previewbox'))
                    )
                )
            )
        ),
        hr(),
        
        h5('History issues'),
        fluidRow(
            ## list all the comments by pages
            ## DT ui
            DTOutput(NS(id, "issues"))
        )
    )
}
#' @importFrom DT renderDT
#' @importFrom magrittr %>%
#' @importFrom shiny markdown
issueServer <- function(id, dataSource, optCrt) {
    moduleServer(id, function(input, output, session) {
        observeEvent(input$comment, {
            output$previewbox <- renderUI(markdown(input$comment))
        })
        observeEvent(input$reply_comment, {
            output$reply_previewbox <- renderUI(markdown(input$reply_comment))
        })
        observeEvent(input$edit_comment, {
            output$edit_previewbox <- renderUI(markdown(input$edit_comment))
        })
        createCommentHTML <- function(comment, parent_comment){
            comment <- unlist(comment)
            actionBtn <- paste0(
                '    <button id="upvote_button_', comment["id"], '" ',
                'class="action-btn upvote" ',
                'onclick="Shiny.onInputChange(\'', NS(id, "upvote_clicked"),
                '\', this.id)"',
                '><i class="far fa-thumbs-up"></i> Upvote</button>',
                '    <button id="edit_button_', comment["id"], '" ',
                'class="action-btn" ',
                'onclick="Shiny.onInputChange(\'', NS(id, "edit_clicked"),
                '\', Math.random()+\'_\'+this.id)"',
                '><i class="fas fa-edit"></i> Edit</button>',
                '<button id="close_button_', comment["id"], '" ',
                'class="action-btn" ',
                'onclick="Shiny.onInputChange(\'', NS(id, "close_clicked"),
                '\', Math.random()+\'_\'+this.id)"',
                '><i class="far fa-check-circle"></i> ',
                if(isTRUE(dataSource()$auth$privilege=='all')){
                    ifelse(comment["open"]=="1", 'Close', 'Open')
                }else{'Close'},
                '</button>',
                '<button id="reply_button_', comment["id"], '" ',
                'class="action-btn" ',
                'onclick="Shiny.onInputChange(\'', NS(id, "reply_clicked"),
                '\', Math.random()+\'_\'+this.id)"',
                '><i class="fas fa-reply"></i> Reply</button>'
            )
            
            if(isTRUE(dataSource()$auth$privilege=='all')){
                actionBtn <- paste0(
                    actionBtn,
                    '<button id="delete_button_', comment["id"], '" ',
                    'class="action-btn" ',
                    'onclick="Shiny.onInputChange(\'', NS(id, "delete_clicked"),
                    '\', Math.random()+\'_\'+this.id)"',
                    '><i class="far fa-trash-alt" style="color:red"></i> Delete</button>'
                )
            }
            if(missing(parent_comment)){
                # Main comment layout
                paste0(
                    '<div class="comment-main">',
                    '  <div class="comment-header">',
                    '    <div style="display: flex; align-items: center;">',
                    '      <div class="comment-title">', comment["title"], '</div>',
                    '      <div class="comment-author">by ', comment["uid"], '</div>',
                    '    </div>',
                    '    <div class="comment-meta">',
                    '      <span class="timestamp">', comment["updated_at"], '</span>',
                    '      <span class="vote"><i class="far fa-thumbs-up"></i> ', comment["vote"], '</span>',
                    '    </div>',
                    '  </div>',
                    '  <div class="comment-content">',
                    '    ', markdown(comment["comment"]),
                    '  </div>',
                    '  <div class="comment-actions">', actionBtn , '  </div>',
                    '</div>'
                )
            }else{
                # Check if this is a nested reply (reply to a reply)
                parent_comment <- unlist(parent_comment)
                is_nested <- parent_comment['pid'] != parent_comment['id']
                
                css_class <- if (is_nested) "comment-nested" else "comment-reply"
                header_class <- if (is_nested) "nested-header" else "reply-header"
                author_class <- if (is_nested) "nested-author" else "reply-author"
                
                # Reply comment layout
                paste0(
                    '<div class="', css_class, '">',
                    '  <div class="reply-header ', header_class, '">',
                    '    <div class="', author_class, '">by ', comment["uid"], '</div>',
                    '    <div class="comment-meta">',
                    '      <span class="timestamp">', comment["updated_at"], '</span>',
                    '      <span class="vote"><i class="far fa-thumbs-up"></i> ', comment["vote"], '</span>',
                    '    </div>',
                    '  </div>',
                    '  <div class="reply-content">',
                    '    ', markdown(comment["comment"]),
                    '  </div>',
                    '  <div class="reply-actions">', actionBtn , '  </div>',
                    '</div>'
                )
            }
        }
        createThreadedComments <- function(comments_df, main_comments) {
            if(nrow(comments_df)==0){
                return(c())
            }
            threaded_html <- c()
            if(missing(main_comments)){
                m <- comments_df$pid == comments_df$id
                main_comments <- comments_df[m, , drop=FALSE]
                addMain <- TRUE
            }else{
                addMain <- FALSE
            }
            if(nrow(main_comments)<1){
                return(c())
            }
            for (i in seq.int(nrow(main_comments))) {
                main_comment <- main_comments[i, , drop=TRUE]
                names(main_comment) <- colnames(comments_df)
                # Add main comment
                if(addMain){
                    this_html <- createCommentHTML(main_comment)
                }else{
                    this_html <- c()
                }
                # Add replies to this main comment
                comment_replies <- 
                    comments_df[comments_df$pid == main_comment$id, , drop=FALSE]
                comment_replies <- 
                    comment_replies[comment_replies$pid != comment_replies$id, 
                                    , drop=FALSE] ## remove self
                if (nrow(comment_replies) > 0) {
                    for (j in seq.int(nrow(comment_replies))) {
                        reply <- comment_replies[j, , drop=TRUE]
                        names(reply) <- colnames(comments_df)
                        this_html <-
                            c(this_html,
                              createCommentHTML(reply, main_comment))
                        reply_reply <-
                            comments_df[comments_df$pid == reply$id, ,
                                        drop=FALSE]
                        if(nrow(reply_reply)>0){
                            this_html <-
                                c(this_html,
                                  createThreadedComments(
                                      comments_df,
                                      comment_replies[j, , drop=FALSE]))
                        }
                    }
                }
                threaded_html <- c(threaded_html,
                                   paste(this_html, collapse = ' '))
            }
            return(threaded_html)
        }
        updateCommentList <- function(){
            output$issues <- renderDT({
                if(isTRUE(dataSource()$auth$privilege=='all')){
                    data_to_display <- 
                        listComments(page_size=.globals$totalComments,
                                     full=TRUE, all=TRUE)
                }else{
                    data_to_display <- 
                        listComments(page_size=1000)
                }
                if(nrow(data_to_display)<1){
                    data.frame('Title'=character(0L),
                               'Name'=character(0L),
                               'Comment'=character(0L),
                               'Actions'=character(0L))
                }else{
                    # # Sort by timestamp and organize by parent-child relationships
                    data_to_display <- split(data_to_display, data_to_display$pid)
                    data_to_display_updated_at <- unlist(lapply(data_to_display, function(.ele){
                        time <- as.POSIXct(.ele$updated_at, format = "%Y-%m-%d %H:%M:%S")
                        max(time)
                    }))
                    data_to_display <- 
                        lapply(data_to_display[order(data_to_display_updated_at,
                                              decreasing = TRUE)],
                               function(.ele){
                                   time <- as.POSIXct(.ele$created_at,
                                                      format = "%Y-%m-%d %H:%M:%S")
                                   .ele <-.ele[order(time), ,
                                               drop=FALSE]
                                   .ele
                                   })
                    data_to_display <- do.call(rbind, data_to_display)
                    data.frame(
                        CommentThread = createThreadedComments(data_to_display),
                        stringsAsFactors = FALSE
                    )
                }
                },# IMPORTANT: Allow HTML rendering for buttons
                escape = FALSE, # IMPORTANT: Allow HTML rendering for buttons
                options = list(
                    ordering = FALSE,
                    lengthMenu = c(5, 10, 25, 50, 100),
                    pageLength = 10 
                ),
                rownames = FALSE,
                colnames = ""
            )
        }
        # Render the paginated comments
        updateCommentList()
        checkLen <- function(n, l, m){
            if(!length(n)) return(TRUE)
            if(nchar(n)<l) return(TRUE)
            if(nchar(n)>m) return(TRUE)
            return(FALSE)
        }
        checkEmail <- function(prefix=''){
            if(is.null(input[[paste0(prefix, "email")]])){
                showNotification("Please input correct email address!",
                                 type = 'error')
                return(FALSE)
            }
            if(isTRUE(!grepl('^[a-zA-Z0-9._%+-]+@[a-zA-Z0-9.-]+\\.[a-zA-Z]{2,}$',
                      input[[paste0(prefix, "email")]]))){
                showNotification("Please input correct email address!",
                                 type = 'error')
                return(FALSE)
            }
            return(TRUE)
        }
        checkForm <- function(prefix=''){
            res <- checkEmail(prefix)
            if(checkLen(input[[paste0(prefix, "uid")]], 3, 50)){
                showNotification("Please input your name ([3, 50] letters).",
                                 type = 'error')
                res <- FALSE
            }
            if(checkLen(input[[paste0(prefix, "title")]],2, 200)){
                showNotification("Please input the issue title ([2, 200] letters).",
                                 type = 'error')
                res <- FALSE
            }
            if(checkLen(input[[paste0(prefix, "comment")]],10, 1000)){
                showNotification("Please input the comment ([10, 1000] letters).",
                                 type = 'error')
                res <- FALSE
            }
            return(res)
        }
        checkInterval <- function(){
            ans <- TRUE
            if(as.numeric(Sys.time()) - as.numeric(input$token) < .globals$commentsIntervals){
                showNotification('System busy! Please try to submit it 10s later.',
                                 type = 'error')
                ans <- FALSE
            }
            updateTextInput(session, 'token', label=NULL,
                            value=as.numeric(Sys.time()))
            return(ans)
        }
        # insert new
        observeEvent(input$submit, {
            new <- checkForm()
            if(new){
                showModal(
                    modalDialog(
                        title = "Confirm Submission",
                        size = 'l',
                        # Original comment preview
                        div(class = "original-comment-preview",
                            div(class = "original-comment-header",
                                div(class = "original-comment-title",
                                    input$title),
                                div(class="comment-author",
                                    paste("by", input$uid)),
                                div('with email'),
                                div(class='comment-email',
                                    input$email)
                            ),
                            tags$i(
                                class = "fas fa-quote-left",
                                style = "color: #667eea; font-size: 14px;"),
                            div(class = "original-comment-content",
                                markdown(input$comment)
                            )
                        ),
                        p("Are you sure you want to submit this issue?"),
                        footer = tagList(
                            actionButton(NS(id, "confirmSubmitBtn"), "Confirm"), # This is the "Yes" button
                            modalButton("Cancel") # This is the "No" button
                        )
                    )
                ) 
            }
        })
        observeEvent(input$confirmSubmitBtn, {
            removeModal()
            ## new issue
            new <- checkForm()
            ## check the intervals
            new <- new & checkInterval()
            ## check the total record in database
            if(countComments()>.globals$totalComments){
                showNotification('Out of storage! Please notify the admin about this issue.',
                                 type = 'error')
                new <- FALSE
            }
            if(new){
                insertComments(uid=input$uid,
                               email=input$email,
                               title=input$title,
                               comment=input$comment,
                               dataset=dataSource()$dataset)
                updateCommentList()
            }
        })
        ## upvote comments
        observeEvent(input$upvote_clicked, {
            id_to_upvote <- as.numeric(gsub("upvote_button_", "", input$upvote_clicked))
            if(is.numeric(id_to_upvote)){
                updateCommentsVote(id_to_upvote)
                updateCommentList()
            }
        })
        ## close comments
        observeEvent(input$close_clicked, {
            id_to_close <- as.numeric(gsub("^.*?close_button_", "", input$close_clicked))
            if(is.numeric(id_to_close)){
                info <- getCommentsById(id_to_close)
                ns <- NS(id)
                keyword <- ifelse(as.logical(info$open), 'Close', 'Open')
                showModal(
                    modalDialog(
                        title = paste("Confirm", keyword),
                        size = 'l',
                        # Original comment preview
                        div(class = "original-comment-preview",
                            div(class = "original-comment-header",
                                div(class = "original-comment-title",
                                    info$title),
                                div(class="comment-author",
                                    paste("by", info$uid)),
                                div('with email'),
                                div(class='comment-email',
                                    textInput(ns('close_email'), 
                                              label = NULL,
                                              placeholder = "used as password"))
                            ),
                            tags$i(
                                class = "fas fa-quote-left",
                                style = "color: #667eea; font-size: 14px;"),
                            div(class = "original-comment-content",
                                markdown(info$comment)
                            )
                        ),
                        p(paste("Are you sure you want to", keyword, "this comment?")),
                        footer = tagList(
                            actionButton(NS(id, "confirmCloseBtn"), keyword), # This is the "Yes" button
                            modalButton("Cancel") # This is the "No" button
                        )
                        
                    )
                )
            }
        })
        observeEvent(input$confirmCloseBtn, {
            if(checkEmail('close_')){
                id_to_close <- as.numeric(gsub("^.*?close_button_", "",
                                               input$close_clicked))
                if(is.numeric(id_to_close)){
                    info <- getCommentsById(id_to_close)
                    if(isTRUE(dataSource()$auth$privilege=='all')){
                        updateComments(id_to_close, 'open', 1-info$open)
                        removeModal()
                        updateCommentList()
                    }else{
                        if(info$email == input$close_email & checkInterval()){
                            updateComments(id_to_close, 'open', 0)
                            removeModal()
                            updateCommentList()
                        }
                    }
                }
            }
        })
        ## delete comments
        observeEvent(input$delete_clicked, {
            id_to_delete <- as.numeric(gsub("^.*?delete_button_", "", input$delete_clicked))
            if(is.numeric(id_to_delete)){
                info <- getCommentsById(id_to_delete)
                ns <- NS(id)
                showModal(
                    modalDialog(
                        title = "Confirm delete",
                        size = 'l',
                        # Original comment preview
                        div(class = "original-comment-preview",
                            div(class = "original-comment-header",
                                div(class = "original-comment-title",
                                    info$title),
                                div(class="comment-author",
                                    paste("by", info$uid))
                            ),
                            tags$i(
                                class = "fas fa-quote-left",
                                style = "color: #667eea; font-size: 14px;"),
                            div(class = "original-comment-content",
                                markdown(info$comment)
                            )
                        ),
                        p("Are you sure you want to delete this comment?"),
                        footer = tagList(
                            actionButton(NS(id, "confirmDeleteBtn"), "Delete"), # This is the "Yes" button
                            modalButton("Cancel") # This is the "No" button
                        )
                        
                    )
                )
            }
        })
        observeEvent(input$confirmDeleteBtn, {
            id_to_delete <- as.numeric(gsub("^.*?delete_button_", "", input$delete_clicked))
            if(is.numeric(id_to_delete)){
                if(isTRUE(dataSource()$auth$privilege=='all') & checkInterval()){
                    deleteComments(id_to_delete)
                    removeModal()
                    updateCommentList()
                }
            }
        })
        ## edit comments
        observeEvent(input$edit_clicked, {
            id_to_edit <- as.numeric(gsub("^.*?edit_button_", "", input$edit_clicked))
            if(is.numeric(id_to_edit)){
                info <- getCommentsById(id_to_edit)
                ns <- NS(id)
                showModal(
                    modalDialog(
                        title = paste("Edit issue: ", info$id),
                        size = 'l',
                        div(
                            # Edit form section
                            div(class = "newcomment-form-container",
                                # Hidden fields
                                div(style='display: none;', 
                                    numericInput(ns('edit_id'),
                                                 label = NULL,
                                                 value = info$id)),
                                # User info row
                                div(class = "newcomment-form-row",
                                    div(class = "newcomment-form-group-modern has-icon",
                                        tags$i(class = "newcomment-input-icon fas fa-user"),
                                        tags$label("for" = ns('edit_uid'),
                                                   class = "required-field", "Your Name"),
                                        div(class = "newcomment-form-control-wrapper",
                                            textInput(ns('edit_uid'), 
                                                      label = NULL,
                                                      value = info$uid)
                                        )
                                    ),
                                    div(class = "newcomment-form-group-modern has-icon",
                                        tags$i(class = "newcomment-input-icon fas fa-envelope"),
                                        tags$label("for" = ns('edit_email'),
                                                   class = "required-field", 
                                                   "Email Address"),
                                        div(class = "newcomment-form-control-wrapper",
                                            textInput(ns('edit_email'), 
                                                      label = NULL,
                                                      placeholder = "used as password")
                                        )
                                    )
                                ),
                                # Comment title
                                div(class = "newcomment-title-input-container",
                                    div(class = "newcomment-form-group-modern",
                                        tags$label("for" = ns('edit_title'),
                                                   class = "required-field",
                                                   "Comment Title"),
                                        div(class = "newcomment-form-control-wrapper",
                                            textInput(ns('edit_title'), 
                                                      label = NULL,
                                                      value = info$title,
                                                      width = "100%")
                                        )
                                    )
                                ),
                                # Comment field
                                div(class = "newcomment-input-container",
                                    tags$label("for" = ns('edit_comment'),
                                               class = "required-field",
                                               "Your Reply"),
                                    div(class = "form-control-wrapper",
                                        textAreaInput(ns('edit_comment'),
                                                      label = NULL,
                                                      value = info$comment,
                                                      rows = 5,
                                                      width = "100%")
                                    )
                                )
                            )
                        ),
                        
                        # Preview section
                        div(class = "newcomment-preview-container",
                            div(class = "newcomment-preview-header",
                                tags$i(class = "fas fa-eye"),
                                "Markdown Comment Preview"
                            ),
                            div(class = "newcomment-preview-content",
                                uiOutput(ns('edit_previewbox'))
                            )
                        ),
                        
                        footer = tagList(
                            actionButton(NS(id, "submitEditBtn"), "Submit"), # This is the "Yes" button
                            modalButton("Cancel") # This is the "No" button
                        )
                    )
                )
            }
        })
        observeEvent(input$submitEditBtn, {
            id_to_edit <- as.numeric(input$edit_id)
            if(is.numeric(id_to_edit)){
                edit <- checkForm(prefix = 'edit_')
                edit <- edit & checkInterval()
                if(edit){
                    info <- getCommentsById(id_to_edit)
                    if(info$email==input$edit_email){
                        removeModal()
                        null <- lapply(c('uid', 'title', 'comment'), function(coln){
                            val <- input[[paste0('edit_', coln)]]
                            if(val!=info[[coln]]){
                                updateComments(info$id, coln, val)
                            }
                        })
                        # refresh table
                        updateCommentList()
                    }else{
                        showNotification("The email address must match exactly with the record.",
                                         type = 'error')
                    }
                }
            }
        })
        ## reply comments
        observeEvent(input$reply_clicked, {
            ns <- NS(id)
            id_to_reply <- as.numeric(gsub("^.*?reply_button_", "", input$reply_clicked))
            if(is.numeric(id_to_reply)){
                info <- getCommentsById(id_to_reply)
                if(nrow(info)){
                    showModal(
                        modalDialog(
                            title = NULL,  # We'll create a custom header
                            size = "xl",
                            
                            # Custom modal content
                            div(
                                # Modal header
                                div(class = "reply-header-content",
                                    div(class = "reply-title",
                                        tags$i(class = "fas fa-reply",
                                               style = "font-size: 20px;"),
                                        "Reply to Comment"
                                    ),
                                    div(class = "reply-subtitle",
                                        paste("Responding to:",
                                              substr(info$uid, 1, 60),
                                              if(nchar(info$uid) > 60){
                                                  "..." } else {""})
                                    )
                                )
                            ),
                            
                            # Modal body
                            div(
                                # Original comment preview
                                div(class = "original-comment-preview",
                                    div(class = "original-comment-header",
                                        tags$i(
                                            class = "fas fa-quote-left",
                                            style = "color: #667eea; font-size: 14px;"),
                                        div(class = "original-comment-title",
                                            info$title)
                                    ),
                                    div(class = "original-comment-content",
                                        markdown(info$comment)
                                    )
                                ),
                                
                                # Reply form section
                                div(class = "newcomment-form-container",
                                    div(class = "newcomment-form-header",
                                        tags$i(class = "fas fa-pen",
                                               style = "color: #667eea;"),
                                        h4(class = "form-section-title",
                                           "Your Reply")
                                    ),
                                    
                                    # Hidden fields
                                    div(style = "display: none;",
                                        numericInput(ns('reply_id'),
                                                     label = NULL,
                                                     value = info$id),
                                        textInput(ns('reply_title'),
                                                  label = NULL,
                                                  value = info$title),
                                        textInput(ns('reply_dataset'),
                                                  label = NULL, 
                                                  value = info$dataset)
                                    ),
                                    
                                    # User info row
                                    div(class = "newcomment-form-row",
                                        div(class = "newcomment-form-group-modern has-icon",
                                            tags$i(class = "newcomment-input-icon fas fa-user"),
                                            tags$label("for" = ns('reply_uid'),
                                                       class = "required-field", "Your Name"),
                                            div(class = "newcomment-form-control-wrapper",
                                                textInput(ns('reply_uid'), 
                                                          label = NULL,
                                                          placeholder = "Enter your name")
                                            ),
                                            div(class = "newcomment-form-info",
                                                tags$i(class = "fas fa-info-circle"),
                                                "We'll email you only if clarification is needed.")
                                        ),
                                        div(class = "newcomment-form-group-modern has-icon",
                                            tags$i(class = "newcomment-input-icon fas fa-envelope"),
                                            tags$label("for" = ns('reply_email'),
                                                       class = "required-field", 
                                                       "Email Address"),
                                            div(class = "newcomment-form-control-wrapper",
                                                textInput(ns('reply_email'), 
                                                          label = NULL,
                                                          placeholder = "your.email@example.com")
                                            ),
                                            div(class = "newcomment-form-info",
                                                tags$i(class = "fas fa-info-circle"),
                                                "Email acts as edit password.")
                                        )
                                    ),
                                    
                                    # Comment field
                                    div(class = "newcomment-input-container",
                                        tags$label("for" = ns('reply_comment'),
                                                   class = "required-field",
                                                   "Your Reply"),
                                        div(class = "form-control-wrapper",
                                            textAreaInput(ns('reply_comment'),
                                                          label = NULL,
                                                          placeholder = "Share your thoughts, provide additional information, or ask follow-up questions...",
                                                          rows = 5,
                                                          width = "100%")
                                        )
                                    )
                                )
                            ),
                            
                            # Preview section
                            div(class = "newcomment-preview-container",
                                div(class = "newcomment-preview-header",
                                    tags$i(class = "fas fa-eye"),
                                    "Markdown Comment Preview"
                                ),
                                div(class = "newcomment-preview-content",
                                    uiOutput(ns('reply_previewbox'))
                                )
                            ),
                            
                            # Custom footer
                            footer = div(class = "reply-actions",
                                         div(class = "newcomment-form-info",
                                             tags$i(class = "fas fa-info-circle"),
                                             "Your reply will be posted publicly"
                                         ),
                                         div(style = "display: flex; gap: 12px;",
                                             actionButton(
                                                 ns('cancelReplyBtn'),
                                                 class = "btn-reply-cancel",
                                                 "Cancel"
                                             ),
                                             actionButton(ns("submitReplyBtn"), 
                                                          tagList(
                                                              tags$i(class = "fas fa-paper-plane"),
                                                              "Post Reply"
                                                          ),
                                                          class = "btn-reply-submit")
                                         )
                            ),
                            
                            # Add modal class for styling
                            div(class = "reply-modal", style = "display: none;")
                            
                        )
                    )
                }
            }
        })
        observeEvent(input$cancelReplyBtn, {
            removeModal()
        })
        observeEvent(input$submitReplyBtn, {
            id_to_reply <- as.numeric(input$reply_id)
            if(is.numeric(id_to_reply)){
                reply <- checkForm(prefix = 'reply_')
                reply <- reply & checkInterval()
                if(reply){
                    removeModal()
                    insertComments(uid=input$reply_uid,
                                   email=input$reply_email,
                                   title=input$reply_title,
                                   comment=input$reply_comment,
                                   dataset=input$reply_dataset,
                                   pid=input$reply_id)
                    # refresh table
                    updateCommentList() 
                }
            }
        })
    })
}

# Plot proportion plot
#' @importFrom ggplot2 ggplot aes .data geom_col ylab coord_flip xlab
#' scale_fill_manual theme
#' @importFrom data.table .SD .N
scProp <- function(
        inpConf,
        inpMeta,
        infoX,
        infoY,
        subsetCellKey,
        subsetCellVal,
        inptyp,
        flipXY,
        labelsFontsize = 24,
        labelsFontFamily = 'Helvetica',
        dataset,
        geneIdMap,
        valueFilterKey,
        valueFilterCutoff,
        valueFilterCutoff2,#not use
        reorder=FALSE,
        orderX,
        orderY) {
    # Prepare ggData
    subsetCellKey <- subsetCellKey[subsetCellKey!="N/A"]
    subsetCellVal <- namedSubsetCellVals(subsetCellKey, subsetCellVal)
    if (sum(lengths(subsetCellVal))) {
        colN <- unique(c(
            inpConf[inpConf$UI == infoX]$ID,
            inpConf[inpConf$UI == infoY]$ID,
            inpConf[inpConf$UI %in% subsetCellKey]$ID))
    } else{
        colN <- c(
            inpConf[inpConf$UI == infoX]$ID,
            inpConf[inpConf$UI == infoY]$ID)
    }
    ggData <- inpMeta[, colN, with = FALSE]
    ggData <- subGrp(ggData, subsetCellKey, subsetCellVal, inpConf)
    subFilterColname <- 'subValue'
    ggData <-
        cbindFilterValues(
            ggData,
            inpConf,
            inpMeta,
            subFilterColname,
            geneIdMap,
            dataset,
            valueFilterKey,
            valueFilterCutoff
        )
    keep <- filterCells(
        ggData,
        valueFilterKey = subFilterColname,
        valueFilterCutoff = valueFilterCutoff,
        inpConf=inpConf)
    
    if(ncol(ggData)==1){
        ggData <- cbind(ggData, ggData) ## X and color grp are same.
    }
    ggData <- ggData[keep, c(1, 2), drop = FALSE]
    colnames(ggData) <- c("X", "grp")
    ggData <- ggData[, list(nCells = .N), by = c("X", "grp")]
    ggData <- ggData[, {
        tot <- sum(.SD$nCells)
        .SD[, list(
            pctCells = 100 * sum(.SD$nCells) / tot,
            nCells = .SD$nCells), by = "grp"]
    },
    by = "X"]
    
    # Do factoring
    if(reorder){
        ggData$X <- factor(ggData$X, levels=orderX)
        ggData$grp <- factor(ggData$grp, levels=orderY)
    }else{
        ggData <- relevelData(ggData, "grp")
        ggData <- relevelData(ggData, "X")
    }
    ggCol <- relevelCol(inpConf, infoY, ggData, "grp")
    
    # Actual ggplot
    if (inptyp == "Proportion") {
        ggOut <- ggplot(ggData, aes(
            .data[["X"]], .data[["pctCells"]], fill = .data[["grp"]])) +
            geom_col() + ylab("Cell Proportion (%)")
    } else {
        ggOut <- ggplot(ggData, aes(
            .data[["X"]], .data[["nCells"]], fill = .data[["grp"]])) +
            geom_col() + ylab("Number of Cells")
    }
    if (flipXY) {
        ggOut <- ggOut + coord_flip()
    }
    ggOut <- ggOut + xlab(infoX) +
        sctheme(
            base_size = labelsFontsize,
            family = labelsFontFamily,
            Xang = 45,
            XjusH = 1) +
        scale_fill_manual("", values = ggCol) +
        theme(legend.position = "right")
    return(ggOut)
}

scCoexp3dUI <- function(id, postfix = 1, subgrp=.globals$subsetgroup[1]) {
    subModuleContainerUI(
        id,
        mainSelectUI = tagList(
            selectInput(
                NS(id, "GeneName1"),
                "Gene 1:",
                choices = NULL,
                width = "100px"
            ),
            selectInput(
                NS(id, "GeneName2"),
                "Gene 2:",
                choices = NULL,
                width = "100px"
            )
        ),
        menuUI = contextMenuCoExprUI(id, plotly = TRUE, group = subgrp),
        contentUI = tagList(
            uiOutput(NS0(id, "GeneExpr3Doup.ui", 1)),
            downloadButton(
                NS(id, 'downloadExpr'),
                "Expression for clicked cell"),
            verbatimTextOutput(NS(id, 'clicked'))
        )
    )
}
scCoexp3dServer <- function(
        pid,
        id,
        dataSource,
        optCrt,
        p_input,
        p_session,
        postfix = 1) {
    moduleServer(id, function(input, output, session) {
        if (is.null(
            p_session$userData$defaults[[dataSource()$dataset]][[id]])) {
            defaults <- list(
                GeneName1 = dataSource()$sc1def$gene1,
                GeneName2 = dataSource()$sc1def$gene2
            )
        } else{
            defaults <-
                p_session$userData$defaults[[dataSource()$dataset]][[id]]
        }
        ## title
        updateSelectizeInput(
            session,
            "GeneName1",
            choices = sort(names(dataSource()$sc1gene)),
            server = TRUE,
            selected = defaults$GeneName1,
            options = list(
                maxOptions = .globals$maxNumGene,
                create = TRUE,
                persist = TRUE,
                render = I(optCrt)
            )
        )
        updateSelectizeInput(
            session,
            "GeneName2",
            choices = sort(names(dataSource()$sc1gene)),
            server = TRUE,
            selected = defaults$GeneName2,
            options = list(
                maxOptions = .globals$maxNumGene,
                create = TRUE,
                persist = TRUE,
                render = I(optCrt)
            )
        )
        
        subModuleMenuObservor(
            id,
            input,
            p_session,
            dataSource,
            c("GeneName1", "GeneName2", "CoExprcol1", "CoExprord1")
        )
        ## plot
        plot3d <- reactive({
            ## for some reason, the postfix changed to FALSE.
            scDRcoex(
                inpConf=dataSource()$sc1conf,
                inpMeta=dataSource()$sc1meta,
                dimRedX=p_input$GeneExprdrX,
                dimRedY=p_input$GeneExprdrY,
                gene1=input$GeneName1,
                gene2=input$GeneName2,
                subsetCellKey=p_input[[paste0("subsetCell",
                                              input[["CellInfosubgrp1"]])]],
                subsetCellVal=
                    getSubsetCellVal(p_input,
                                     group=input[["CellInfosubgrp1"]]),
                dataset=dataSource()$dataset,
                geneIdMap=dataSource()$sc1gene,
                plotType="3D",
                pointSize=p_input$GeneExprsiz,
                GeneExprDotCol=input$CoExprcol1,
                GeneExprDotOrd=input$CoExprord1,
                labelsFontsize=p_input$GeneExprfsz,
                labelsFontFamily=p_input$GeneExprfml,
                plotAspectRatio=p_input$GeneExprasp,
                keepXYlables=p_input$GeneExprtxt,
                valueFilterKey = p_input$filterCell,
                valueFilterCutoff = p_input$filterCellVal,
                valueFilterCutoff2 = p_input$filterCellVal2,
                hideFilterCell = input$CoExprhid1
            )
        })
        source <- NS0(NS(pid, id), "GeneExpr3Doup", 1)
        output$GeneExpr3Doup1 <- renderPlotly({
            ggplotly(plot3d()) %>% event_register("plotly_click")
        })
        output$GeneExpr3Doup.ui1 <- renderUI({
            plotlyOutput(
                source,
                height = .globals$pList1[p_input$GeneExprpsz])
        })
        output$downloadExpr <- exprDownloadHandler(
            dataSource()$sc1gene,
            dataSource()$dataset,
            dataSource()$sc1meta)
        output$clicked <- plotly3d_click(session)
    })
}

# sqlite operation
#' @importFrom DBI dbConnect dbDisconnect dbWriteTable
#'  dbGetQuery dbSendQuery dbListTables dbClearResult
#'  sqlInterpolate dbListFields dbAppendTable
#' @importFrom RSQLite SQLite
getDBconn <- function(){
    dbConnect(SQLite(),
              dbname = file.path(.globals$app_path,
                                 .globals$credential_path))
}
connectDB <- function(FUN, ...){
    con <- getDBconn()
    on.exit(dbDisconnect(con))
    FUN(conn = con, ...)
}
sendNoreplyQueryToDB <- function(...){
    conn <- getDBconn()
    on.exit(dbDisconnect(conn))
    res <- dbSendQuery(conn = conn, ...)
    dbClearResult(res)
}
#' @importFrom shinymanager read_db_decrypt
isEncrypted <- function(){
    db <- connectDB(read_db_decrypt, name = .globals$credentialTableName)
    identical(names(db), c("value", "iv"))
}
getCredential <- function(){
    res <- connectDB(read_db_decrypt, 
                     .globals$credentialTableName,
                     .globals$passphrase)
}
tableExists <- function(tableName){
    tableName %in% connectDB(dbListTables)
}
getConfigTable <- function(){
    res <- connectDB(read_db_decrypt, .globals$configTableName)
}
replaceNULL <- function(x, by=NA){
    ifelse(is.null(x), by, x)
}
createConfigTable <- function(appconf){
    appData <- lapply(appconf, FUN = function(.ele){
        if(!is(.ele, 'APPconf')){
            stop('Please check the folder privilege')
        }
        c(title=.ele$title,
          id=.ele$id,
          species=.ele$species,
          type=.ele$type,
          markers=paste(markers(.ele), collapse = .globals$configTableSep),
          keywords=paste(.ele$keywords, collapse = .globals$configTableSep),
          groupCol=paste(.ele$groupCol, collapse = .globals$configTableSep),
          ref_bib=trimBib(replaceNULL(.ele$ref$bib)),
          ref_doi=replaceNULL(.ele$ref$doi),
          ref_pmid=replaceNULL(.ele$ref$pmid),
          ref_title=replaceNULL(.ele$ref$entry$title),
          ref_author=paste(.ele$ref$entry$author,
                           collapse = .globals$configTableSep),
          ref_year=replaceNULL(.ele$ref$entry$year),
          ref_journal=replaceNULL(.ele$ref$entry$journal),
          ref_abstract=replaceNULL(.ele$ref$entry$abstract))
    })
    appData <- do.call(rbind, appData)
    if(length(dim(appData))==2){
        colnames(appData) <- c('title', 'id', 'species', 'type', 'markers',
                               'keywords', 'groupCol', 'ref_bib', 'ref_doi',
                               'ref_pmid', 'ref_title', 'ref_author',
                               'ref_year', 'ref_journal', 'ref_abstract')
        appData <- as.data.frame(appData)
        appData$locker <- vapply(appData$id, FUN = checkLocker,
                                 FUN.VALUE = logical(1L))
        connectDB(dbWriteTable, name = .globals$configTableName,
                  value = appData, overwrite = TRUE)
    }else{
        stop('No data available.')
    }
}

updateConfigTable <- function(appconf){
    if(missing(appconf)){
        appconf <- getAppConfObj(privilege = 'all')
        createConfigTable(appconf)
    }else{
        if(is(appconf, 'APPconf')){
            appconf <- list(ele=appconf)
        }
        lapply(appconf, function(.ele){
            if(!is(.ele, 'APPconf')){
                stop('Please check the folder privilege')
            }
            query <- paste0('UPDATE ', 
                           .globals$configTableName,
                           ' SET `title`="', .ele$title, '",',
                           ' `species`="', .ele$species, '",',
                           ' `type`="', .ele$type, '",',
                           ' `markers`="',
                           paste(markers(.ele),
                                 collapse = .globals$configTableSep), '",',
                           ' `keywords`="', 
                           paste(.ele$keywords,
                                 collapse = .globals$configTableSep), '",',
                           ' `groupCol`="',
                           paste(.ele$groupCol,
                                 collapse = .globals$configTableSep), '",',
                           ' `ref_bib`="',
                           magicQuote(trimBib(replaceNULL(.ele$ref$bib))), '",',
                           ' `ref_doi`="',
                           replaceNULL(.ele$ref$doi), '",',
                           ' `ref_pmid`="',
                           replaceNULL(.ele$ref$pmid), '",',
                           ' `ref_title`="',
                           replaceNULL(.ele$ref$entry$title), '",',
                           ' `ref_author`="',
                           paste(.ele$ref$entry$author, 
                                 collapse = .globals$configTableSep), '",',
                           ' `ref_year`="',
                           replaceNULL(.ele$ref$entry$year), '",',
                           ' `ref_journal`="',
                           replaceNULL(.ele$ref$entry$journal), '",',
                           ' `ref_abstract`="',
                           replaceNULL(.ele$ref$entry$abstract), '"',
                           ' WHERE id="', .ele$id, '"')
            sendNoreplyQueryToDB(statement = query)
        })
    }
}

updateConfigTblKey <- function(key, feild, value){
    if(missing(key) || missing(feild)){
        stop('Not proper query statement.')
    }
    query <- paste0('UPDATE ', 
                   .globals$configTableName,
                   ' SET ', feild, '="', value, '"',
                   ' WHERE id="', key, '"')
    sendNoreplyQueryToDB(statement = query)
}
updateConfigTblLocker <- function(key, value){
    updateConfigTblKey(key = key, feild = 'locker', value = as.numeric(value))
}

checkKeyFromConfig <- function(key, feild, unique=TRUE){
    if(missing(key) || length(key)==0){
        # list full references
        query <- paste('SELECT', feild, 'FROM', .globals$configTableName)
    }else{
        # list references by key
        if(length(key)==1){
            query <- paste0('SELECT ', feild, ' FROM ',
                            .globals$configTableName,
                            ' WHERE id="', key, '"')
        }else{
            query <- paste0('SELECT ', feild, ' FROM ',
                            .globals$configTableName,
                            ' WHERE id in ("',
                            paste(key, collapse = '","'), '")')
        }
    }
    res <- unlist(connectDB(dbGetQuery, statement = query))
    if(unique){
        unique(sort(res))
    }else{
        res
    }
}
# list one or full references
listReferences <- function(key){
    ref <- checkKeyFromConfig(key, feild='ref_bib')
    ref <- ref[ref!=""]
    ref
}
makeSortedUnique <- function(x){
    if(length(names(x))==length(x)){
        x <- x[order(names(x))]
    }else{
        x <- sort(x)
    }
    x <- x[!duplicated(x)]
    x
}
# list all datasets
listDatasets <- function(key, privilege='', named=FALSE){
    if(is.null(privilege)){
        privilege <- ''
    }
    ds <- checkKeyFromConfig(key, feild='id', unique=FALSE)
    ds <- unname(ds)
    if(named){
        dsn <- checkKeyFromConfig(key, feild = 'title', unique = FALSE)
        names(ds) <- dsn
    }
    if(privilege=='all'){
        return(makeSortedUnique(ds))
    }
    locker <- checkKeyFromConfig(key, feild = 'locker', unique=FALSE)
    if(any(locker)){
        keep <- mapply(function(.ds, .locker){
            !.locker || grepl(.ds, privilege, fixed = TRUE)
        }, ds, locker)
        return(makeSortedUnique(ds[keep]))
    }else{
        return(makeSortedUnique(ds))
    }
}
# list all species
listSpecies <- function(key){
    checkKeyFromConfig(key, feild='species')
}

intToBin <- function(x) {
    y <- as.integer(x)
    class(y) <- "binmode"
    y <- as.character(y)
    dim(y) <- dim(x)
    y
}

ip2bin <- function(x){
    x <- strsplit(x, "\\.")
    x <- do.call(rbind, x)
    mode(x) <- "integer"
    x <- intToBin(x)
    mode(x) <- "integer"
    d <- dim(x)
    x <- sprintf("%08d", x)
    x <- matrix(x, nrow = d[1], ncol = d[2], byrow = FALSE)
    x <- apply(x, 1, paste, collapse="")
    y <- strsplit(x, "")
    y <- do.call(rbind, y)
    mode(y) <- "integer"
    apply(y, 1, function(.y) sum(.y * 2^rev((seq_along(.y)-1))))
}

touchIPtable <- function(){
    if(!is.null(.globals$IPlocationFilename)){
        if(file.exists(file.path(.globals$app_path,
                                 .globals$IPlocationFilename))){
            iptable <- readRDS(file.path(.globals$app_path,
                                         .globals$IPlocationFilename))
            connectDB(dbWriteTable, name = .globals$IPlocationTablename,
                      value = iptable, overwrite = TRUE)
            query <- paste0('CREATE UNIQUE INDEX `from` ON ',
                            .globals$IPlocationTablename, '(`from`)')
            sendNoreplyQueryToDB(statement = query)
        }
    }
}
getIPtable <- function(ips){
    if(!tableExists(.globals$IPlocationTablename)){
        return(data.frame(ip=unique(ips)))
    }
    ips <- unique(ips)
    ips <- ips[!is.na(ips)]
    ips <- ips[ips!='']
    if(length(ips)==0){
        return(data.frame(ip=ips))
    }
    ipb <- ip2bin(ips)
    res <- lapply(ipb, function(.ip){
        query <- paste0('SELECT * FROM ',
                        .globals$IPlocationTablename,
                        ' INNER JOIN (SELECT MAX(`from`) AS `start` FROM ',
                        .globals$IPlocationTablename,
                        ' WHERE `from` <= ', .ip,
                        ') AS f ON ', .globals$IPlocationTablename,
                        '.`from` = f.`start`',
                        ' WHERE `to` >= ', .ip)
        connectDB(dbGetQuery, statement = query)
    })
    names(res) <- as.character(ips)
    res <- do.call(rbind, res)
    res$from <- rownames(res)
    res$to <- NULL
    res$start <- NULL ## remove the start column introduced by the inner join
    colnames(res)[colnames(res)=='from'] <- 'ip'
    res
}
touchVisitorTable <- function(){
    counter <- NULL
    if(!tableExists(.globals$counterTableName)){
        counter <- read.delim(.globals$counterFilename, header = TRUE)
    }else{
        query <- paste('SELECT * FROM',
                       .globals$counterTableName)
        counter <- 
            connectDB(dbGetQuery,
                      statement = query)
    }
    if(!tableExists(.globals$IPlocationTablename)){
        touchIPtable()
    }
    if(length(counter)){
        if(!all(c('latitude', 'longitude', 'region', 'city') %in%
                colnames(counter))){
            counterIP <- getIPtable(counter$ip)
            counter <- merge(counter, counterIP, by='ip', all.x = TRUE)
            counter <- counter[order(counter$date), ]
            connectDB(dbWriteTable, name = .globals$counterTableName,
                      value = counter, overwrite = TRUE)
        }
    }
}
## visitor table
updateVisitorTable <- function(input, output, session){
    touchVisitorTable()
    ## update visitor stats
    update_visitor <- function(){
        req(input$remote_addr)
        current <- Sys.time()
        ip <- isolate(input$remote_addr)
        agent <- isolate(input$remote_agent)
        ## check ip and time not within 10 min
        query <- paste0('SELECT `date` FROM ',
                        .globals$counterTableName,
                        ' WHERE `date` BETWEEN "',
                        current, '" AND "', current+600,
                        '" AND `ip`="', ip, '"')
        res <- connectDB(dbGetQuery, statement = query)
        if(nrow(res)==0){
            ip2loc <- getIPtable(ip)
            ip2loc$date <- as.character(current)
            ip2loc$agent <- agent
            th <- paste(colnames(ip2loc), collapse = "`,`")
            bd <- paste(ip2loc[1, ], collapse = '","')
            query <- paste0('INSERT INTO ', .globals$counterTableName,
                            ' (`', th, '`) VALUES ("', bd, '")')
            sendNoreplyQueryToDB(statement = query)
        }
    }
    observeEvent(input$remote_addr, update_visitor())
    output$total_visitor <- renderPlot({
        counter <- listVisitors(summary=TRUE)
        ggplot(counter, aes(x=.data[["time"]], y=.data[["total"]])) +
            geom_bar(stat = "identity", fill="darkorchid4") +
            theme_minimal() + xlab("") + ylab("visitor counts") +
            theme(axis.text.x = element_text(angle = 90, vjust = 0.5, hjust=1))
    })
}
listVisitors <- function(summary=FALSE, ipCounter=FALSE){
    if(summary){
        current <- Sys.time()
        # select ~730 day data 63072000 = 730*60*60*24
        query <- paste0('SELECT strftime("%Y-%m", `date`) AS `time`,',
                        ' count(`ip`) AS total,',
                        ' count(DISTINCT `ip`) AS uniqueIP FROM ',
                        .globals$counterTableName,
                        ' WHERE `date` BETWEEN "',
                        current - 63072000, '" AND "', current,
                        '" GROUP BY `time`')
    }else{
        if(ipCounter){
            query <- paste0('SELECT count(`ip`) AS total, ',
                            ' count(DISTINCT `ip`) AS uniqueIP FROM ',
                            .globals$counterTableName)
        }else{
            query <- paste0('SELECT * FROM ',
                            .globals$counterTableName)
        }
    }
    counter <- connectDB(dbGetQuery, query)
}

## comments table
touchCommentTable <- function(){
    createCommentTable <- function(){
        sql <- paste('CREATE TABLE IF NOT EXISTS', .globals$commentsTableName, 
                     '(id INTEGER PRIMARY KEY,',
                     'uid TEXT NOT NULL,',
                     'email TEXT NOT NULL,',
                     'pid INTEGER DEFAULT 0,',
                     'title TEXT NOT NULL,',
                     'comment TEXT NOT NULL,',
                     'dataset TEXT,',
                     'vote INTEGER NOT NULL DEFAULT 0,',
                     'open INTEGER NOT NULL DEFAULT 1,',
                     "created_at TEXT NOT NULL DEFAULT (strftime('%Y-%m-%d %H:%M:%S', 'now', 'localtime')),",
                     "updated_at TEXT NOT NULL DEFAULT (strftime('%Y-%m-%d %H:%M:%S', 'now', 'localtime')))")
        sendNoreplyQueryToDB(statement=sql)
    }
    if(!tableExists(.globals$commentsTableName)){
        createCommentTable()
    }else{
        column_names <- connectDB(dbListFields, name=.globals$commentsTableName)
        if(!all(c('id', 'uid', 'email', 'pid', 'title',
                  'comment', 'dataset', 'vote', 'open',
                  'created_at', 'updated_at') %in%
                column_names)){
            availableComments <- listComments(page_size = .globals$totalComments,
                                              full = TRUE, all=TRUE,
                                              touch = FALSE)
            sendNoreplyQueryToDB(paste0("DROP TABLE IF EXISTS ",
                                        .globals$commentsTableName))
            createCommentTable()
            if(nrow(availableComments)>0){
                if(!'pid' %in% colnames(availableComments)){
                    availableComments$pid <- availableComments$id
                }
                if(!'vote' %in% colnames(availableComments)){
                    availableComments$vote <- 0
                }
                connectDB(dbAppendTable, name=.globals$commentsTableName,
                          value=availableComments)
            }
        }
    }
}
countComments <- function(){
    sql <- paste('SELECT COUNT(*) FROM', .globals$commentsTableName)
    connectDB(dbGetQuery, sql)[[1]]
}
distinctCommentsTitles <- function(){
    touchCommentTable()
    sql <- paste("SELECT DISTINCT title FROM",
                 .globals$commentsTableName,
                 "ORDER BY updated_at DESC")
    connectDB(dbGetQuery, sql)$title
}
listComments <- function(page_size, full=FALSE, all=FALSE, touch=TRUE){
    if(touch) touchCommentTable()
    col <- ifelse(full,
                  '*',
                  'id, uid, title, comment, created_at, updated_at, pid, vote')
    where <- ifelse(all, 
                    "", " WHERE open=1")
    query <- paste0("SELECT ", col, " FROM ", .globals$commentsTableName,
                    where,
                    " ORDER BY updated_at DESC LIMIT ", page_size)
    res <- connectDB(dbGetQuery, query)
    if(any(is.na(res$pid))){
        res$pid[is.na(res$pid)] <- res$id[is.na(res$pid)]
    }
    if(any(res$pid==0)){
        res$pid[res$pid==0] <- res$id[res$pid==0]
    }
    res
}
getCommentsById <- function(id){
    query <- paste0("SELECT * FROM ",
                    .globals$commentsTableName,
                    " WHERE id='", id,"'")
    connectDB(dbGetQuery, query)
}
updateComments<- function(id, coln, val){
    con <- getDBconn()
    on.exit(dbDisconnect(con))
    sql <- paste0('UPDATE ', .globals$commentsTableName,
                  " SET ?coln = ?val,",
                  " updated_at = strftime('%Y-%m-%d %H:%M:%S', 'now', 'localtime')",
                  " WHERE id=?id")
    query <- sqlInterpolate(
        conn = con, sql,
        coln=coln, val=val, id=id
    )
    sendNoreplyQueryToDB(statement=query)
}
updateCommentsVote <- function(id){
    stopifnot(is.numeric(id))
    sql <- paste0('UPDATE ', .globals$commentsTableName,
                  " SET `vote` = `vote` + 1",
                  " WHERE id='", id, "'")
    sendNoreplyQueryToDB(statement=sql)
}
insertComments <- function(uid, email, title, comment, dataset, pid){
    touchCommentTable()
    con <- getDBconn()
    on.exit(dbDisconnect(con))
    if(missing(pid)){
        pid <- 0
    }
    sql <- paste0('INSERT INTO ', .globals$commentsTableName,
                  ' (`uid`, `email`, `title`, `comment`, `dataset`, `pid`) ',
                  'VALUES (?uid, ?email, ?title, ?comment, ?dataset, ?pid)')
    query <- sqlInterpolate(
        conn = con, sql,
        uid = uid,
        email = email,
        title = title,
        comment = comment,
        dataset = dataset,
        pid = pid
    )
    res <- dbSendQuery(conn = con, statement = query)
    dbClearResult(res)
}
deleteComments <- function(id){
    sql <- paste("WITH RECURSIVE to_delete(id) AS (",
    "SELECT id FROM", .globals$commentsTableName, "WHERE id=", id,
    "UNION ALL SELECT t.id FROM", .globals$commentsTableName,
    "t INNER JOIN to_delete d ON t.pid = d.id)",
    "DELETE FROM", .globals$commentsTableName,
    "WHERE id IN (SELECT id FROM to_delete)")
    sendNoreplyQueryToDB(sql)
    # sql <- paste0("DELETE FROM ", .globals$commentsTableName,
    #               " WHERE id IN '", id, "' OR pid = '", id, "'")
    # sendNoreplyQueryToDB(statement=sql)
}
## gene table
## gene name, expressed datasets
touchGeneTable <- function(updateDB=FALSE){
    if(updateDB || !tableExists(.globals$geneSymbolTableName)){
        datasets <- listDatasets()
        symbols <- lapply(datasets, function(.ele) {
            names(readData("sc1gene", .ele))
        })
        symbols <- data.frame(dataset = rep(datasets, lengths(symbols)),
                              symbol = unlist(symbols))
        symbols$expr <- NA
        connectDB(dbWriteTable, name = .globals$geneSymbolTableName,
                  value = symbols, overwrite = TRUE)
    }
}
listGeneSymbols <- function(genes, datasets, like = FALSE, checkExpr = FALSE){
    query <- paste0('SELECT DISTINCT `symbol` FROM ',
                    .globals$geneSymbolTableName)
    where <- NULL
    if(!missing(datasets)){
        datasets <- datasets[!is.null(datasets)]
        datasets <- datasets[!is.na(datasets)]
        if(length(datasets)==1){
            where <- c(where, paste0('`dataset` = "', datasets, '"'))
        }else{
            if(length(datasets)>1){
                where <- c(where,
                           paste0('`dataset` IN ("', 
                                  paste(datasets, collapse = '", "'), '")'))
            }
        }
    }
    if(!missing(genes)){
        genes <- genes[!is.null(genes)]
        genes <- genes[!is.na(genes)]
        genes <- tolower(genes)
        if(length(genes)==1){
            if(like){
                where <- c(where, paste0('LOWER(`symbol`) LIKE "', genes, '"'))
            }else{
                where <- c(where, paste0('LOWER(`symbol`) = "', genes, '"'))
            }
        }else{
            if(length(genes)>1){
                where <- c(where, paste0('LOWER(`symbol`) IN ("', 
                           paste(genes, collapse = '", "'), '")'))
            }
        }
    }
    if(checkExpr){
        ## default is NULL
        ## when user do search, if all expr is 0, the value will be set to 0
        ## and the updated value will not be NULL anymore.
        where <- c(where, '`expr` is NULL')
    }
    if(length(where)){
        query <- paste(query, 'WHERE', paste(where, collapse = ' AND '))
    }
    res <- connectDB(dbGetQuery, statement = query)
    res$symbol
}

setGeneExprForData <- function(symbol, dataset, expr){
    if(length(symbol)!=length(dataset)){
        return(NULL)
    }
    mapply(FUN=function(.symbol, .dataset, .expr){
        query <- paste0('UPDATE ', .globals$geneSymbolTableName,
                        ' SET `expr` = ', .expr,
                        ' WHERE `symbol` = "', .symbol, '"',
                        ' AND `dataset` = "', .dataset, '"')
        sendNoreplyQueryToDB(statement = query)
    }, symbol, dataset, expr)
}

touchGenename2Symbol <- function(){
    if(!tableExists(.globals$gn2symTableName)){
        gn2sym <- readRDS(
            system.file('extdata', 'gn2sym.rds', package = 'scRNAseqApp'))
        db <- rbind(
            data.frame(gene=gn2sym$unique, name=names(gn2sym$unique)),
            data.frame(gene=unlist(gn2sym$multiple),
                       name=rep(names(gn2sym$multiple),
                                lengths(gn2sym$multiple))))
        db <- unique(db)
        connectDB(dbWriteTable, name = .globals$gn2symTableName,
                  value = db, overwrite = TRUE)
    }
}

mapGeneSymbols <- function(genes){
    genes <- tolower(genes)
    if(length(genes)==1){
        where <- paste0('LOWER(`gene`)="', genes,
                        '" OR LOWER(`name`)="', genes, '"')
    }else{
        genes <- paste(genes, collapse = '","')
        where <- paste0('LOWER(`gene`) IN ("', genes,
                        '") OR LOWER(`name`) IN ("', genes, '")')
    }
    query <- paste0('SELECT `gene` FROM ', .globals$gn2symTableName,
                    ' WHERE ', where)
    rs <- connectDB(dbGetQuery, statement = query)
}

## search key
touchSearchTable <- function(){
    createSearchTable <- function(){
        sql <- paste('CREATE TABLE IF NOT EXISTS', .globals$searchTableName, 
                     '(word TEXT PRIMARY KEY,',
                     'count INTEGER DEFAULT 0)')
        sendNoreplyQueryToDB(statement=sql)
    }
    if(!tableExists(.globals$searchTableName)){
        createSearchTable()
    }
}
updateSearchTable <- function(words){
    if(isTRUE(words!="")){
        touchSearchTable()
        con <- getDBconn()
        on.exit(dbDisconnect(con))
        
        sql <- paste0("INSERT INTO ", .globals$searchTableName,
                      " (word, count) VALUES (?word, 1)",
                      " ON CONFLICT(word) DO UPDATE SET count = count + 1")
        query <- sqlInterpolate(
            conn = con, sql,
            word = words
        )
        res <- dbSendQuery(conn = con, statement = query)
        dbClearResult(res) 
    }
}
listSearchTable <- function(limit=100){
    touchSearchTable()
    query <- paste0("SELECT * FROM ", .globals$searchTableName,
                    " ORDER BY count DESC LIMIT ", limit)
    connectDB(dbGetQuery, query)
}

subsetGeneExprUI <- function(id) {
    tabPanel(
        value = id,
        htmlOutput(NS(id, 'GeneExpr')),
        tabsubTitleUI(
            id,
            'GeneExpr',
            description = paste(
                "In this tab, users can visualise gene expressions of",
                "two groups side-by-side on low-dimensional representions."
            )
        ),
        fluidRow(
            column(
                3,
                dimensionReductionUI(id),
                h4("Information to show"),
                selectInput(
                    NS(id, "GeneName"),
                    "Gene name:",
                    choices = NULL) %>% helper1(category = "geneName"),
                selectInput(
                    NS(id, "CellInfo"),
                    "Cell information to show:",
                    choices = NULL)
            ),
            column(
                4,
                subsetCellByInfoUI(id),
                subsetCellByFilterUI(id)),
            column(
                5,
                graphicsControlUI(id), br(),
                cellInfoTblUI(id, 1))
        ),
        fluidRow(
            column(
                6,
                style = "border-right: 2px solid black",
                htmlOutput(NS0(id, "subPlotTitle", 1)),
                fluidRow(column(6, uiOutput(
                    NS0(id, "GeneExprgrp.ui", 1)
                )),
                column(
                    6, geneExprPlotControlUI(id, postfix = 1)
                )),
                geneExprDotPlotUI(id, 1)
            ),
            column(
                6,
                htmlOutput(NS0(id, "subPlotTitle", 2)),
                fluidRow(column(6, uiOutput(
                    NS0(id, "GeneExprgrp.ui", 2)
                )),
                column(
                    6, geneExprPlotControlUI(id, postfix = 2)
                )),
                geneExprDotPlotUI(id, 2)
            )
        )
    )
}
subsetGeneExprServer <- function(id, dataSource, optCrt) {
    moduleServer(id, function(input, output, session) {
        ## title
        output$GeneExpr <-
            renderUI({
                HTML(paste("Subset", dataSource()$terms["GeneExpr"]))
            })
        ## subtitle
        output$GeneExprSubTitle <-
            renderUI({
                h4(paste(
                    "Subset gene",
                    dataSource()$terms['expression'],
                    "on dimension reduction"
                ))
            })
        ## input column 1
        ### Dimension Reduction
        updateDimRedSelInputPair(session, input, dataSource)
        ### Information to show
        #### gene name to plot
        selectedGene <- dataSource()$sc1def$gene1
        if (!is.null(dataSource()$genelist)) {
            selectedGene <- dataSource()$genelist[1]
        }
        updateSelectizeInput(
            session,
            "GeneName",
            choices = sort(names(dataSource()$sc1gene)),
            server = TRUE,
            selected = selectedGene,
            options = list(
                maxOptions = .globals$maxNumGene,
                create = TRUE,
                persist = TRUE,
                render = I(optCrt)
            )
        )
        #### cell information to plot
        updateSelectInput(
            session,
            "CellInfo",
            "Cell information to show:",
            choices = dataSource()$sc1conf[dataSource()$sc1conf$grp == TRUE]$UI,
            selected = dataSource()$sc1def$grp1
        )
        
        ## input column 2
        updateSubsetCellUI(id, input, output, session, dataSource)
        updateFilterCellUI(id, optCrt, input, output, session, dataSource)
        
        ## plot region
        ### update the Color Range to make two plots comparable
        inpColRange <- reactive({
            scDRgene(
                inpConf=dataSource()$sc1conf,
                inpMeta=dataSource()$sc1meta,
                dimRedX=input$GeneExprdrX,
                dimRedY=input$GeneExprdrY,
                gene1=input$GeneName,
                subsetCellKey=c(input$CellInfo, input$subsetCell),
                subsetCellVal=getSubsetCellVal(
                    input,
                    list(c(input$GeneExprsub1b, input$GeneExprsub2b)),
                    input$CellInfo),
                dataset=dataSource()$dataset,
                geneIdMap=dataSource()$sc1gene,
                pointSize=input$GeneExprsiz,
                gradientCol=input$GeneExprcol1,
                GeneExprDotOrd=input$GeneExprord1,
                labelsFontsize=input$GeneExprfsz,
                labelsFontFamily=input$GeneExprfml,
                plotAspectRatio=input$GeneExprasp,
                keepXYlables=input$GeneExprtxt,
                inpPlt=input$GeneExprtype1,
                inpXlim=if (input$GeneExprxlimb1 %% 2 == 0)
                    0
                else
                    input$GeneExprxlim1,
                inpColRange = TRUE,
                valueFilterKey = input$filterCell,
                valueFilterCutoff = input$filterCellVal,
                valueFilterCutoff2 = input$filterCellVal2,
                hideFilterCell = input$GeneExprhid1
            )
        })
        ## plots
        getSubGroupName <- function(dataSource, input) {
            sub <- strsplit(
                dataSource()$sc1conf[
                    dataSource()$sc1conf$UI == input$CellInfo]$fID,
                "\\|")
            if (length(sub)) {
                sub <- sub[[1]]
            }
            sub
        }
        updateSubsetGeneExprPlot(
            1,
            getSubGroupName,
            optCrt,
            inpColRange,
            id,
            input,
            output,
            session,
            dataSource
        )
        updateSubsetGeneExprPlot(
            2,
            getSubGroupName,
            optCrt,
            inpColRange,
            id,
            input,
            output,
            session,
            dataSource
        )
        
        ## cell stats
        observeEvent(input$GeneExprsub1b, updateGeneExprDT())
        observeEvent(input$GeneExprsub2b, updateGeneExprDT())
        updateGeneExprDT <- function() {
            output$GeneExpr.dt1 <- renderDT({
                ggData <- scDRnum(
                    inpConf=dataSource()$sc1conf,
                    inpMeta=dataSource()$sc1meta,
                    inpCellInfo=c(
                        input$CellInfo,
                        input$GeneExprsub1b,
                        input$GeneExprsub2b
                    ),
                    gene=input$GeneName,
                    inpsubName=input$subsetCell,
                    inpsubValue=getSubsetCellVal(input),
                    dataset=dataSource()$dataset,
                    geneIdMap=dataSource()$sc1gene,
                    inpsplt=input$GeneExprsplt1
                )
                datatable(
                    ggData,
                    rownames = FALSE,
                    extensions = "Buttons",
                    options = list(
                        pageLength = -1,
                        dom = "tB",
                        buttons = c("copy", "csv", "excel")
                    )
                ) %>%
                    formatRound(
                        columns = c("pctExpress"),
                        digits = 2)
            })
        }
    })
}

plotBubbleHeatmapUI <- function(id) {
    tabPanel(
        value = id,
        HTML("Bubbleplot / Heatmap"),
        h4("Gene expression bubbleplot / heatmap"),
        "In this tab, users can visualise the gene expression patterns of ",
        "multiple genes grouped by categorical cell information",
        "(e.g. library / cluster).",
        br(),
        "The normalised expression are averaged,",
        "log-transformed and then plotted.",
        br(),
        br(),
        fluidRow(
            column(
                3,
                style = "border-right: 2px solid black",
                textAreaInput(
                    NS(id, "genelist"),
                    HTML(paste(
                        "List of gene names <br /> (Max", 
                        .globals$maxHeatmapGene,
                        "genes, <br />separated by , or ; or newline):"
                    )),
                    height = "200px",
                    value = NULL
                ) %>%
                    shinyhelper::helper(
                        type = "inline",
                        size = "m",
                        fade = TRUE,
                        title = "List of genes to plot on bubbleplot / heatmap",
                        content = c(
                            "Input genes to plot",
                            "- Maximum 500 genes
                            (due to plot space limitations)",
                            "- Genes should be separated by comma,
                            semicolon or newline"
                        )
                    ),
                xaxisCellInfoUI(id),
                yaxisCellInfoUI(id),
                subsetCellByInfoUI(id, mini = TRUE),
                sliderInput(
                    NS(id, "filterVal"),
                    "Filter the cells by value",
                    min = 0,
                    max = 10,
                    value = 0
                ),
                radioButtons(
                    NS(id, "plottyp"),
                    "Plot type:",
                    choices = c("Bubbleplot", "Heatmap", "Violin"),
                    selected = "Bubbleplot",
                    inline = TRUE
                ),
                checkboxInput(
                    NS(id, "plotscl"),
                    "Scale gene expression",
                    value = TRUE),
                checkboxInput(
                    NS(id, "plotrow"),
                    "Cluster rows (genes)",
                    value = TRUE),
                checkboxInput(
                    NS(id, "plotcol"),
                    "Cluster columns (samples)",
                    value = FALSE),
                checkboxInput(
                    NS(id, "plotflp"),
                    "Flip X/Y", value = FALSE),
                conditionalPanel(
                    condition = paste0("input.plottyp == 'Heatmap'"),
                    ns = NS(id),
                    checkboxInput(
                        NS(id, "plotall"),
                        "All cells (time consuming)",
                        value = FALSE)
                ),
                conditionalPanel(
                    condition = "input.plottyp == 'Violin'",
                    ns = NS(id),
                    checkboxInput(
                        NS(id, "plotpts"), "Show data points",
                        value = FALSE),
                    checkboxInput(
                        NS(id, "addnoise"), "Add noise",
                        value = TRUE
                    )
                ),
                conditionalPanel(
                    condition = "input.plottyp != 'Heatmap'",
                    ns = NS(id),
                    sliderInput(
                        NS(id, "plotsiz"), "Data point size:",
                        min = 0, max = 4, value = 1.25, step = 0.25)
                ),
                conditionalPanel(
                    condition = "input.plottyp != 'Violin'",
                    ns = NS(id),
                    conditionalPanel(
                        condition = "input.plotrow==1",
                        ns = NS(id),
                        radioButtons(NS(id, 'row_dend_side'),
                                     'Row dendrogram side (gene)',
                                     choices = c('left', 'right'),
                                     selected = 'left',
                                     inline = TRUE),
                        sliderInput(NS(id, 'row_dend_width'),
                                     'Width of row dendrogram (mm)',
                                     value = 10,
                                    min = 0, max = 100, step = 0.1)
                    ),
                    conditionalPanel(
                        condition = "input.plotcol==1",
                        ns = NS(id),
                        radioButtons(NS(id, 'column_dend_side'),
                                     'Column dendrogram side (sample))',
                                     choices = c('top', 'bottom'),
                                     selected = 'top',
                                     inline = TRUE),
                        sliderInput(NS(id, 'column_dend_height'),
                                    'Height of column dendrogram (mm)',
                                    value = 10,
                                    min = 0, max = 100, step = 0.1)
                    ),
                    radioButtons(NS(id, 'legend_side'),
                                  "Legend side",
                                  choices = c('right', 'bottom'),
                                  selected = 'bottom',
                                  inline = TRUE),
                    actionButton(
                        NS(id, "userbreaks"),
                        "Set ColorKey range"),
                    conditionalPanel(
                        condition = "input.userbreaks % 2 ==1",
                        ns = NS(id),
                        numericInput(NS(id, "colorb1"), "min cutoff", value = -10),
                        numericInput(NS(id, "colorb2"), "max cutoff", value = 10),
                        actionButton(NS(id, "setcolrg"), "Apply colorkey range")
                    )
                ),
                conditionalPanel(
                    condition = "input.plotcol % 2 != 1",
                    ns = NS(id),
                    checkboxInput(
                        NS(id, "plotord"),
                        "Reorder the contents", value = FALSE
                    ),
                    conditionalPanel(
                        condition = "input.plotord % 2 == 1",
                        ns=NS(id),
                        uiOutput(outputId = NS(id, "plotXord"))
                    )
                ),
                conditionalPanel(
                    condition = "input.CellInfoY != ''",
                    ns = NS(id),
                    checkboxInput(
                        NS(id, "plotsord"),
                        "Reorder the spliter", value = FALSE
                    ),
                    conditionalPanel(
                        condition = "input.plotsord % 2 == 1",
                        ns=NS(id),
                        uiOutput(outputId = NS(id, "plotSord"))
                    )
                ),
                br(),
                boxPlotControlUI(id, withPoints = FALSE, withColor = TRUE)
            ),
            column(9, h4(htmlOutput(
                NS(id, "oupTxt")
            )), geneExprDotPlotUI(id))
        )
    )
}

plotBubbleHeatmapServer <- function(id, dataSource, optCrt) {
    moduleServer(id, function(input, output, session) {
        colrg <- reactiveVal(NA)
        ## input column
        genelist <- dataSource()$sc1def$genes
        if (!is.null(dataSource()$genelist)) {
            if (length(dataSource()$genelist) > 1) {
                genelist <- dataSource()$genelist
            } else{
                genelist <- c(dataSource()$genelist, genelist)
            }
        }
        updateTextAreaInput(
            session, "genelist",
            value = paste0(genelist, collapse = ", "))
        updateSelectInput(
            session,
            "CellInfoX",
            "Group by:",
            choices = getGroupUI(dataSource),
            selected = dataSource()$sc1def$grp1
        )
        updateSelectInput(
            session,
            "CellInfoY",
            "Split by:",
            choices = c("N/A", getGroupUI(dataSource)),
            selected = NA
        )
        updateSubsetCellUI(id, input, output, session, dataSource, addNA = TRUE)
        
        observeEvent(input$plotflp, {
            if(input$plotflp){
                updateRadioButtons(session,
                                   "row_dend_side",
                                   choices = c("top", "bottom"),
                                   selected = "top")
                updateRadioButtons(session,
                                   "column_dend_side",
                                   choices = c("left", "right"),
                                   selected = "left")
            }else{
                updateRadioButtons(session,
                                   "row_dend_side",
                                   choices = c("left", "right"),
                                   selected = "left")
                updateRadioButtons(session,
                                   "column_dend_side",
                                   choices = c("top", "bottom"),
                                   selected = "top")
            }
        }, ignoreInit = FALSE)
        
        observeEvent(input$userbreaks, {
            rg <- scBubbHeat(
                inpConf = dataSource()$sc1conf,
                inpMeta = dataSource()$sc1meta,
                inp = input$genelist,
                inpGrp = input$CellInfoX,
                grpKey = input$subsetCell,
                grpVal = getSubsetCellVal(input),
                inpGrp1c = input$filterVal,
                inpPlt = input$plottyp,
                dataset = dataSource()$dataset,
                geneIdMap = dataSource()$sc1gene,
                inpScl = input$plotscl,
                inpRow = input$plotrow,
                row_dend_side = input$row_dend_side,
                row_dend_width = unit(input$row_dend_width, 'mm'),
                inpCol = input$plotcol,
                column_dend_side = input$column_dend_side,
                column_dend_height = unit(input$column_dend_height, 'mm'),
                legend_side = input$legend_side,
                inpcols = input$plotcols,
                pointSize = input$plotpts,
                labelsFontsize = input$plotfsz,
                labelsFontFamily = input$plotfml,
                flipXY = input$plotflp,
                plotAllCells = input$plotall,
                legendTitle = dataSource()$terms['expression'],
                returnColorRange = TRUE
            )
            rg <- round(rg, digits = 2)
            updateNumericInput(
                session,
                "colorb1",
                value = rg[2],
                min = floor(rg[1]),
                max = ceiling(rg[4]),
                step = .01
            )
            updateNumericInput(
                session,
                "colorb2",
                value = rg[3],
                min = ceiling(rg[1]),
                max = ceiling(rg[4]),
                step = .01
            )
            
        })
        observeEvent(input$setcolrg, {
            colrg(sort(c(
                input$colorb1, input$colorb2
            )))
        })
        
        observeEvent(input$CellInfoX,
                     updateCheckboxInput(session, 'plotord', value = FALSE))
        updateRankList(
            input, output, dataSource, "CellInfoX", "plotXord",
            NS(id, "cellinfoXorder"))
        updateRankList(
            input, output, dataSource, "CellInfoY", "plotSord",
            NS(id, "cellinfoSorder"))
        
        ## update the ui
        output$oupTxt <- renderUI({
            geneList <- scGeneList(input$genelist, dataSource()$sc1gene)
            if (nrow(geneList) > 50) {
                HTML("More than 50 input genes! Please reduce the gene list!")
            } else {
                oup <- paste0(
                    nrow(geneList[geneList$present == TRUE]),
                    " genes OK and will be plotted")
                if (nrow(geneList[geneList$present == FALSE]) > 0) {
                    oup <- paste0(
                        oup,
                        "<br/>",
                        nrow(geneList[geneList$present == FALSE]),
                        " genes not found (",
                        paste0(
                            geneList[geneList$present == FALSE]$gene,
                            collapse = ", "),
                        ")"
                    )
                }
                HTML(oup)
            }
        })
        
        ## plot region
        ### plots
        plot1 <- reactive({
            ht <- scBubbHeat(
                inpConf = dataSource()$sc1conf,
                inpMeta = dataSource()$sc1meta,
                inp = input$genelist,
                inpGrp = input$CellInfoX,
                grpKey = input$subsetCell,
                grpVal = getSubsetCellVal(input),
                inpGrp1c = input$filterVal,
                inpPlt = input$plottyp,
                dataset = dataSource()$dataset,
                geneIdMap = dataSource()$sc1gene,
                inpScl = input$plotscl,
                inpRow = input$plotrow,
                row_dend_side = input$row_dend_side,
                row_dend_width = unit(input$row_dend_width, 'mm'),
                inpCol = input$plotcol,
                column_dend_side = input$column_dend_side,
                column_dend_height = unit(input$column_dend_height, 'mm'),
                legend_side = input$legend_side,
                inpcols = input$plotcols,
                pointSize = if(input$plottyp=='Bubbleplot'){
                    input$plotsiz
                }else{
                    if(input$plotpts %% 2 ==0) 0 else input$plotsiz
                },
                labelsFontsize = input$plotfsz,
                labelsFontFamily = input$plotfml,
                flipXY = input$plotflp,
                plotAllCells = input$plotall,
                colorBreaks = 
                    if (input$userbreaks %% 2 == 0 && !is.na(colrg()[1]))
                        NA
                    else
                        colrg(),
                legendTitle = dataSource()$terms['expression'],
                reorder=input$plotord,
                orderX = input$cellinfoXorder,
                splitBy=input$CellInfoY,
                sreorder = input$plotsord,
                orderS = input$cellinfoSorder,
                addnoise = input$addnoise
            )
        })
        observeEvent(
            input$plottyp,
            updateGeneExprDotPlotUI(
                postfix = 1,
                id,
                input,
                output,
                session,
                plot1,
                .globals$pList3[input$plotpsz],
                dataSource()$dataset,
                input$plottyp,
                input$CellInfoX,
                handlerFUN = if (input$plottyp == "Violin") {
                    plotsDownloadHandler
                } else{
                    heatmapDownloadHandler
                }
            )
        )
    })
}

uploadUI <- function (id) {
    ns <- NS(id)
    tagList(column(
        width = 10,
        offset = 1,
        fluidRow(
            actionButton(
                ns('upload'),
                label = "Upload",
                icon = icon("upload")
            ),
            textOutput(ns('message'))
        ),
        fluidRow(
            column(
                width = 6,
                fileInput(
                    ns("file"),
                    label = "Seurat object",
                    multiple = FALSE
                ),
                textInput(
                    ns("title"),
                    label = "Title for the data"
                ),
                textInput(
                    ns("dir"),
                    label = "data folder (must be unique)"
                ),
                checkboxInput(
                    ns("locker"),
                    label = "Require privilege for the data",
                    value = TRUE
                ),
                selectInput(
                    ns("datatype"),
                    label = "Data type",
                    choices = c("scRNAseq", "scATACseq", "scMultiome"),
                    selected = "scRNAseq"
                ),
                textInput(
                    ns("gene1"),
                    label = "Primary default gene to show"
                ),
                textInput(
                    ns("gene2"),
                    label = "Secondary default gene to show"
                ),
                div(class = "consoleOutput",
                    verbatimTextOutput(ns(
                        'consoleOutput'
                    )))
            ),
            column(
                width = 6,
                selectInput(
                    ns("species"),
                    label = "Species",
                    choices = .globals$supported_organisms,
                    selected = "Homo sapiens"
                ),
                textInput(
                    ns("species2"),
                    label = "Other species"
                ),
                selectInput(
                    ns("gexAssay"),
                    label = "Gene expression assay",
                    choices = c("SCT", "RNA"),
                    selected = "SCT"
                ),
                selectInput(
                    ns("gexSlot"),
                    label = "Slot in single-cell assay",
                    choices = c("data", "scale.data", "counts"),
                    selected = "data"
                ),
                selectInput(
                    ns("atacAssay"),
                    label = "ATAC peaks assay",
                    choices = c("ATAC", "peaks"),
                    selected = NULL
                ),
                selectInput(
                    ns("atacSlot"),
                    label = "Slot in ATAC assay",
                    choices = c("data", "scale.data", "counts"),
                    selected = "data"
                ),
                checkboxGroupInput(
                    ns("meta_to_include"),
                    label = "Columns to include from the metadata"),
                textAreaInput(
                    ns("multigene"),
                    label = "Default genes to show in bubbleplot/heatmap"),
                textAreaInput(
                    ns("keywords"),
                    label = "The key words for this work"),
                textAreaInput(
                    ns("reference"),
                    label = "BibTeX string"),
                textInput(
                    ns("doi"),
                    label = "DOI"),
                textInput(
                    ns("pmid"),
                    label = "PubMed ID")
            )
        ),
        hr(),
        fluidRow(
            column(
                width = 3,
                h4("Assign cell cycle"),
                actionButton(
                    ns("cellcycle"),
                    label = "CellCycleScoring"),
                actionButton(
                    ns("tricycle"),
                    label = "Tricycle")
            ),
            column(
                width = 3,
                h4("Assign cell type"),
                selectInput(
                    ns("celldex"),
                    label = NULL,
                    choices = c(
                        "HumanPrimaryCellAtlasData",
                        "BlueprintEncodeData",
                        "MouseRNAseqData",
                        "ImmGenData",
                        "DatabaseImmuneCellExpressionData",
                        "NovershternHematopoieticData",
                        "MonacoImmuneData"
                    ),
                    selected = "HumanPrimaryCellAtlasData"
                ),
                actionButton(
                    ns("singler"),
                    label = "singleR")
            ),
            column(
                width = 3,
                h4("Trajectories"),
                actionButton(
                    ns("slingshot"),
                    label = "slingshot")
            ),
            column(
                width = 3,
                h4("Cell communications"),
                p("N/A")
            )
        )
    ))
}

#' @importFrom tools file_ext
#' @importFrom SeuratObject Reductions Idents Assays DefaultAssay GetAssayData
#'  `DefaultAssay<-` VariableFeatures Misc `Misc<-` Embeddings `Idents<-`
#' @importFrom Seurat FindAllMarkers FindVariableFeatures ScaleData
#'  CellCycleScoring as.SingleCellExperiment
#' @importFrom RefManageR GetBibEntryWithDOI GetPubMedByID
#' @importFrom utils head
#' @importFrom methods slot
uploadServer <- function(id) {
    moduleServer(id, function(input, output, session) {
        global <- reactiveValues(
            filepath = NULL,
            seu = NULL,
            config = NULL,
            ref = NULL,
            markers = NULL,
            tmpf = tempfile()
        )
        getSeuObj <- function() {
            seu <- global$seu
            DefaultAssay(seu) <-
                ifelse("SCT" %in% Assays(seu), "SCT", "RNA")
            seu
        }
        seuReductionMethod <- function(seu, forMonocole = TRUE) {
            reduction_method <- Reductions(seu)
            getReductionMethod(reduction_method, forMonocole)
        }
        saveMisc <- function(slot) {
            misc <- Misc(global$seu, slot)
            writeMisc(misc, input$dir, slot)
        }
        observeEvent(input$file, {
            adminProcess({
                file <- input$file
                ext <- tolower(file_ext(file$datapath))
                filename <- sub(paste0(".", ext), "", basename(file$name))
                uniqueFilename <-
                    make.names(
                        c(dir(.globals$datafolder), filename),
                        unique = TRUE,
                        allow_ = TRUE)[
                            length(dir(.globals$datafolder)) + 1]
                req(file)
                validate(need(
                    ext == "rds",
                    "Please upload a seurat object saved in rds file"
                ))
                isolate(global$filepath <- file$datapath)
                isolate(global$seu <- readRDS(file$datapath))
                validate(need(
                    is(global$seu, "Seurat"),
                    "Please upload a seurat object"
                ))
                message("Handling data")
                updateTextInput(
                    session, "dir",
                    value = uniqueFilename)
                updateTextInput(
                    session, "title",
                    value = filename)
                cellInfo <- colnames(global$seu[[]])
                updateCheckboxGroupInput(
                    session,
                    "meta_to_include",
                    label = "Columns to include from the metadata",
                    choices = cellInfo,
                    selected = cellInfo
                )
                assays <- Assays(global$seu)
                validate(need(
                    any(c("SCT", "RNA") %in% assays),
                    "Please upload a seurat object with 'SCT' or 'RNA' assay"
                ))
                if (!DefaultAssay(global$seu) %in% c("SCT", "RNA")) {
                    DefaultAssay(global$seu) <-
                        match.arg(
                            assays,
                            choices = c("SCT", "RNA"),
                            several.ok = TRUE
                        )[1]
                }
                defaultAssay <- DefaultAssay(global$seu)
                updateSelectInput(
                    session,
                    "gexAssay",
                    choices = assays,
                    selected = defaultAssay)
                message("Check scale.data slot")
                if (length(extAssayData(
                    global$seu, slot = "scale.data")) == 0) {
                    global$seu <- FindVariableFeatures(
                        global$seu,
                        selection.method = "vst",
                        nfeatures = 1000
                    )
                    global$seu <- ScaleData(global$seu)
                }
                top2 <- head(VariableFeatures(global$seu), 2)
                if (length(top2) == 2) {
                    updateTextInput(session, "gene1", value = top2[1])
                    updateTextInput(session, "gene2", value = top2[2])
                } else{
                    updateTextInput(
                        session, "gene1",
                        value = rownames(global$seu)[1])
                    updateTextInput(
                        session, "gene2",
                        value = rownames(global$seu)[2])
                }
                message("Find all markers")
                if (!is.null(Misc(global$seu, "markers"))) {
                    ## the markers is available at Misc(seu, "markers") slot
                    markers <- Misc(global$seu, "markers")
                    if (is.list(markers) && !is.data.frame(markers)) {
                        markers <- as.data.frame(markers[[1]])
                    }
                } else{
                    if (!is.factor(Idents(global$seu))) {
                        grp <- cellInfo[
                            grepl(
                                'cluster|cell(.*)type',
                                cellInfo,
                                ignore.case = TRUE)]
                        grp_d <- adist('celltype', grp)
                        Idents(global$seu) <- grp[which.min(grp_d)][1]
                    }
                    markers <- FindAllMarkers(
                        global$seu,
                        only.pos = TRUE,
                        min.pct = .25,
                        logfc.threshold = .25
                    )
                    if (length(markers)) {
                        Misc(global$seu, "markers") <- markers
                    }
                }
                markers <- split(markers, markers$cluster)
                markers <-
                    lapply(markers, head, n = min(5, ceiling(50 / length(
                        markers
                    ))))
                markers <- lapply(markers, function(.ele)
                    (
                        if (!is.null(.ele$gene)) {
                            return(.ele$gene)
                        } else{
                            return(rownames(.ele))
                        }))
                markers <- unique(unlist(markers))
                message("Update all inputs")
                if (length(markers) > 1) {
                    updateTextAreaInput(
                        session, "multigene",
                        value = markers)
                } else{
                    updateTextAreaInput(
                        session,
                        "multigene",
                        value = rownames(global$seu)[
                            seq.int(min(20, nrow(global$seu)))])
                }
            }, "Uploading data", "Preprocessing Done!")
        })
        observeEvent(input$meta_to_include, {
            global$config <-
                createConfig(
                    global$seu,
                    meta.to.include = unique(input$meta_to_include))
        })
        observeEvent(input$gexAssay, {
            if (is(global$seu, "Seurat")) {
                slots <- c("data", "scale.data", "counts")
                d <- lapply(slots, function(.ele){
                    extAssayData(object = global$seu, slot=.ele)
                    })
                d <- lapply(d, nrow)
                d <-
                    vapply(
                        d,
                        FUN = function(.d)
                            .d > 0,
                        FUN.VALUE = logical(1L)
                    )
                slots <- slots[d]
                if (length(slots) == 0) {
                    adminMsg(
                        "There is no slots named as data, scale.data or counts",
                        type = "error")
                } else{
                    updateSelectInput(
                        session,
                        "gexSlot",
                        choices = slots,
                        selected = slots[1])
                }
            }
        })
        observeEvent(
            input$doi,
            updateRefById(
                "reference",
                "doi",
                GetBibEntryWithDOI,
                input,
                output,
                session
            )
        )
        observeEvent(input$pmid, {
            global$ref <- updateRefById(
                "reference",
                "pmid",
                GetPubMedByID,
                input,
                output,
                session)
        })
        observeEvent(input$multigene, {
            global$markers <- strsplit(input$multigene, "\\s+|,|;")[[1]]
        })
        observeEvent(input$upload, {
            if (!is.null(global$seu)) {
                adminProcess({
                    dir.create(file.path(.globals$datafolder, input$dir))
                    message("set data config file")
                    updateAppConf(input, reactive({
                        global
                    }))
                    message("makeShinyApp")
                    makeShinyFiles(
                        global$seu,
                        scConf = global$config,
                        assayName = input$gexAssay,
                        gexSlot = input$gexSlot,
                        atacAssayName = input$atacAssay,
                        atacSlot = input$atacSlot,
                        appDir = file.path(.globals$datafolder, input$dir),
                        defaultGene1 = input$gene1,
                        defaultGene2 = input$gene2,
                        default.multigene = global$markers
                    )
                    message("set misc files")
                    for (slot in names(Misc(global$seu))) {
                        saveMisc(slot)
                    }
                    message("Check file LOCKER")
                    if (input$locker) {
                        setLocker(input$dir)
                    }
                    message("reset the inputs")
                    updateTextInput(session, "dir",
                                    value = "")
                    updateTextInput(session, "title",
                                    value = "")
                    updateCheckboxGroupInput(
                        session,
                        "meta_to_include",
                        label = "Columns to include from the metadata",
                        choices = character(0),
                        selected = character(0)
                    )
                    updateTextInput(session, "gene1", value = "")
                    updateTextInput(session, "gene2", value = "")
                    updateTextAreaInput(session, "multigene",
                                        value = "")
                    updateTextInput(session, "doi", value = "")
                    updateTextInput(session, "pmid", value = "")
                    updateTextInput(session, "reference", value = "")
                }, "Create data by ShinyCell::makeShinyApp", "Upload done!")
            }
        })
        ## cell cycle
        observeEvent(input$cellcycle, {
            adminProcess({
                cc.genes.list <- readRDS(
                    system.file(
                        "extdata", "cc.genes.list.rds",
                        package = "scRNAseqApp")
                )
                cc.genes <- cc.genes.list[[input$species]]
                isENS <- vapply(
                    cc.genes,
                    FUN = function(.ele) {
                        .ele <- unlist(.ele)
                        sum(.ele %in% rownames(global$seu))
                    },
                    FUN.VALUE = numeric(1L)
                )
                cc.genes <- cc.genes[[which.max(isENS)]]
                seu <- getSeuObj()
                isolate(
                    global$seu <- CellCycleScoring(
                        seu,
                        s.features = cc.genes$s.genes,
                        g2m.features = cc.genes$g2m.genes,
                        set.ident = FALSE
                    )
                )
                cellInfo <- colnames(global$seu[[]])
                updateCheckboxGroupInput(
                    session,
                    "meta_to_include",
                    choices = cellInfo,
                    selected = cellInfo
                )
            }, "Starting CellCycleScoring", "Cell phase updated!")
        })
        observeEvent(input$tricycle, {
            if (!input$species %in% c("Homo sapiens", "Mus musculus")) {
                adminMsg("Only support human and mouse data.", type = "error")
                return()
            }
            askNamespace("tricycle", "SummarizedExperiment")
            adminProcess({
                seu <- getSeuObj()
                exp <- as.SingleCellExperiment(seu)
                gname.type <- ifelse(
                    grepl("ENS", rownames(exp)[1]),
                    "ENSEMBL",
                    "SYMBOL")
                species <- ifelse(
                    input$species == "Mus musculus",
                    "mouse",
                    "human")
                tricyle <- addTricycle(
                    exp = exp,
                    gname.type = gname.type,
                    species = species,
                    meta = seu[[]]
                )
                seu$tricyclePosition <- tricyle$tricyclePosition
                seu$CCStage <- tricyle$CCStage
                isolate(global$seu <- seu)
                cellInfo <- colnames(global$seu[[]])
                updateCheckboxGroupInput(
                    session,
                    "meta_to_include",
                    choices = cellInfo,
                    selected = cellInfo
                )
            }, "Starting Tricycle!", "Tricycle done!")
        })
        ## cell type
        observeEvent(input$singler, {
            askNamespace("SingleR", "celldex", "SummarizedExperiment")
            if (!input$species %in% c(
                "Homo sapiens",
                "Mus musculus",
                "Danio rerio")) {
                adminMsg(
                    "Only support human, mouse and fish data.",
                    type = "error")
                return()
            }
            adminProcess({
                seu <- getSeuObj()
                try(attachNamespace("celldex"))
                ref <-
                    get(input$celldex,
                        envir = as.environment("package:celldex"))()
                sce <- as.SingleCellExperiment(seu)
                pred <- SingleR::SingleR(
                    test = sce,
                    ref = ref,
                    labels = ref$label.fine
                )
                ## assign singler to meta data
                seu$SingleR_labels <- pred$labels
                isolate(global$seu <- seu)
                cellInfo <- colnames(global$seu[[]])
                updateCheckboxGroupInput(
                    session,
                    "meta_to_include",
                    choices = cellInfo,
                    selected = cellInfo
                )
            }, "Starting SingleR!", "SingleR done!")
        })
        observeEvent(input$slingshot, {
            adminProcess({
                askNamespace("slingshot")
                seu <- getSeuObj()
                misc_slingshot <- NULL
                reduction_method <- seuReductionMethod(seu, FALSE)
                dimred <- Embeddings(seu, reduction = reduction_method)
                # by group points
                grp_ids <- lapply(getGrpIDs(global$config), function(id) {
                    seu[[]][[id]]
                })
                lineages <- addSlingshot(dimred, grp_ids)
                message("Assign slingshot to miscellaneous data")
                ## assign slingshot to meta data
                Misc(global$seu, "slingshot") <- lineages
            }, "Starting SlingShot", "SlingShot done!")
        })
    })
}

# Get gene list
#' @importFrom data.table data.table
scGeneList <- function(inp, geneIdMap) {
    geneList <- data.table(gene = unique(trimws(strsplit(
        inp,
        ",|;|\n|\r|\\s+"
    )[[1]])),
    present = TRUE)
    geneList <- geneList[geneList$gene !="", ]
    geneList[!geneList$gene %in% names(geneIdMap)]$present <- FALSE
    return(geneList)
}

# Plot gene expression on dimred
#' @importFrom ggplot2 ggplot aes .data geom_point xlab ylab guides
#' guide_colorbar scale_color_gradientn coord_fixed scale_y_discrete
#' scale_x_continuous xlim
#' @importFrom ggridges geom_density_ridges theme_ridges
scDRgene <- function(
        inpConf,
        inpMeta,
        dimRedX,
        dimRedY,
        gene1,
        subsetCellKey,
        subsetCellVal,
        dataset,
        geneIdMap,
        pointSize,
        gradientCol,
        GeneExprDotOrd,
        labelsFontsize,
        labelsFontFamily,
        plotAspectRatio,
        keepXYlables,
        inpPlt = "Dotplot",
        inpXlim,
        inpColRange = 0,
        valueFilterKey,
        valueFilterCutoff,
        valueFilterCutoff2,
        hideFilterCell = FALSE,
        geneType = c('gene', 'coor'),
        ...) {
    if (gene1[1] == "") {
        return(ggplot())
    }
    geneType <- match.arg(geneType)
    if(geneType=='gene'){
        if (is.na(geneIdMap[gene1])) {
            return(ggplot())
        }
        if (is.null(geneIdMap[gene1])) {
            return(ggplot())
        }
    }
    
    # Prepare ggData
    subFilterColname <- 'subValue'
    subGrpColname <- 'sub'
    exprColname <- 'val'
    subsetCellKey <- subsetCellKey[subsetCellKey!="N/A"]
    subsetCellVal <- namedSubsetCellVals(subsetCellKey, subsetCellVal)
    ggData <- inpMeta[, c(
        inpConf[inpConf$UI == dimRedX]$ID,
        inpConf[inpConf$UI == dimRedY]$ID,
        inpConf[inpConf$UI %in% subsetCellKey]$ID),
        with = FALSE]
    cnid <- if(ncol(ggData)>2) 3 else 0
    ggData <-
        cbindFilterValues(
            ggData,
            inpConf,
            inpMeta,
            subFilterColname,
            geneIdMap,
            dataset,
            valueFilterKey,
            valueFilterCutoff,
            valueFilterCutoff2
        )
    
    if(geneType=='gene'){
        ggData[[exprColname]] <- read_exprs(
            dataset,
            geneIdMap[gene1],
            valueOnly = TRUE)
    }else{
        coord <- strsplit(gene1[1], '[:-]')[[1]]
        stopifnot('Wrong coordinates format'=length(coord)==3)
        names(coord) <- c("seqnames", "start", "end")
        coord <- as.list(coord)
        coord[['start']] <- as.integer(coord[['start']])
        coord[['end']] <- as.integer(coord[['end']])
        coordExp <- readATACdataByCoor(
            dataset,
            coord = coord,
            cells = inpMeta$sampleID)
        if(length(coordExp)==0){
            return(ggplot())
        }
        ggData[[exprColname]] <- rowSums(coordExp)
    }
    
    if (any(ggData[[exprColname]] < 0)) {
        ggData[ggData[[exprColname]] < 0][[exprColname]] <- 0
    }
    
    keep <- filterCells(
        ggData[, -c(1, 2)],
        subsetCellKey,
        subsetCellVal,
        subFilterColname,
        valueFilterCutoff,
        valueFilterCutoff2,
        inpConf=inpConf
    )
    
    ## make the first subsetCellKey as sub
    colnames(ggData)[c(1, 2)] <- c("X", "Y")
    if(cnid>2) colnames(ggData)[cnid] <- subGrpColname
    rat <- getRatio(ggData)
    bgCells <- sum(!keep) > 0
    
    if (bgCells) {
        ggData2 <- ggData[!keep]
        ggData <- ggData[keep]
    }
    ggData <- orderGeneExpr(ggData, GeneExprDotOrd, exprColname)
    
    if (is.logical(inpColRange[1])) {
        return(range(ggData[[exprColname]]))
    }
    # Actual ggplot
    if (inpPlt == "Dotplot") {
        if (length(inpColRange) == 1) {
            inpColRange <- c(min(
                0, min(inpColRange, na.rm = TRUE),
                na.rm = TRUE),
                inpColRange)
        }
        if (inpColRange[2] > 0) {
            ggData[ggData[[exprColname]] > inpColRange[2], exprColname] <-
                inpColRange[2]
        }
        ggOut <- ggXYplot(ggData)
        if (bgCells) {
            ggOut <- labelBackgroundCells(
                ggOut,
                ggData2,
                pointSize,
                color = "snow2",
                shape = 16,
                hide = hideFilterCell)
        }
        ggOut <- pointPlot(
            ggOut = ggOut,
            pointSize = pointSize,
            fontSize = labelsFontsize,
            labelsFontFamily = labelsFontFamily,
            dimRedX = dimRedX,
            dimRedY = dimRedY,
            keepXYlables = keepXYlables) +
            guides(color = guide_colorbar(barwidth = 15))
        if (inpColRange[2] > 0) {
            ggOut <- ggOut +
                scale_color_gradientn(
                    gene1,
                    colours = availableThemes(gradientCol),
                    limits = inpColRange)
        } else{
            ggOut <- ggOut +
                scale_color_gradientn(
                    gene1,
                    colours = availableThemes(gradientCol))
        }
        ggOut <- fixCoord(ggOut, plotAspectRatio, rat)
    } else{
        ## ridgePlot
        ggData[[subGrpColname]] <- factor(
            ggData[[subGrpColname]],
            levels = rev(sortLevels(as.character(
                unique(ggData[[subGrpColname]])
            ))))
        ggOut <- ggplot(
            ggData,
            aes(
                x = .data[[exprColname]],
                y = .data[[subGrpColname]],
                fill = .data[[subGrpColname]]
            )) +
            geom_density_ridges(scale = 4, show.legend = FALSE) +
            theme_ridges() +
            scale_y_discrete(expand = c(0.01, 0)) +
            scale_x_continuous(expand = c(0, 0)) +
            ylab("Subsets") +
            xlab("Expression Level")
        if (length(inpXlim) == 2) {
            ggOut <- ggOut + xlim(inpXlim)
        }
    }
    
    return(ggOut)
}

scCoexpUI <- function(id, postfix = 1, subgrp=.globals$subsetgroup[1]) {
    subModuleContainerUI(
        id,
        mainSelectUI = tagList(
            selectInput(
                NS(id, "GeneName1"),
                "Gene 1:",
                choices = NULL,
                width = "100px"
            ),
            selectInput(
                NS(id, "GeneName2"),
                "Gene 2:",
                choices = NULL,
                width = "100px"
            )
        ),
        menuUI = contextMenuCoExprUI(id, group=subgrp),
        contentUI = geneExprDotPlotUI(id, postfix = postfix)
    )
}
scCoexpServer <- function(
        pid,
        id,
        dataSource,
        optCrt,
        p_input,
        p_session,
        interactive,
        postfix = 1) {
    moduleServer(id, function(input, output, session) {
        if (is.null(
            p_session$userData$defaults[[dataSource()$dataset]][[id]])) {
            defaults <- list(
                GeneName1 = dataSource()$sc1def$gene1,
                GeneName2 = dataSource()$sc1def$gene2
            )
        } else{
            defaults <-
                p_session$userData$defaults[[dataSource()$dataset]][[id]]
        }
        ## title
        updateSelectizeInput(
            session,
            "GeneName1",
            choices = sort(names(dataSource()$sc1gene)),
            server = TRUE,
            selected = defaults$GeneName1,
            options = list(
                maxOptions = .globals$maxNumGene,
                create = TRUE,
                persist = TRUE,
                render = I(optCrt)
            )
        )
        updateSelectizeInput(
            session,
            "GeneName2",
            choices = sort(names(dataSource()$sc1gene)),
            server = TRUE,
            selected = defaults$GeneName2,
            options = list(
                maxOptions = .globals$maxNumGene,
                create = TRUE,
                persist = TRUE,
                render = I(optCrt)
            )
        )
        
        subModuleMenuObservor(
            id,
            input,
            p_session,
            dataSource,
            c("GeneName1", "GeneName2", "CoExprcol1", "CoExprord1")
        )
        ## plot
        plot1 <- reactive({
            scDRcoex(
                inpConf=dataSource()$sc1conf,
                inpMeta=dataSource()$sc1meta,
                dimRedX=p_input$GeneExprdrX,
                dimRedY=p_input$GeneExprdrY,
                gene1=input$GeneName1,
                gene2=input$GeneName2,
                subsetCellKey=p_input[[paste0("subsetCell",
                                              input[[paste0("CellInfosubgrp",
                                                            postfix)]])]],
                subsetCellVal=
                    getSubsetCellVal(p_input,
                                     group=input[[paste0("CellInfosubgrp",
                                                         postfix)]]),
                dataset=dataSource()$dataset,
                geneIdMap=dataSource()$sc1gene,
                plotType="2D",
                pointSize=p_input$GeneExprsiz,
                GeneExprDotCol=input$CoExprcol1,
                GeneExprDotOrd=input$CoExprord1,
                labelsFontsize=p_input$GeneExprfsz,
                labelsFontFamily=p_input$GeneExprfml,
                plotAspectRatio=p_input$GeneExprasp,
                keepXYlables=p_input$GeneExprtxt,
                valueFilterKey = p_input$filterCell,
                valueFilterCutoff = p_input$filterCellVal,
                valueFilterCutoff2 = p_input$filterCellVal2,
                hideFilterCell = input$CoExprhid1
            )
        })
        updateSubModulePlotUI(
            postfix,
            pid,
            id,
            input,
            output,
            session,
            interactive,
            plot1,
            .globals$pList1[p_input$GeneExprpsz],
            dataSource()$dataset,
            p_input$GeneExprdrX,
            p_input$GeneExprdrY,
            input$GeneName1,
            input$GeneName2
        )
    })
}

---
title: "scRNAseqApp Guide"
author: "Jianhong Ou"
bibliography: bibliography.bib
csl: nature.csl
vignette: >
  %\VignetteIndexEntry{scRNAseqApp Vignette}
  %\VignetteEncoding{UTF-8}
  %\VignetteEngine{knitr::rmarkdown}
output:
  BiocStyle::html_document:
    toc: true
    toc_float: true
    toc_depth: 4
runtime: shiny
---

```{r, echo=FALSE, results="hide", warning=FALSE, message=FALSE}
suppressPackageStartupMessages({
    library(scRNAseqApp)
})
knitr::opts_chunk$set(warning=FALSE, message=FALSE, fig.width=5, fig.height=3.5)
```

# Introduction

Single-cell RNA sequencing (scRNA-seq) is a powerful technique to study gene
expression, cellular heterogeneity, and cell states within samples in
single-cell level. The development of scRNA-seq shed light to address the
knowledge gap about the cell types, cell interactions, and key genes involved
in biological process and their dynamics.

To increase the re-usability and reproducibility of scientific findings,
more and more publishers require raw data and detailed descriptions of how the
data were analyzed. However, difficulties arise due to the highly concise
descriptions of analysis and the differences of the computing environments.
Furthermore, to precisely meet the publishing requirement, the communication of
the bioinformatician with researchers is a time-consuming step.
Multiple interactive visualization tools were developed to provide the
researchers access to the details of the data. Those tools include,
but not limited to,
alona[@franzen2020alona],
ASAP[@gardeux2017asap],
Asc-Seurat[@pereira2021asc],
BingleSeq[@dimitrov2020bingleseq],
CellView[@bolisetty2017cellview],
cellxgene VIP[@li2022cellxgene],
Cerebro[@hillje2020cerebro],
CHARTS[@bernstein2021charts],
ChromSCape[@prompsy2020interactive],
Cirrocumulus[@li2020cumulus],
CReSCENT[@mohanraj2020crescent],
Cytosplore Viewer[@tasic2018shared],
Granatum[@zhu2017granatum],
InterCellar[@interlandi2022intercellar],
iS-CellR[@patel2018cellr],
iSEE[@rue2018isee],
loom-viewer,
Loupe Cell Browser,
PIVOT[@zhu2018pivot],
SC1[@moussa2021sc1],
SCANNER[@cai2022scanner],
scClustViz[@innes2018scclustviz],
SCope[@davie2018single],
scSVA[@tabaka2019scsva],
scVI[@lopez2018deep],
SeuratV3Wizard[@yousif2020nasqar]/NASQAR,
ShinyArchRUiO[@sharma2022shinyarchr],
ShinyCell[@ouyang2021shinycell],
singleCellTK[@hong2022comprehensive],
Single Cell Explorer (scExplorer)[@feng2019single],
Single Cell Interactive Application (SCiAp)[@moreno2021user] and
UCSC Cell Browser[@speir2021ucsc],
SPRING[@weinreb2018spring],
WASP[@hoek2021wasp], and
Vitessce[@keller2021vitessce].

The basic information of the tools are list in the following tables
(the table was created at 12/06/2022):

| Tool               | platform       | plot type                              |
|:-------------------|:---------------|:---------------------------------------|
| alona              | web-based      | scatters, bar                          |
| ASAP               | web-based      | scatters                               |
| Asc-Seurat     | shiny package  | scatters, heatmap, violin, dot, trajectory |
| BingleSeq          | shiny package  | scatters, heatmap, violin, ridge       |
| CellView           | shiny package  | scatters                               |
| cellxgene VIP      | web-based      | scatters, bar, and visualization plugin|
| Cerebro/cerebroApp | shiny package  | scatters, 3D scatters                  |
| CHARTS             | web-based      | scatters, bar                          |
| ChromSCape         | shiny package  | scatters, heatmap                      |
| Cirrocumulus       | python package |                                        |
| CReSCENT           | web-based      | scatters, violin                       |
| Cytosplore Viewer  | Cytosplore     | scatters, phylogeny                    |
| Granatum           | shiny package  | scatters, bar, trajectory, ppi         |
| InterCellar        | shiny package  | dot, ppi, circos, radar, pie           |
| iS-CellR           | shiny package  | scatters, heatmap, violin, bar, dot    |
| iSEE               | shiny package  | scatters, heatmap, violin, bar         |
| loom-viewer        | python package |                                        |
| Loupe Cell Browser | Desktop        | scatters                               |
| PIVOT              | shiny package  | scatters, heatmap, violin, bar, pie    |
| SC1                | web-based      | scatters, heatmap, violin, bar         |
| SCANNER            | web-based      | scatters                               |
| scClustViz         | shiny package  | scatters, heatmap, violin, bar, dot    |
| scExplorer         | web-based      | scatters, heatmap                      |
| SCope              | web-based      | scatters                               |
| scSVA              | shiny package  | scatters, 3D scatters                  |
| scVI               | python package | scatters, heatmap, violin, bar         |
| seuratv3wizard     | web-based      | scatters                               |
| ShinyArchRUiO      | shiny package  | scatters, heatmap, track               |
| ShinyCell          | shiny package  | scatters, heatmap, violin, bar         |
| SCiAp[^1]          | galaxy         |                                        |
| singleCellTK   | shiny package  | scatters, heatmap, violin, dot, trajectory |
| SPRING             | python package | scatters                               |
| UCSC Cell Browser  | web-based      | scatters                               |
| Vitessce           | python package | scatters, heatmap                      |
| WASP               | shiny package  | scatters, heatmap                      |
[^1]: short name for Single Cell Interactive Application

| Tool           | languages  | license | starts | watching | forks | citation |
|:---------------|:-----------|:--------|:-------|:---------|:------|:---------|
| alona          | python     | GPL-3   | 12     | 3        | 5     | 24       |
| ASAP           | Java,R,Python  | GPL-3 | 18   | 6        | 8     | 88       |
| Asc-Seurat     | R          | GPL-3   | 12     | 2        | 6     | 10       |
| BingleSeq      | R          | MIT     | 18     | 2        | 6     | 4        |
| CellView       | R          | MIT     | 16     | 9        | 8     | 8        |
| cellxgene VIP  | python,R,JavaScript  | MIT    | 81   | 6 | 24    | 10       |
| Cerebro/cerebroApp | R,JavaScript,C   | MIT    | 79   | 7 | 18    | 41       |
| CHARTS         | python     | MIT     | 2      | 6        | 0     | 7        |
| ChromSCape     | R          | GPL-3   | 11     | 2        | 4     | 13       |
| Cirrocumulus   | JavaScript,Python  | BSD-3   | 38    | 7 | 9     | 78       |
| CReSCENT       | R,Perl     | GPL-3   | 8      | 1        | 4     | 11       |
| Cytosplore Viewer  | java,javascript  |  |     |          |       |          |
| Granatum       | R          | Apache2 | 18     | 4        | 11    | 65       |
| InterCellar    | R          | MIT     | 7      | 1        | 3     | 4        |
| iS-CellR       | R          | GPL-3   | 21     | 6        | 6     | 15       |
| iSEE           | R          | MIT     | 201    | 14       | 39    | 41       |
| loom-viewer    | python,JavaScript   | BSD-2   | 32   | 9 | 6     |          |
| Loupe Cell Browser |        |         |        |          |       |          |
| PIVOT          | R          |         | 27     | 6        | 15    | 27       |
| SC1            | R          |         |        |          |       | 5        |
| SCANNER        | R          |         | 0      | 2        | 1     | 1        |
| scClustViz     | R          | MIT     | 41     | 12       | 10    | 36       |
| scExplorer     | JavaScript,Python   | GPL-3   | 7    | 3 | 6     | 16       |
| SCope          | python,JavaScript   | GPL-3   | 60   | 8 | 14    | 438[^2]  |
| scSVA          | R          | GPL-3   | 20     | 6        | 7     | 8        |
| scVI           | python     | BSD-3   | 840    | 27       | 263   | 787      |
| seuratv3wizard | R          | GPL-3   | 29     | 8        | 13    | 28[^3]   |
| ShinyArchRUiO  | R          | GPL-3   | 11     | 2        | 4     | 3        |
| ShinyCell      | R          | GPL-3   | 70     | 9        | 23    | 28       |
| SCiAp          |            |         |        |          |       | 16       |
| singleCellTK   | R          | MIT     | 105    | 12       | 61    | 6        |
| SPRING         | python,matlab,JavaScript |    | 59   | 10   | 29 | 250      |
| UCSC Cell Browser | JavaScript,Python,R  | GPL-3   | 3   | 1 | 39 | 46       |
| Vitessce       | JavaScript,Python,R     | MIT     | 92  | 6 | 23 | 2        |
| WASP           | R,python   |         | 2      | 1        | 0     | 3        |
[^2]: the package contribute partial to the citation
[^3]: citation is from NASQAR

| Tool              | source code                                              |
|:------------------|:---------------------------------------------------------|
| alona             | https://github.com/oscar-franzen/adobo/                  |
| ASAP              | https://github.com/DeplanckeLab/ASAP                     |
| Asc-Seurat        | https://github.com/KirstLab/asc_seurat/                  |
| BingleSeq         | https://github.com/dbdimitrov/BingleSeq/                 |
| CellView          | https://github.com/mohanbolisetty/CellView               |
| cellxgene   VIP   | https://github.com/interactivereport/cellxgene_VIP       |
| Cerebro/cerebroApp| https://github.com/romanhaa/Cerebro                      |
| CHARTS            | https://github.com/stewart-lab/CHARTS                    |
| ChromSCape        | https://github.com/vallotlab/ChromSCape                  |
| Cirrocumulus      | https://github.com/lilab-bcb/cirrocumulus                |
| CReSCENT          | https://github.com/pughlab/crescent                      |
| Cytosplore Viewer |                                                          |
| Granatum          | https://github.com/lanagarmire/Granatum                  |
| InterCellar       | https://github.com/martaint/InterCellar                  |
| iS-CellR          | https://github.com/immcore/iS-CellR                      |
| iSEE              | https://github.com/iSEE/iSEE                             |
| loom-viewer       | https://github.com/linnarsson-lab/loom-viewer            |
| Loupe Cell Browser|                                                          |
| PIVOT             | https://github.com/kimpenn/PIVOT                         |
| SC1               |                                                          |
| SCANNER           | https://github.com/GuoshuaiCai/scanner                   |
| scClustViz        | https://github.com/BaderLab/scClustViz                   |
| scExplorer        | https://github.com/d-feng/scExplorer                     |
| SCope             | https://github.com/aertslab/Scope                        |
| scSVA             | https://github.com/klarman-cell-observatory/scSVA        |
| scVI              | https://github.com/scverse/scvi-tools                    |
| seuratv3wizard    | https://github.com/nasqar/seuratv3wizard                 |
| ShinyArchRUiO     | https://github.com/EskelandLab/ShinyArchRUiO             |
| ShinyCell         | https://github.com/SGDDNB/ShinyCell                      |
| SCiAp             |                                                          |
| singleCellTK      | https://github.com/compbiomed/singleCellTK               |
| SPRING            | https://github.com/AllonKleinLab/SPRING_dev              |
| UCSC Cell Browser | https://github.com/ucscGenomeBrowser/cellBrowser         |
| Vitessce          | https://github.com/vitessce/vitessce                     |
| WASP              | https://github.com/andreashoek/wasp                      |

| Tool              | demo                                                     |
|:------------------|:---------------------------------------------------------|
| alona             | https://alona.panglaodb.se/                              |
| ASAP              | https://asap.epfl.ch/                                    |
| Asc-Seurat        |                                                          |
| BingleSeq         |                                                          |
| CellView          | https://mbolisetty.shinyapps.io/CellView/                |
| cellxgene   VIP   | https://cellxgenevip-ms.bxgenomics.com/                  |
| Cerebro/cerebroApp|                                                          |
| CHARTS            | https://charts.morgridge.org/                            |
| ChromSCape        | https://vallotlab.shinyapps.io/ChromSCape/               |
| Cirrocumulus      |                                                          |
| CReSCENT          | https://crescent.cloud/                                  |
| Cytosplore Viewer |                                                          |
| Granatum          | http://granatum.dcmb.med.umich.edu:8103/                 |
| InterCellar       |                                                          |
| iS-CellR          |                                                          |
| iSEE              | https://marionilab.cruk.cam.ac.uk/iSEE_allen/            |
| loom-viewer       |                                                          |
| Loupe Cell Browser|                                                          |
| PIVOT             |                                                          |
| SC1               | https://sc1.engr.uconn.edu/                              |
| SCANNER           | https://www.thecailab.com/scanner/                       |
| scClustViz        | https://innesbt.shinyapps.io/scclustvizdemoapp/          |
| scExplorer        | http://singlecellexplorer.org                            |
| SCope             | https://scope.aertslab.org/                              |
| scSVA             |                                                          |
| scVI              |                                                          |
| seuratv3wizard    | https://nasqar.abudhabi.nyu.edu/SeuratV3Wizard/          |
| ShinyArchRUiO     | https://cancell.medisin.uio.no/ShinyArchR.UiO/           |
| ShinyCell         | http://shinycell1.ddnetbio.com/                          |
| SCiAp             | https://humancellatlas.usegalaxy.eu/                     |
| singleCellTK      | https://sctk.bu.edu/                                     |
| SPRING.           | https://kleintools.hms.harvard.edu/tools/                |
|                   | springViewer_1_6_dev.html?datasets/mouse_HPCs/           |
|                   | basal_bone_marrow/full[^4]                               |
| UCSC Cell Browser | https://cells.ucsc.edu/                                  |
| Vitessce          | http://vitessce.io/                                      |
| WASP              |                                                          |

| Tool              | visualization tutorial                                   |
|:------------------|:---------------------------------------------------------|
| alona             | https://alona.panglaodb.se/faq.html                      |
| ASAP              | https://asap.epfl.ch/home/tutorial?t=fca                 |
| Asc-Seurat        | https://asc-seurat.readthedocs.io/en/latest/index.html   |
| BingleSeq   | https://github.com/dbdimitrov/BingleSeq/blob/master/README.md  |
| CellView          |                                                          |
| cellxgene VIP|https://interactivereport.github.io/cellxgene_VIP/tutorial/docs|
| Cerebro/cerebroApp| https://romanhaa.github.io/cerebroApp/                   |
| CHARTS      | https://github.com/stewart-lab/CHARTS/blob/master/README.md    |
| ChromSCape  | https://vallotlab.github.io/ChromSCape/articles/vignette.html  |
| Cirrocumulus      | https://cirrocumulus.readthedocs.io/en/latest/           |
| CReSCENT          | https://pughlab.github.io/crescent-frontend/             |
| Cytosplore Viewer | https://viewer.cytosplore.org/                           |
| Granatum | https://github.com/lanagarmire/Granatum/blob/master/doc/          |
|                   | Granatum_manual_0.92.pdf[^4]                             |
| InterCellar       |                                                          |
| iS-CellR          |                                                          |
| iSEE              |                                                          |
| loom-viewer       | https://github.com/linnarsson-lab/loom-viewer            |
| Loupe Cell Browser|                                                          |
| PIVOT | https://rawgit.com/qinzhu/PIVOT/master/inst/app/www/manual_file.html |
| SC1               |                                                          |
| SCANNER           |                                                          |
| scClustViz        |                                                          |
| scExplorer        | http://singlecellexplorer.org/tutorial.html              |
| SCope             | https://github.com/aertslab/SCope/blob/master/README.md  |
| scSVA             | https://github.com/klarman-cell-observatory/scSVA/blob/  |
|                   | master/docs/index.md[^4]                                 |
| scVI           | https://docs.scvi-tools.org/en/stable/user_guide/index.html |
| seuratv3wizard|https://github.com/nasqar/seuratv3wizard/blob/master/README.md|
| ShinyArchRUiO     |                                                          |
| ShinyCell         | https://github.com/SGDDNB/ShinyCell/blob/master/README.md|
| SCiAp             |                                                          |
| singleCellTK      |                                                          |
| SPRING            | https://kleintools.hms.harvard.edu/tools/spring.html     |
| UCSC Cell Browser|https://cellbrowser.readthedocs.io/en/master/interface.html|
| Vitessce          | http://vitessce.io/docs/                                 |
| WASP              | https://github.com/andreashoek/wasp/blob/main/README.md  |
[^4]: URL shown in multiple lines

# Motivation

Based on [_ShinyCell_](https://github.com/SGDDNB/ShinyCell),
The _scRNAseqApp_ package is developed with multiple highly interactive
visualizations of how cells and subsets of cells cluster behavior.
The end users can discover the expression of genes in multiple interactive
manners with highly customized filter conditions by selecting metadata
supplied with the publications and download the ready-to-use results for
republishing. 

# Quick start

Here is an example using _scRNAseqApp_ with a subset of scRNA-seq data.

## Installation
First, install _scRNAseqApp_ and other packages required to run 
the examples.
Please note that the example dataset used here is from a small subset of
PBMC[@satija2015spatial].
Additional package are also required for enhancement functions such 
trajectory analysis or cell communication analysis.

```{r, eval=FALSE}
library(BiocManager)
BiocManager::install("scRNAseqApp")
```

## Load library

```{r}
library(scRNAseqApp)
```

## Initial the database

```{r}
publish_folder=tempdir()
scInit(app_path=publish_folder)
```

## Start shiny app
```{r}
scRNAseqApp(app_path = publish_folder)
```

# Create a new data

There are two ways to create a new data.

- from the administrator mode
- use R session

## by administrator

Log in to admin by `Switch User` tab and click `administrator button` in the 
right-bottom corner of screen. Click `UploadData` and upload a Seurat object.

## via R session

There are two steps to create a new data via R session.
First, create the config file with description of the data and
second create the data from a Seurat object.

```{r}
library(Seurat)
appconf <- createAppConfig(
            title="pbmc_small_test",
            destinationFolder = "pbmc_small_test",
            species = "Homo sapiens",
            doi="10.1038/nbt.3192",
            datatype = "scRNAseq",
            abstract = 'Put the description of the data here.')
createDataSet(
    appconf,
    pbmc_small,
    datafolder=file.path(publish_folder, "data"))
dir(file.path(publish_folder, 'data'))
```

# Add downloadable file

If you have files intended for user download, kindly save them in the
`www/download` folder.
The application will then generate a list, differentiating between files and
user folder names, for convenient access to download options.
The file named as `readme` in text format in each folder will be used as 
descriptions for the files.

# Distribute to a shiny server

There are two steps to distribute to a shiny server. First, install the package
in the server as root user. Second, in a R session run `scInit()` after load 
the `scRNAseqApp` library. If you initialed the app offline, copy the app folder
to the shiny server.

Note: the following files need to be writable for shiny: `www/database.sqlite`,
`www/counter.tsv` and the app `www` folder should also be writable because the 
user manager is depend on SQLite, and the SQLite needs to be able to 
create a journal file in the same directory as the DB, 
before any modifications can take place. 
The journal is used to support transaction rollback.

# SessionInfo

```{r}
sessionInfo()
```


# References

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/scInit.R
\name{scInit}
\alias{scInit}
\title{Create a scRNAseqApp project}
\usage{
scInit(
  app_path = getwd(),
  root = "admin",
  password = "scRNAseqApp",
  datafolder = "data",
  overwrite = FALSE,
  app_title = "scRNAseq Database",
  app_description =
    "This database is a collection of\\n        single cell RNA-seq data.",
  passphrase = NULL
)
}
\arguments{
\item{app_path}{path, a directory where do you want to create the app}

\item{root}{character(1), the user name for administrator}

\item{password}{character(1), the password for administrator}

\item{datafolder}{the folder where saved the dataset for the app}

\item{overwrite}{logical(1), overwrite the \code{app_path} if there is
a project.}

\item{app_title, app_description}{character(1). The title and
description of the home page.}

\item{passphrase}{A password to protect the data inside the database.}
}
\value{
no returns. This function will copy files to \code{app_path}
}
\description{
To run scRNAseqApp, you need to first create a directory
which contains the required files.
}
\examples{
if(interactive()){
    scInit()
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/makeShinyFiles.R
\name{extractFragmentNameMapList}
\alias{extractFragmentNameMapList}
\title{Extract fragment name map}
\usage{
extractFragmentNameMapList(obj, atacAssayName)
}
\arguments{
\item{obj}{input single-cell object for Seurat (v3+)}

\item{atacAssayName}{assay in single-cell data object to use for plotting
open chromatin.}
}
\value{
A list of fragment name map.
}
\description{
Try to extract cell name information. It will be used to map the fragment cell
name to the object cell name.
}
\examples{
library(Seurat)
#library(Signac)
#fnm <- extractFragmentNameMapList(atac_small, 'peaks')

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/createDataset.R
\name{createDataSet}
\alias{createDataSet}
\title{Create a dataset
Create a dataset from a Seurat object. The function will try to find
the markers in the Misc data named as 'markers'.
The misc data should be output of function \code{FindAllMarkers}.}
\usage{
createDataSet(
  appconf,
  seu,
  config,
  contrast,
  assayName,
  gexSlot = c("data", "scale.data", "counts"),
  atacAssayName,
  atacSlot = c("data", "scale.data", "counts"),
  LOCKER = FALSE,
  datafolder = "data",
  default.symbol = "rownames",
  theme = "Paired",
  binSize = 1,
  fragmentNameMapList
)
}
\arguments{
\item{appconf}{a list object represent the information about the dataset}

\item{seu}{a Seurat object}

\item{config}{config file for makeShinyFiles}

\item{contrast}{The contrast group}

\item{assayName}{assay in single-cell data object to use for plotting
gene expression, which must match one of the following:
\itemize{
\item{Seurat objects}: "RNA" or "integrated" assay,
default is "RNA"
}}

\item{gexSlot}{layer in single-cell assay to plot.
Default is to use the "data" layer}

\item{atacAssayName}{assay in single-cell data object to use for plotting
open chromatin.}

\item{atacSlot}{layer in single-cell atac assay to plot.
Default is to use the "data" layer}

\item{LOCKER}{Set locker if the file is required login}

\item{datafolder}{app data folder}

\item{default.symbol}{character(1L) specifying the default rownames to be used. If use default, the gene symbols will be the row names of the assay. If one column name of the meta.feature of the assay is supplied, the function will try to extract the symbols from the meta.feature slot of the assay.}

\item{theme}{color theme. default is "Paired" from ColorBrewer palettes.}

\item{binSize}{number of bps for each bin for ATAC fragment coverage. Used
to reduce the file size of bigwig.}

\item{fragmentNameMapList}{list of named character vector.
The name map list must be the same order as the fragment list in the object.
For each element of the list,
the names of the vector are the name of the fragment and
the vector contains the cell names (column names of the assay).
You can try \link{extractFragmentNameMapList}.}
}
\value{
The updated Seurat object.
}
\description{
Create a dataset
Create a dataset from a Seurat object. The function will try to find
the markers in the Misc data named as 'markers'.
The misc data should be output of function \code{FindAllMarkers}.
}
\examples{
library(Seurat)
if(interactive()){
    appconf <- createAppConfig(
        title="pbmc_small",
        destinationFolder = "pbmc_small",
        species = "Homo sapiens",
        doi="10.1038/nbt.3192",
        datatype = "scRNAseq")
    createDataSet(appconf, pbmc_small, datafolder=tempdir())
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/APPconf-class.R
\docType{class}
\name{APPconf-class}
\alias{APPconf-class}
\alias{APPconf}
\title{Class \code{"APPconf"}}
\usage{
APPconf(...)
}
\arguments{
\item{\dots}{Each argument in \dots becomes an slot in the new
\code{"APPconf"}-class.}
}
\value{
A APPconf object.
}
\description{
An object of class \code{"APPconf"} represents
the metadata for a dataset.
}
\section{Slots}{

\describe{
\item{\code{title}}{character(1). Title of the data}

\item{\code{id}}{character(1). Folder name of the data}

\item{\code{species}}{character(1). species}

\item{\code{ref}}{Reference information in a list with element bib, doi, pmid
and entry. Entry must be an object of \link[utils]{bibentry}}

\item{\code{type}}{character(1). Type of the data, scRNAseq or scATACseq.}

\item{\code{markers}}{list. A list of data.frame represents cell
markers.}

\item{\code{keywords}}{character. A vector of characters represents the
keywords of the study.}

\item{\code{groupCol}}{character. The key group column name to separate
the cells.}
}}

\examples{
appconf <- readRDS(system.file("extdata", "data",
    "pbmc_small", "appconf.rds", package="scRNAseqApp"))
appconf
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/APPconf-methods.R
\name{APPconf-methods}
\alias{APPconf-methods}
\alias{show,APPconf-method}
\alias{$,APPconf-method}
\alias{$<-,APPconf-method}
\alias{[[,APPconf,ANY,ANY-method}
\alias{[[<-,APPconf,ANY,ANY,ANY-method}
\alias{[,APPconf,ANY,ANY,ANY-method}
\alias{as.list,APPconf-method}
\alias{as.character,APPconf-method}
\alias{markers,APPconf-method}
\alias{markers}
\alias{lapply,APPconf-method}
\alias{unlist,APPconf-method}
\title{The methods for \link{APPconf-class}}
\usage{
\S4method{show}{APPconf}(object)

\S4method{$}{APPconf}(x, name)

\S4method{$}{APPconf}(x, name) <- value

\S4method{[[}{APPconf,ANY,ANY}(x, i, j, ..., exact = TRUE)

\S4method{[[}{APPconf,ANY,ANY,ANY}(x, i, j, ...) <- value

\S4method{[}{APPconf,ANY,ANY,ANY}(x, i, j, ..., drop = TRUE)

\S4method{as.list}{APPconf}(x, ...)

\S4method{as.character}{APPconf}(x, ...)

\S4method{markers}{APPconf}(x)

\S4method{lapply}{APPconf}(X, FUN, ...)

\S4method{unlist}{APPconf}(x, recursive = TRUE, use.names = TRUE)
}
\arguments{
\item{object}{an object of APPconf}

\item{x}{APPconf object.}

\item{name}{A literal character string or a name (possibly backtick quoted).}

\item{value}{value to replace.}

\item{i, j}{indices specifying elements to extract or replace.}

\item{\dots}{Named or unnamed arguments to form a signature.}

\item{exact}{see \link[base]{Extract}}

\item{drop}{see \link[base]{drop}}

\item{X}{an APPconf object.}

\item{FUN}{function used by \code{lapply}}

\item{recursive, use.names}{function used by \link[base:unlist]{unlist}}
}
\value{
A named character vector.
}
\description{
The assessment and replacement methods for \link{APPconf-class}
}
\examples{
appconf <- readRDS(system.file("extdata", "data",
    "pbmc_small", "appconf.rds", package="scRNAseqApp"))
appconf
appconf$title
appconf[["title"]]
as.list(appconf)
as.character(appconf)
markers(appconf)
lapply(appconf, print)
unlist(appconf)
}
\concept{APPconf}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/createDataset.R
\name{createAppConfig}
\alias{createAppConfig}
\title{Create a metadata to describe the dataset}
\usage{
createAppConfig(
  title,
  destinationFolder,
  species,
  doi,
  pmid,
  bibentry,
  datatype = c("scRNAseq", "scATACseq", "scMultiome"),
  markers,
  keywords,
  abstract
)
}
\arguments{
\item{title}{The title of the dataset}

\item{destinationFolder}{The destination folder name of the dataset without
the root folder of the datasets. The data will be saved as
\code{appdataFolder/destinationFolder}}

\item{species}{The species of the dataset}

\item{doi, pmid}{The DOI or PMID of the reference}

\item{bibentry}{An object of bibentry}

\item{datatype}{character(1). Type of the data, scRNAseq, scATACseq or
scMultiome.}

\item{markers}{A list of data.frame with gene symbols as rownames or
a character vector.}

\item{keywords}{The keywords for the dataset.
For example the condition, cell type, tissue information
The keywords will be used for whole database search}

\item{abstract}{The abstract of the reference.}
}
\value{
An object of \link{APPconf} object
}
\description{
The function will return a APPconf object which contain
the reference, keywords for the dataset.
}
\examples{
if(interactive()){
    config <- createAppConfig(
        title="pbmc_small",
        destinationFolder = "pbmc_small",
        species = "Homo sapiens",
        doi="10.1038/nbt.3192",
        datatype = "scRNAseq")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/createDataset.R
\name{createSeuFromCellRanger}
\alias{createSeuFromCellRanger}
\title{load data from cellRanger}
\usage{
createSeuFromCellRanger(outsFolder)
}
\arguments{
\item{outsFolder}{the outs folder of cellRanger}
}
\value{
An SeuratObject
}
\description{
load data from cellRanger
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/createDataset.R
\name{createSeuFromMatrix}
\alias{createSeuFromMatrix}
\title{load data from a count matrix}
\usage{
createSeuFromMatrix(matrix, meta, genes, cluster, ...)
}
\arguments{
\item{matrix}{count matrix}

\item{meta}{cell-level meta data}

\item{genes}{character. gene names, will be the rownames of the matrix}

\item{cluster}{the cluster coordinates}

\item{...}{The parameter passed to read.delim when read cluster file.}
}
\value{
An SeuratObject
}
\description{
load data from a count matrix
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/availableThemes.R
\name{availableThemes}
\alias{availableThemes}
\title{Available themes of color sets}
\usage{
availableThemes(output = "name")
}
\arguments{
\item{output}{The theme names or 'name' to print all available themes.}
}
\value{
A palette.
}
\description{
The color themes inspired by modern web design palettes
}
\examples{
availableThemes()

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/app.R
\name{scRNAseqApp}
\alias{scRNAseqApp}
\title{scRNAseqApp main function}
\usage{
scRNAseqApp(
  app_path = getwd(),
  datafolder = "data",
  defaultDataset = "pbmc_small",
  windowTitle = "scRNAseq/scATACseq database",
  favicon = system.file("assets", "img", "favicon.ico", package = "scRNAseqApp"),
  banner = system.file("assets", "img", "banner.png", package = "scRNAseqApp"),
  footer = tagList(HTML("&copy;"), "2024 -", format(Sys.Date(), "\%Y"),
    "jianhong@morgridge"),
  maxRequestSize = 1073741824,
  timeout = 30,
  theme = bs_theme(bootswatch = "lumen"),
  use_bs_themer = FALSE,
  showHelpVideo = FALSE,
  ...
)
}
\arguments{
\item{app_path}{path, a directory where do you want to create the app}

\item{datafolder}{the folder where saved the dataset for the app}

\item{defaultDataset}{default dataset for the app.}

\item{windowTitle}{The title that should be displayed by the browser window.}

\item{favicon}{The favicon for the page.}

\item{banner}{The banner image.}

\item{footer}{The footer html contents.}

\item{maxRequestSize}{Maximal upload file size. Default is 1G.}

\item{timeout}{Timeout session (minutes) before logout if sleeping.
Default to 30. 0 to disable.}

\item{theme}{A theme.}

\item{use_bs_themer}{logical(1). Used to determine the theme.}

\item{showHelpVideo}{logical(1) or character(1).
Show help videos in homepage or not. If an url is provided, the
url will be embedded as a iframe element.}

\item{...}{parameters can be passed to shinyApp except ui and server.}
}
\value{
An object that represents the app.
}
\description{
create a scRNAseqApp once the initialization is done.
}
\examples{
if(interactive()){
    app_path=tempdir()
    scInit(app_path=app_path)
    setwd(app_path)
    scRNAseqApp()
}
}
